/* ----------------------------------------------------------- */
/*                   HTK  -  HMM Toolkit                       */
/* ----------------------------------------------------------- */
/*         Copyright: Steve Young      December 1990           */
/*         Cambridge University Engineering Department         */
/*         Trumpington Street, Cambridge, CB2 1PZ              */
/*                                                             */
/*         Co-authored by Valtcho Valtchev                     */
/*                                                             */
/* Permission is hereby granted to copy, use and modify this   */
/* software provided that this banner notice is not removed    */
/*                                                             */
/* ----------------------------------------------------------- */
/*         File: HGraf.c:   Minimal Graphics Interface         */
/* ----------------------------------------------------------- */

#include <stdio.h>         /* Standard Libraries */
#include <math.h>

#include <X11/Xlib.h>      /* the X11 stuff makes string.h also available */
#include <X11/Xutil.h>
#include <X11/Xos.h>

#include "HShell.h"
#include "HMath.h"
#include "HGraf.h"

/* ---------------------- UNIX/X-Windows Version ------------ */

/*
   This module provides a minimal graphics facility.  It assumes
   that an implementation can provide a single fixed window onto 
   which the usual graphics and text can be painted.  No event
   handling is required or assumed, hence in multiple window
   environments damage to the HGraf window may not be repaired.
      
*/


Display    *display;      /* pointer to display structure */
Window      win;          /* the window */
int         screen;
GC          gc;           /* graphics context */
XSizeHints  size_hints;   /* hints structure for the windowa manager */

XEvent report;


char *display_name = NULL;             /* if not set the $DISPLAY is used */
char *window_name  = "HGraf: ";        /* name of the window */
char *icon_name    = "HGraf";          /* name of the icon */
char *default_font = "9x15";


/*------------------------- FONTS SECTION-----------------------*/


#define NO_OF_FONTS 6
#define FONTS_AVAILABLE 6

char *font_names[NO_OF_FONTS] = {
      "-adobe-courier-medium-o-normal--8-80-75-75-m-50-iso8859-1",
      "-adobe-courier-medium-o-normal--10-100-75-75-m-60-iso8859-1",
      "-adobe-courier-medium-o-normal--12-120-75-75-m-70-iso8859-1",
      "-adobe-courier-medium-o-normal--14-140-75-75-m-90-iso8859-1",
      "-adobe-courier-medium-o-normal--18-180-75-75-m-110-iso8859-1",
      "-adobe-courier-medium-o-normal--24-240-75-75-m-150-iso8859-1"  };

int   font_sizes[NO_OF_FONTS] = {8, 10, 12, 14, 18, 24};

XFontStruct *font_info[NO_OF_FONTS];
XFontStruct *current_font;

int  HFont; 

/*--------------------------------------------------------------*/


int  x_pos = 0, y_pos = 0;            /* initial position of the window */


/* ---- locally used functions ----*/

void LoadX11Font(char *font_name, XFontStruct **font_info);

void InstallFonts(void);

void UnInstallFonts(void);

unsigned long x11color(char *colorname);


void TermGraf()
{
  int exiting;

  if (display != NULL) {
    printf("\nClick on display to exit ...");
    exiting = FALSE;
    while (!exiting) {
      XNextEvent(display, &report);
      switch (report.type) {
      case Expose:
   break;
      case ButtonPress:
      case KeyPress:
        UnInstallFonts();
   XFreeGC(display, gc);
   XCloseDisplay(display);
        exiting = TRUE;
      default:
   break;
      } /* end switch */
    } /* end while */

    printf("\n\n\n");

  }
}

/* EXPORT-> InitHGraf: Create a window of given size and initialise */
void InitHGraf(int width, int height, char *name)
{

  unsigned int display_width, display_height;
  unsigned int valuemask    = 0;
  unsigned int line_width   = 1;
  unsigned int border_width = 4;
  
  int line_style = LineSolid;
  int join_style = JoinRound;
  int cap_style  = CapRound;

  char buf[128];

  XGCValues values;

  /* connect to server */
  if ((display = XOpenDisplay(display_name)) == NULL) {
    fprintf(stderr, "HGraf: cannot connect to X server %s\n",
       XDisplayName(display_name));
    exit(-1);
  }

  /* get screen size */
  screen = DefaultScreen(display);

  /* width and height may be set according to the above values */
  display_width  = DisplayWidth(display, screen);
  display_height = DisplayHeight(display, screen);
 
  /* create window - black on white background */
  win = XCreateSimpleWindow(display, RootWindow(display, screen),
                            x_pos, y_pos, width, height, border_width,
                            BlackPixel(display, screen),
             WhitePixel(display, screen) );

  /* init size hint property for window manager */
  size_hints.flags = PPosition | PSize | PMinSize;
  size_hints.x = x_pos;
  size_hints.y = y_pos;
  size_hints.width  = width;
  size_hints.height = height;
  size_hints.min_width  = 350;
  size_hints.min_height = 250;

  /* set properties for window manager */
  strcpy(buf, window_name);
  strcat(buf, name);
  XSetStandardProperties(display, win, buf, icon_name, NULL,
          NULL, 0, &size_hints);


  /* select events */
  XSelectInput(display, win, ExposureMask | KeyPressMask | ButtonPress);

  /* create graphics context */
  gc = XCreateGC(display, win, valuemask, &values);

  /* load fonts */
  InstallFonts();

  /* set default line attributes */
  XSetLineAttributes(display, gc, line_width, line_style,
           cap_style, join_style);

  HSetColour(BLACK);   /* default color */
  HSetTextSize( 12);   /* default text size */
  HSetXMode( GCOPY);   /* default transfer mode */

  XMapWindow(display, win);  

  /* get first EXPOSE event - not allowed to draw before it */
  do 
    XNextEvent(display, &report);
  while (report.type != Expose);

}

/* EXPORT-> HSetColour: Set current colour to c */
void HSetColour(Colour c)
{
  static char *XColArray[] = {"white", "yellow", "gold", "red",
               "magenta", "plum", "blue", "light blue",
               "green", "forest green", "firebrick", "brown",
               "light grey", "grey", "dark slate grey",
               "black" };

  /* set the foreground color in gc */
  XSetForeground(display, gc, x11color(XColArray[c]));
}

/* EXPORT-> HSetGrey: Set current colour to the specified grey level*/
void HSetGrey(int level)
{
  Colormap cmap;
  XColor grey_def, white_def, black_def;
  short RGB_val, grey_step;

  cmap  = DefaultColormap(display, screen);

  /* get the white pixel value - max brightness for RGB */
  white_def.pixel = WhitePixel(display, screen);
  XQueryColor(display, cmap, &white_def);
  /* get the black pixel value - min brightness for RGB */
  black_def.pixel = BlackPixel(display, screen);
  XQueryColor(display, cmap, &black_def);

  /* divide the black - white range by the no. of grey levels*/
  grey_step = (white_def.red - black_def.red)/MAXGREYLEVEL;

  /* calculate the values for RGB (all the same) for grey level <level> */  
  RGB_val = white_def.red - level*grey_step; 
  
  grey_def.red   = RGB_val;
  grey_def.green = RGB_val;
  grey_def.blue  = RGB_val;

  /* find the nearest available or allocate a new cell if possible */
  if (!XAllocColor(display, cmap, &grey_def)) {
    fprintf(stderr, "HGraf: all colorcells allocated\n");
    exit(0);
  }

  /* set the foreground color in gc */
  XSetForeground(display, gc, grey_def.pixel);
}

/* EXPORT-> HSetXMode: Set current transfer mode */
void HSetXMode(XferMode m)
{ 
  int TFuncX11[3] = {GXcopy, GXor, GXxor};
  XSetFunction(display, gc, TFuncX11[((int) m)]);
}

/* EXPORT-> HDrawLine: Draw line from (x0,y0) to (x1,y1) */
void HDrawLine(int x0,int y0,int x1,int y1)
{
  XDrawLine(display, win, gc, x0, y0, x1, y1);
}

/* EXPORT-> HDrawRect: Draw rectangle */
void HDrawRect(int x0,int y0,int x1,int y1)
{ 
  unsigned int rw, rh;
  /* calculate width and height */
  rw = abs(x1 - x0);  
  rh = abs(y1 - y0); 
  XDrawRectangle(display, win, gc, x0, y0, rw, rh); 
}

/* EXPORT-> HFillRect: Draw filled rectangle  */
void HFillRect(int x0,int y0,int x1,int y1)
{
  unsigned int rw, rh;
  /* calculate width and height */
  rw = 1 + abs(x1 - x0);
  rh = 1 + abs(y1 - y0);
  /* width and height have to be incremented */
  XFillRectangle(display, win, gc, x0, y0, rw, rh);
}

/* EXPORT-> HDrawArc: Draw arc from stAngle thru arcAngle degrees */
void HDrawArc(int x0,int y0,int x1,int y1,int stAngle,int arcAngle)
{
  unsigned int rw, rh;
  /* calculate width and height */
  rw = abs(x1 - x0);
  rh = abs(y1 - y0);
  /* the angles are signed integers in 64ths of a degree */
  stAngle *=64;
  arcAngle*=64;
  XDrawArc( display, win, gc, x0, y0, rw, rh, stAngle, arcAngle);
}

/* EXPORT-> HFillArc: Draw filled arc from stAngle thru arcAngle degrees */
void HFillArc(int x0,int y0,int x1,int y1,int stAngle,int arcAngle)
{
  unsigned int rw, rh;
  /* calculate width and height */
  rw = abs(x1 - x0);
  rh = abs(y1 - y0);
  /* the angles are signed integers in 64ths of a degree */
  stAngle *=64;
  arcAngle*=64;
  XFillArc( display, win, gc, x0, y0, rw, rh, stAngle, arcAngle);
}

/* EXPORT-> HSetTextSize: Set font size in points */
void HSetTextSize(int size)
{  
  int i, index;
  int d, min_d;

  min_d = 1000;

  for (i=0; i < FONTS_AVAILABLE; i++) {
    d=abs(font_sizes[i] - size);
    if (d < min_d) {
      min_d = d;
      index = i;
    }
  }
  current_font = font_info[index];
  XSetFont(display, gc, current_font->fid); 
}

/* EXPORT-> HDrawText: Draw the given text string */
void HDrawText(int x,int y, char *s)
{
  XDrawString(display, win, gc, x, y, s, strlen(s));
}

/* EXPORT-> HTextWidth: Returns the width of the given string */
int HTextWidth(char *s)
{
  return XTextWidth(current_font, s, strlen(s));
}

/* EXPORT-> HPlotVector: plot vector v in given rectangle */
void HPlotVector(int x0,int y0,int x1,int y1, 
                 Vector v, int st, int en, float ymax, float ymin)
{
  Real yScale, yOffset, xInc, x;
  int  xOld, yOld, ix, iy, i;

  if (st >= en || st < 1 || en > VectorSize(v))
    fprintf(stderr, "HGraf: Plot indices %d -> %d out of range", st, en);
 
  x = (x1 - x0 - 1);
  xInc = x/(en - st);
  yScale = (y1 - y0)/(ymin - ymax);
  yOffset = y0 - ymax*yScale;
  x = x0;
  xOld = x; yOld = v[st]*yScale + yOffset;
  for (i = st+1; i <= en; i++){
    x += xInc; ix = x;
    iy = v[i]*yScale + yOffset;
    XDrawLine(display, win, gc, xOld, yOld, ix, iy);
    xOld = ix; yOld = iy;
  } 
}

/* EXPORT-> HSpoolGraf: start saving an image of window in fname */
void HSpoolGraf(char *fname)
{
}


void LoadX11Font(char *font_name, XFontStruct **font_info)

{
  if ((*font_info = XLoadQueryFont(display, font_name)) == NULL) {
    fprintf(stderr, "HGraf: cannot open %s font\n", font_name);
    exit (-1);
  }
}


void InstallFonts(void)
{
  int  i;
  for (i=0; i < FONTS_AVAILABLE; i++) {
    LoadX11Font(font_names[i], &(font_info[i]));
  }
}


void UnInstallFonts(void)
{
  int  i;
  for (i=0; i < FONTS_AVAILABLE; i++) {
    XUnloadFont(display, font_info[i]->fid);
  }
}


unsigned long x11color(char *colorname) 
{
  Colormap cmap;
  XColor exact_def;
  int depth;

  depth = DisplayPlanes(display, screen);
  cmap  = DefaultColormap(display, screen);

  /* check if name is the X11 database */
  if (!XParseColor(display, cmap, colorname, &exact_def)) {
    fprintf(stderr, "HGraf: color name %s not in X11 database\n", colorname);
    exit(0);
  }

  /* find the nearest available or allocate a new cell if possible */
  if (!XAllocColor(display, cmap, &exact_def)) {
    fprintf(stderr, "HGraf: all colorcells allocated\n");
    exit(0);
  }
  return exact_def.pixel;
}
