/* ----------------------------------------------------------- */
/*                   HTK  -  HMM Toolkit                       */
/* ----------------------------------------------------------- */
/*         Copyright: Steve Young      October 1989            */
/*         Cambridge University Engineering Department         */
/*         Trumpington Street, Cambridge, CB2 1PZ              */
/*                                                             */
/* Permission is hereby granted to copy, use and modify this   */
/* software provided that this banner notice is not removed    */
/*                                                             */
/* ----------------------------------------------------------- */
/*         File: HLabel.h:   Speech Label File Input           */
/* ----------------------------------------------------------- */

/* 
   This module allows an internal data structure representing a (possibly
   multilevel) transcription to be constructed by reading labels from an
   external file.  A number of single level input formats are supported but
   multilevel input and all output is limited to the default HTK format.
*/

typedef struct _NameCell{ /* Hash Table Linked List Item */
   char * name;            /* Label Name */
   int aux;                /* User flags etc*/
   struct _NameCell *next; /* Chain */
}NameCell;

typedef NameCell *LabId;  /* Internal representation of names */
  
typedef struct {           /* Information for each label */
   LabId labid;            /* Id of label */
   long start,end;         /* Start and end times in 100ns units */
   float score;            /* Score eg. logP */
}Label;

typedef struct _LabList { /* Sequence of labels at one level */
   short nLabels;          /* Number at this level */
   Label *labels;          /* Array[0..nLabels-1]Of Label */
   struct _LabList *next;  /* Next (lower) level if any */
}LabList;

typedef LabList *Transcription;   /* Pointer to level 0 LabList */

void InitLabel(void);
/*
   Initialises hash table - must be called before using any of
   the routines in this module.
*/

LabId GetLabId(char *name, int insert);
/*
   Lookup given name in hash table and return its id.  If it
   is not there and insert is true then insert the new name
   otherwise return NULL.
*/

void SetLFormat(FileFormat fmat);
/*
   Set the current format to fmat.  This will override the
   environment variable HLFORMAT.  Note that the format only
   affects reading all output is in HTK format.
*/

FileFormat GetLFormat(void);
/*
   Return the current format from the environment variable HLFORMAT, or
   the override specified in last call to SetLFormat.
*/

Transcription LOpen(char *fname);
/*
   Reads the labels stored in file fname according to the current format
   as defined by GetLFormat and build the internal representation 
   specified by the types LabList and Label.
*/

void LSave(char *fname, Transcription t);
/* 
   Save the given transcription in file fname using HTK format.
*/

void LClose(Transcription t);
/*
   Reclaim the storage allocated to t.
*/

Transcription LClone(Transcription t);
/*
   Return a copy of transcription t.
*/

int NumLevels(Transcription t);
/*
   Return the number of levels in transcription t.
*/

LabList *GetLevel(Transcription t, int n);
/*
   Return a pointer to the LabList for level n in transcription t.
   Levels are indexed 0,1,2,....
*/

int NumCases(LabId x, Transcription t, int n);
/* 
   Return the number of occurrences of label x in level n of
   transcription t.
*/

Label *GetCase(LabId x, int m, Transcription t, int n);
/*
   Return a pointer to the Label struct for the m'th 
   occurrence of label x in level n of transcription t.
*/

void LTriStrip(int enab);
/*
   When enab is set, all triphone labels with the form A-B+C
   are converted to B on read in by LOpen.
*/

void TriStrip(char *s);
/* 
   Remove contexts of form A- and +B from s 
*/

LabId *LoadIdList(char *fname, int *nLabs);
/*
   Create an array of label ids indexed 1..P by reading
   the label names in fname.  The array returned is of
   size P+1 with the final element containing NULL.  The
   number of labels P is returned in *nLabs.
*/

int FindLabId(LabId id, LabId *idList);
/*
   Find the index of id in idList.  Return 0 if not found.
*/

Boolean ReadLabel(FILE *f, char *buf);
/*
   Read the next label from f into buf.  A label is any
   sequence of printing chars.  Returns false if no label
   found.
*/

/* ------------------------------------------------------------------ */
/*                     End of File: HLabel.h                          */
/* ------------------------------------------------------------------ */



