/* ----------------------------------------------------------- */
/*                   HTK  -  HMM Toolkit                       */
/* ----------------------------------------------------------- */
/*         Copyright: Steve Young      October 1989            */
/*         Cambridge University Engineering Department         */
/*         Trumpington Street, Cambridge, CB2 1PZ              */
/*                                                             */
/* Permission is hereby granted to copy, use and modify this   */
/* software provided that this banner notice is not removed    */
/*                                                             */
/* ----------------------------------------------------------- */
/*         File: HMath.h:   Math Support                       */
/* ----------------------------------------------------------- */

#define PI   3.1415927
#define TPI  6.2831853     /* P*2 */
#define LZERO  -1.0E10     /* ~log(0) */
#define LSMALL -0.5E10     /* log values < LSMALL are set to LZERO */
#define MINEARG -708.3     /* lowest exp() arg */
#define MINLARG 2.45E-308  /* lowest log() arg */

typedef double Real;    /* Choose this type for optimum fpu speed */
typedef Real LogReal;
typedef float *Vector;     /* vector[1..size]   */
typedef float **Matrix;    /* matrix[1..nrows][1..ncols] */



void InitMath(int useLogs);
/*
   Initialise the module,  useLogs must be set if log
   arithmetic is to be performed
*/

Vector CreateVector(int size);
/*
   Create and return a vector of size components
*/

void FreeVector(Vector v);
/*
   Free the memory allocated for vector v
*/

void ZeroVector(Vector v);
/*
   Zero the elements of v
*/

void FloatMultVector(Vector v, float w);
/*
  Multiply each element of v by  w   
*/

void CopyVector(Vector v1, Vector v2);
/*
   Copy v1 into v2; sizes must be the same
*/

int VectorSize(Vector v);
/*
   Return the number of components in vector v
*/

void ReadVector(FILE *stream, Vector v);
void ReadlnVector(FILE *stream, Vector v);
void BReadVector(FILE *stream, Vector v);
/*
   Read vector v from stream using e format or binary
*/

void WriteVector(FILE *stream, Vector v);
void WritelnVector(FILE *stream, Vector v);
void BWriteVector(FILE *stream, Vector v);
/*
   Write vector v to stream using e format and default precision
   or binary
*/

void ShowVector(char * title,Vector v,int maxTerms);
/*
   Print the title followed by upto maxTerms elements of v
*/

Matrix CreateMatrix(int nrows,int ncols);
/*
   Create and return a matrix with nrows rows and ncols columns
*/

void FreeMatrix(Matrix m);
/*
   Free the space occupied by matrix m
*/

void ZeroMatrix(Matrix m);
/*
   Zero the elements of m
*/

void ZeroLogMatrix(Matrix m);
/*
   Zero the elements of log m
*/

void CopyMatrix(Matrix m1, Matrix m2);
/*
   Copy matrix m1 to m2 
*/

int NumRows(Matrix m);
int NumCols(Matrix m);
/*
   Return the number of rows/cols in matrix m
*/

void ReadMatrix(FILE *stream, Matrix m);
void ReadlnMatrixColByCol(FILE *stream, Matrix m);
void ReadSymMatrix(FILE *stream, Matrix m);
void BReadMatrix(FILE *stream, Matrix m);
/*
   Read matrix from stream into m using e format or binary.
   Sym version expects m to be in upper triangular form
*/
   
void WriteMatrix(FILE *stream, Matrix m);
void WritelnMatrix(FILE *stream, Matrix m);
void WriteSymMatrix(FILE *stream, Matrix m);
void BWriteMatrix(FILE *stream, Matrix m);
void WritelnMatrixColByCol(FILE *stream, Matrix m);

/*
    write matrix to stream using e format and default precision
    or binary. Sym version expects m to be in upper triangular
    form
*/

void ShowMatrix(char * title,Matrix m,int maxCols,int maxRows);
/*
   Print the title followed by upto maxCols elements of upto
   maxRows rows of m.
*/

LogReal LAdd(LogReal x, LogReal y);
/*
   Return x+y where x and y are stored as logs, 
   sum < LSMALL is floored to LZERO 
*/

LogReal LSub(LogReal x, LogReal y);
/*
   Return x-y where x and y are stored as logs, 
   sum < LSMALL is floored to LZERO 
*/

float   L2F(LogReal x);
/*
   Convert log(x) to real, result is floored to 0.0 if x < LSMALL 
*/

LogReal CovInvert(Matrix c, Matrix invc);
/*
   Computes inverse of c in invc and returns the log of Det(c).
*/

LogReal CovDet(Matrix c);
/*
   Returns log of Det(c)
*/

float RandomValue(void);
/*
   Return a random number in range 0.0->1.0 with uniform distribution
*/

float GaussDeviate(float mu, float sigma);
/*
   Return a random number with a N(mu,sigma) distribution
*/

/* ------------------------ End of HMath.h ------------------------- */
