/* ----------------------------------------------------------- */
/*                   HTK  -  HMM Toolkit                       */
/* ----------------------------------------------------------- */
/*         Copyright: Steve Young      October 1989            */
/*         Last Update:                September 1990          */
/*                                                             */
/*         Cambridge University Engineering Department         */
/*         Trumpington Street, Cambridge, CB2 1PZ              */
/*                                                             */
/* Permission is hereby granted to copy, use and modify this   */
/* software provided that this banner notice is not removed    */
/*                                                             */
/* ----------------------------------------------------------- */
/*         File: HModel.c  HMM Model Definition Data Type      */
/* ----------------------------------------------------------- */

#include <stdio.h>         /* Standard Libraries */
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <math.h>
#include <errno.h>
#include <stdarg.h>

#include "HShell.h"
#include "HSpIO.h"
#include "HMath.h"
#include "HModel.h"

#ifdef MPW
   #pragma segment hmods
#endif

/*
   Principle function of this model is to map between external textual
   representation for hmm definitions and internal data structure.
   External format is
   
      hmmdef = <BeginHMM>
                  options
                  stateinfo {stateinfo}
                  transP
               <EndHMM>
            
      options = { <NumStates>   n |             -- default 5
                  <NumMixtures> m |             -- default 1
                  <VecSize>     v |             -- default 12
                  <TieMaster>     |             -- Master for tied mixes
                  <TieSlave>    f |             -- Slave, master is f
               (<lpc>|<lpcepstra>| <fbank>|
                <mfcc>|<lprefc>|<lpdelcep>) |   -- default <lpdelcep>
               (<nullD>|<poissonD>|<gammaD>) |  -- default <nullD>
               (<diagC> | <fullC>)              -- default <diagC>
               }
               
      stateinfo = <State> s                     -- state id 2..n
                     mixture {mixture}          -- mixtures 1..m
                     [duration]                 -- omitted if <nullD>
                     [label]
                     
      mixture  =  [ <Mixture> n f]              -- f is mixture weight
                     <Mean> f f f f ...         -- vector of v means
                     [<Variance>                -- <diagC> case
                            f f f ...  |        --    v variances
                      <InvCovar>                -- <fullC> case
                           f f f f ...          --    v x v inverse covariance
                             f f f ...                matrix in upper triangular
                               f f ...                form
                                     f]
                     [<GConst> f]               -- precomputed const part of b(x)
                     
      duration = <Duration> f [f]               -- <poissonD> [<gammaD>] cases
      
      label    = <label> "string"
      
      transP   = <TransP>                       -- n x n matrix
                     f f f f ....
                     f f f f ....
   where n, m, s and v are integers and f are floats. Note that all 
   tokens aside from numbers and strings are written within angle brackets.  
   Numbers and strings conform to usual C conventions.
   
   Note that for the full covariance case, the INVERSE covariance matrix is
   stored since this is needed for both recognition and training.  Training 
   also needs the covariance matrix and this may be stored in aux.  
   GConst is that part of the Gaussian output probability that can be 
   precomputed ie the part that doesnt depend on the observation vector x.
   
   For tied mixtures, one hmm def is the Master containing all the output 
   distributions.  All the other defs are slaves in which the means and
   covariances are omitted.  Each slave names the master in the <TieSlave>
   option.
   
*/

/* ------------------------ Symbol Conversions ------------------------ */

#define MAXSYMLEN 40
#define MAXSTRLEN 80
static char *symMap[] = { 
               "BEGINHMM", "ENDHMM", "NUMSTATES","NUMMIXTURES",
               "VECSIZE", "NULLD",
               "POISSOND", "GAMMAD", "DIAGC", "FULLC", "STATE",
               "MIXTURE", "MEAN", "VARIANCE", "INVCOVAR", "GCONST" ,
               "DURATION", "LABEL", "TIEMASTER", "TIESLAVE", "TRANSP",
               "SAMPKIND" };
               
enum _Symbol {BEGINHMM, ENDHMM, NUMSTATES, NUMMIXTURES,
              VECSIZE , NDUR,
              PDUR, GDUR, DCOV, FCOV, STATE, 
              MIXTURE, MEAN, VARIANCE, INVCOVAR, GCONST,
              DURATION, LABEL, TIEMASTER, TIESLAVE, TRANSP, SAMPKIND };
              
typedef enum _Symbol Symbol;

static char defName[100];     /* current model for err messages */
static SampKind theSampKind;  /* samp kind when sym==SAMPKIND */

static Symbol GetSym(FILE *f, char *exp)
{
   char c, buf[MAXSYMLEN], cc[20];
   int i;
   Symbol s;
   
   while (isspace(c=fgetc(f)));        /* Read symbol string into buf */
   if (c != '<') {
      if (c == EOF) 
         HError(2,"%s: EOF when %s expected",defName,exp);
      cc[0] = c; cc[1]='\0';
      fgets(buf,10,f); strcat(cc,buf);
      HError(2,"%s: Symbol %s expected [got %s]",defName,exp,cc);
   }
   i=0;
   while ((c=fgetc(f)) != '>' && i<MAXSYMLEN) 
      buf[i++] = islower(c)?toupper(c):c;
   buf[i] = '\0';
   if (c != '>')
      HError(2,"%s: > missing in %s when looking for %s",defName,buf,exp);
   
   for (s=BEGINHMM; s<=TRANSP; s++)       /* Look symbol up in symMap */
      if (strcmp(symMap[s],buf) == 0) 
         return s;                        /* if not there it may be a sampkind */
   if ((theSampKind = Str2SampKind(buf)) != ANON)
      return SAMPKIND;
      
   HError(2,"%s: Unknown symbol %s when looking for %s",defName,buf,exp);
}

static void ChkSym(FILE *f, Symbol s)
{
   char *ss;
   
   ss = symMap[s];
   if (GetSym(f,ss) != s)
      HError(2,"%s: Symbol %s expected",defName,ss);
}

static int GetInt(FILE *f)
{
   int n;
   
   if (fscanf(f,"%d",&n) != 1)
      HError(2,"%s: Integer Expected",defName);
   return n;
}

static float GetFloat(FILE *f)
{
   float x;
   
   if (fscanf(f,"%e",&x) != 1)
      HError(2,"%s: Float Expected",defName);
   return x;
}

static char * GetString(FILE *f)
{
   char c, buf[MAXSTRLEN], *s;
   int i;
   
   while (isspace(c=fgetc(f)));        /* Read string into buf */
   if (c != '"')
      HError(2,"String expected");
   i=0;
   while ((c=fgetc(f)) != '"' && i<MAXSTRLEN) 
      buf[i++] = c;
   buf[i] = '\0';
   if (c != '"')
      HError(2,"%s: String terminator > missing in %s",defName,buf);
   s = NewString(strlen(buf));
   return strcpy(s,buf);
}

/* ------------------------- Create/Close Model --------------------------- */   


/* EXPORT-> HCreate: Create a data structure in memory for a hmm def */
void HCreate(int numStates, int numMixtures, int vecSize, SampKind pkind,
            CovKind ckind, DurKind dkind, TieStatus tstat, HMMDef *hmm)
{
   int n,m;
   StateInfo *sptr;
   OutProb *optr;
   
   hmm->numStates = numStates; hmm->numMixtures = numMixtures;
   hmm->vecSize = vecSize;     hmm->pkind = pkind;
   hmm->ckind = ckind;         hmm->dkind = dkind;
   hmm->tstat = tstat;   
   if ((sptr = (StateInfo*)calloc(numStates-2,sizeof(StateInfo)) ) == NULL)
      HError(2,"Cannot allocate state info array");
   hmm->stateinfo = sptr-2;   /* stateinfo is indexed 2..numStates-1 */
   for (n=2; n<=numStates-1; n++) {
      if ((optr = (OutProb*)calloc(numMixtures,sizeof(OutProb)) ) == NULL)
         HError(2,"Cannot allocate Outprob array");
      sptr->outp = optr-1;    /* outprobs is indexed 1..numMixtures */
      for (m=1; m<=numMixtures; m++) {
         optr->weight = 0.0;
         if (tstat!=FreeT) {
            optr->mean = optr->covariance.var = NULL;
         } else {
            optr->mean = CreateVector(vecSize);
            if (ckind == FULLC)
               optr->covariance.inv = CreateMatrix(vecSize,vecSize);
            else
               optr->covariance.var = CreateVector(vecSize);
         }
         optr->gConst = LZERO;
         optr->aux = NULL; 
         optr++;
      }
      sptr->dur1 = sptr->dur2 = 0.0;
      sptr->label = NULL;
      sptr++;
   }
   hmm->transP = CreateMatrix(numStates,numStates);
}

/* EXPORT-> HClose: Free memory allocated to hmm */
void HClose(HMMDef *hmm)
{
   int n,m;
   StateInfo *sptr;
   OutProb *optr;

   sptr = hmm->stateinfo+2;
   for (n=2; n<=hmm->numStates-1; n++) {
      optr = sptr->outp+1; 
      for (m=1; m<=hmm->numMixtures; m++) {
         if (hmm->tstat == FreeT){
            FreeVector(optr->mean);
            if (hmm->ckind == FULLC)
               FreeMatrix(optr->covariance.inv);
            else
               FreeVector(optr->covariance.var);
         }
         if (optr->aux != NULL)
            free(optr->aux);
         optr++;
      }
      if (sptr->label != NULL)
         free(sptr->label);
      free(sptr->outp+1);
      sptr++;
   }
   free(hmm->stateinfo+2);
   FreeMatrix(hmm->transP);
   if (hmm->tstat != FreeT) 
      free(hmm->tieMaster);
}

/* --------------------------- Tied Mixtures ----------------------------- */

static int numTMix;
static int vsTMix;
static CovKind ckTMix;
static OutProb *tMix = NULL;      /* array[1..numTMix]of OutProb */
static char *tMasterName = NULL;  /* name of Master for Tied Mixes */

/* TMCreate: create tied mixture storage */
static void TMCreate(char *tname, int numMixtures, int vecSize, CovKind ckind)
{
   int m;
   OutProb *optr;
   char buf[100];
   
   numTMix = numMixtures; ckTMix = ckind; vsTMix = vecSize;
   NameOf(tname,buf);
   tMasterName = strcpy(NewString(strlen(buf)),buf);
   if ((optr = (OutProb*)calloc(numMixtures,sizeof(OutProb)) ) == NULL)
      HError(2,"Cannot allocate Outprob array");   
   tMix=optr-1;
   for (m=1; m<=numMixtures; m++) {
      optr->weight = 0.0;
      optr->mean = CreateVector(vecSize);
      if (ckind == FULLC)
         optr->covariance.inv = CreateMatrix(vecSize,vecSize);
      else
         optr->covariance.var = CreateVector(vecSize);
      optr->gConst = LZERO;
      optr->aux = NULL; 
      optr++;
   }
}

/* EXPORT->TMClose: release storage occupied by tied mixtures */
void TMClose(void)
{
   int m;
   OutProb *optr;

   optr = tMix; 
   for (m=1; m<=numTMix; m++) {
      optr++;
      FreeVector(optr->mean);
      if (ckTMix == FULLC)
         FreeMatrix(optr->covariance.inv);
      else
         FreeVector(optr->covariance.var);
   }
   free(tMix+1);
   free(tMasterName);
   tMix = NULL;
   tMasterName = NULL;
}

/* TMJoin: join the given HMM to the tied mixtures */
static void TMJoin(HMMDef *hmm, int nmix, CovKind ck, int vSize)
{
   int n,m;
   StateInfo *sptr;
   OutProb *optr,*tptr;
   
   if (tMix==NULL)
      HError(2,"No Master to Join to");
   if (nmix!=numTMix || ck!=ckTMix || vsTMix!=vSize)
      HError(2,"Master and slave not joinable");
   for (n=2;n<hmm->numStates;n++){
      sptr = hmm->stateinfo + n;
      tptr = tMix; optr = sptr->outp;
      for (m=1;m<=nmix;m++){
         ++tptr; ++optr;
         optr->mean = tptr->mean;
         optr->covariance = tptr->covariance;
      }
   }
}

/* ChkSlaveName: check that given name matches current tie master */
static void ChkSlaveName(char *s)
{
   if (strcmp(s,tMasterName)!=0)
      HError(2,"MisMatched Master %s and Slave %s",tMasterName,s);
}

/* ---------------------------- Input Model ------------------------------ */

static Symbol sym;               /* the current input symbol */
static int numStates;            /* HMM option parameters */
static int numMixtures;
static int vecSize;
static SampKind pkind;
static DurKind dkind;
static CovKind ckind;
static TieStatus tstat;
static char *slaveTMName;

/* InitOptions: set options to default values */
static void InitOptions(void)
{
   numStates = 5; numMixtures = 1; vecSize = 12;
   pkind = LPDELCEP; dkind = NULLD; ckind = DIAGC;
   tstat = FreeT; 
   if (slaveTMName != NULL){
      free(slaveTMName); slaveTMName = NULL;
   }
}

/* GetOptions: read in the HMM option info */
void GetOptions(FILE *f)
{
   while ((sym=GetSym(f,"State or trans mat")) != STATE) 
      switch (sym) {
         case NUMSTATES:   numStates=GetInt(f); break;
         case NUMMIXTURES: numMixtures=GetInt(f); break;
         case VECSIZE:     vecSize=GetInt(f); break;
         case SAMPKIND:    pkind = theSampKind; break;
         case NDUR:        dkind = NULLD; break;
         case PDUR:        dkind = POISSOND; break;
         case GDUR:        dkind = GAMMAD; break;
         case DCOV:        ckind = DIAGC; break;
         case FCOV:        ckind = FULLC; break;
         case TIEMASTER:   tstat = MasterT; break;
         case TIESLAVE:    tstat=SlaveT; 
                           slaveTMName = GetString(f);
                           break;
         default: 
            HError(2,"%s: Ilegal Option Symbol %s",defName, symMap[sym]);
      }
}

/* GetTransMat: read in the transition matrix */
static void GetTransMat(FILE *f,HMMDef *hmm, Boolean logTrans)
{
   int i,j;
   Vector v;
   
   if (sym != TRANSP)
      HError(2,"%s: <transP> expected",defName);
   ReadMatrix(f,hmm->transP);
   if(logTrans)
      for (i=1;i<numStates;i++){
         v=hmm->transP[i];
         for (j=1;j<=numStates;j++)
            v[j] = v[j]==0.0 ? LZERO : log(v[j]);
      }
   v = hmm->transP[numStates];
   for (j=1;j<=numStates;j++)
      v[j] = logTrans ? LZERO : 0.0;
}

/* CheckMixes: check consistency of mixture distributions */
static void CheckMixes(HMMDef *hmm)
{
   int i,m;
   StateInfo *sptr;
   float sum;
   OutProb *optr;
   
   if (numMixtures == 1) return;
   for (i=2; i<numStates; i++){
      sptr = hmm->stateinfo +i;
      sum = 0.0;
      for (m=1; m<=numMixtures; m++){
         optr = sptr->outp + m;
         if (optr->weight>MINMIX){
            if (optr->mean == NULL)
               HError(2,"%s: No mean vector, s=%d, m=%d",defName,i,m);
            if (optr->covariance.inv == NULL)
               HError(2,"%s: No covariance, s=%d, m=%d",defName,i,m);
            sum += optr->weight;
            if (optr->gConst == LZERO)
               HError(2,"%s: GConst not set, s=%d, m=%d",defName,i,m);
         }
      }
      if (sum < 0.99 || sum > 1.01)
         HError(2,"%s: Mix Weight Sum out of range at s=%d",defName,i);
   }
}
         
/* DoState: process a single HMM state definition */
static void DoState(FILE *f, HMMDef *hmm, TieStatus tstat)
{
   int n,m;
   StateInfo *sptr;
   OutProb *optr;
   float weight;
   
   n=GetInt(f);
   sptr = hmm->stateinfo + n;
   sym = GetSym(f,"StatInfo");
   while (sym == MIXTURE || sym == MEAN) {
      if (sym==MIXTURE) {
         m = GetInt(f); weight = GetFloat(f);
         sym = GetSym(f,"Mean or next tied mix");
      } else {
         if (tstat!=FreeT)
            HError(2,"%s: Tied Mixes must have explicit mix weights",defName);
         m = 1; weight = 1.0;
      }
      optr = sptr->outp + m;
      optr->weight = weight;
      if (tstat == FreeT || (tstat==MasterT && n==2)){
         if (sym!=MEAN)
            HError(2,"%s: <Mean> Expected[got %s]",defName,symMap[sym]);
         ReadVector(f,optr->mean);
         sym = GetSym(f,"Covariance");
         if (ckind == DIAGC) {
            if (sym != VARIANCE) 
               HError(2,"%s: <Variance> expected [got %s]",defName,symMap[sym]);
            ReadVector(f,optr->covariance.var);
         } else {
            if (sym != INVCOVAR) 
               HError(2,"%s: <InvCovar> expected[got %s]",defName,symMap[sym]);
            ReadSymMatrix(f,optr->covariance.inv);
         }
         sym=GetSym(f,"GConst or next mix");
         if (sym==GCONST){
            optr->gConst = GetFloat(f);
            sym=GetSym(f,"Next mix");
         } else {
            if (ckind == DIAGC)
               FixDiagGConst(optr);
            else
               FixFullGConst(optr,-CovDet(optr->covariance.inv));
         }
         if (tstat==MasterT)
            tMix[m].gConst = optr->gConst;
      } else {
         optr->gConst = tMix[m].gConst;
      }
   }
   if (sym==DURATION) {
      if (dkind == NULLD)
         HError(2,"%s: Cannot specify duration if NULLD",defName);
      sptr->dur1 = GetFloat(f);
      if (dkind == GAMMAD)
         sptr->dur2 = GetFloat(f);
      sym = GetSym(f,"Label or next state");
   }
   if (sym==LABEL) {
      sptr->label = GetString(f);
      sym = GetSym(f,"Next State");
   }
}


/* H_Open: create a hmm def from contents of fname, if logTrans
           is set then transition matrix is converted to logs */
static void H_Open(char *fname, HMMDef *hmm, Boolean logTrans)
{
   FILE *f;

   if ((f=fopen(fname,"r")) == NULL)
      HError(2,"Cannot open HMM def %s",fname);
   strcpy(defName,fname);
   ChkSym(f,BEGINHMM);
   InitOptions();
   GetOptions(f);
   if (tstat==MasterT){       /* New Tied Master */
      if(tMix != NULL) TMClose();
      TMCreate(fname,numMixtures,vecSize,ckind);
   }
   if (tstat==SlaveT) 
      ChkSlaveName(slaveTMName);
   HCreate(numStates,numMixtures,vecSize,pkind,ckind,dkind,tstat,hmm);
   if (tstat!=FreeT) {
      TMJoin(hmm,numMixtures,ckind,vecSize);
      hmm->tieMaster = strcpy(NewString(strlen(slaveTMName)),slaveTMName);
   }
   while (sym == STATE) 
      DoState(f,hmm,tstat);
   GetTransMat(f,hmm,logTrans);
   ChkSym(f,ENDHMM);
   fclose(f);
   CheckMixes(hmm);
}

/* EXPORT-> HOpen: Read model definition from fname and create HMM def */
void HOpen(char *fname, HMMDef *hmm)
{
   H_Open(fname,hmm,TRUE);
}

/* EXPORT-> HOpen2: as for HOpen but no conversion of transP to logs */
void HOpen2(char *fname, HMMDef *hmm)
{
   H_Open(fname,hmm,FALSE);
}

/* ---------------------------- Output Model ------------------------------ */

#ifdef MPW
typedef struct  {
    unsigned long fdType;          /*the type of the file*/
    unsigned long fdCreator;       /*file's creator*/
    unsigned short fdFlags;        /*flags ex. hasbundle,invis,locked, etc. */
    long fdLocation;               /*file's location in folder*/
    short fdFldr;                  /*folder containing file*/
}FInfo;

/* SetHMMFileInfo: Sets file creator to HTK; file kind to TEXT */
static void SetHMMFileInfo(char *fn)
{
   FInfo inf;

   if (getfinfo(fn,0,&inf))
      HError(2,"Cannot Get File Info for %s",fn);
   inf.fdType='T'<<24|'E'<<16|'X'<<8|'T';
   inf.fdCreator='H'<<24|'T'<<16|'K'<<8|' ';
   inf.fdFlags |= 6;
   if(setfinfo(fn,0,&inf))
      HError(2,"Cannot Set File Info for %s",fn);
}
#endif

static void H_Save(char *fname, HMMDef *hmm, int logTrans)
{
   int n,m,i,j;
   StateInfo *sptr;
   OutProb *optr;
   char buf1[20],buf2[20],buf3[20];
   FILE *f;
   Vector v;
   float rSum;
   
   if ((f=fopen(fname,"w")) == NULL)
      HError(2,"Cannot open hmm save file %s",fname);
   fprintf(f,"<BeginHMM>\n");
   fprintf(f,"<NumStates> %d  ", hmm->numStates);
   fprintf(f,"<NumMixtures> %d  ", hmm->numMixtures);
   fprintf(f,"<VecSize> %d\n", hmm->vecSize);
   fprintf(f,"<%s> <%s> <%s>\n", CovKind2Str(hmm->ckind,buf1),
      DurKind2Str(hmm->dkind,buf2), SampKind2Str(hmm->pkind,buf3));
   switch(hmm->tstat){
      case FreeT: 
         break;
      case SlaveT:
         fprintf(f,"<TieSlave> \"%s\"\n",hmm->tieMaster);
         break;
      case MasterT:
         fprintf(f,"<TieMaster>\n");
         break;
   }
   for (n=2; n<=hmm->numStates-1; n++) {
      fprintf(f,"<State> %d\n",n); 
      sptr = hmm->stateinfo+n;
      optr = sptr->outp;
      for (m=1; m<=hmm->numMixtures; m++) { 
         optr++;
         if (optr->weight>MINMIX || hmm->numMixtures == 1 
                                 || (hmm->tstat == MasterT && n == 2) ) {
            if (hmm->numMixtures > 1)
               fprintf(f,"<Mixture> %d %e\n",m,optr->weight);
            if (hmm->tstat == FreeT ||  (hmm->tstat == MasterT && n == 2)){
               fprintf(f,"<Mean>\n");
               WriteVector(f,optr->mean);
               if ( hmm->ckind==DIAGC) {
                  fprintf(f,"<Variance>\n");
                  WriteVector(f,optr->covariance.var);
               } else {
                  fprintf(f,"<InvCovar>\n");
                  WriteSymMatrix(f,optr->covariance.inv);
               }
               if (optr->gConst > LSMALL)
                  fprintf(f,"<GConst> %e\n",optr->gConst);
            }
         }
      }
      if (hmm->dkind !=NULLD) {
         fprintf(f,"<Duration>");
         switch (hmm->dkind) {
            case POISSOND: fprintf(f,"%e\n",sptr->dur1); break;
            case GAMMAD:   fprintf(f,"%e %e\n",sptr->dur1,sptr->dur2); break;
         }
      }
      if (sptr->label != NULL)
         fprintf(f,"<Label>  \"%s\"\n",sptr->label);
   }
   fprintf(f,"<TransP>\n");

   /* DR - the following code fixes a bug in which HSave had the side effect
      of returning the unlogged transition matrix to the calling routine if
      logTrans was set */

   v = CreateVector(hmm->numStates);

   for (i=1;i<hmm->numStates;i++){       
       CopyVector(hmm->transP[i], v);
       rSum = 0.0;         
      for (j=1; j<=hmm->numStates;j++) {
         if (logTrans) v[j] = L2F(v[j]);
         rSum += v[j];
      }
      for (j=1;j<=hmm->numStates;j++)
         fprintf(f,"%e ",v[j]/rSum);
      fprintf(f,"\n");
   }

   FreeVector(v);

   for (j=1;j<=hmm->numStates;j++)
     fprintf(f,"%e ",0.0);
   fprintf(f,"\n");

   fprintf(f,"<EndHMM>\n");
   fclose(f);
#ifdef MPW
    SetHMMFileInfo(fname);
#endif
}

/* EXPORT-> HSave: Save the model hmm in file fname */
void HSave(char *fname, HMMDef *hmm)
{
   H_Save(fname,hmm,1);
}

/* EXPORT-> HSave2: as for HSave but transP not assumed log format */
void HSave2(char *fname, HMMDef *hmm)
{
   H_Save(fname,hmm,0);
}

/* ---------------------------- Print Model ------------------------------ */

/* EXPORT-> HPrint: Print the model hmm according to specified flags */
void HPrint(HMMDef *hmm, int flags, ...)
{
   int stState,enState,i,j,n,m;
   Vector v;
   Matrix c;
   StateInfo *sptr;
   OutProb *optr;
   va_list ap;
   char buf1[20],buf2[20],buf3[20];
   
   va_start(ap,flags);
   if (flags & (HPMEANS|HPVARS|HPDURS|HPLABELS)){
      stState = va_arg(ap,int);
      enState = va_arg(ap,int);
      if (stState<2 || enState>=hmm->numStates)
         HError(2,"Illegal state range in hprint");
   }
   va_end(ap);
   if (flags & HPDETAILS) {
      printf("Num States   = %3d\n", hmm->numStates);
      printf("Num Mixtures = %3d\n", hmm->numMixtures);
      printf("Vector Size  = %3d\n", hmm->vecSize);
      printf("Features     = %s %s %s\n",
         CovKind2Str(hmm->ckind,buf1),DurKind2Str(hmm->dkind,buf2),
         SampKind2Str(hmm->pkind,buf3));
      switch(hmm->tstat){
         case FreeT: 
            break;
         case SlaveT:
            printf("Tied Mixture Slave [Master=%s]\n",hmm->tieMaster);
            break;
         case MasterT:
            printf("Tied Mixture Master\n");
            break;
      }

   }
   if (flags & (HPMEANS|HPVARS|HPDURS|HPLABELS))
      for (n=stState; n<=enState; n++) {
         printf("State %d\n",n); 
         sptr = hmm->stateinfo+n;
         if (flags & (HPMEANS|HPVARS)) {
            optr = sptr->outp;
            for (m=1; m<=hmm->numMixtures; m++) { 
               optr++;
               if (optr->weight>MINMIX || hmm->numMixtures == 1) {
                  if (hmm->numMixtures>1)
                     printf(" Mixture %d:  weight %8.3f\n",m,optr->weight);
                  if (flags &HPMEANS) {
                     v = optr->mean;
                     printf("  Means:\n");
                     for (i=1;i<=hmm->vecSize;i++)
                        printf("%8.3f",v[i]);
                     printf("\n");
                  }
                  if (flags & HPVARS) {
                     if ( hmm->ckind==DIAGC) {
                        v = optr->covariance.var;
                        printf("  Variances:\n");
                        for (i=1;i<=hmm->vecSize;i++)
                           printf("%8.3f",v[i]);
                        printf("\n");
                     } else {
                        c = optr->covariance.inv;
                        printf("  Inverse Covariance Matrix:\n");
                        for (i=1;i<=hmm->vecSize;i++) {
                           for (j=1;j<=hmm->vecSize;j++)
                              printf("%8.3f",c[i][j]);
                           printf("\n");
                        }
                     }
                     printf("  Gauss Constant:  ");
                     if (optr->gConst > LSMALL)
                        printf("%e\n",optr->gConst);
                     else
                        printf("not set\n");
                  }
               }
            }
         }
         if (flags & HPDURS) {
            printf("  Duration Parameters: ");
            switch (hmm->dkind) {
               case NULLD:    printf("None\n"); break;
               case POISSOND: printf("%8.3f\n",sptr->dur1); break;
               case GAMMAD:   printf("%8.3f %8.3f\n",sptr->dur1,sptr->dur2); break;
            }
         }
         if (flags & HPLABELS)
            printf("  Label: %s\n", sptr->label == NULL  ? "None" : sptr->label);
      }
   if (flags & HPTRANS) {
      printf("Transition Matrix:\n");
      for (i=1;i<=hmm->numStates;i++) {
         for (j=1;j<=hmm->numStates;j++)
            printf("%8.3f",flags&LOGTRANS?L2F(hmm->transP[i][j]):hmm->transP[i][j]);
         printf("\n");
      }
   }  
}

/* ----------------- Output Probability Calculation ---------------------- */

static Vector tMixProb = NULL;         /* precomputed tied mix probs */
static Vector tMixLogProb = NULL;      /* precomputed tied mix log probs */

static LogReal DOutP(Vector x, int vecSize, OutProb *optr)
{
   int i;
   Real sum,xmm;
   
   sum = optr ->gConst;
   for (i=1;i<=vecSize;i++) {
      xmm=x[i]-optr->mean[i];
      sum += xmm*xmm/optr->covariance.var[i];
   }
   return -0.5*sum;
}

static LogReal FOutP(Vector x, int vecSize, OutProb *optr)
{
   Real xmm[100], sum;
   int i,j;
   Vector crow;
   
   for (i=1;i<=vecSize;i++)
      xmm[i] = x[i]-optr->mean[i];
   sum = 0.0;
   for (i=1;i<vecSize;i++) {
      crow = optr->covariance.inv[i];
      for (j=i+1;j<=vecSize;j++)
         sum += xmm[i]*xmm[j]*crow[j];
    }
   sum *= 2;
   sum += optr ->gConst;
   for (i=1;i<=vecSize;i++)
      sum += xmm[i]*xmm[i]*optr->covariance.inv[i][i];
   return -0.5*sum;
}

/* EXPORT-> PreCompute: precomputes prob (log) of vector x for mix set */
void PreCompute(Vector x)
{
   OutProb *optr;
   LogReal y;
   int m;
   
   if (tMixProb==NULL)
      tMixProb = CreateVector(numTMix);
   if (tMixLogProb==NULL)
      tMixLogProb = CreateVector(numTMix);
   optr = tMix;
   for (m=1;m<=numTMix;m++){
      ++optr;
      y = (ckTMix==DIAGC) ? DOutP(x,vsTMix,optr) : FOutP(x,vsTMix,optr);
      tMixProb[m] = (y>MINEARG) ? exp(y) : 0.0;
      tMixLogProb[m] = y;
   }
}

/* EXPORT-> OutP: Returns probability (log) of vector x for given state */
LogReal OutP(Vector x, HMMDef *hmm, int state)
{
   int m;
   LogReal bx,px;
   OutProb *optr;
   Real y;
   
   optr = (hmm->stateinfo+state)->outp+1;
   
   if (hmm->numMixtures == 1)    /* Single Mixture Case */
      return (hmm->ckind==DIAGC) 
         ? DOutP(x,hmm->vecSize,optr) : FOutP(x,hmm->vecSize,optr);
         
   if (tMixProb != NULL) { /* Must be tied mixtures */
      y = 0.0;                      /* Tied Mixture Case */
      for (m=1; m<=hmm->numMixtures; m++) { 
         if (optr->weight>MINMIX)   
            y += optr->weight*tMixProb[m];
         optr++;
      }
      bx = (y>MINLARG)?log(y):LZERO;
   } else {
      bx = LZERO;                   /* Multi Mixture Case */
      for (m=1; m<=hmm->numMixtures; m++) {
         if (optr->weight>MINMIX) {   
            px=(hmm->ckind==DIAGC) 
               ? DOutP(x,hmm->vecSize,optr) : FOutP(x,hmm->vecSize,optr);
            bx = LAdd(bx,log(optr->weight)+px);
         }
         optr++;
      }
   }
   return bx;
}


/* EXPORT-> MOutP: Returns prob ([DR added] log) of vector x for given state & given mixture */
LogReal MOutP(Vector x, HMMDef *hmm, int state,int mixture)
{
   OutProb *optr;
   
   if (tMixLogProb!=NULL)
      return tMixLogProb[mixture];
   optr = (hmm->stateinfo+state)->outp+mixture; 
   return (hmm->ckind==DIAGC) 
          ? DOutP(x,hmm->vecSize,optr) : FOutP(x,hmm->vecSize,optr);
}
         
/* EXPORT->FixDiagGConst: Sets gConst for given state in DIAGC case */
void FixDiagGConst(OutProb *optr)
{
   Real sum;
   int i,n;
   LogReal z;
   
   n=VectorSize(optr->covariance.var);
   sum = n*log(TPI);
   for (i=1; i<=n; i++){
      z = log(optr->covariance.var[i]);
      if (errno==EDOM || errno==ERANGE){
         perror("FixDiagGConst");   
         errno = 0;  /* may not be legal on some machines */
      }
      sum += z;
   }
   optr->gConst = sum;
}

/* EXPORT->FixFullGConst: Sets gConst for given state in FULLC case */
/*    note that ldet is passed as a parameter rather than computed  */
/*    here since it will often be known as a byproduct of inverting */
/*    the covariance matrix   */
void FixFullGConst(OutProb *optr, LogReal ldet)
{
   optr->gConst = NumRows(optr->covariance.inv)*log(TPI) + ldet;
}

/* EXPORT->FixGConsts: Sets all gConsts in give hmm */
void FixGConsts(HMMDef *hmm)
{
   int n,m;
   StateInfo *sptr;
   OutProb *optr;
   
   sptr = hmm->stateinfo+1;
   for (n=2; n<hmm->numStates; n++) {
      sptr++;
      optr = sptr->outp;
      for (m=1; m<=hmm->numMixtures; m++) {
         optr++;
         if (optr->weight > MINMIX)
            if (hmm->ckind == DIAGC)
               FixDiagGConst(optr);
            else
               FixFullGConst(optr,-CovDet(optr->covariance.inv));
      }
   }
}

/* ----------------------- Enum Conversions ------------------------------ */

/* EXPORT-> DurKind2Str: Return string representation of enum DurKind */
char *DurKind2Str(DurKind dkind, char *buf)
{
   static char *durmap[] = {"NULLD","POISSOND","GAMMAD"};
   return strcpy(buf,durmap[dkind]);
}


/* EXPORT-> CovKind2Str: Return string representation of enum CovKind */
char *CovKind2Str(CovKind ckind, char *buf)
{
   static char *covmap[] = {"DIAGC","FULLC"};
   return strcpy(buf,covmap[ckind]);
}

/* ------------------------- End of HModel.c --------------------------- */








