/* ----------------------------------------------------------- */
/*                   HTK  -  HMM Toolkit                       */
/* ----------------------------------------------------------- */
/*         Copyright: Steve Young      October 1989            */
/*         Cambridge University Engineering Department         */
/*         Trumpington Street, Cambridge, CB2 1PZ              */
/*                                                             */
/* Permission is hereby granted to copy, use and modify this   */
/* software provided that this banner notice is not removed    */
/*                                                             */
/* ----------------------------------------------------------- */
/*         File: HModel.h  HMM Model Definition Data Type      */
/* ----------------------------------------------------------- */

/* 
   The following types define the in-memory representation of a 
   HMM, for the corresponding external text file format see HModel.c.
   Note that all probabilities are stored as logs internally.
*/

#define MINMIX  1.0E-3         /* Min usable mixture weight */

enum _CovKind {DIAGC, FULLC};
typedef enum _CovKind CovKind;

enum _DurKind {NULLD, POISSOND, GAMMAD};
typedef enum _DurKind DurKind;

enum _TieStatus {FreeT,MasterT,SlaveT};
typedef enum _TieStatus TieStatus;

typedef union {
   Vector var;    /* if ckind == DIAGC */
   Matrix inv;    /* if ckind == FULLC */
} Covariance;

typedef struct {           /* 1 of these per mixture per state */
   float weight;
   Vector mean;            /* Output means */
   Covariance covariance;  /* Output covariances */
   float gConst;           /* Precomputed component of b(x) */
   void *aux;              /* Temp work space */
} OutProb;

typedef struct {           /* 1 of these per state */
   OutProb *outp;          /* -> array[1..numMix] of Outprob */
   char *label;            /* Label if any */
   float dur1,dur2;        /* Duration parms, if any */
} StateInfo;

typedef struct {
   int numStates;          /* includes entry and exit states */
   int numMixtures;        /* number of Gaussian mixtures */
   int vecSize;            /* dimension of observation vectors */
   SampKind pkind;         /* kind of obs vector components */
   StateInfo *stateinfo;   /* -> array[2..numStates-1] of StateInfo */  
   Matrix transP;          /* transition matrix (logs) */
   CovKind ckind;          /* kind of covariance */
   DurKind dkind;          /* kind of duration model */
   char *tieMaster;        /* name of file holding tied mixture defs */
   TieStatus tstat;        /* for tied mixtures */
} HMMDef;

void HCreate(int numStates, int numMixtures, int vecSize, SampKind pkind,
             CovKind ckind, DurKind dkind, TieStatus tstat, HMMDef *hmm);
/*
   Create a data structure in memory to store a new hmm definition
*/

void HOpen(char *fname, HMMDef *hmm);
void HOpen2(char *fname, HMMDef *hmm);
/*
   Read model definition from file fname and create equivalent representation
   in memory.  Format of external definition is defined in .c file. HOpen2
   is same as HOpen except that it does not convert transition matrix to
   log real form.
*/

void HSave(char *fname, HMMDef *hmm);
void HSave2(char *fname, HMMDef *hmm);
/*
   Save the model hmm in file fname. HSave2 is same as HOpen except
   that it does not assume trans matrix is in log form.
*/

void HClose(HMMDef *hmm);
/*
   Free memory allocated to hmm
*/

void TMClose(void);
/* 
   Release global storage allocated for a set of tied mixture HMM's
*/

enum HPFLAGS {HPDETAILS = 1, HPTRANS = 2, HPLABELS = 4, 
              HPMEANS = 8, HPVARS = 16, HPDURS = 32, LOGTRANS = 64};
#define HPALL 0xff
void HPrint(HMMDef *hmm, int flags, ...);
/*
   Print info on hmm to stdout.  Flags select info to print, if any
   state info is wanted then flags arg should be followed by ids of first 
   and last state to print eg HPrint(hmm,HPMEANS|HPVARS,4,6) will print 
   the means and variances for states 4 5 6.  LOGTRANS must be
   specified if transition matrix is in log form.
*/

LogReal OutP(Vector x, HMMDef *hmm, int state);
LogReal MOutP(Vector x, HMMDef *hmm, int state, int mixture);
/*
   Returns (log) probability of vector x given output distribution
   parameters in model hmm for given state. MOutP gives individual
   mixture prob.
*/

void PreCompute(Vector x);
/*
   Precomputes log prob of vector x for stored tied mixture set 
*/

void FixDiagGConst(OutProb *optr);
/*
   Sets gConst for given state in DIAGC case
*/

void FixFullGConst(OutProb *optr, LogReal ldet);
/*
   Sets gConst for given state in FULLC case.  ldet is the log of
   the determinant of the covariance matrix. 
*/

void FixGConsts(HMMDef *hmm);
/*
   Sets the gConst values in all states of given hmm
*/

char *DurKind2Str(DurKind dkind, char *buf);
char *CovKind2Str(CovKind ckind, char *buf);
/*
   Useful utility routines to convert enum types to strings.
   In each case, string is stored in *buf and pointer to buf
   is returned
*/

/* ------------------------- End of HModel.h --------------------------- */
