/* ----------------------------------------------------------- */
/*                   HTK  -  HMM Toolkit                       */
/* ----------------------------------------------------------- */
/*         Copyright: Steve Young      October 1989            */
/*         Cambridge University Engineering Department         */
/*         Trumpington Street, Cambridge, CB2 1PZ              */
/*                                                             */
/* Permission is hereby granted to copy, use and modify this   */
/* software provided that this banner notice is not removed    */
/*                                                             */
/* ----------------------------------------------------------- */
/*         File: HParse.h:   Network Defn Parser/Builder       */
/* ----------------------------------------------------------- */

/*
   This module provides a single routine which parses the set
   of rewrite rules in the supplied text file and builds an
   equivalent representation in memory.
   
   The syntax rules for the textual definition of the net are
   given in HParse.c
   
   The internal representation is a set of nodes containing the
   name of the associated model, an array of pointers to all 
   successor nodes and an array of pointers to all predecessors.
   In addition, all nodes are joined together into a single linked
   list to simplify processing.
   
   As with HMM's themselves, the network contains a single
   entry node called ENTRY and a single exit node called EXIT.
*/

typedef struct _Node *Link;
typedef Link LinkSet[];

typedef struct _Node{
   LabId modelName;
   short numSucc;    /* number of successor nodes */
   short numSuccMax; /* num succ node slots allocated */
   short numPred;    /* number of predecessor nodes */
   short numPredMax; /* num pred node slots allocated */
   LinkSet *succ;    /* -> array[0..numSuccMax-1]of Link */
   LinkSet *pred;    /* -> array[0..numPredMax-1]of Link */
   Link chain;       /* simple linked list of all nodes */  
   void *user;       /* for attaching user defined data */
} Node;

typedef struct {
   Link entryNode;
   Link exitNode;
   Link chain;
} Network;


Network ParseNetwork(char *fname);
/*
   Returns the network defined within the file fname
*/

void FreeNetwork(Network *network);
/*
   Frees all storage allocated for given network
*/

void PrintNetwork(Network *network);
/*
   Prints the name of each node in the network and its 
   successors/predecessors
*/

Link CreateNode(LabId name, Link *chain, int numSuccMax, int numPredMax);
/*
   CreateNode: create node and link it into given chain 
*/

void FreeNode(Link p);
/*
   FreeNode: release storage allocated to p 
*/

void ShrinkNode(Link p);
/*
   ShrinkNode:  reduces sizes of LinkSets to minimum 
*/

int NotJoined(Link a, Link b);
/*
   NotJoined: returns true if a and b are not linked 
*/

void JoinNodes(Link a, Link b);
/*
   JoinNodes: join node a -> node b 
*/

void RemPredLink(Link pred, Link p);
/*
   RemPredLink: remove reference to pred in node p 
*/

void RemSuccLink(Link succ, Link p);
/*
   RemSuccLink: remove reference to succ in node p 
*/




/* ------------------------ End of HParse.h ------------------------- */

