/* ----------------------------------------------------------- */
/*                   HTK  -  HMM Toolkit                       */
/* ----------------------------------------------------------- */
/*         Copyright: Steve Young      October 1989            */
/*         Last Update:                September 1990          */
/*                                                             */
/*         Cambridge University Engineering Department         */
/*         Trumpington Street, Cambridge, CB2 1PZ              */
/*                                                             */
/* Permission is hereby granted to copy, use and modify this   */
/* software provided that this banner notice is not removed    */
/*                                                             */
/* ----------------------------------------------------------- */
/*      File: HSigP.c:   Signal Processing Routines            */
/* ----------------------------------------------------------- */

#include <stdio.h>         /* Standard Libraries */
#include <stdlib.h>
#include <math.h>

#include "HShell.h"        /* HTK Libraries */
#include "HMath.h"
#include "HSigP.h"

#if defined MPW
   #pragma segment hmods
#endif

/*
   This module provides a set of basic speech signal processing
   routines, it is used primarily by HCode to convert raw speech
   data into parameterised form
*/

/* --------------- Windowing and PreEmphasis ---------------------*/

static int hamWinSize=0;         /* Size of current Hamming window */
static Real hamWin[MAXFRAMESIZE];   /* Current Hamming window */
static int cepWinL=0;         /* Size of current cepstral weight window */
static Real cepWin[MAXFEATURES]; /* Current cepstral weight window */

void GenHamWindow (int frameSize)
{
   int i;
   Real a;
   
   a = TPI / (frameSize - 1);
   for (i=0;i<frameSize;i++)
      hamWin[i] = 0.54 - 0.46 * cos(a*i);
   hamWinSize = frameSize;
}

/* EXPORT->Ham: Apply Hamming Window to Speech frame s */
void Ham (Vector s)
{
   int i,frameSize;
   
   frameSize=VectorSize(s);
   if (hamWinSize != frameSize)
      GenHamWindow(frameSize);
   for (i=0;i<frameSize;i++)
      s[i+1] *= hamWin[i];
}

void GenCepWin (int cepLiftering)
{
   int i;
   Real a, Lby2;
   
   a = PI/cepLiftering;
   Lby2 = cepLiftering/2.0;
   for (i=1;i<=MAXFEATURES;i++)
      cepWin[i-1] = 1.0 + Lby2*sin(i * a);
   cepWinL = cepLiftering;
}  

/* EXPORT->WeightCepstrum: Apply cepstral weighting to c */
void WeightCepstrum (Vector c, int start, int count, int cepLiftering)
{
   int i,j;
   
   if (cepWinL != cepLiftering)
         GenCepWin(cepLiftering);
   j = start;
   for (i=0;i<count;i++)
      c[j++] *= cepWin[i];
}

/* EXPORT->UnWeightCepstrum: Undo cepstral weighting of c */
void UnWeightCepstrum(Vector c, int start, int count, int cepLiftering)
{
   int i,j;
   
   if (cepWinL != cepLiftering)
         GenCepWin(cepLiftering);
   j = start;
   for (i=0;i<count;i++)
      c[j++] /= cepWin[i];
}

/* EXPORT->PreEmphasise: pre-emphasis signal in s */
void PreEmphasise (Vector s, float k)
{
   int i;
   Real preE;
   
   preE = k;
   for (i=VectorSize(s);i>=2;i--)
      s[i] -= s[i-1]*preE;
   s[1] *= 1.0-preE;
}

/* ------------ LPC and LP Cepstral Analysis ---------- */

/* AutoCorrelate: store autocorrelation coef 0 to p in r */
void AutoCorrelate (Vector s, Real r[], int p, int frameSize)
{
   Real sum;
   int   i,j;

   for (i=0;i<=p;i++) {
      sum = 0.0;
      for (j=1;j<=frameSize-i;j++)
         sum += s[j]*s[j+i];
      r[i] = sum;
   }
}

/* EXPORT->Wave2LPC: Calculate LPCoef in a & RefC in k */
void Wave2LPC (Vector s, Vector a, Vector k, float *re, float *te)
{
   Real thisA[MAXFEATURES+1]; /* Current LP filter coefficients */
   Real newA[MAXFEATURES+1];  /* New LP filter coefficients */
   Real r[MAXFEATURES+1];     /* AutoCorrelation Sequence */
   Real ki;                   /* Current Reflection Coefficient */
   Real E;                    /* Prediction Error */
   int   i,j,p,frameSize;

   if (a!=NULL) p=VectorSize(a); else
      if (k!=NULL) p=VectorSize(k); else
         HError(2,"Null a and k vectors in WaveToLPC");
   frameSize=VectorSize(s);
   AutoCorrelate(s,r,p,frameSize);
   *te = E = r[0];
   for (i=1;i<=p;i++) {
      ki = r[i];              /* Calc next reflection coef */
      for (j=1;j<i;j++)
         ki = ki + thisA[j] * r[i - j];
      ki = ki / E;   
      if (k!=NULL) k[i] = ki;
      E *= 1 - ki*ki;         /* Update Error */
      newA[i] = -ki;          /* Calc new filter coef */
      for (j=1;j<i;j++)
         newA[j] = thisA[j] - ki * thisA[i - j];
      for (j=1;j<=i;j++)   
         thisA[j] = newA[j];
   }
   if (a!=NULL)
      for (i=1;i<=p;i++)
         a[i]=thisA[i];
   *re = E;
}

/* EXPORT->LPC2RefC: transfer from filter to ref coef */
void LPC2RefC(Vector a, Vector k)
{
   Real thisA[MAXFEATURES+1]; /* Current LP filter coefficients */
   Real newA[MAXFEATURES+1];  /* New LP filter coefficients */
   int i,j,p;
   Real ki,x;
   
   p=VectorSize(a);
   for (i=1;i<=p;i++)  thisA[i]=a[i];
   for (i=p;i>=1;i--)  { 
      ki = -thisA[i];
      k[i] = ki;
      x = 1 - ki*ki;
      for (j=1;j<i;j++) 
         newA[j] = (thisA[j] + ki * thisA[i - j]) / x;
      for (j=1;j<i;j++) 
         thisA[j] = newA[j];
   }

}

/* EXPORT->RefC2LPC: transfer from ref coef to filter */
void RefC2LPC (Vector k, Vector a)
{
   Real thisA[MAXFEATURES+1]; /* Current LP filter coefficients */
   Real newA[MAXFEATURES+1];  /* New LP filter coefficients */
   int i,j,p;
   Real ki;
   
   p=VectorSize(k);
   for (i=1;i<=p;i++) { 
      ki = k[i];
      newA[i] = -ki;
      for (j=1;j<i;j++) 
         newA[j] = thisA[j] - ki * thisA[i - j];
      for (j=1;j<=i;j++)   
         thisA[j] = newA[j];
   };
   for (i=1;i<=p;i++)  a[i]=thisA[i];
}


/* EXPORT->LPC2Cepstrum: transfer from lpc to cepstral coef */
void LPC2Cepstrum (Vector a, Vector c)
{
   int i,n,p;
   Real sum;
   
   p=VectorSize(a);
   for (n=1;n<=p;n++)  { 
      sum = 0.0;
      for (i=1;i<n;i++) 
         sum = sum + (n - i) * a[i] * c[n - i];
      c[n] = -(a[n] + sum / n);
   }

}

/* ----------------------- FFT's --------------------------- */

/* EXPORT-> FVec2Spectrum: cvt feature vector f to a spectrum */
void FVec2Spectrum (Vector f, Vector s)
{
   int i,p,n;
   
   p=VectorSize(f); n=VectorSize(s);
   s[1] = 1.0;
   for (i=1;i<=p;i++) 
      s[i+1] = f[i];
   for (i=p+2;i<=n;i++) 
      s[i] = 0.0;
   Realft(s);

}

/* EXPORT-> FFT: apply fft/invfft to complex s */
void FFT(Vector s, int invert)
{
   int ii,jj,n,nn,mmax,m,j,istep,i;
   Real wtemp,wr,wpr,wpi,wi,theta;
   Real tempr,tempi,x;
   
   n=VectorSize(s);
   nn=n / 2; j = 1;
   for (ii=1;ii<=nn;ii++) {
      i = 2 * ii - 1;
      if (j>i) {
         tempr = s[j]; tempi = s[j + 1];
         s[j] = s[i];  s[j + 1] = s[i + 1];
         s[i] = tempr; s[i + 1] = tempi;
      }
      m = n / 2;
      while (m >= 2  && j > m) {
         j -= m; m /= 2;
      }
      j += m;
   };
   mmax = 2;
   while (n > mmax) {
      istep = 2 * mmax; theta = 6.28318530717959 / mmax;
      if (invert) theta = -theta;
      x = sin(0.5 * theta);
      wpr = -2.0 * x * x; 
      wpi = sin(theta); wr = 1.0; wi = 0.0;
      for (ii=1; ii<=mmax/2; ii++) {
         m = 2 * ii - 1;
         for (jj = 0; jj<=(n - m) / istep;jj++) {
            i = m + jj * istep;
            j = i + mmax;
            tempr = wr * s[j] - wi * s[j + 1];
            tempi = wr * s[j + 1] + wi * s[j];
            s[j] = s[i] - tempr;
            s[j + 1] = s[i + 1] - tempi;
            s[i] = s[i] + tempr;
            s[i + 1] = s[i + 1] + tempi;
         }
         wtemp = wr;
         wr = wr * wpr - wi * wpi + wr;
         wi = wi * wpr + wtemp * wpi + wi;
      }
      mmax = istep;
   }
   if (invert)
      for (i = 1;i<=n;i++) 
         s[i] = s[i] / nn;
   
}

/* EXPORT-> Realft: apply fft to real s */
void Realft (Vector s)
{
   Real wr, wi, wpr, wpi, wtemp, theta, x;
   int n,i, i1, i2, i3, i4;
   Real c1, c2, h1r, h1i, h2r, h2i, wrs, wis;

   n=VectorSize(s) / 2;
   theta = 6.28318530717959 / (2.0 * n);
   c1 = 0.5; c2 = -0.5;
   FFT(s, 0);
   x = sin(0.5 * theta);
   wpr = -2.0 * x * x;
   wpi = sin(theta); wr = 1.0 + wpr; wi = wpi;
   for (i=2; i<n/2; i++) {
      i1 = i + i - 1;      i2 = i1 + 1;
      i3 = n + n + 3 - i2; i4 = i3 + 1;
      wrs = wr; wis = wi;
      h1r = c1 * (s[i1] + s[i3]);
      h1i = c1 * (s[i2] - s[i4]);
      h2r = -c2 * (s[i2] + s[i4]);
      h2i = c2 * (s[i1] - s[i3]);
      s[i1] = h1r + wrs * h2r - wis * h2i;
      s[i2] = h1i + wrs * h2i + wis * h2r;
      s[i3] = h1r - wrs * h2r + wis * h2i;
      s[i4] = -h1i + wrs * h2i + wis * h2r;
      wtemp = wr;
      wr = wr * wpr - wi * wpi + wr;
      wi = wi * wpr + wtemp * wpi + wi;
   }
   h1r = s[1];
   s[1] = h1r + s[2];
   s[2] = 0.0;
}
   
/* EXPORT-> SpecModulus: store modulus of s in m */
void SpecModulus(Vector s, Vector m)
{
   int i,j;
   Real x,y;
   
   for (i=1;i<=VectorSize(s)/2;i++) {
      j=i+i; x=s[j-1]; y=s[j];
      m[i]=sqrt(x*x + y*y);
   }
}

/* EXPORT-> SpecLogModulus: store log modulus of s in m */
void SpecLogModulus(Vector s, Vector m, int invert)
{
   int i,j;
   Real x,y;
   
   for (i=1;i<=VectorSize(s)/2;i++) {
      j=i+i; x=s[j-1]; y=s[j];
      x=0.5*log(x*x + y*y);
      m[i] = invert ? -x : x;
   }
}

/* EXPORT-> SpecPhase: store phase of s in m */
void SpecPhase(Vector s, Vector m)
{
   int i,j;
   Real ph,re,im;
   
   for (i=1;i<=VectorSize(s)/2;i++) {
      j=i+i;
      re=s[j-1]; im=s[j];
      if (re==0.0) 
         ph = (im>=0.0) ? PI/2.0 : -PI/2.0;
      else {
         ph=atan(im/re);
         if (ph<0.0 && re<0.0)
            ph += PI;
         else if (ph>0.0 && im<0.0)
            ph -= PI;
      }
      m[i]=ph;
   }
}


/* ------------------------ End of HSigP.c ------------------------- */
