/* ----------------------------------------------------------- */
/*                   HTK  -  HMM Toolkit                       */
/* ----------------------------------------------------------- */
/*         Copyright: Steve Young      October 1989            */
/*         Last Update:                September 1990          */
/*                                                             */
/*         Cambridge University Engineering Department         */
/*         Trumpington Street, Cambridge, CB2 1PZ              */
/*                                                             */
/* Permission is hereby granted to copy, use and modify this   */
/* software provided that this banner notice is not removed    */
/*                                                             */
/* ----------------------------------------------------------- */
/*         File: HSpIO.c:   Speech Data File Input/Output      */
/* ----------------------------------------------------------- */


#include <stdio.h>         /* Standard Libraries */
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <unistd.h>

#include "HShell.h"
#include "HMath.h"
#include "HSigP.h"
#include "HSpIO.h"

#ifdef MPW
   #pragma segment hmods
#endif

#define CREATECHUNKSIZE 10000


/*
THIS MODULE ASSUMES THAT short'S ARE 2 BYTES AND long'S ARE 4 BYTES

This module provides an interface to speech data files both raw
and parameterised stored in a variety of formats.  It is assumed
that all files consist of a (possibly null) header followed by 
a sequence of samples.  The latter may be vector or scalar.

The type FileFormat lists the currently supported formats 
which are:

   NOHEAD      - no head at all
   ALIEN       - alien ie unknown
   SPED        - sjy's original Mac file format
   HTK         - the preferred HMM Toolkit format
   TIMIT       - the TIMIT CD-ROM database
   NIST        - the NIST Resource Management database
   SCRIBE      - the UK SCRIBE CD-ROM database
   
Each data file can have the following sample kinds

   WAVEFORM    - scalar samples (usually raw speech data)
   LPC         - linear prediction coefficients
   LPREFC      - linear prediction reflection coefficients
   LPCEPSTRA   - lp derived cepstral coefficients
   LPDELCEP    - lp derived cepstral coefficients + delta coef
   IREFC       - lp ref coefficients coded as 16bit shorts
   MFCC        - mel-frequency cepstral coefficients
   FBANK       - log filterbank channel energies
   ANON        - none of the above
 
For each of the above (except WAVEFORM), the constant HASENERGY may
be added to the SampKind to indicate that the log energy has been
appended, also the constant HASPITCH can be added to indicate that
an f0 estimate has been appended.

Note, however, that apart from the scalar/vector distinction, none
the HMM tools apart from HCode depends on the sample kind.

Formats TIMIT, NIST and SCRIBE are expected to be common
across machines and can only be used for WAVEFORM data.
No attempt has been made to make parameterised data portable.
The SpOpen routine automatically byte swaps these WAVEFORM
files as necessary.

The basic algorithm for determining which file format to use is
as follows

   a) if explicitly defined by SetFormat then use that else
   b) if HFORMAT is defined in environment then use that else
   c) format is HTK

If the format is ALIEN, the following environment variables 
must be set

   HDSIZE      - size of header in bytes
   SAMPSIZE    - size of a sample in bytes
   SAMPPERIOD  - sample period in 100ns units
   SAMPKIND    - one of the allowed sample kinds

The header of an alien file is simply skipped, the needed format
data being provided by the above.  The number of samples is computed
from the file size.

The NOHEAD format is treated the same as ALIEN except that a header
size of zero is assumed. 

When a data file is opened with SpOpen, the entire contents of the 
file are read into memory, byte swapped if necessary, and the file
itself is closed.  The DataFile structure holds all information
relevant to open data files.  No client should modify this structure.
External to this module, all data files are sequences of samples
indexed from 0 to nSamples-1.

If the source file format is IREFC it will be automatically converted
to LPREFC.  The IREFC format is supported simply to conserve external
data storage.

If the environment variable HCOERCE is set then the source file 
will be converted (where possible) to the sample kind specified by
HCOERCE.  If this transformation involves the 
mapping LPC -> LPCEPSTRA then HLIFTER may be set to apply
cepstral liftering (see the -l option in HCode).  If this
transformation involves the mapping LPCEPSTRA -> LPDELCEP then
the delta window may be set by HDELTA (see the -q option in HCode).

Note that default buffering is used for all i/o.  Appropriate
calls to setvbuf within SpOpen and SpSave may improve performance 
considerably depending on the local environment.

*/

/* ------------------- Header Structures --------------------- */

#define SPEDHDSIZE 116           /* SPED File Header */
typedef struct {  
   long nSamples;
   short sampSize;
   long sampPeriod;
   char dummy[40];
   unsigned char sampKind;
   /* more irrelevant stuff */
} SPEDhdr;

typedef struct {              /* TIMIT File Header */
   short hdrSize;
   short version;
   short numChannels;
   short sampRate;
   long  nSamples;
} TIMIThdr;

typedef struct {              /* HTK File Header */
   long nSamples;
   long sampPeriod;
   short sampSize;
   short sampKind;
} HTKhdr;

/* ------------------- Detect Byte Ordering ------------------ */

extern char *getenv();

/* VAXOrder: returns true if machine has VAX ordered bytes */
static Boolean VAXOrder(void)
{
     short x, *px;
     unsigned char *pc;

     px = &x;
     pc = (unsigned char *) px;
     *pc = 1; *(pc+1) = 0;         /* store bytes 1 0 */
     return x==1;          /* does it read back as 1? */
}

/* MustSwap: returns true if given format must be byte swapped
   on this machine */
static Boolean MustSwap(FileFormat ff)
{
   char *bos;
   
   bos = getenv("HBORDER");
   if (bos != NULL) {         /* force required order */
      if (strcmp(bos,"VAX") == 0)
         return !VAXOrder();
      else
         return VAXOrder();
   } else
      switch(ff) {
         case TIMIT: return !VAXOrder();
         case NIST:  return !VAXOrder();
         case SCRIBE: return FALSE;
         default: return FALSE;
      }
}

/* ------------------- Setting File Format ------------------- */

static int formatSet = 0;          /* Flag indicating format set */
static FileFormat format = HTK;
static FileFormat hcoerce = ANON;   /* target of coercion if any */

static char *fmtmap[] = {
   "NOHEAD", "SPED", "HTK", "TIMIT", "NIST", 
   "SCRIBE", "ALIEN"
};
static char *smkmap[] = {"WAVEFORM", 
                        "LPC", "LPREFC", "LPCEPSTRA",
                        "LPDELCEP", "IREFC", "MFCC", 
                        "FBANK","ANON"};

/* EXPORT-> SetFormat: set format and override environ setting */
void SetFormat(FileFormat fmat)
{
   format = fmat; 
   formatSet = 1;
}

/* --------------- Enum to String Conversions ---------------- */

char *Format2Str(FileFormat format, char *buf)
{
   return strcpy(buf,fmtmap[format]);
}

char *SampKind2Str(SampKind kind, char *buf)
{
   strcpy(buf,smkmap[BaseSampKind(kind)]);
   if (HasEnergy(kind))
      strcat(buf,"_E");
   if (HasPitch(kind))
      strcat(buf,"_P");
   return buf;
}

/* ------------- Communication with Environment -------------- */

/* EXPORT->Str2Format: Convert string representation to enum FileFormat */
FileFormat Str2Format(char *fmt)
{
   FileFormat i = -1;
   char *s;
 
   do {
      s=fmtmap[++i]; 
      if (strcmp(fmt,s) == 0) break;
   } while (strcmp("ALIEN",s)!=0);
   return i;
}

/* EXPORT->Str2SampKind: Convert string representation to enum SampKind */
SampKind Str2SampKind(char *smk)
{
   SampKind i = -1;
   char *s,buf[255];
   Boolean hasE,hasP;
   
   hasE=hasP=FALSE;
   strcpy(buf,smk);
   while ((s=strrchr(buf,'_')) != NULL) {
      switch(*(s+1)){
         case 'E': hasE = TRUE; break;
         case 'P': hasP = TRUE; break;
         default: HError(2,"unknown sampkind qualifier %s",smk);
      }
      *s = '\0';
   }
 
   do {
      s=smkmap[++i]; 
      if (strcmp(buf,s) == 0) break;
   } while (strcmp("ANON",s)!=0);
   if (i != ANON){
      if (hasE) i |= HASENERGY;
      if (hasP) i |= HASPITCH;
   }
   return i;
}
   
/* EXPORT-> GetFormat: get format from SetFormat or environ or return deflt */
FileFormat GetFormat()
{
   char *fmt;
   
   if (formatSet) return format;
   if ((fmt = getenv("HFORMAT")) == NULL) return format;
   return Str2Format(fmt);
}

/* GetSampKind: get sample kind from environ or return default */
static SampKind GetSampKind(char * evName)
{
   char *smk;
   
   if ((smk = getenv(evName)) == NULL) return ANON;
   return Str2SampKind(smk);
}

/* GetEnvValue: Get an integer value from environment */
static long GetEnvValue(char *s)
{
   char *intstr;
   
   if ((intstr=getenv(s)) == NULL)
      HError(2,"Need Integer Environment Var %s",s);
   if (strlen(intstr)>0 && isdigit(intstr[0]) )
      return atol(intstr);
   else
      HError(2,"%s must be an integer",s);
}

/* --------------- Format Dependent Routines ----------------- */

static void SwapShort(short *p)
{
   char temp,*q;
   
   q = (char*) p;
   temp = *q; *q = *(q+1); *(q+1) = temp;
}

static void SwapLong(long *p)
{
   char temp,*q;
   
   q = (char*) p;
   temp = *q; *q = *(q+3); *(q+3) = temp;
   temp = *(q+1); *(q+1) = *(q+2); *(q+2) = temp;
}

            /* ---- SPED ---- */

static void GetSPEDHeaderInfo(FILE *dataStream, DataFile *data)
{
   SPEDhdr hdr;
   int n = sizeof hdr;
   
   rewind(dataStream);
   if (fread(&hdr, 1, n, dataStream) != n)
      HError(2,"Cannot read SPED format header");
   if (hdr.sampSize <= 0 || hdr.sampSize > 256 || hdr.nSamples < 0 ||
       hdr.sampKind < 0 || hdr.sampKind >= ANON || hdr.sampPeriod <= 0 ||
       hdr.sampPeriod > 1000000)
       HError(2,"Bad Numbers in SPED format header");
   data->nSamples   = hdr.nSamples;
   data->sampSize   = hdr.sampSize;
   data->sampKind   = hdr.sampKind;
   data->sampPeriod = hdr.sampPeriod*10;
   data->hdrSize = SPEDHDSIZE;
}

            /* ---- HTK ---- */

static void GetHTKHeaderInfo(FILE *dataStream, DataFile *data)
{
   HTKhdr hdr;
   int n = sizeof hdr;
   
   rewind(dataStream);
   if (fread(&hdr, 1, n, dataStream) != n)
      HError(2,"Cannot read HTK format header");
   if (hdr.sampSize <= 0 || hdr.sampSize > 256 || hdr.nSamples < 0 ||
       hdr.sampKind < 0 || BaseSampKind(hdr.sampKind) >= ANON || 
       hdr.sampPeriod <= 0 || hdr.sampPeriod > 1000000)
       HError(2,"Bad Numbers in HTK format header");
   data->nSamples   = hdr.nSamples;
   data->sampSize   = hdr.sampSize;
   data->sampKind   = hdr.sampKind;
   data->sampPeriod = hdr.sampPeriod;
   data->hdrSize = n;
}

static void PutHTKHeaderInfo(FILE *dataStream, DataFile *data)
{
   HTKhdr hdr;
   int n = sizeof hdr;
   
   hdr.nSamples   = data->nSamples;
   hdr.sampSize   = data->sampSize;
   hdr.sampKind   = data->sampKind;
   hdr.sampPeriod = data->sampPeriod;
   rewind(dataStream);
   if (fwrite(&hdr, 1, n, dataStream) != n)
      HError(2,"Cannot write HTK format header");
}

            /* ---- TIMIT ---- */

static void GetTIMITHeaderInfo(FILE *dataStream, DataFile *data)
{
   TIMIThdr hdr;
   int n = sizeof hdr;
   Boolean bswap; 
   
   bswap = !VAXOrder();       /* TIMIT is VAX ordered */
   rewind(dataStream);
   if (fread(&hdr, 1, n, dataStream) != n)
      HError(2,"Cannot read TIMIT format header");
   if (bswap) {
      SwapShort(&hdr.hdrSize);
      SwapShort(&hdr.version);         /* Not Used */
      SwapShort(&hdr.numChannels);     /* Not Used */
      SwapShort(&hdr.sampRate);
      SwapLong(&hdr.nSamples);
   }
   if (hdr.nSamples < 0 || hdr.sampRate < 0 || hdr.numChannels < 0 ||
       hdr.numChannels > 8 )
       HError(2,"Bad Numbers in TIMIT format header");
   data->nSamples   = hdr.nSamples;
   data->sampSize   = 2;
   data->sampKind   = WAVEFORM;
   data->sampPeriod = hdr.sampRate*10 / 4;
   data->hdrSize = n;
}

            /* ---- NIST ---- */
            
static char cNIST;

static char * GetNISTToken(FILE *f,char *buf)
{
   int i=0;
   
   while (isspace(cNIST)) cNIST=fgetc(f);
   do {
      buf[i++] = cNIST; cNIST=fgetc(f);
   } while(!isspace(cNIST));
   buf[i] = '\0';
   return buf;
}

static void NISTSkipLine(FILE *f)
{
   while (cNIST != '\012') cNIST=fgetc(f);
   cNIST=fgetc(f);
}

static int GetNISTIVal(FILE *f)
{
   char buf[100];
   
   if (strcmp(GetNISTToken(f,buf),"-i") != 0)
      HError(2,"NIST type indicator -i expected");
   return atoi(GetNISTToken(f,buf));
}

static void GetNISTHeaderInfo(FILE *dataStream, DataFile *data)
{
   char token[100],*lab;
   long nS,sR,sS;
   
   rewind(dataStream); 
   cNIST = ' '; nS=sR=sS=-1;
   lab = GetNISTToken(dataStream,token);           /* Check NIST label */
   if (strlen(lab)>4) *(lab+4) = '\0';
   if (strcmp(lab,"NIST") !=0)
      HError(2,"NIST header label missing");
   NISTSkipLine(dataStream);
   data->hdrSize = atoi(GetNISTToken(dataStream,token)); /* header #bytes */
   NISTSkipLine(dataStream);
   while (strcmp(GetNISTToken(dataStream,token),"end_head")!=0){
      if (strcmp(token,"sample_count") == 0)    /* objects */
         nS = GetNISTIVal(dataStream);
      else if (strcmp(token,"sample_rate") == 0)
         sR = GetNISTIVal(dataStream);
      else if (strcmp(token,"sample_n_bytes") == 0)
         sS = GetNISTIVal(dataStream);
      NISTSkipLine(dataStream);
   }
   if (sS != 2)
      HError(2,"Sample size = %d in NIST header",sS);
   if (nS == -1)
      HError(2,"Num samples undefined in NIST header");
   if (sR == -1)
      HError(2,"Num samples undefined in NIST header");
   data->nSamples = nS;
   data->sampSize = sS;
   data->sampPeriod = 1.0E7 / (float)sR;
   data->sampKind = WAVEFORM;
}
            /* ---- SCRIBE ---- */
            
static void GetSCRIBEHeaderInfo(FILE *dataStream, DataFile *data)
{
   rewind(dataStream);
   data->hdrSize = 0;
   HError(2,"Sorry - SCRIBE data format still undefined");
}


/* ---------------------- Some General Utility Routines --------------- */

/* EXPORT->BaseSampKind: return the basic sample kind without qualifiers */
SampKind BaseSampKind(SampKind kind)
{
   return kind & BASEMASK;
}

/* EXPORT->HasEnergy: returns true if log energy included */
Boolean HasEnergy(SampKind kind)
{
   return (kind & HASENERGY) != 0;
}

/* EXPORT->HasPitch: returns true if f0 included */
Boolean HasPitch(SampKind kind)
{
   return (kind & HASPITCH) != 0;
}

/* EXPORT->SampNComp: returns num comps in parm vector */
int SampNComp(DataFile *data)
{
   int n;
   
   if (BaseSampKind(data->sampKind) == IREFC)
      n = data->sampSize / sizeof(short);
   else
      n = data->sampSize / sizeof(float);
   return n;
}

/* EXPORT->SampDimen: returns dimensions of parm vector */
int SampDimen(DataFile *data)
{
   int n;
   
   n = SampNComp(data);
   if (BaseSampKind(data->sampKind) == LPDELCEP) n /= 2;
   if (HasEnergy(data->sampKind)) --n;
   if (HasPitch(data->sampKind)) --n;
   return n;
}

/* ---------------------- Parameter Transforms --------------- */

/* XformPossible: returns true if proposed xform is possible */
static int XformPossible(SampKind src, SampKind tgt)
{
   static short xmap[8][8] = {
      { 0, 0, 0, 0, 0, 0, 0, 0},    /* src = WAVEFORM */
      { 0, 1, 1, 1, 1, 0, 0, 0},    /* src = LPC */
      { 0, 1, 1, 1, 1, 0, 0, 0},    /* src = LPREFC */
      { 0, 0, 0, 1, 1, 0, 0, 0},    /* src = LPCEPSTRA */
      { 0, 0, 0, 0, 1, 0, 0, 0},    /* src = LPDELCEP */
      { 0, 0, 0, 0, 0, 1, 0, 0},    /* src = IREFC */
      { 0, 0, 0, 0, 0, 0, 1, 0},    /* src = MFCC */
      { 0, 0, 0, 0, 0, 0, 0, 0},    /* src = ANON */
   };
   
   return xmap[src][tgt];
}

/* Xform2LPREFC: LPC -> LPREFC */
static void Xform2LPREFC(DataFile *data)
{
   Vector a,k;
   int i,n,nbytes,size;
   float *p;
   
   n = SampDimen(data);
   nbytes = n * sizeof(float);
   size = SampNComp(data);
   a = CreateVector(n); 
   k = CreateVector(n);
   p = (float *)data->data;
   for (i=0;i<data->nSamples;i++){
      memcpy(a+1,p,nbytes);
      LPC2RefC(a,k);
      memcpy(p,k+1,nbytes);
      p += size;
   }  
   FreeVector(a); FreeVector(k);
   data->sampKind += LPREFC-LPC;
}

/* Xform2LPC: LPREFC -> LPC */
static void Xform2LPC(DataFile *data)
{
   Vector a,k;
   int i,n,nbytes,size;
   float *p;
   
   n = SampDimen(data);
   nbytes = n * sizeof(float);
   size = SampNComp(data);
   a = CreateVector(n); 
   k = CreateVector(n);
   p = (float *)data->data;
   for (i=0;i<data->nSamples;i++){
      memcpy(k+1,p,nbytes);
      RefC2LPC(k,a);
      memcpy(p,a+1,nbytes);
      p += size;
   }  
   FreeVector(a); FreeVector(k);
   data->sampKind += LPC-LPREFC;
}

/* Xform2LPCEPSTRA: LPC -> LPCEPSTRA */
static void Xform2LPCEPSTRA(DataFile *data)
{
   Vector a,c;
   int i,n,nbytes,size,lifter=0;
   float *p;
   
   if (getenv("HLIFTER") != NULL)  
      lifter = GetEnvValue("HLIFTER");
   n = SampDimen(data);
   nbytes = n * sizeof(float);
   size = SampNComp(data);
   a = CreateVector(n); 
   c = CreateVector(n);
   p = (float *)data->data;
   for (i=0;i<data->nSamples;i++){
      memcpy(a+1,p,nbytes);
      LPC2Cepstrum(a,c);
      if (lifter>0)
         WeightCepstrum(c,1,n,lifter);
      memcpy(p,c+1,nbytes);
      p += size;
   }  
   FreeVector(a); FreeVector(c);
   data->sampKind += LPCEPSTRA-LPC;
}

/* Xform2LPDELCEP: LPCEPSTRA -> LPDELCEP */
static void Xform2LPDELCEP(DataFile *data)
{
   void *p;
   int i,lasti,j,t,n,n2,delwin=2;
   float *fp1,*fp2, *back, *forw;
   float sum, sigmaT2;
   
   /* First make space for delta coef */
   data->dataSize *= 2;
   if ( (p=(void *)realloc(data->data,data->dataSize)) == NULL)
      HError(2,"Unable to create space for Delta Coef");
   data->data = p;
   n = SampNComp(data); n2 = n+n;
   fp1 = fp2 = (float *) p;
   fp1 += data->nSamples*n;
   fp2 += data->nSamples*n2;
   for (i=0;i<data->nSamples;i++){
      fp2 -= n2; fp1 -= n;
      memcpy(fp2,fp1,data->sampSize);
   }
   data->sampSize *= 2;
   
   /* Now compute the deltas */
   if (getenv("HDELTA") != NULL)  
      delwin = GetEnvValue("HDELTA");
   sigmaT2 = 0.0;
   for (t=1;t<=delwin;t++)
      sigmaT2 += t*t;
   sigmaT2 *= 2.0;
   fp1 = (float *)p;                      /* fp1 will step thru ceps */
   fp2 = fp1 + n;                         /* fp2 will step thru dels */
   lasti = data->nSamples - delwin;
   for (i=0;i<delwin;i++) {               /* leading edge - use simple */
      for (j=0;j<n;j++){                  /* first order difference */
         *fp2 = *(fp1+n2) - *fp1;
         ++fp1; ++fp2;
      }
      fp1 += n; fp2 += n;
   }
   for (i=delwin;i<lasti;i++){                  /* Full regression */
      for (j=0;j<n;j++){
         back = forw = fp1; sum = 0.0;
         for (t=1;t<=delwin;t++) {
            back -= n2; forw += n2;
            sum += t * (*forw - *back);
         }
         *fp2 = sum / sigmaT2;
         ++fp1; ++fp2;
      }
      fp1 += n; fp2 +=n;
   }
   for (i=lasti;i<data->nSamples;i++) {   /* trailing edge - use simple */
      for (j=0;j<n;j++){                  /* first order difference */
         *fp2 = *fp1 - *(fp1-n2);
         ++fp1; ++fp2;
      }
      fp1 += n; fp2 += n;
   }
   data->sampKind += LPDELCEP-LPCEPSTRA;
}

/* DeleteQualifier: delete last qual from each parm vector 
      - dont bother to reclaim storage */
static void DeleteQualifier(DataFile *data)
{
   int oldSize,newSize,i;
   char *p1,*p2;
   
   p1 = p2 = (char *)data->data;
   oldSize = data->sampSize;
   newSize = oldSize - sizeof(float);
   for (i=1;i<=data->nSamples;i++){
      memcpy(p2,p1,newSize);
      p1 += oldSize;
      p2 += newSize;
   }
   data->sampSize = newSize;
}

/* XformParms: Convert to format given by hcoerce */
static void XformParms(DataFile *data)
{
   SampKind srckind,tgtkind;
   
   srckind = data->sampKind; tgtkind = hcoerce;
   if (srckind == tgtkind) return;
   if (!HasEnergy(srckind) && HasEnergy(tgtkind) ||
            !HasPitch(srckind) && HasPitch(tgtkind) )
       HError(2,"Cannot create qualifier data");
   if (HasPitch(srckind) && !HasPitch(tgtkind)){
      DeleteQualifier(data);
      data->sampKind -= HASPITCH;
   }
   if (HasEnergy(srckind) && !HasEnergy(tgtkind)){
      DeleteQualifier(data);
      data->sampKind -= HASENERGY;
   }
   srckind = BaseSampKind(srckind);
   tgtkind = BaseSampKind(tgtkind);
   if (srckind == tgtkind) return;
   if (XformPossible(srckind,tgtkind)){
      if (srckind == LPC && tgtkind == LPREFC) {
         Xform2LPREFC(data); return;
      }
      if (srckind == LPREFC) {
         Xform2LPC(data); srckind = LPC;
      }
      if (tgtkind == srckind) return;
      if (srckind == LPC) {
         Xform2LPCEPSTRA(data); srckind = LPCEPSTRA;
      }
      if (tgtkind == srckind) return;
      if (srckind == LPCEPSTRA) {
         Xform2LPDELCEP(data); srckind = LPDELCEP;
      }
      if (tgtkind == srckind) return;
      HError(2,"Xform dead end!");
   } else
      HError(2,"Auto Transform from %s to %s not possible",
         smkmap[srckind],smkmap[tgtkind]);
}

/* ---------------------- Data Loading ----------------------- */

/* DataSpaceNeeded: returns number of bytes of memory space needed
   to store given data file.  This is normally same as source file
   but is xfloatsize/shortsize for IREFC and x2 for LPCEPSTRA ->
   LPDELCEP */
static int DataSpaceNeeded(DataFile *data)
{
   int n;
   
   n = data->nSamples * data->sampSize;
   if (BaseSampKind(data->sampKind) == IREFC)
      n = n * sizeof(float) / sizeof(short);
   if (BaseSampKind(data->sampKind) != LPDELCEP && hcoerce == LPDELCEP)
      n *= 2;
   return n;
}

/* CvtShort2Float: Converts the contents of the given data file */
static void CvtShort2Float(DataFile *data)
{
   int i,count;
   short *sp;
   float *fp;
   
   count = data->dataSize / sizeof(float);
   sp = (short *)data->data + count; 
   fp = (float *)data->data + count;
   for (i=0;i<count;i++)
      *--fp = *--sp / 16368.0;
   data->sampSize = data->sampSize*sizeof(float) / sizeof(short);
}

/* DataBytes: Returns number of bytes in file excluding header */
static long DataBytes(FILE *dataStream, DataFile *data)
{
   long fileLen;
   
   if (fseek(dataStream,0,SEEK_END))
      HError(2,"Cannot seek to end of data file");
   if ((fileLen = ftell(dataStream)) == -1L)
      HError(2,"Cannot read data file end position");
   return fileLen+1-data->hdrSize;
}

/* LoadData: Loads samples from stream into Data File */
static void LoadData(FILE *dataStream, DataFile *data)
{
   int unusedSamples;
   
   /* Check size of data area consistent with nSamples */
   unusedSamples = DataBytes(dataStream, data) / data->sampSize 
                     - data->nSamples;
   if (unusedSamples < 0 || unusedSamples > 1)
      HError(0,"Unused space in data file for %d samples",unusedSamples);
   
   data->dataSize = DataSpaceNeeded(data);
   data->data = (void *)malloc(data->dataSize);
   if (data->data == NULL)
      HError(2,"Cannot allocate memory for file data");
   if (fseek(dataStream,data->hdrSize,SEEK_SET))
      HError(2,"Cannot seek to start of file data area");      
   if (fread(data->data, data->sampSize, data->nSamples, dataStream) 
         != data->nSamples)
      HError(2,"Cannot read data into memory");
   if (BaseSampKind(data->sampKind) == IREFC) {
      CvtShort2Float(data);
      data->sampKind += LPREFC-IREFC;
   }
   if (hcoerce != ANON && data->sampKind != WAVEFORM)
      XformParms(data);
}

/* ------------------- Data File Opening --------------------- */

/* MakeHeaderInfo: set up the header info in supplied DataFile using */
/* values supplied by the environment, compute nSamples from file size */
static void MakeHeaderInfo(FILE *dataStream, int hdrSize, DataFile *data)
{
   data->hdrSize    = hdrSize;
   data->sampPeriod = GetEnvValue("SAMPPERIOD");
   data->sampSize   = GetEnvValue("SAMPSIZE");
   data->sampKind   = GetSampKind("SAMPKIND");
   data->nSamples   = DataBytes(dataStream,data) / data->sampSize;
}

/* EXPORT-> SpOpen: open file fname */
void SpOpen(char *fname, DataFile *data)
{
   FILE *dataStream;
   short *p;
   long i;
   
   /* Open read only stream for fname */
   if ((dataStream = fopen(fname, "rb")) == NULL)
      HError(2,"Cannot open data file %s",fname);
      
   /* Get Format via SetFormat, Env or default */
   data->format = GetFormat();
   
   /* Set hcoerce flag if HCOERCE set in environment */
    hcoerce = GetSampKind("HCOERCE");
    
   /* Use header to set up DataFile info,  for ALIEN and NOHEAD */
   /* this info must be set in environment  */
   switch(data->format) {
      case ALIEN: 
         MakeHeaderInfo(dataStream, GetEnvValue("HDSIZE"), data);
         break;
      case NOHEAD:
         MakeHeaderInfo(dataStream, 0, data);
         break;
      case SPED:  
         GetSPEDHeaderInfo(dataStream,data);
         break;
      case HTK:
         GetHTKHeaderInfo(dataStream,data);
         break;
      case TIMIT: 
         GetTIMITHeaderInfo(dataStream,data);
         break;
      case NIST: 
         GetNISTHeaderInfo(dataStream,data);
         break;
      case SCRIBE: 
         GetSCRIBEHeaderInfo(dataStream,data);
         break;
      default:    /* this shouldnt happen! */
         HError(2,"Unknown format %s (SpOpen BUG)",data->format);
   }
   
   /* Copy Data from file into memory */
   LoadData(dataStream, data);
   
   /* If necessary byte swap the data, this is only allowed for */
   /* WAVEFORM data */
   if (data->sampKind == WAVEFORM)
      if (MustSwap(data->format)) {     /* Byte Swapping */
         p = data->data;
         for (i=1; i<=data->nSamples; i++)
            SwapShort(p++); 
      }
   
   if (fclose(dataStream) == EOF)
      HError(2,"Unable to close file %s",fname);
}

/* ---------------- Data File Closing ------------------------ */

/* EXPORT->SpClose: close the given data file and free storage */
void SpClose(DataFile *data)
{
   free(data->data);
   data->data = NULL;
}

/* ------------- Data File Creation and Saving --------------- */

/* EXPORT->SpCreate: create the specified data file */
void SpCreate(long sampPeriod, int  sampSize, 
              SampKind sampKind, DataFile *data)
{
   data->nSamples = 0;
   data->sampPeriod = sampPeriod;
   data->sampSize = sampSize;
   data->sampKind = sampKind;
   data->format = HTK;
   data->hdrSize = sizeof(HTKhdr);
   data->dataSize = CREATECHUNKSIZE;
   if ((data->data = (void *)malloc(data->dataSize)) == NULL)
      HError(2,"Cannot allocate memory for file data");
};

#ifdef MPW
typedef struct  {
    unsigned long fdType;          /*the type of the file*/
    unsigned long fdCreator;       /*file's creator*/
    unsigned short fdFlags;        /*flags ex. hasbundle,invis,locked, etc. */
    long fdLocation;               /*file's location in folder*/
    short fdFldr;                  /*folder containing file*/
}FInfo;

/* SetSpFileInfo: Sets file creator to HTK; file kind to DATA */
static void SetSpFileInfo(char *fn)
{
   FInfo inf;

   if (getfinfo(fn,0,&inf))
      HError(2,"Cannot Get File Info for %s",fn);
   inf.fdType='D'<<24|'A'<<16|'T'<<8|'A';
   inf.fdCreator='H'<<24|'T'<<16|'K'<<8|' ';
   inf.fdFlags |= 6;
   if(setfinfo(fn,0,&inf))
      HError(2,"Cannot Set File Info for %s",fn);
}
#endif

/* EXPORT->SpSave: save the given data file to the ext file fname */
void SpSave(char *fname, DataFile *data)
{
   FILE *dataStream;
   
   if ((dataStream = fopen(fname, "wb")) == NULL)
      HError(2,"Cannot create data file %s for save",fname);
   PutHTKHeaderInfo(dataStream,data);
   if (fseek(dataStream,data->hdrSize,SEEK_SET))
      HError(2,"Cannot seek to start of file data area for save to %s",fname);      
   if (fwrite(data->data, data->sampSize, data->nSamples, dataStream) 
         != data->nSamples)
      HError(2,"Cannot write data to file %s",fname);
   if (fclose(dataStream) == EOF)
      HError(2,"Unable to close save file %s",fname);
#ifdef MPW
   SetSpFileInfo(fname);
#endif
}

/* ---------------- Data Access Routines --------------------- */

/* EXPORT->GetSample: get the index'th sample and store at bufp */
void GetSample(DataFile *data, long index, void *bufp)
{
   memcpy(bufp, (char *)data->data + index*data->sampSize, 
                 data->sampSize);
}

/* EXPORT->PutSample: put the given sample at index'th pos in data */
void PutSample(DataFile *data, long index, void *bufp)
{
   long bytesNeeded,size1,size2;
   void *p;
   
   if (index+1 > data->nSamples) { /* extending data file */
      data->nSamples = index+1;
      bytesNeeded = data->nSamples*data->sampSize;
      if (bytesNeeded > data->dataSize) { /* need to extend buffer */
         size1 = bytesNeeded;
         size2 = data->dataSize + CREATECHUNKSIZE;
         bytesNeeded = (size1>size2) ? size1 : size2;
         if ((p = (void *)realloc(data->data,bytesNeeded)) == NULL)
            HError(2,"Unable to grow data file to %ld bytes",bytesNeeded);
         data->dataSize = bytesNeeded;
         data->data = p;
      }
   }
   memcpy((char *)data->data+index*data->sampSize,bufp,data->sampSize);
}

/* ----------------------------------------------------------- */
/*                      END:  HSpIO.c                          */
/* ----------------------------------------------------------- */
