/* ----------------------------------------------------------- */
/*                   HTK  -  HMM Toolkit                       */
/* ----------------------------------------------------------- */
/*         Copyright: Steve Young      October 1989            */
/*         Cambridge University Engineering Department         */
/*         Trumpington Street, Cambridge, CB2 1PZ              */
/*                                                             */
/* Permission is hereby granted to copy, use and modify this   */
/* software provided that this banner notice is not removed    */
/*                                                             */
/* ----------------------------------------------------------- */
/*         File: HSpIO.h:   Speech Data File Input/Output      */
/* ----------------------------------------------------------- */


enum _FileFormat{
        NOHEAD, 
        SPED,              /* almost defunct */
        HTK,               /* default for all parameterised files */
        TIMIT,             /* TIMIT database */
        NIST,              /* NIST databases eg RM1 */
        SCRIBE,            /* UK Scribe databases */
        ALIEN              /* Unknown */
};
typedef enum _FileFormat FileFormat;

enum _SampKind{
      WAVEFORM,                           /* Raw speech waveform */
      LPC,LPREFC,LPCEPSTRA,LPDELCEP,      /* LP-based Coefficients */
      IREFC,                              /* Ref Coef in 16 bit form */
      MFCC,                               /* Mel-Freq Cepstra */
      FBANK,                              /* Filter Bank */
      ANON};
typedef enum _SampKind SampKind;
#define HASENERGY 0100           /* Added if log energy included */
#define HASPITCH  0200           /* Added if f0 included */
#define BASEMASK  077            /* Mask to remove qualifiers */

typedef struct {
   long nSamples;       /* No of samples in file */
   long sampPeriod;     /* Sample period in 100ns units */
   int  sampSize;       /* Sample size in bytes */
   int  hdrSize;        /* Header size in bytes */
   FileFormat format;   /* Format of file */
   SampKind sampKind;   /* Kind of samples stored in data */
   long dataSize;       /* Num bytes allocated for data */
   void *data;          /* Pointer to actual data */
}DataFile;

void SpOpen(char *fname, DataFile *data);
/*
   Open the file fname, set up the relevent info in data then
   load the actual speech data into memory.  Format of file is
   determined by either SetFormat or environment variable HFORMAT.
   If format is ALIEN then the env variable HDSIZE must be set.
   If format is ALIEN or NOHEAD then the env variables SAMPSIZE,
   SAMPPERIOD and SAMPKIND must be set
*/

void SpClose(DataFile *data);
/*
   Free memory buffer allocated for data
*/

void SpCreate(long sampPeriod, int  sampSize, 
              SampKind sampKind, DataFile *data);
/*
   Create an empty data file in data with supplied characteristics.
*/

void SpSave(char *fname, DataFile *data);
/*
   Store data file in external file fname in HTK format
*/

void SetFormat(FileFormat fmat);
/*
   Set the current file format to fmat.  This will affect subsequent
   calls to SpOpen, once used this routine overrides env variable
   HFORMAT
*/

FileFormat GetFormat(void);
/* 
   Get format from environment, or format override specified by
   SetFormat.
*/

void GetSample(DataFile *data, long index, void *bufp);
/*
   Get the index sample in data and copy into bufp.  Index values
   run from 0 to data->nSamples-1
*/

void PutSample(DataFile *data, long index, void *bufp);
/*
   Store sample in bufp at index sample in data file, if index denotes
   a sample beyond current end of data file then data file is extended
   so that index is last sample
*/

char *Format2Str(FileFormat format, char *buf);
FileFormat Str2Format(char *fmt);
char *SampKind2Str(SampKind kind, char *buf);
SampKind Str2SampKind(char *smk);
/*
   Useful utility routines to convert between enum types & strings.
   In enum to string cases, string is stored in *buf and 
   pointer to buf is returned
*/

SampKind BaseSampKind(SampKind kind);
Boolean HasEnergy(SampKind kind);
Boolean HasPitch(SampKind kind);
/* 
   Functions to separate base sample kind from qualifiers 
*/

int SampDimen(DataFile *data);
/*
   Returns dimensions of sample vectors excluding additional 
   info such as log energy
*/

int SampNComp(DataFile *data);
/*
   Returns number of components in sample vectors including 
   additional info such as log energy
*/

/* ------------------------ End of HSpIO.h ------------------------- */
