/* ----------------------------------------------------------- */
/*                   HTK  -  HMM Toolkit                       */
/* ----------------------------------------------------------- */
/*         Copyright: David Rainton      June 1991             */
/*                                                             */
/*                                                             */
/* Permission is hereby granted to copy, use and modify this   */
/* software provided that this banner notice is not removed    */
/*                                                             */
/* ----------------------------------------------------------- */
/*         File: HSubS.c  HMM Subspace Definition Data Type    */
/* ----------------------------------------------------------- */

#include <stdio.h>         /* Standard Libraries */
#include <string.h>
#include <ctype.h>
#include <math.h>
#include <errno.h>
#include <stdarg.h>

#include "HShell.h"
#include "HSpIO.h"
#include "HMath.h"
#include "HSubS.h"
#include "HModel.h"


/*-------------Local Utility routines------------------------*/

static int GetInt(FILE *f)
{
   int n;
   
   if (fscanf(f,"%d",&n) != 1)
      HError(2,"GetInt: Integer Expected");
   return n;
}

static float GetFloat(FILE *f)
{
   float x;
   
   if (fscanf(f,"%e",&x) != 1)
      HError(2,"GetFloat: Float Expected");
   return x;
}

/*------------Reading the external subspace specification---*/

/* creates and empty subspace structure */

void SubSCreate (int vsz, SubSDef *SubS)
{
 
  SubS->vsz = vsz; 
  SubS->mean = CreateVector(SubS->vsz);
  SubS->eig  = CreateMatrix(SubS->vsz, SubS->vsz);

} /* end of SubSCreate() */


/* this reads and loads  a subspace definition file */ 
void SubSOpen(char *fname, SubSDef *SubS)
{
 FILE *fp;
 int i,j;
 
 if ((fp = fopen (fname, "r")) == NULL) 
   HError(1, "unable to open file name %s", fname);

 SubS->vsz  = GetInt(fp);  
 SubS->mean = CreateVector(SubS->vsz);
 SubS->eig  = CreateMatrix(SubS->vsz, SubS->vsz);

 ReadlnVector(fp, SubS->mean); 
 ReadlnMatrixColByCol(fp, SubS->eig); 

 fclose(fp);

} /* end SubSOpen() */

void SubSWrite (char *fname, SubSDef SubS)
{
 FILE *fp; 
 int i,j;

 if ((fp = fopen (fname, "w")) == NULL) 
   HError(1, "unable to open file name %s", fname);

 if (fprintf (fp, "%d\n", SubS.vsz) != 0)
   HError(1, "unable to write SubS.vsz (%d) to  %s", SubS.vsz, fname);

 WritelnVector(fp, SubS.mean);
 WritelnMatrixColByCol(fp, SubS.eig);
  
 fclose (fp);

} /* end of SubSWrite() */

void PrintSubS (SubSDef SubS)
{
 int i,j;

 if (fprintf (stdout, "%d\n", SubS.vsz) != 0)
   HError(1, "unable to write SubS.vsz (%d) to stdout", SubS.vsz);

 WritelnVector(stdout, SubS.mean);
 WritelnMatrixColByCol(stdout, SubS.eig);
  

} /* end of SubSWrite() */


/*---------------------Data rotation routines--------------------*/

/* this rotates the vector in[1...vsz] using SubS to produce out[1...vsz] */
void RotateVector(Vector in, Vector out, SubSDef SubS)
{
 int i,j;

  ZeroVector(out);

  for (i=1; i<=SubS.vsz; i++) 
    for (j=1; j<=SubS.vsz; j++) out[i] += in[j]*SubS.eig[j][i];
  

} /* end RotateVector */


/* this rotates in place a sequence of data vectors */

void RotateData(int start, int stop, DataFile *data, SubSDef SubS)
{
  int index;
  Vector buf, buf1;

  buf = CreateVector(data->sampSize); /* vector buf[0] - buf[data->sampSize]*/  
  buf1 = CreateVector(data->sampSize);

  for (index=start; index<=stop; index++) {
    GetSample(data, index, buf+1);                                     /* vectorcopy data[index] -> buf+1 */
    RotateVector (buf, buf1, SubS);                                  /* rotate buf -> buf1 */
    memcpy(data->data+index*data->sampSize, buf1+1, data->sampSize); /* copy buf1 -> data[index]*/
  } 
   
  FreeVector (buf);
  FreeVector (buf1);

} /* end RotateData() */

void UpdateCov (Vector x, Vector m, Matrix c, int n, int sz)

{
    Vector buf;
    int   i, j;

    if (n==1)  CopyVector (x, m); /* with 1 sample mean = sample, covariance is zero */
    else {  /* for 2nd, 3rd, ... samples update recursively mean and variance */

      buf = CreateVector(sz);

      for (i=1; i<=sz ;i++)
      {
  	buf[i] = m[i];
	m[i] = (n*m[i] + x[i]) / (n+1);
      }
      for (i = 1 ; i <= sz ; i++)
  	for (j = 1 ; j <= sz ; j++)
  	    c[i][j] = (n*(c[i][j] + buf[i]*buf[j]) + x[i]*x[j]) / (n+1) - m[i]*m[j];

      FreeVector(buf);
    } /* end else */
}


static float at,bt,ct;
#define PYTHAG(a,b) ((at=fabs(a)) > (bt=fabs(b)) ? \
(ct=bt/at,at*sqrt(1.0+ct*ct)) : (bt ? (ct=at/bt,bt*sqrt(1.0+ct*ct)): 0.0))

static float maxarg1,maxarg2;
#define MAX(a,b) (maxarg1=(a),maxarg2=(b),(maxarg1) > (maxarg2) ?\
	(maxarg1) : (maxarg2))
#define SIGN(a,b) ((b)<0 ? -fabs(a) : fabs(a))

void SvdCmp(Matrix a, Vector w, Matrix v)
{
	int flag,i,its,j,jj,k,l,nm;
	float c,f,h,s,x,y,z;
	float anorm=0.0,g=0.0,scale=0.0;
	Vector rv1;
        int m, n;

        m = NumRows(a);
        n = NumCols(a);
 
	if (m < n) 
         HError(1, "SVDCMP: You must augment A with extra zero rows");
	rv1=CreateVector(n);
	for (i=1;i<=n;i++) {
		l=i+1;
		rv1[i]=scale*g;
		g=s=scale=0.0;
		if (i <= m) {
			for (k=i;k<=m;k++) scale += fabs(a[k][i]);
			if (scale) {
				for (k=i;k<=m;k++) {
					a[k][i] /= scale;
					s += a[k][i]*a[k][i];
				}
				f=a[i][i];
				g = -SIGN(sqrt(s),f);
				h=f*g-s;
				a[i][i]=f-g;
				if (i != n) {
					for (j=l;j<=n;j++) {
						for (s=0.0,k=i;k<=m;k++) s += a[k][i]*a[k][j];
						f=s/h;
						for (k=i;k<=m;k++) a[k][j] += f*a[k][i];
					}
				}
				for (k=i;k<=m;k++) a[k][i] *= scale;
			}
		}
		w[i]=scale*g;
		g=s=scale=0.0;
		if (i <= m && i != n) {
			for (k=l;k<=n;k++) scale += fabs(a[i][k]);
			if (scale) {
				for (k=l;k<=n;k++) {
					a[i][k] /= scale;
					s += a[i][k]*a[i][k];
				}
				f=a[i][l];
				g = -SIGN(sqrt(s),f);
				h=f*g-s;
				a[i][l]=f-g;
				for (k=l;k<=n;k++) rv1[k]=a[i][k]/h;
				if (i != m) {
					for (j=l;j<=m;j++) {
						for (s=0.0,k=l;k<=n;k++) s += a[j][k]*a[i][k];
						for (k=l;k<=n;k++) a[j][k] += s*rv1[k];
					}
				}
				for (k=l;k<=n;k++) a[i][k] *= scale;
			}
		}
		anorm=MAX(anorm,(fabs(w[i])+fabs(rv1[i])));
	}
	for (i=n;i>=1;i--) {
		if (i < n) {
			if (g) {
				for (j=l;j<=n;j++)
					v[j][i]=(a[i][j]/a[i][l])/g;
				for (j=l;j<=n;j++) {
					for (s=0.0,k=l;k<=n;k++) s += a[i][k]*v[k][j];
					for (k=l;k<=n;k++) v[k][j] += s*v[k][i];
				}
			}
			for (j=l;j<=n;j++) v[i][j]=v[j][i]=0.0;
		}
		v[i][i]=1.0;
		g=rv1[i];
		l=i;
	}
	for (i=n;i>=1;i--) {
		l=i+1;
		g=w[i];
		if (i < n)
			for (j=l;j<=n;j++) a[i][j]=0.0;
		if (g) {
			g=1.0/g;
			if (i != n) {
				for (j=l;j<=n;j++) {
					for (s=0.0,k=l;k<=m;k++) s += a[k][i]*a[k][j];
					f=(s/a[i][i])*g;
					for (k=i;k<=m;k++) a[k][j] += f*a[k][i];
				}
			}
			for (j=i;j<=m;j++) a[j][i] *= g;
		} else {
			for (j=i;j<=m;j++) a[j][i]=0.0;
		}
		++a[i][i];
	}
	for (k=n;k>=1;k--) {
		for (its=1;its<=30;its++) {
			flag=1;
			for (l=k;l>=1;l--) {
				nm=l-1;
				if (fabs(rv1[l])+anorm == anorm) {
					flag=0;
					break;
				}
				if (fabs(w[nm])+anorm == anorm) break;
			}
			if (flag) {
				c=0.0;
				s=1.0;
				for (i=l;i<=k;i++) {
					f=s*rv1[i];
					if (fabs(f)+anorm != anorm) {
						g=w[i];
						h=PYTHAG(f,g);
						w[i]=h;
						h=1.0/h;
						c=g*h;
						s=(-f*h);
						for (j=1;j<=m;j++) {
							y=a[j][nm];
							z=a[j][i];
							a[j][nm]=y*c+z*s;
							a[j][i]=z*c-y*s;
						}
					}
				}
			}
			z=w[k];
			if (l == k) {
				if (z < 0.0) {
					w[k] = -z;
					for (j=1;j<=n;j++) v[j][k]=(-v[j][k]);
				}
				break;
			}
			if (its == 500) 
                         HError (1, "No convergence in 500 SVDCMP iterations");
			x=w[l];
			nm=k-1;
			y=w[nm];
			g=rv1[nm];
			h=rv1[k];
			f=((y-z)*(y+z)+(g-h)*(g+h))/(2.0*h*y);
			g=PYTHAG(f,1.0);
			f=((x-z)*(x+z)+h*((y/(f+SIGN(g,f)))-h))/x;
			c=s=1.0;
			for (j=l;j<=nm;j++) {
				i=j+1;
				g=rv1[i];
				y=w[i];
				h=s*g;
				g=c*g;
				z=PYTHAG(f,h);
				rv1[j]=z;
				c=f/z;
				s=h/z;
				f=x*c+g*s;
				g=g*c-x*s;
				h=y*s;
				y=y*c;
				for (jj=1;jj<=n;jj++) {
					x=v[jj][j];
					z=v[jj][i];
					v[jj][j]=x*c+z*s;
					v[jj][i]=z*c-x*s;
				}
				z=PYTHAG(f,h);
				w[j]=z;
				if (z) {
					z=1.0/z;
					c=f*z;
					s=h*z;
				}
				f=(c*g)+(s*y);
				x=(c*y)-(s*g);
				for (jj=1;jj<=m;jj++) {
					y=a[jj][j];
					z=a[jj][i];
					a[jj][j]=y*c+z*s;
					a[jj][i]=z*c-y*s;
				}
			}
			rv1[l]=0.0;
			rv1[k]=f;
			w[k]=x;
		}
	}
	FreeVector(rv1);
}

#undef SIGN
#undef MAX
#undef PYTHAG



/*** End of HSubS.c ***/















