/* ----------------------------------------------------------- */
/*                   HTK  -  HMM Toolkit                       */
/* ----------------------------------------------------------- */
/*         Copyright: David Rainton    September 1991          */
/*         Last Update:                                        */
/*                                                             */
/*         ATR Interpreting Telephony Research Labs            */
/*         Japan                                               */
/*                                                             */
/* Permission is hereby granted to copy, use and modify this   */
/* software provided that this banner notice is not removed    */
/*                                                             */
/*  D. Rainton - Modification History                          */
/*                                                             */
/*                                                             */
/* ----------------------------------------------------------- */
/*         File: HUtilities.c: Misclaneous routines for GPD    */
/* ----------------------------------------------------------- */



#include <stdio.h>         /* Standard Libraries */
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <math.h>
#include <errno.h>
#include <stdarg.h>

#include "HShell.h"       /* HTK Libraries */
#include "HSpIO.h"
#include "HMath.h"
#include "HModel.h"
#include "HLabel.h"
#include "HUtilities.h"


/* EXPORT -> OpenFile: opens the specified file  */
FILE *OpenFile(char *name, char *mode)
{
 FILE *fptr;

 if((fptr = fopen(name,mode)) == NULL)
   HError(2, "Cannot open file %s for %s",name,mode);

 return(fptr);
}


/* EXPORT -> GetChkedInt: used to check that integer command line arguments are within the 
                specified range min->max */
int GetChkedInt(int min, int max, char * swtname)
{
   int val;
   
   if (NextArg() != INTARG) 
      HError(1,"Integer Arg Required for %s option",swtname);
   val=GetIntArg();
   if (val<min || val>max)  
      HError(1,"Integer arg out of range(%d - %d) in %s option",min,max,swtname); 
   return val;
}

/* EXPORT -> GetChkedFlt: used to check that floating point command line arguments are within the 
                specified range min->max*/
float GetChkedFlt(float min, float max, char * swtname)
{
   float val;
   
   if (NextArg() != INTARG && NextArg() != FLOATARG) 
      HError(1,"Float Arg Required for %s option",swtname);
   val=GetFltArg();
   if (val<min || val>max) 
      HError(1,"Float arg out of range (%f - %f) in %s option",min,max,swtname); 
   return val;
}

/* EXPORT -> OpenAndCheckData: open and check a given data file for consistency with vSize and SK*/
void OpenAndCheckData(DataFile *data, char *fn, int vSize, SampKind SK)
{
   SpOpen(fn,data);      
   if (data->sampSize / sizeof(float) != vSize)
      HError(2,"Incompatible Vec Size in file %s",fn);
   if (data->sampKind != SK)
      HError(2,"Incompatible Sample Kind in file %s",fn);
}


/* EXPORT -> CheckData: check a given data file for consistency with vSize and SK*/
void CheckData(DataFile *data, char *fn, int vSize, SampKind SK)
{
   if (data->sampSize / sizeof(float) != vSize)
      HError(2,"Incompatible Vec Size in file %s",fn);
   if (data->sampKind != SK)
      HError(2,"Incompatible Sample Kind in file %s",fn);
}

/* EXPORT -> LoadHMM: load specified hmm, adding hmmExt if set */
void LoadHMM(char *name, HMMDef *def, char *hmmDir, char*hmmExt, HMMParameters *hmmP)
{
   char fn[255];
   int len;
   
   if (hmmDir==NULL)
      fn[0]='\0';
   else {
      strcpy(fn,hmmDir);
      len = strlen(fn);
      if (fn[len-1]!=PATHCHAR) {
         fn[len]=PATHCHAR;
         fn[++len]='\0';
      }
   }
   strcat(fn,name);
   if (hmmExt!=NULL) {
      strcat(fn,".");
      strcat(fn,hmmExt);
   }
   SPINBALL
   HOpen(fn,def);
   if (hmmP->vSize==0) {
      hmmP->vSize = def->vecSize;
      hmmP->M = def->numMixtures;
      hmmP->CK = def->ckind; hmmP->SK = def->pkind;
      hmmP->isTied = def->tstat != FreeT;
   } else {
      if(hmmP->vSize!=def->vecSize)
         HError(2,"HMMs have incompatible vecSizes");
      if(hmmP->M!=def->numMixtures)
         HError(2,"HMMs have incompatible numMixtures");
      if (hmmP->CK!=def->ckind)
         HError(2,"HMMs have different covariance kinds");
      if (hmmP->SK!=def->pkind)
         HError(2,"HMMs have different sample kinds");
      if (hmmP->isTied != (def->tstat != FreeT))
         HError(2,"Mixed Free and Tied Distributions");
   }
   if (hmmP->maxStates<def->numStates)
      hmmP->maxStates = def->numStates;
}


extern int  trace;           /* Trace level */

/* EXPORT -> LoadHMMDefs: create HMMdefList and load it */
HMMDef * LoadHMMDefs(LabId *HMMidList, int H, char *hmmDir, char *hmmExt, HMMParameters *hmmP)
{
   int h;
   HMMDef *defList, *d;
   
   if ((defList = (HMMDef *)malloc((H+1)*sizeof(HMMDef))) == NULL)
      HError(2,"Unable to create hmm def array");
   --defList;
   for (h=1;h<=H;h++){
      d = defList+h;
      LoadHMM(HMMidList[h]->name, d, hmmDir, hmmExt, hmmP);
      if(trace >= 1) {
       printf("Loaded phoneme HMM %s\n", HMMidList[h]->name);
       fflush(stdout);
      }
   }

   return (defList);
}

/* EXPORT -> MallocHMMDefList: malloc space for empty HMMdefList and return pointer to the calling routine */
HMMDef * MallocHMMDefList(HMMDef * HMMdefList, int H)
{
   int h;
   HMMDef *defList, *d;
   
   if ((defList = (HMMDef *)malloc((H+1)*sizeof(HMMDef))) == NULL)
      HError(2,"Unable to create hmm def array");
   --defList;
   for (h=1;h<=H;h++) {
     d = HMMdefList+h;
     HCreate(d->numStates, d->numMixtures, d->vecSize, d->pkind, d->ckind, d->dkind, d->tstat,
              defList+h); 
   }

   return (defList);
}

/* EXPORT -> SaveHMMDefs: save the list of HMM definitions in HMMdefList to the directory hmmDir */
void SaveHMMDefs(HMMDef *HMMdefList, LabId *HMMidList, char *hmmDir, char *hmmExt, int H)
{
  int i;
  char hmmfn[256];

  for(i=1;i<=H;i++) {
   MakeFN(HMMidList[i]->name, hmmDir,hmmExt,hmmfn);
   if (trace>2) printf("Writing out HMM %s\n", hmmfn);
   HSave(hmmfn, HMMdefList+i);
  }
 
}


#define LEVELSEP '/'


/* EXPORT -> PrintTranscription1: print a label file transcription to stdout */
void PrintTranscription1(Transcription t, Boolean printscore)
{
   LabList *p;
   int i;
   Label *l;
   
   p = (LabList *) t;
   while (p != NULL){
      for (i = 0, l=p->labels; i < p->nLabels; i++,l++){
         printf("%4d ",l->start);
         printf("%4d ",l->end);
         printf("%6s ",l->labid->name);
         if(printscore) printf("%10.4f",l->score);
         printf("\n");
         fflush(stdout);   
      }
      if (p->next != NULL)
         printf("%c \n",LEVELSEP);
      p = p->next;
   }
   fflush(stdout);
}

/* EXPORT -> PrintTranscription: print a label file transcription to stdout */
void PrintTranscription(Transcription t)
{
 PrintTranscription1(t, TRUE);
}

/* EXPORT -> MixtureExists: check to see if a mixture exits */
Boolean MixtureExists(HMMDef *hmm, int state, int mixture)

{
  if (hmm == NULL) 
    HError(1,"HMM does not exist");
  if (hmm->stateinfo == NULL) 
    HError(1,"HMM State %d does not exist", state); 
  if(hmm->stateinfo[state].outp[mixture].weight>MINMIX) return(TRUE);
  else return(FALSE);
}

/* EXPORT -> return x^2 */
double sqr(double x)
{
 return(pow(x,2.0));
}

/* EXPORT -> NormaliseVector: normalise a vector */
void NormaliseVector(Vector x, int size)

{
 int i;
 double sum=0.0;

 for(i=1;i<=size;i++) sum += sqr(x[i]);

 if (sum == 0.0) 
   HError(1, "Cannot normalise a zero vector"); 
 else { 
   sum = sqrt(sum);
   for(i=1;i<=size;i++)
     x[i] /= sum; 
 }
}


/* EXPORT -> ClampVector: if a vector element falls below threshold set to threshold*/
void ClampVector(Vector x, int size, float threshold, float clampvalue)
{
 int i;
  
 for(i=1;i<=size;i++) if (x[i] < threshold) x[i]=clampvalue; 
}

/* EXPORT -> ClampVectorToZero: if a vector element falls below threshold set to zero*/
void ClampVectorToZero(Vector x, int size, float threshold)
{
 ClampVector(x,size,threshold,0.0);
}


/* EXPORT -> AddVectors: returns the result v1+v2 in v1*/ 
void AddVectors(Vector v1, Vector v2)
{
 int i;
 if(VectorSize(v1) != VectorSize(v2)) HError(1,"Cannot add vectors of different size"); 
  
 for(i=1;i<=VectorSize(v1);i++) v1[i] += v2[i];
}

/* EXPORT -> SubtractVectors: returns the result v1-v2 in v1*/ 
void SubtractVectors(Vector v1, Vector v2)
{
 int i;
 if(VectorSize(v1) != VectorSize(v2)) HError(1,"Cannot subtract vectors of different size"); 
  
 for(i=1;i<=VectorSize(v1);i++) v1[i] -= v2[i];
}


/* EXPORT -> AddMatrices: returns the result m1+m2 in m1*/ 
void AddMatrices(Matrix m1, Matrix m2)
{
 int i,j;
 int nr, nc;
 
 nr = NumRows(m1);
 nc = NumCols(m1);

 if (nr != NumRows(m2)) HError(1,"Cannot add matrices with differing number of rows"); 
 if (nc != NumCols(m2)) HError(1,"Cannot add matrices with differing number of columns"); 
  
 for(i=1;i<=nr;i++) for(j=1;j<=nc;j++) m1[i][j] += m2[i][j];
}

/* EXPORT -> VectorMultiply: multiplies the vector v by the constant k*/ 
void VectorMultiply(Vector v, double k)
{
 int i;
  
 for(i=1;i<=VectorSize(v);i++) v[i] *= k;
}

/* EXPORT -> VectorSubtract: subtracts the constant k from each element of the vector v*/ 
void VectorSubtract(Vector v, double k)
{
 int i;
  
 for(i=1;i<=VectorSize(v);i++) v[i] -= k;
}


/* EXPORT -> VectorAdd: adds the constant k to each element of the vector v*/ 
void VectorAdd(Vector v, double k)
{
 int i;
  
 for(i=1;i<=VectorSize(v);i++) v[i] += k;
}


/* EXPORT -> VectorDivide: divides the vector v by the constant k*/ 
void VectorDivide(Vector v, double k)
{
 int i;

 if (k == 0.0) HError(1,"Cannot divide by zero");
 for(i=1;i<=VectorSize(v);i++) v[i] /= k;
}


/* EXPORT -> VectorL2F: Convert log(v) to real. Result is
   floored to 0.0 if x < LSMALL */ 
void VectorL2F(Vector v)
{
 int i;

 for(i=1;i<=VectorSize(v);i++) L2F(v[i]);
}

/* EXPORT -> Vector2Norm: Returns the (sum_1^N v[i]^2)^(1/2)  */
double Vector2Norm(Vector v)
{
 int i;
 double sum=0.0;

 for(i=1;i<=VectorSize(v);i++) sum += v[i]*v[i];
 sum = sqrt(sum);

 return sum;

}

/* EXPORT -> Vector1Norm: Returns the sum_1^N |v[i]| */
double Vector1Norm(Vector v)
{
 int i;
 double sum=0.0;

 for(i=1;i<=VectorSize(v);i++) sum += fabs(v[i]);

 return sum;

}

/* EXPORT -> VectorMax: Returns max(v[i]) */
double VectorMax(Vector v)
{
 int i;
 double max= -1*MINLARG;

 for(i=1;i<=VectorSize(v);i++) if(v[i]>max) max=v[i];

 return max;

}

/* EXPORT -> VectorAbsMax: Returns max(abs(v[i])) */
double VectorAbsMax(Vector v)
{
 int i;
 double absmax=0.0;

 for(i=1;i<=VectorSize(v);i++) if(fabs(v[i])>absmax) absmax=fabs(v[i]);

 return absmax;

}




/*EXPORT -> DotProduct: Returns the dot product of the vectors v1 and v2 */
double DotProduct(Vector v1, Vector v2)
{
 int i;
 double dot=0.0;
 
 if(VectorSize(v1) != VectorSize(v2))
 HError(1,"Cannot compute the dot product for unequal sized vectors");

 for(i=1;i<=VectorSize(v1);i++) dot += v1[i]*v2[i];

 return dot;

}


/* EXPORT-> DRHPrint: Print the model hmm according to specified flags */
void DRHPrint(HMMDef *hmm, int flags, ...)
{
   int stState,enState,i,j,n,m;
   Vector v;
   Matrix c;
   StateInfo *sptr;
   OutProb *optr;
   va_list ap;
   char buf1[20],buf2[20],buf3[20];
   
   va_start(ap,flags);
   if (flags & (HPMEANS|HPVARS|HPDURS|HPLABELS)){
      stState = va_arg(ap,int);
      enState = va_arg(ap,int);
      if (stState<2 || enState>=hmm->numStates)
         HError(2,"Illegal state range in hprint");
   }
   va_end(ap);
   if (flags & HPDETAILS) {
      printf("Num States   = %3d\n", hmm->numStates);
      printf("Num Mixtures = %3d\n", hmm->numMixtures);
      printf("Vector Size  = %3d\n", hmm->vecSize);
      printf("Features     = %s %s %s\n",
         CovKind2Str(hmm->ckind,buf1),DurKind2Str(hmm->dkind,buf2),
         SampKind2Str(hmm->pkind,buf3));
      switch(hmm->tstat){
         case FreeT: 
            break;
         case SlaveT:
            printf("Tied Mixture Slave [Master=%s]\n",hmm->tieMaster);
            break;
         case MasterT:
            printf("Tied Mixture Master\n");
            break;
      }

   }
   if (flags & (HPMEANS|HPVARS|HPDURS|HPLABELS))
      for (n=stState; n<=enState; n++) {
         printf("State %d\n",n); 
         sptr = hmm->stateinfo+n;
         if (flags & (HPMEANS|HPVARS)) {
            optr = sptr->outp;
            for (m=1; m<=hmm->numMixtures; m++) { 
               optr++;
               if (/* optr->weight>MINMIX || hmm->numMixtures == 1 */ TRUE) {
                  if (hmm->numMixtures>1)
                     printf(" Mixture %d:  weight %8.3f\n",m,optr->weight);
                  if (flags &HPMEANS) {
                     v = optr->mean;
                     printf("  Means:\n");
                     for (i=1;i<=hmm->vecSize;i++)
                        printf("%8.3f",v[i]);
                     printf("\n");
                  }
                  if (flags & HPVARS) {
                     if ( hmm->ckind==DIAGC) {
                        v = optr->covariance.var;
                        printf("  Variances:\n");
                        for (i=1;i<=hmm->vecSize;i++)
                           printf("%8.3f",v[i]);
                        printf("\n");
                     } else {
                        c = optr->covariance.inv;
                        printf("  Inverse Covariance Matrix:\n");
                        for (i=1;i<=hmm->vecSize;i++) {
                           for (j=1;j<=hmm->vecSize;j++)
                              printf("%8.3f",c[i][j]);
                           printf("\n");
                        }
                     }
                     printf("  Gauss Constant:  ");
                     if (optr->gConst > LSMALL)
                        printf("%e\n",optr->gConst);
                     else
                        printf("not set\n");
                  }
               }
            }
         }
         if (flags & HPDURS) {
            printf("  Duration Parameters: ");
            switch (hmm->dkind) {
               case NULLD:    printf("None\n"); break;
               case POISSOND: printf("%8.3f\n",sptr->dur1); break;
               case GAMMAD:   printf("%8.3f %8.3f\n",sptr->dur1,sptr->dur2); break;
            }
         }
         if (flags & HPLABELS)
            printf("  Label: %s\n", sptr->label == NULL  ? "None" : sptr->label);
      }
   if (flags & HPTRANS) {
      printf("Transition Matrix:\n");
      for (i=1;i<=hmm->numStates;i++) {
         for (j=1;j<=hmm->numStates;j++)
            printf("%8.3f",flags&LOGTRANS?L2F(hmm->transP[i][j]):hmm->transP[i][j]);
         printf("\n");
      }
   }  
}


/* EXPORT -> Sigmoid: returns (1/(1+exp(-a(x-b))) */
double Sigmoid (double x, double a, double b)
{
 
 return( 1 / (1 + exp(-a*(x-b))) );

}

/* EXPORT -> DeltaSigmoid: returns the first differential of Sig(x,a,b) */
double DeltaSigmoid (double x, double a, double b)
{
double s;
s = Sigmoid(x,a,b);
 return( s*s*a*exp(-a*(x-b)) );

}




