/* ----------------------------------------------------------- */
/*                   HTK  -  HMM Toolkit                       */
/* ----------------------------------------------------------- */
/*        Copyright: Steve Young      October 1989             */
/*        Last Update:                September 1990           */
/*                                                             */
/*        Cambridge University Engineering Department          */
/*        Trumpington Street, Cambridge, CB2 1PZ               */
/*                                                             */
/*  Permission is hereby granted to copy, use and modify this  */
/*  software provided that this banner notice is not removed   */
/*                                                             */
/* ----------------------------------------------------------- */
/*      File: HDbase.c:   Training Database Management         */
/* ----------------------------------------------------------- */


/*
   This module provides for the storage and retrieval of training
   data.  It is intended for use in 'isolated word' styles of 
   training where all stored observation sequences correspond to
   the same segment.
   
   Each sequence put into the dbase is stored in a memory buffer.
   This memory buffer is grown until either the request to realloc
   is not met or MAXPAGESIZE is reached.  At that point, paging 
   starts and memory buffers are swapped to and from disc.
   
   All sequences must be stored before attempting to read back any
   of the data.
   
   Note that default buffering is used for all i/o.  An appropriate
   call to setvbuf within PutSequence may improve performance 
   considerably depending on the local environment.
*/


#include <stdio.h>         /* Standard Libraries */
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <unistd.h>

#include "HShell.h"
#include "HMath.h"
#include "HSpIO.h"
#include "HDbase.h"

#if defined MPW
   #pragma segment hmods
#endif

#define MAXSEQUENCES 2000           /* Max number of sequences to store */
#define SEGSIZE      8192           /* Num floats in a segment */
#define MAXPAGESIZE  100000         /* Max number of floats per page */
typedef long Address;
#define PAGE(A) ((A >> 24) & 0xff)  /* Page number is in 4th byte of address */
#define INDEX(A) (A & 0xffffff)     /* Index is in 1st 3 bytes of address */

typedef struct {
   short len;
   Address addr;
} Entry;

static Entry itab[MAXSEQUENCES];    /* Holds address of each seq in vbuf */
static int itnext;                  /* index of next free entry in itab */

static float *vbuf;                 /* Dynamic storage area for vectors */
static int vbnext;                  /* Index of next free float in vbuf */

static int pageSize;                /* Num floats per page */
static int curPage;                 /* Current page in core */
static FILE * dbFile = NULL;        /* File for backing store */

static SampKind kind;               /* Kind of stored data */
static int vSize;                   /* Vector Size */


static void SavePage(void)
{
   if (fwrite(vbuf,sizeof(float),pageSize,dbFile) != pageSize)
      HError(2,"Unable to write HDbase page %d",curPage);
}

static void NewPage(void)
{
   curPage++; vbnext = 0;  
}

static void LoadPage(int pagenum)
{
   if (getenv("HDBSWAP") != NULL)
      HError(0,"HDBase - swapping in page %d",pagenum);
   if (fseek(dbFile,pageSize*sizeof(float)*pagenum,SEEK_SET) != 0)
      HError(2,"Cannot seek to disk page %d in HDbase",pagenum);
   if (fread(vbuf,sizeof(float),pageSize,dbFile) != pageSize)
      HError(2,"Unable to read page %d in HDbase",pagenum);
   curPage = pagenum;
}

/* EXPORT->InitDbase: Initialise the database */
void InitDbase(int vecSize, SampKind sampKind)
{
   if ((vbuf = (float *)malloc(SEGSIZE * sizeof(float)) ) == NULL)
      HError(2,"Unable to allocate first segment in HDbase");
   vbnext = 0;
   itnext = 0;
   pageSize = SEGSIZE; curPage = 0;
   kind = sampKind, vSize = vecSize;
}

/* EXPORT->CloseDbase: close down the database */
void CloseDbase(void)
{
   free(vbuf);
   fclose(dbFile);
}

/* EXPORT->PutSequence: Store the seq covering the time span start to end */
void PutSequence(long start, long end, DataFile *data)
{
   long stidx, endidx, i, len, spaceNeeded;
   int *ip;
   void *p;
   
   if (data->sampKind != kind || data->sampSize/sizeof(float) !=vSize)
      HError(2,"Attempt to put incompatible data into HDbase");
   if (itnext == MAXSEQUENCES)
      HError(2,"Max sequence limit reached in HDbase");
   stidx = start / data->sampPeriod;
   endidx = end / data->sampPeriod; 
   if (endidx>=data->nSamples) endidx = data->nSamples-1;
   len = endidx-stidx+1;
   spaceNeeded = len * (vSize + 1);
   
   if (spaceNeeded > (pageSize-vbnext)) {    /* In core page is full */
      if (dbFile == NULL) {   /* havent started paging yet */
         if (pageSize+SEGSIZE < MAXPAGESIZE &&
            (p = (void *)realloc(vbuf,(pageSize+SEGSIZE)*sizeof(float))) != NULL) {
            pageSize += SEGSIZE;       /* in core page successfully extended */
            vbuf = (float *)p;
         } else {             /* memory exhausted, start paging */
            if ((dbFile = tmpfile()) == NULL) 
               HError(2,"Cannot open HDbase temp file");
            errno=0;   /* some compilers leave errno set after tmpfile */
            SavePage(); NewPage();
         }
      } else {
         SavePage(); NewPage();
      }
   }
   itab[itnext].addr = (curPage <<24) + vbnext; /* Copy Sequence */
   itab[itnext].len = len;
   for (i=stidx;i<=endidx;i++) {
      ip = (int *)vbuf+vbnext++; *ip = vSize;
      GetSample(data,i,vbuf+vbnext);
      vbnext += vSize;
   }
   itnext++;
}

/* EXPORT->FixDbase: Freeze the database */
void FixDbase(void)
{
   if (dbFile != NULL) {
      SavePage(); LoadPage(0);
   }
}
   
/* EXPORT->NumSequences: Total number of sequences currently stored */
int  NumSequences()
{
   return itnext;
}

/* EXPORT->SequenceLength: Return number of vectors in index'th sequence */
int  SequenceLength(int index)
{
   return itab[index].len;
}

/* EXPORT->GetDBVector: Return the i'th vector in the index'th sequence */
Vector GetDBVector(int index, int i)
{
   Address a;
   int vbpage;
   
   a = itab[index].addr;
   vbpage = PAGE(a);
   if (vbpage != curPage) 
      LoadPage(vbpage);
   return vbuf+(INDEX(a) + (vSize+1)*i);
}
   
   
/* ------------------------ End of HDbase.c ------------------------- */
