/* ----------------------------------------------------------- */
/*                   HTK  -  HMM Toolkit                       */
/* ----------------------------------------------------------- */
/*         Copyright: Steve Young      October 1989            */
/*         Last Update:                September 1990          */
/*                                                             */
/*         Cambridge University Engineering Department         */
/*         Trumpington Street, Cambridge, CB2 1PZ              */
/*                                                             */
/* Permission is hereby granted to copy, use and modify this   */
/* software provided that this banner notice is not removed    */
/*                                                             */
/* ----------------------------------------------------------- */
/*         File: HLabel.c:   Speech Label File Input           */
/* ----------------------------------------------------------- */

#include <stdio.h>         /* Standard Libraries */
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include "HShell.h"
#include "HSpIO.h"
#include "HLabel.h"

#if defined MPW
   #pragma segment hmods
#endif

/*
 File formats supported by this module:
 
      TIMIT - text files with each label on a line in the form
                  start-sample end-sample  label   
              (eg 456 7899 sh). 
              
      SPED  - a sequence of structs of type SPEDLab (see below)

      HTK   - a HTK transcription consists of one or more
              levels where each level is separated by a / character
                 <transcription> = <level> {"/" <level>}
              each level is a sequence of lines of text, each defining
              a label
                 <level> = <label> { <label> }
              each label consists of optional start and end times in
              units of 100ns followed by a name followed by an
              optional score  
               <label> = [<start> <end>] <name> [<score>] [";" <comment>]
              The start and end times are ints, the score can be
              int or float.  The name must begin with a non-digit
              character other than the level separator.

   Format selection only operates on input and works in a similar
   way to HSpIO.  The default is HTK, this can be overridden by
   defining the environment variable HLFORMAT, this in turn can
   be overridden by a call to SetLFormat.

*/

/* ----------------- Format Handling Routines --------- */
extern char *getenv();

static short formatSet = FALSE;     /* Flag indicating format set */
static FileFormat format = HTK;     /* Current label file format */
      
/* EXPORT-> SetLFormat: set format and override environ setting */
void SetLFormat(FileFormat fmat)
{
   format = fmat; 
   formatSet = TRUE;
}

/* EXPORT-> GetLFormat: get format from SetLFormat or environ or return default
 */
FileFormat GetLFormat()
{

   char *fmt;
   
   if (formatSet) return format;
   if ((fmt = getenv("HLFORMAT")) == NULL) return format;
   return Str2Format(fmt);
}

/* ---------------- Label Name Hashing ----------------- */

#define HASHSIZE 101      /* size of hash table */

static NameCell *hashtab[HASHSIZE];  /* the actual table */

/* Hash: return a hash value for given label name */
static unsigned Hash(char *name)
{
   unsigned hashval;

   for (hashval=0; *name != '\0'; name++)
      hashval = *name + 31*hashval;
   return hashval%HASHSIZE;
}

/* NewCell: return a pointer to a new NameCell */
static NameCell *NewCell(char *name)
{
   char *s;
   NameCell *p;

   if ((s=(char *)malloc(strlen(name)+1)) == NULL)
      HError(2,"Cannot malloc space for Label Name in NewCell");
   strcpy(s,name);
   if ((p=(NameCell *)malloc(sizeof(NameCell))) == NULL)
      HError(2,"Cannot malloc space for NameCell");
   p->name = s; p->next = NULL; p->aux = 0;
   return p;
}

/* EXPORT->InitLabel: initialise module */
void InitLabel(void)
{
   int i;

   for (i=0;i<HASHSIZE;i++)
      hashtab[i] = NULL;
}

/* EXPORT->GetLabId: return id of given name */
LabId GetLabId(char *name, int insert)
{
   int h;
   NameCell *p;

   h = Hash(name); p = hashtab[h];
   if (p==NULL) {  /* special case - this slot empty */
      if (insert)
         p=hashtab[h]=NewCell(name);
      return p;
   }
   do{             /* general case - look for name */
      if (strcmp(name,p->name) == 0)
         return p; /* found it */
      p = p->next;
   } while (p != NULL);
   if (insert){    /* name not stored */
      p = NewCell(name);
      p->next = hashtab[h];
      hashtab[h] = p; 
   }
   return p;
}

/* ----------------- TriPhone Stripping ------------------- */

static short stripTriPhones = FALSE;   /* Enable triPhone stripping */

/* EXPORT->TriStrip: Remove contexts of form A- and +B from s */
void TriStrip(char *s)
{
   char buf[100],*p;
   
   if ((p = strchr(s,'-')) == NULL) p = s; else ++p;
   strcpy(buf,p);
   if ((p = strrchr(buf,'+')) != NULL) 
      *p = '\0';
   strcpy(s,buf);
}

/* ----------------- LabList Allocation ------------------- */

#define CHUNKSIZE 100
static int numChunks;      /* .. in current list */

/* CreateLabList: create and return a lablist with one chunk */
static LabList *CreateLabList()
{
   LabList *l;
   
   if ((l=(LabList *)malloc(sizeof(LabList))) == NULL)
      HError(2,"Cannot create LabList header");
   l->nLabels = 0; l->next = NULL;
   if ((l->labels = (Label *)malloc(sizeof(Label)*CHUNKSIZE)) == NULL)
      HError(2,"Cannot create LabList body");
   numChunks = 1; 
   return l;
}

/* AddLabel: append given label to given lab list */
static void AddLabel(LabList *l, Label x)
{
   int newSize;
   
   if (l->nLabels == numChunks * CHUNKSIZE) {
      ++numChunks;      /* grow label list */
      newSize =  sizeof(Label)*CHUNKSIZE*numChunks;
      if ((l->labels = (Label *)realloc(l->labels,newSize)) == NULL)
         HError(2,"Cannot grow lab list");
   }
   l->labels[l->nLabels++] = x;
}

/* ShrinkLabList: remove unused label slots from current lab list */
static void ShrinkLabList(LabList *l)
{
   int newSize;
   
   if (l->nLabels == numChunks*CHUNKSIZE) return;
   newSize = sizeof(Label) * l->nLabels;
   if ((l->labels = (Label *)realloc(l->labels,newSize)) == NULL)
      HError(2,"Cannot shrink label list");
}

/* AppendLabList: append lab list l to transcription t */
static void AppendLabList(LabList *l, Transcription t)
{
   LabList *p = t;
   
   while(p->next != NULL)
      p = p->next;
   p->next = l; l->next = NULL;
}

/* ----------------- Transcription Operations ------------------- */

/* EXPORT->NumLevels: return number of levels in transcription t */
int NumLevels(Transcription t)
{
   LabList *p = t;
   int count = 0;
   
   while (p!=NULL) {
      p=p->next; ++count;
   }
   return count;
}

/* EXPORT->GetLevel: return a pointer ot the n'th level of t */
LabList *GetLevel(Transcription t, int n)
{
   LabList *p = t;
   int count = n;
   
   while (p!=NULL && count>0) {
      p=p->next; --count;
   }
   if (p==NULL) 
      HError(2,"Cant get transcription level %d",n);
   return p;   
}

/* EXPORT->NumCases: return num occurrences of label x in level n of t */
int NumCases(LabId x, Transcription t, int n)
{
   Label *p;
   int i,count = 0;
   LabList *l;
   
   l = GetLevel(t,n); p = l->labels;
   for (i=0; i<l->nLabels; i++){
      if (p->labid == x) count++;
      p++;
   }
   return count;
}

/* EXPORT->GetCase: returns m'th occurrence of label x at level n of t */
Label *GetCase(LabId x, int m, Transcription t, int n)
{
   Label *p;
   int i, count = 0;
   LabList *l;
   
   l=GetLevel(t,n);
   p = l->labels;
   for (i=0; i<l->nLabels; i++){
      if (p->labid == x) ++count;
      if (count == m) break;
      p++;
   }
   if (count != m)
      HError(2,"Cannot find %d'th occurrence of %s",m,x->name);
   return p;
}

/* -------------------- HTK Label Format ------------------ */

#define LEVELSEP '/'
#define COMMCHAR  ';'

static char curch;
static enum {TRINT,TRFLT,TRSTR,TREOL,TRLEV,TREOF} trSym;
static int trInt;
static float trFlt;
static char trStr[80];

/* GetTrSym: get next symbol from f */
static void GetTrSym(FILE *f)
{
   int i;
   
   trInt = 0; trFlt = 0.0; trStr[0]='\0';
   while (curch == ' ' || curch == '\t')  /* Skip Space */
      curch = fgetc(f);
   if (curch == COMMCHAR){             /* skip comment */
      while (curch != '\n' && curch != EOF) curch = fgetc(f);
   }

   switch (curch) {
      case '\n': 
         curch = fgetc(f);
         trSym = TREOL; 
         break;
      case LEVELSEP:
         while (isspace(curch=fgetc(f)));
         trSym = TRLEV; 
         break;
      case EOF:
         trSym = TREOF;
         break;
      default:
         i=0;
         do {
            trStr[i++] = curch; curch = fgetc(f);
         } while (isgraph(curch));
         trStr[i]='\0';
         if (isdigit(trStr[0]) || trStr[0] == '-'){  /* int or float */
            if (strchr(trStr,'.')==NULL) {
               trSym = TRINT; 
               sscanf(trStr,"%d",&trInt);
            } else {
               trSym = TRFLT;
               sscanf(trStr,"%f",&trFlt);
            }
         } else {
            if (stripTriPhones) TriStrip(trStr);
            trSym = TRSTR;
         }
         break;
   }
}

/* -------------------- HTK Label Format ------------------ */

/* LoadHTKLabels: load a HTK transcription */
static Transcription LoadHTKLabels(FILE *f)
{
   Transcription t;
   LabList *p;
   Label l;
   
   p=CreateLabList(); t = (Transcription)p; curch = ' ';
   GetTrSym(f);
   while (trSym==TRINT || trSym==TRSTR){
      l.start = 0; l.end = 0; l.score = 0.0;
      if (trSym==TRINT) {
         l.start = trInt; GetTrSym(f);
         if (trSym==TRINT) {
            l.end = trInt; GetTrSym(f);
         }
      }
      if (trSym != TRSTR)
         HError(2,"Label Name Expected");
      l.labid = GetLabId(trStr,TRUE);
      GetTrSym(f);
      if (trSym==TRINT | trSym==TRFLT){
         l.score = (trSym==TRINT)?trInt:trFlt;
         GetTrSym(f);
      }
      if (trSym == TREOL)
         GetTrSym(f);
      else
         HError(2,"End of Line Expected in HTK Label File");
      AddLabel(p,l);
      if (trSym == TRLEV){
         GetTrSym(f);
         ShrinkLabList(p);
         p=CreateLabList();
         AppendLabList(p,t);
      }
   }
   if (trSym != TREOF)
      HError(2,"Junk at end of HTK transcription");
   ShrinkLabList(p);
   return t;
}

/* --------------- Foreign Label Formats --------------------- */

/* LoadTIMITLabels: load a TIMIT transcription */
static Transcription LoadTIMITLabels(FILE *f)
{
   LabList *p;
   Label l;
   
   p = CreateLabList(); curch = ' ';
   GetTrSym(f);
   while (trSym == TRINT){    
      l.start = trInt*625;    /* sample rate is 16KHz */
      GetTrSym(f);
      if (trSym != TRINT) 
         HError(2,"End Time expected in TIMIT Label File");
      l.end   = trInt*625;
      GetTrSym(f);
      if (trSym != TRSTR) 
         HError(2,"Label Name expected in TIMIT Label File");
      l.labid = GetLabId(trStr,TRUE);
      l.score = 0.0;
      AddLabel(p,l);
      GetTrSym(f);
      if (trSym == TREOL)
         GetTrSym(f);
   }
   ShrinkLabList(p);
   return (Transcription) p;
}

/* LoadSPEDLabels: load a SPED transcription */

typedef struct { 
   char name[8];
   long startTime;
   long endTime;
} SPEDlab;

static Transcription LoadSPEDLabels(FILE *f)
{
   LabList *p;
   int i,len;
   char lab[80];
   Label l;
   SPEDlab slab;

   p = CreateLabList();
   while (1) {
      if (fread(&slab,sizeof(SPEDlab),1,f) != 1)
         break;
      l.start = slab.startTime*10;
      l.end   = slab.endTime*10;
      len = slab.name[0];
      if (len>=8)
         HError(2,"Bad Label in SPED file");
      for (i=0;i<len;i++)
         lab[i]=slab.name[i+1];
      lab[len] = '\0';
      if (stripTriPhones) TriStrip(lab);
      l.labid = GetLabId(lab,TRUE); 
      l.score = 0.0;
      AddLabel(p,l);
   }
   ShrinkLabList(p);
   return (Transcription)p;   
}

/* ------------------ Label File Opening/Closing -------------------- */

/* EXPORT->LOpen: Load transcription in fname and return it */
Transcription LOpen(char * fname)
{
   FILE *f;
   Transcription t;
   char *mode;
   FileFormat fmat;
   
   fmat = GetLFormat();  
   mode = "r";
   if (fmat == SPED) mode = "rb";
   if ((f=fopen(fname,mode)) == NULL)
      HError(2,"Unable to open label file %s",fname);
   switch (fmat) {
      case TIMIT:    t = LoadTIMITLabels(f); break;
      case HTK:      t = LoadHTKLabels(f); break;
      case SPED:     t = LoadSPEDLabels(f); break;
      default:
         HError(2,"Illegal label file format [%d]",fmat);
   }
   fclose(f);
   return t;
}
   
/* EXPORT->LClose: Release all storage allocated for t */
void LClose(Transcription t)
{
   LabList *p,*q;
   
   p = (LabList *) t;
   while (p != NULL){
      free(p->labels);
      q = p; p = p->next;
      free(q);
   }
}

/* EXPORT->LSave: Save transcription in fname using HTK format */

#define SAVLST 1
#define SAVLEN 2
#define SAVLSC 4

void LSave(char *fname, Transcription t)
{
   LabList *p;
   int i,flags;
   Label *l;
   FILE *f;
   
   if ((f=fopen(fname,"w")) == NULL)
      HError(2,"Unable to create label file %s",fname);
   p = (LabList *) t;
   while (p != NULL){
      flags = 0;     /* see which optional fields need saving */
      for (i = 0, l=p->labels; i < p->nLabels; i++,l++){
         if (l->start != 0) flags |= SAVLST;
         if (l->end   != 0) flags |= SAVLEN;
         if (l->score != 0.0) flags |= SAVLSC;
      }
      for (i = 0, l=p->labels; i < p->nLabels; i++,l++){
         if (SAVLST&flags) 
            fprintf(f,"%d ",l->start);
         if (SAVLEN&flags) 
            fprintf(f,"%d ",l->end);
         fprintf(f,"%s ",l->labid->name);
         if (SAVLSC&flags)
            fprintf(f,"%f",l->score);
         fprintf(f,"\n");
      }
      if (p->next != NULL)
         fprintf(f,"%c \n",LEVELSEP);
      p = p->next;
   }
   fclose(f);
}

/* EXPORT->LClone: clone a transcription and return it */
Transcription LClone(Transcription t)
{
   LabList *p,*q;
   int n;
   
   p = (LabList *)t; 
   if (p == NULL) return NULL;
   if ((q = (LabList *)malloc(sizeof(LabList))) == NULL)
      HError(2,"Cannot clone LabList header");
   q->nLabels = p->nLabels;
   n = q->nLabels*sizeof(Label);
   if ((q->labels = (Label *)malloc(n)) == NULL)
      HError(2,"Cannot clone Label List");
   memcpy(q->labels,p->labels,n);
   q->next = LClone(p->next);
   return (Transcription) q;  
}

/* EXPORT->LTriStrip: enable triphone stripping */
void LTriStrip(int enab)
{
   stripTriPhones = enab;
}

/* ---------------------- Id List Routines -------------------------- */

/* EXPORT->LoadIdList: Create an array of label ids indexed 1..P */
LabId *LoadIdList(char *fname, int *nLabs)
{
   FILE *f;
   LabId *p;
   char buf[80];
   const int chunk = 50;
   int n,size;
   
   if ((f = fopen(fname,"r")) == NULL)
      HError(2,"Cannot open id list file %s",fname);
   if ((p = (LabId *)malloc((chunk+1)*sizeof(LabId))) == NULL)
      HError(2,"Cannot create initial id list array");
   n = 0; size = chunk;
   while (ReadLabel(f,buf)){
      if (n >= size) {     /* need to expand array */
         size += chunk;
         if ((p = (LabId *)realloc(p,(size+1)*sizeof(LabId))) == NULL)
            HError(2,"Cannot extend id list array");
      }
      p[n++] = GetLabId(buf,TRUE);        
   }
   p[n] = NULL;   /* add final NULL */
   if (n < size)  /* shrink array to just that required */
      if ((p = (LabId *)realloc(p,(n+1)*sizeof(LabId))) == NULL)
         HError(2,"Cannot shrink id list array");
   *nLabs = n;
   fclose(f);
   return p-1;    /* external indices run from 1.. */
}

/* EXPORT->FindLabId: find and return the index of id in idList.  
   Return 0 if not found */
int FindLabId(LabId id, LabId *idList)
{
   int i = 1;
   
   while (idList[i] != NULL) {
      if (idList[i] == id) return i;
      ++i;
   }
   return 0;
}

/* EXPORT->ReadLabel: into buf from file f and return TRUE if ok */
Boolean ReadLabel(FILE *f, char *buf)
{
   char c;
   
   c = fgetc(f);
   while (isspace(c)) c = fgetc(f);
   if (c==EOF || !isgraph(c)) return FALSE;
   do {
      *buf++ = c; c = fgetc(f);
   } while ( !isspace(c) && c != EOF);
   *buf = '\0';
   return TRUE;
}


/* ------------------------ End of HLabel.c ------------------------- */
