/* ----------------------------------------------------------- */
/*                   HTK  -  HMM Toolkit                       */
/* ----------------------------------------------------------- */
/*         Copyright: Steve Young      October 1989            */
/*         Last Update:                September 1990          */
/*                                                             */
/*         Cambridge University Engineering Department         */
/*         Trumpington Street, Cambridge, CB2 1PZ              */
/*                                                             */
/* Permission is hereby granted to copy, use and modify this   */
/* software provided that this banner notice is not removed    */
/*                                                             */
/* ----------------------------------------------------------- */
/*         File: HMath.c   Math Support Module                 */
/* ----------------------------------------------------------- */

/*
This library provides math support in the following three areas

   a) Log Arithmetic
   b) Random Deviates
   c) Vector/Matrix Storage and Operators
   
Note that the type Real should be set in HMath.h to give optimum fpu
performance.  For example, on a 68881 fpu using long double types avoids
the need for conversions.
*/

#include <stdio.h>         /* Standard Libraries */
#include <stdlib.h>
#include <time.h>
#include <math.h>

#include <errno.h>

#define RAND_MAX 2147483648

#include "HShell.h"        /* HKT Libraries */
#include "HMath.h"

#if defined MPW
   #pragma segment hmods
#endif

/* -------------------- Vector and Matrix Handling ----------------- */

/*
   Vectors are pointers to arrays of float (ie float*); matrices are pointers
   to an array of vectors (ie float**).  All indexing is v[1..n] and 
   m[1..r][1..c].  The actual size of each vector is stored in v[0].  For 
   matrices, the row lengths are stored in every row since they are 
   genuine vectors, the number of columns is stored in m[0].
*/

/* EXPORT-> CreateVector:  Allocate space for vector v[1..size] */
Vector CreateVector(int size)
{
   float *v;
   int *i;
   
   v = (float *)malloc((unsigned)(size+1)*sizeof(float));
   if (v==NULL) 
      HError(2,"Cannot create vector with %d elements",size);
   i = (int *) v; *i = size;
   return v;
}

/* EXPORT-> FreeVector: Free space allocated for vector v[1..size] */
void FreeVector(Vector v)
{
   free((char *)v);
}

/* EXPORT-> CopyVector: copy v1 into v2; sizes must be the same */
void CopyVector(Vector v1, Vector v2)
{
   int i,*size; 
   
   size = (int *) v1;
   for (i=1; i <= *size; i++) 
      v2[i] = v1[i];
}

/* EXPORT-> VectorSize: returns number of components in v */
int VectorSize(Vector v)
{
   int *i;
   
   i = (int *) v;
   return *i;
}

/* EXPORT-> ReadVector: read vector from stream using e format */ 
void ReadVector(FILE *stream, Vector v)
{
   int i,*size;
   
   size = (int *) v;
   SPINBALL
   for (i=1; i <= *size; i++) 
      if (fscanf(stream,"%e",++v) != 1) 
         HError(2,"Reading vector component %d",i);
}

/* EXPORT-> WriteVector: write vector v to stream using e format 
   and default precision */
void WriteVector(FILE *stream, Vector v)
{
   int i,*size;
   
   size = (int *) v;
   SPINBALL
   for (i=1; i <= *size; i++) 
      fprintf(stream,"%e ",v[i]);
   fputc('\n',stream);
}

/* EXPORT-> BReadVector: read vector from binary stream */ 
void BReadVector(FILE *stream, Vector v)
{
   int n,*size;
   
   size = (int *) v;
   n = *size;
   if (fread(v+1,sizeof(float),n,stream) != n)
      HError(2,"BRead Vector failed");
   SPINBALL
}

/* EXPORT-> BWriteVector: write vector v to binary stream  */
void BWriteVector(FILE *stream, Vector v)
{
   int n,*size;
   
   size = (int *) v;
   n = *size;
   if (fwrite(v+1,sizeof(float),n,stream) != n)
      HError(2,"BWrite Vector failed");
   SPINBALL
}

/* EXPORT->ZeroVector: Zero the elements of v */
void ZeroVector(Vector v)
{
   int i,n;
   
   n=VectorSize(v);
   for (i=1;i<=n;i++)
      v[i]=0.0;
}

/* Export->ShowVector: show the vector v preceded by title */
void ShowVector(char * title,Vector v,int maxTerms)
{
   int i,maxi;
   
   maxi = VectorSize(v);
   if (maxi>maxTerms) maxi=maxTerms;
   printf("%s",title);  
   for (i=1;i<=maxi;i++)  printf("%.2f ",v[i]);
   printf("...\n");
}

/* EXPORT-> CreateMatrix:  Allocate space for matrix m[1..nrows][1..ncols] */
Matrix CreateMatrix(int nrows,int ncols)
{
   int *i,j;
   float **m;
   
   m = (float **)malloc((unsigned)(nrows+1)*sizeof(float *));
   if (m==NULL) 
      HError(2,"Cannot create matrix pointers with %d rows",nrows);
   i = (int *) m; *i = nrows;
   for (j=1;j<=nrows; j++) 
      m[j] = CreateVector(ncols);
   return m;
}

/* EXPORT-> FreeMatrix: Free space allocated for matrix m */
void FreeMatrix(Matrix m)
{
   int i,*nrows;
   
   nrows = (int *) m;
   for (i=1; i <= *nrows; i++)
      free((char *) m[i]);
   free((char *) m);
}

/* EXPORT-> CopyMatrix: copy matrix m1 to m2 */
void CopyMatrix(Matrix m1, Matrix m2)
{
   int i,*nrows;
   
   nrows = (int *) m1;
   for (i=1; i <= *nrows; i++)
      CopyVector(m1[i],m2[i]);
}

/* EXPORT-> NumRows: number of rows in matrix m */
int NumRows(Matrix m)
{
   int *nrows;
   
   nrows = (int *) m;
   return *nrows;
}

/* EXPORT-> NumCols: number of columns in matrix m */
int NumCols(Matrix m)
{
   int *ncols;
   
   ncols = (int *) m[1];
   return *ncols;
}

/* EXPORT-> ReadMatrix: read matrix from stream into m using e format */
void ReadMatrix(FILE *stream, Matrix m)
{
   int i,*nrows;
   
   nrows = (int *) m;
   for (i=1; i <= *nrows; i++)
      ReadVector(stream,m[i]);
}

/* EXPORT-> ReadSymMatrix: read symmetric matrix in upper triangular
   form from stream into m using e format */
void ReadSymMatrix(FILE *stream, Matrix m)
{
   int i,j,*nrows,size;
   
   nrows = (int *) m; size = *nrows;
   for (i=1; i <= size; i++) {
      SPINBALL
      for (j=i; j <= size; j++) {
         if (fscanf(stream,"%e",&m[i][j]) != 1) 
            HError(2,"Reading component %d of row %d",j,i);
         if (i != j) m[j][i] = m[i][j];
      }
   }
}

/* EXPORT-> WriteMatrix: write matrix to stream using e format
   and default precision */
void WriteMatrix(FILE *stream, Matrix m)
{
   int i,*nrows;
   
   nrows = (int *) m;
   for (i=1; i <= *nrows; i++)
      WriteVector(stream,m[i]);
}

/* EXPORT-> WriteSymMatrix: write symmetric matrix to stream in
   upper triangular form using e format and default precision */
void WriteSymMatrix(FILE *stream, Matrix m)
{
   int i,j,*nrows,size;
   
   nrows = (int *) m; size = *nrows;
   for (i=1; i <= size; i++) {
      SPINBALL
      for (j=1; j<i; j++)
         fprintf(stream," ");
      for (j=i; j <= size; j++) {
         fprintf(stream,"%e ",m[i][j]);
      }
      fputc('\n',stream);
   }
}

/* EXPORT-> BReadMatrix: read matrix from binary stream into m  */
void BReadMatrix(FILE *stream, Matrix m)
{
   int i,*nrows;
   
   nrows = (int *) m;
   for (i=1; i <= *nrows; i++)
      BReadVector(stream,m[i]);
}

/* EXPORT-> BWriteMatrix: write matrix to binary stream  */
void BWriteMatrix(FILE *stream, Matrix m)
{
   int i,*nrows;
   
   nrows = (int *) m;
   for (i=1; i <= *nrows; i++)
      BWriteVector(stream,m[i]);
}

/* EXPORT->ZeroMatrix: Zero the elements of m */
void ZeroMatrix(Matrix m)
{
   int i,j,nr,nc;
   
   nr=NumRows(m); nc=NumCols(m);
   for (i=1;i<=nr;i++)
      for (j=1;j<=nc;j++) m[i][j]=0.0;
}

/* Export->ShowMatrix: show the matrix m preceded by title */
void ShowMatrix(char * title,Matrix m,int maxCols,int maxRows)
{
   int margin,i,j,k;
   int maxi,maxj;
   
   margin = strlen(title);
   maxi = NumRows(m);
   if (maxi>maxRows) maxi=maxRows;
   maxj = NumCols(m);
   if (maxj>maxCols) maxj = maxCols;
   for (i=1;i<=maxi;i++) {
      if (i==1) printf("%s",title);
         else for (k=1;k<=margin;k++) putchar(' ');
      for (j=1;j<=maxj;j++)
         printf("%.2f ",m[i][j]);
      printf("...\n");
   }
   for (k=1;k<=margin;k++) putchar(' ');
   printf("...\n");
}

/* -------------------- Matrix Operations ---------------------- */

static void Ludcmp(Matrix a, int *indx, int *sign)  /* Num Recipes p43 */
{
   int i,imax,j,k,n;
   Real big,dum,sum,temp;
   Vector vv,tmp;
   
   n = NumRows(a);
   vv = CreateVector(n);
   *sign = 1;
   for (i=1; i<=n; i++) {
      big = 0.0;
      for (j=1; j<=n; j++)
         if ((temp = fabs(a[i][j])) > big )
            big = temp;
      if (big == 0.0)
         HError(2,"Singular matrix in ludcmp");
      vv[i] = 1.0/big;
   }
   for (j=1; j<=n; j++) {
      for (i=1; i<j; i++) {
         sum = a[i][j];
         for (k=1; k<i; k++)
            sum -= a[i][k]*a[k][j];
         a[i][j]=sum;
      }
      big=0.0;
      for (i=j; i<=n; i++) {
         sum = a[i][j];
         for (k=1; k<j; k++)
            sum -= a[i][k]*a[k][j];
         a[i][j]=sum;
         if ( (dum=vv[i]*fabs(sum)) >=big) {
            big = dum; imax = i;
         }
      }
      if (j != imax) {
         tmp = a[imax]; a[imax] = a[j]; a[j] = tmp;
         *sign = -(*sign);
         vv[imax]=vv[j];
      }
      indx[j]=imax;
      if (a[j][j] == 0.0)
         HError(2,"Matrix is singular");
      if (j != n) {
         dum = 1.0/a[j][j];
         for (i=j+1; i<=n;i++)
            a[i][j] *= dum;
      }
   }
   FreeVector(vv);
}

static void Lubksb(Matrix a, int *indx, float b[])
{
   int i,ii=0,ip,j,n;
   Real sum;
   
   n=NumRows(a);
   for (i=1;i<=n;i++) {
      ip=indx[i]; sum=b[ip]; b[ip]=b[i];
      if (ii)
         for (j=ii;j<=i-1;j++)
            sum -=a[i][j]*b[j];
      else
         if (sum) ii=i;
      b[i]=sum;
   }
   for (i=n; i>=1; i--) {
      sum=b[i];
      for (j=i+1; j<=n; j++)
         sum -=a[i][j]*b[j];
      b[i]=sum/a[i][i];
   }
}        
         
/* EXPORT-> CovInvert: puts inverse of c in invc, returns log(Det(c)) */
LogReal CovInvert(Matrix c, Matrix invc)
{
   Matrix a;
   float col[100],ldet;
   Real x;  int sign;
   int n,i,j,indx[100];
   
   n=NumRows(c);
   a=CreateMatrix(n,n);
   CopyMatrix(c,a);              /* Make a copy of c */
   Ludcmp(a,indx,&sign);            /* Do LU Decomposition */
   for (j=1; j<=n; j++) {     /* Invert matrix */
      for (i=1; i<=n; i++)
         col[i]=0.0;
      col[j]=1.0;
      Lubksb(a,indx,col);
      for (i=1; i<=n; i++)
         invc[i][j] = col[i];
   }  
   ldet = 0.0;                /* Calc log(det(c)) */
   for (i=1; i<=n; i++) {
      x = a[i][i];
      if (x<0.0) sign = -sign;
      ldet += log(fabs(x));
   }
   if (sign<0)
      HError(0,"Determinant of cov matrix is negative in CovInvert");
   FreeMatrix(a);
   return ldet;
}

/* EXPORT->CovDet: Returns log(Det(c)) */
LogReal CovDet(Matrix c)
{
   Matrix a;
   LogReal ldet;
   Real x;
   int n,i,sign,indx[100];
   
   n=NumRows(c);
   a=CreateMatrix(n,n);
   CopyMatrix(c,a);              /* Make a copy of c */
   Ludcmp(a,indx,&sign);         /* Do LU Decomposition */
   ldet = 0.0;                   /* Calc log(det(c)) */
   for (i=1; i<=n; i++) {
      x = a[i][i];
      if (x<0.0) sign = -sign;
      ldet += log(fabs(x));
   }
   if (sign<0)
      HError(0,"Determinant of cov matrix is negative in CovDet");
   FreeMatrix(a);
   return ldet;
}
   
/* -------------------- Log Arithmetic ---------------------- */

/*
   The type LogReal is for representing real numbers on a log 
   scale.  LZERO is used for log(0) in log arithmetic, any log 
   real value <= LSMALL is considered to be zero.
*/

static LogReal minLogExp;

/* EXPORT-> LAdd: Return sum x + y on log scale, 
   sum < LSMALL is floored to LZERO */
LogReal LAdd(LogReal x, LogReal y)
{
   LogReal temp,diff,z;
   
   if (x<y) {
      temp = x; x = y; y = temp;
   }
   diff = y-x;
   if (diff<minLogExp) 
      return  (x<LSMALL)?LZERO:x;
   else {
      z = exp(diff);
      if (errno==EDOM || errno==ERANGE){
         perror("LAdd");   
         errno = 0;  /* may not be legal on some machines */
      }
      return x+log(1.0+z);
   }
}

/* EXPORT-> LSub: Return diff x - y on log scale, 
   diff < LSMALL is floored to LZERO */
LogReal LSub(LogReal x, LogReal y)
{
   LogReal temp,diff,z;
   
   if (x<y) {
      temp = x; x = y; y = temp;
   }
   diff = y-x;
   if (diff<minLogExp) 
      return  (x<LSMALL)?LZERO:x;
   else {
      z = exp(diff);
      if (errno==EDOM || errno==ERANGE){
         perror("LSub");   
         errno = 0;  /* may not be legal on some machines */
      }
      return x+log(1.0-z);
   }
}

/* EXPORT-> L2F: Convert log(x) to real, result is
   floored to 0.0 if x < LSMALL */
float   L2F(LogReal x)
{
   return (x<LSMALL) ? 0.0 : exp(x);
}

/* -------------------- Random Numbers ---------------------- */

/* EXPORT->RandInit: Initialise random number generators */
void RandInit(void)
{
   unsigned int seed;      

   seed = (unsigned int) clock();
   srand(seed);
}

/* EXPORT-> RandomValue:  */
float RandomValue(void)
{
   return (float)rand()/RAND_MAX;
}

/* EXPORT-> GaussDeviate: random number with a N(mu,sigma) distribution */
float GaussDeviate(float mu, float sigma)
{
   Real fac,r,v1,v2,x;
   static int gaussSaved = 0; /* GaussDeviate generates numbers in pairs */
   static float gaussSave;    /* 2nd of pair is remembered here */


   if (gaussSaved) {
      x = gaussSave; gaussSaved = 0;
   }
   else {
      do {
         v1 = 2.0*(float)rand()/RAND_MAX - 1.0;
         v2 = 2.0*(float)rand()/RAND_MAX - 1.0;
         r = v1*v1 + v2*v2;
      }
      while (r>=1.0);
      fac = sqrt(-2.0*log(r)/r);
      gaussSaved = 1;
      gaussSave = v1*fac;
      x = v2*fac;
   }
   return x*sigma+mu;
}

/* --------------------- Initialisation ---------------------- */

/* EXPORT-> InitMath: initialise this module */

void InitMath(int useLogs)
{
   RandInit();
   minLogExp = -log(-LZERO);
}

/* ----------------------------------------------------------- */
/*                      END:  HMath.c                          */
/* ----------------------------------------------------------- */
