/* ----------------------------------------------------------- */
/*                   HTK  -  HMM Toolkit                       */
/* ----------------------------------------------------------- */
/*         Copyright: Steve Young      October 1989            */
/*         Last Update:                September 1990          */
/*                                                             */
/*         Cambridge University Engineering Department         */
/*         Trumpington Street, Cambridge, CB2 1PZ              */
/*                                                             */
/* Permission is hereby granted to copy, use and modify this   */
/* software provided that this banner notice is not removed    */
/*                                                             */
/* ----------------------------------------------------------- */
/*         File: HParse.c:   Network Defn Parser/Builder       */
/* ----------------------------------------------------------- */

#include <stdio.h>         /* Standard Libraries */
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include "HShell.h"
#include "HSpIO.h"
#include "HLabel.h"
#include "HParse.h"

#if defined MPW
   #pragma segment hmods2
#endif

/*
   This module provides a single routine which parses the set
   of rewrite rules in the supplied text file and builds an
   equivalent representation in memory.
   
   The syntax rules for the textual definition of the net are -

      model = char{char}   -- any sequence of characters except the meta
                              chars {}[]<>|=$();*\/   - the latter must be
                              escaped using backslash
      variable = $char{char}

      factor = "(" expr ")" |    -- simple factoring
               "{" expr "}" |    -- 0 or more repetitions
               "<" expr ">" |    -- 1 or more repetitions
               "[" expr "]" |    -- 0 or 1 repetition ie optional
               "<<" expr ">>" |  -- triphone loop
               model |           -- the name of a HMM 
               variable          -- must be already defined
               
      sequence = factor {factor}          -- sequences
      
      expr   = sequence {"|" sequence}    -- alternatives

      subnet = variable "=" expr ";"      -- define a variable

      network = {subnet} "(" expr ")"     -- the network itself

   All variables must be defined by a subnet definition before being
   used in an expression (this prohibits recursion and makes it easier
   to implement). C style comments may be placed anywhere in the text
   
   The network build process proceeds in 3 steps:
   
      1) the parser builds a network using additional 'glue'
         nodes and largish link vectors
      2) the net is scanned and all glue nodes removed
      3) the net is scanned and link vectors realloc'ed down
         to the required size
         
   Variables encountered in RHS's are expanded immediately.
         
*/

/* ---------------------------- Node Handling ------------------------ */

#define LINKCHUNKSIZE 5
#define MAXIDENT 31
typedef char Ident[MAXIDENT+1];


/* CreateLinkSet: allocate space for a LinkSet with size slots */
static LinkSet *CreateLinkSet(int size)
{
   LinkSet *p;
   
   if ((p = (LinkSet *) malloc(size*sizeof(Link))) == NULL)
      HError(2,"Cannot allocate LinkSet[%d]",size);
   return p;
}

/* ResizeLinkSet: change number of slots in oldLinks to newSize slots */
static LinkSet *ResizeLinkSet(LinkSet *oldLinks, int newSize)
{
   LinkSet *p;
   
   if (newSize==0){
      if (oldLinks != NULL) free(oldLinks);
      return NULL;
   }
   if ((p = (LinkSet *)realloc(oldLinks,newSize*sizeof(Link))) == NULL)
      HError(2,"Cannot resize LinkSet[%d]",newSize);
   return p;
}

/* PrModelName: print name of model p to stdout */
static void PrModelName(Link p)
{
   char *s = p->modelName->name;
   Ident name;
   
   if (s==NULL) 
      strcpy(name,"****");
   else if (strlen(s) > MAXIDENT)
      strcpy(name,"????");
   else
      strcpy(name,s);
   printf("%s[%03d] ",name,(int)p % 1000);
}

/* PrintLinkSet: print first n slots of given LinkSet to stdout */
static void PrintLinkSet(int n, LinkSet *p)
{
   int i;

   if (n==0)
      printf(" ---");
   else
      for (i=0;i<n;i++) PrModelName((*p)[i]);
   printf("\n");
}

/* RemSuccLink: remove reference to succ in node p */
static void RemSuccLink(Link succ, Link p)
{
   Link *a;
   short found = FALSE;
   int i = 0,j;
   
   a = (Link *)p->succ;
   while (i<p->numSucc && !found) {
      i++;
      found = *a++ == succ;
   }
   if (!found)
      HError(2,"Cannot find succ link to remove"); 
   p->numSucc--;
   for (j=i;j<=p->numSucc;j++,a++) 
      *(a-1) = *a;
}

/* RemPredLink: remove reference to pred in node p */
static void RemPredLink(Link pred, Link p)
{
   Link *a;
   short found = 0;
   int i = 0,j;
   
   a = (Link *)p->pred;
   while (i<p->numPred && !found) {
      i++;
      found = *a++ == pred;
   }
   if (!found)
      HError(2,"Cannot find Pred link to remove"); 
   p->numPred--;
   for (j=i;j<=p->numPred;j++,a++)
      *(a-1) = *a;
}

/* CreateNode: create node and link it into given chain */
static Link CreateNode(LabId name, Link *chain, int numSuccMax, int numPredMax)
{
   Link p;
   
   if ((p = (Link)malloc(sizeof(Node))) == NULL)
      HError(2,"Cannot allocate Node %s",name);
   p->modelName=name;
   p->numSucc=0; p->numPred=0;
   p->numPredMax=numPredMax;
   p->numSuccMax=numSuccMax;
   p->succ = CreateLinkSet(numSuccMax);
   p->pred = CreateLinkSet(numPredMax);
   p->chain = *chain; *chain=p;
   p->user = NULL;
   return p;
}

/* FreeNode:  release storage allocated to p */
static void FreeNode(Link p)
{
   if (p->succ!=NULL) free(p->succ);
   if (p->pred!=NULL) free(p->pred);
   free(p);
}

/* ShrinkNode:  reduces sizes of LinkSets to minimum */
static void ShrinkNode(Link p)
{
   SPINBALL
   p->succ = ResizeLinkSet(p->succ,p->numSucc);
   p->numSuccMax = p->numSucc;
   p->pred = ResizeLinkSet(p->pred,p->numPred);
   p->numPredMax = p->numPred;
}

/* NotJoined: returns true if a and b are not linked */
static int NotJoined(Link a, Link b)
{
   int i;
   Link *succ;
   
   succ = (Link *)a->succ;
   for (i=1; i<=a->numSucc;i++)
      if (b == *succ++) return FALSE;
   return TRUE;
}

/* JoinNodes: join node a -> node b */
static void JoinNodes(Link a, Link b)
{
   int n;
   
   if (NotJoined(a,b)) {
      if (b->numPred==b->numPredMax) {
         n = b->numPredMax+LINKCHUNKSIZE;
         b->pred = ResizeLinkSet(b->pred,n);
         b->numPredMax=n;
      }
      if (a->numSucc==a->numSuccMax) {
         n = a->numSuccMax+LINKCHUNKSIZE;
         a->succ = ResizeLinkSet(a->succ,n);
         a->numSuccMax=n;
      }
      (*b->pred)[b->numPred++]=a; 
      (*a->succ)[a->numSucc++]=b;
   }
}

/* PrintNode: print node p to stdout */
static void PrintNode(Link p)
{
   PrModelName(p); printf("\n");
   printf("  Pred: "); PrintLinkSet(p->numPred,p->pred);
   printf("  Succ: "); PrintLinkSet(p->numSucc,p->succ);
   fflush(stdout);
}
   
/* --------------------------- Network Handling ---------------------- */

/* CloneNetwork: returns a clone of given network.  Works by attaching
   a copy of each node in prototype to user slot, then replacing each
   link to node p in the copies by a link to p->user */
static Network CloneNetwork(Network prototype)
{
   Network clone;
   Link p = prototype.chain;
   Link *old,*new,q;
   int i;
   
   clone.chain=NULL; 
   SPINBALL
   while (p!=NULL) {    /* copy all nodes of prototype */
      q = CreateNode(p->modelName,&clone.chain,
                              p->numSuccMax,p->numPredMax);
      q->numPred = p->numPred;
      q->numSucc = p->numSucc;
      (Link)p->user = q; 
      p=p->chain;
   }
   SPINBALL
   p = prototype.chain; /* now fix links */
   while (p!=NULL) {
      q=(Link)p->user;
      old=(Link *)p->pred; new=(Link *)q->pred;
      for (i=1; i<=p->numPred; i++) {
         *new++ = (*old)->user; old++;
      }
      old=(Link *)p->succ; new=(Link *)q->succ;
      for (i=1; i<=p->numSucc; i++) {
         *new++ = (*old)->user; old++;
      }
      p=p->chain;
   }
   clone.entryNode = (Link)prototype.entryNode->user;
   clone.exitNode = (Link)prototype.exitNode->user;
   return clone;
}
   
/* ------------------------ Variable (SubNet) Defs ------------------- */

typedef struct _SubNetDef{
   LabId netName;
   Network network;
   struct _SubNetDef *next;
}SubNetDef;

static SubNetDef *defs;

/* InitSubNetDefs: set the list of subnet defs to empty */
static void InitSubNetDefs(void)
{
   defs = NULL;
}

/* GetSubNetDef: returns the subnet corresponding to name */
static Network *GetSubNetDef(LabId name)
{
   SubNetDef *p = defs;
   short found = FALSE;
   
   while (p!=NULL && !found) {
      found = p->netName == name;
      if (!found) p=p->next;
   }
   return found ? &p->network : NULL;
}

/* DefineSubNet: define a subnet with given name */
static void DefineSubNet(LabId name, Link entryNode, Link exitNode, Link chain)
{
   SubNetDef *p;
   
   if ((p=(SubNetDef *)malloc(sizeof(SubNetDef)))==NULL)
      HError(2,"Unable to allocate space for subnet def");
   p->netName=name;
   p->network.entryNode = entryNode;
   p->network.exitNode = exitNode;
   p->network.chain = chain;
   p->next = defs; defs = p;
}

/* FreeSubNetDefs: free all storage used by subnet definitions */
static void FreeSubNetDefs(void)
{
   SubNetDef *p;
   
   while (defs!=NULL)
   {
      FreeNetwork(&defs->network);
      p=defs->next; free(defs); defs=p;
   }
}

/* ------------ Special Triphone Loop Builder ------------- */

typedef struct{
   LabId l;
   LabId r;
   LabId m;
}SplitName;

/* SplitTriName: splits name n of form A-B+C in to 3 parts 
   if either context is missing then the corresponding part 
   is null */
static void SplitTriName(LabId n, SplitName *x)
{
   char buf[64],*p;

   x->l = x->r = NULL;
   strcpy(buf,n->name);
   if ((p=strrchr(buf,'+')) != NULL) { /* Right Context found */
      *p++ = '\0';
      x->r = GetLabId(p,TRUE);
   }
   if ((p=strchr(buf,'-')) != NULL) {  /* Left Context found */
      *p++ = '\0';
      x->m = GetLabId(p,TRUE);
      x->l = GetLabId(buf,TRUE);
   } else 
      x->m = GetLabId(buf,TRUE);
}

/* DoesMatch: returns true if name s is equal to snet or is a
              member of subnet snet */
static int DoesMatch(LabId s, LabId snet)
{
   Network *subnet;
   Link a;
   int i;

   if (snet == NULL) return TRUE;      /* No context match required */
   if ((subnet=GetSubNetDef(snet)) == NULL)  /* No subnet so use simple match */
      return s == snet;
   else {
      a = subnet->entryNode;
      while ((*a->succ)[0]->modelName == NULL)
         a = (*a->succ)[0];
      for (i=0;i<a->numSucc;i++)
         if (s == (*a->succ)[i]->modelName)
            return TRUE;
      return FALSE;
   }
}
   

/* MakeTriLoop:  builds a context dependent triphone loop around
      given expression which must be a simple list of alternatives.
      Each triphone in list must have a name of the form L-X+R 
      where L and R are either simple names or the names of
      subnetworks which again must be simple lists.  Each triphone 
      L1-X1+R1 is looped back to all phones L2-X2+R2 for which X1 
      is in L2 and X2 is in R1.  Either the left or right context
      may be missing in which case simple right or left context
      dependence is implemented.
*/
static void MakeTriLoop(Link a)
{
   int i,j;
   Link this, p;
   SplitName *trilist,*tl,*tli,*tlj;

   while ((*a->succ)[0]->modelName == NULL)
      a = (*a->succ)[0];
   if((trilist = (SplitName *)malloc(a->numSucc*sizeof(SplitName)))==NULL)
      HError(2,"Unable to allocate list of triphones");
   for (i=0;i<a->numSucc;i++){   /* Build split name list */
      tl = trilist+i;
      this = (*a->succ)[i];
      SplitTriName(this->modelName,tl);
   }
   for (i=0;i<a->numSucc;i++){   /* join models */
      SPINBALL
      tli = trilist+i;
      this = (*a->succ)[i];
      for (j=0;j<a->numSucc;j++){
         tlj = trilist+j;
         p = (*a->succ)[j];
         if (DoesMatch(tli->m,tlj->l) && DoesMatch(tlj->m,tli->r))
            JoinNodes(this,p);
      }
   }
}
      
/* ----------------------- Scanner ------------------------ */

#define ESCAPE '\\'
#define MAXLINE  255    /* max length of an input line */

enum _Symbol{NAMESYM, VARSYM, LPARSYM, RPARSYM, LBRACESYM, 
             RBRACESYM, LANGSYM, RANGSYM, LBRAKSYM, RBRAKSYM, 
             LTRISYM,RTRISYM,EQSYM, SEMISYM, BARSYM, EOFSYM};
             
             
static LabId ident;                 /* Current identifier, if any */
static char inlyne[MAXLINE+1];      /* Current input line */
static char ch;                     /* Current character */
static enum _Symbol symbol;         /* Current symbol */
static int curpos;                  /* Current position in input line */
static int curlen;                  /* Current length of input line */
static FILE *f;                     /* Input stream */

/* InitScan: initialise scanner to read from fname */
static void InitScan(char *fname)
{
   if ((f=fopen(fname,"r")) == NULL)
      HError(2,"Cannot open Network Defn file %s",fname);
   curlen=curpos=1; ch=' '; inlyne[0]=' '; inlyne[1]='\0';
}

/* PGetCh: get next input character -> ch */
static void PGetCh(void)
{
   if (curpos>=curlen) {
      if (fgets(inlyne,MAXLINE,f)==NULL) {
         ch=EOF; return;
      }
      curpos=0; curlen=strlen(inlyne);
      if (inlyne[curlen-1] < ' ')
         inlyne[curlen-1]=' ';   /* change newline,if any to space */
      else {
         inlyne[curlen++]=' ';      /* otherwise append space */
         inlyne[curlen]='\0';
      }
   }
   ch = inlyne[curpos++];
}

/* PGetIdent: get identifer -> ident */
static void PGetIdent(void)
{
   int i=0;
   Ident id;
   
   SPINBALL
   do {
      if (ch==ESCAPE) PGetCh();
      if (i<MAXIDENT) id[i++]=ch;
      PGetCh();
   } while ( ch!=' ' && ch!='{' && ch!='}' && ch!='[' && ch!=']' &&
             ch!='<' && ch!='>' && ch!='(' && ch!=')' && ch!='=' && 
             ch!=';' && ch!='|' && ch!='/');
   id[i]='\0';
   ident = GetLabId(id,TRUE);
}

/* PGetSym: get next symbol -> symbol */
static void PGetSym(void)
{
   while (isspace(ch) || (ch=='/' && inlyne[curpos]=='*') ) {
      if (isspace(ch))  /* skip space */
         PGetCh();
      else {            /* skip comment */
         PGetCh(); PGetCh();
         while (!(ch=='*' && inlyne[curpos]=='/')) PGetCh();
         PGetCh(); PGetCh();        
      }
   }
   switch (ch) {
      case '$': PGetCh(); PGetIdent(); symbol=VARSYM; break;
      case '(': PGetCh(); symbol=LPARSYM; break;
      case ')': PGetCh(); symbol=RPARSYM; break;
      case '<': PGetCh(); 
                if (ch=='<') {symbol=LTRISYM; PGetCh();} else
                  symbol=LANGSYM; break;
      case '>': PGetCh();  
                if (ch=='>') {symbol=RTRISYM; PGetCh();} else
                  symbol=RANGSYM; break;
      case '{': PGetCh(); symbol=LBRACESYM; break;
      case '}': PGetCh(); symbol=RBRACESYM; break;
      case '[': PGetCh(); symbol=LBRAKSYM; break;
      case ']': PGetCh(); symbol=RBRAKSYM; break;
      case '=': PGetCh(); symbol=EQSYM; break;
      case ';': PGetCh(); symbol=SEMISYM; break;
      case '|': PGetCh(); symbol=BARSYM; break;
      case EOF: symbol=EOFSYM; break;
      default:  PGetIdent(); symbol=NAMESYM; break;
   }
}

/* ------------------------- Errors ------------------------ */

/* ParseError: Print a parser error message then die */
static void ParseError(int errn)
{
   int i;
   
   fprintf(stderr,"%s\n",inlyne);
   for (i=1; i<curpos-1; i++) fputc(' ',stderr);
   fprintf(stderr, "^  *** Error %d\n",errn);
   switch(errn) {
      case 1: HError(1,"Unexpected EOF");
      case 2: HError(1,"Garbage at end of file");
      case 3: HError(1,"Factor expected");
      case 4: HError(1,"variable expected");
      case 5: HError(1,"= expected");
      case 6: HError(1,"; expected");
      case 7: HError(1,") expected");
      case 8: HError(1,"] expected");
      case 9: HError(1,"} expected");
      case 10: HError(1,"> expected");
      case 11: HError(1,">> expected");
      default: HError(1,"Unknown error number!!!");
   }
}

/* --------------------------- Parser -------------------------------- */

/*
   Each parser routine has two parameters: hd and tl.  As each
   routine completes successfully, it uses these parameters to
   return the head and tail of a subnetwork representing the
   structure just parsed.
*/

Link curChain;       /* current network being built */
LabId curId;         /* name of current variable def, if any */

static void PExpr(Link *hd, Link *tl);

/* PModel: parse a model name and create a node for it */
static void PModel(Link *hd, Link *tl)
{
   Link p;
   
   p=CreateNode(ident,&curChain,LINKCHUNKSIZE,LINKCHUNKSIZE);
   *hd = *tl = p;
   PGetSym();
}

/* PVariable: parse a variable name and return a clone of its def */
static void PVariable(Link *hd, Link *tl)
{
   Network *proto, clone;
   Link p;
   
   if((proto=GetSubNetDef(ident))==NULL)
      HError(2,"Variable %s is undefined",ident);
   clone=CloneNetwork(*proto);
   *hd=clone.entryNode; *tl=clone.exitNode;
   p=clone.chain;    /* join clone to current chain */
   while (p->chain!=NULL) p=p->chain;
   p->chain=curChain; curChain=clone.chain;
   PGetSym();
}

/* PGroup: parse ( .. ) */
static void PGroup(Link *hd, Link *tl)
{
   PGetSym();
   PExpr(hd,tl);
   if (symbol!=RPARSYM) ParseError(7);
   PGetSym();
}


/* POption: parse [ .. ] */
static void POption(Link *hd, Link *tl)
{
   Link eTail;
   
   PGetSym();
   PExpr(hd, &eTail);
   *tl = CreateNode(NULL,&curChain,LINKCHUNKSIZE,LINKCHUNKSIZE);
   JoinNodes(*hd,*tl);
   JoinNodes(eTail,*tl);
   if (symbol!=RBRAKSYM) ParseError(8);
   PGetSym();
}

/* PRepetition0: parse { .. } */
static void PRepetition0(Link *hd, Link *tl)
{
   Link eHead, eTail;
   
   PGetSym();
   PExpr(&eHead, &eTail);
   *hd = *tl = CreateNode(NULL,&curChain,LINKCHUNKSIZE,LINKCHUNKSIZE);
   JoinNodes(*tl,eHead);
   JoinNodes(eTail,*tl);
   if (symbol!=RBRACESYM) ParseError(9);
   PGetSym();
}

/* PRepetition1: parse < .. > */
static void PRepetition1(Link *hd, Link *tl)
{
   PGetSym();
   PExpr(hd, tl);
   JoinNodes(*tl,*hd);
   if (symbol!=RANGSYM) ParseError(10);
   PGetSym();
}

/* PTriloop: parse << .. >> */
static void PTriloop(Link *hd, Link *tl)
{
   PGetSym();
   PExpr(hd, tl);
   MakeTriLoop(*hd);
   if (symbol!=RTRISYM) ParseError(11);
   PGetSym();
}

/* PFactor: parse a factor */
static void PFactor(Link *hd, Link *tl)
{
   switch(symbol) {
      case NAMESYM:     PModel(hd,tl);       break;
      case VARSYM:      PVariable(hd,tl);    break;
      case LPARSYM:     PGroup(hd,tl);       break;
      case LBRAKSYM:    POption(hd,tl);      break;
      case LBRACESYM:   PRepetition0(hd,tl); break;
      case LANGSYM:     PRepetition1(hd,tl); break;
      case LTRISYM:     PTriloop(hd,tl);     break;
      default: ParseError(3);
   }
}

/* PSequence: parse a sequence of factors */
static void PSequence(Link *hd, Link *tl)
{
   Link hd2, tl2;
   
   PFactor(hd,tl);
   while (symbol==NAMESYM || symbol==VARSYM || symbol==LBRAKSYM || 
          symbol==LBRACESYM || symbol==LPARSYM || 
          symbol==LTRISYM || symbol==LANGSYM) {
      PFactor(&hd2, &tl2);
      JoinNodes(*tl,hd2);
      *tl=tl2;
   }
}

/* PExpr: parse an expression */
static void PExpr(Link *hd, Link *tl)
{
   Link hd2, tl2;
   
   PSequence(hd,tl);
   if (1 || symbol==BARSYM){     /* need some glue */
      hd2 = *hd, tl2 = *tl;
      *hd=CreateNode(NULL,&curChain,LINKCHUNKSIZE,LINKCHUNKSIZE);
      *tl=CreateNode(NULL,&curChain,LINKCHUNKSIZE,LINKCHUNKSIZE);
      JoinNodes(*hd,hd2);
      JoinNodes(tl2,*tl);
   }
   while (symbol==BARSYM) {
      PGetSym();
      PSequence(&hd2, &tl2);
      JoinNodes(*hd,hd2);
      JoinNodes(tl2,*tl);
   }
}

/* PSubNet: parse a subnet def and store it */
static void PSubNet()
{
   Link hd,tl;
   
   if (symbol != VARSYM) ParseError(4);
   curId=ident;
   PGetSym();
   if (symbol != EQSYM) ParseError(5);
   PGetSym();
   curChain=NULL;
   PExpr(&hd,&tl);
   if (symbol != SEMISYM) ParseError(6);
   PGetSym();
   DefineSubNet(curId,hd,tl,curChain);
}

/* PNetwork: parse a complete network definition */
static void PNetwork(Link *hd, Link *tl)
{
   Link entryNode,exitNode;
   
   InitSubNetDefs();
   while (symbol != LPARSYM) {
      PSubNet();
   }
   PGetSym();
   curId=NULL; curChain=NULL;
   entryNode = CreateNode(GetLabId("ENTER",TRUE),&curChain,LINKCHUNKSIZE,1);
   exitNode = CreateNode(GetLabId("EXIT",TRUE),&curChain,1,LINKCHUNKSIZE); 
   PExpr(hd,tl);
   JoinNodes(entryNode,*hd);
   JoinNodes(*tl,exitNode);
   *hd=entryNode; *tl=exitNode;
   if (symbol != RPARSYM) ParseError(7);
   PGetSym();
   if (symbol!=EOFSYM) ParseError(2);
   FreeSubNetDefs();
}

/* ------------------------ Clean Up Network ------------------------- */

/* RemoveGlue: remove all glue nodes from the given network */
static void RemoveGlue(Network *net)
{
   Link p,q;
   Link *a,*b;
   int i,j;
   
   p=net->chain; net->chain=NULL; 
   while (p!=NULL) {
      if (p->modelName==NULL) {     /* then its a glue node */
         SPINBALL
         a=(Link *)p->pred;      
         for (i=1;i<=p->numPred;i++){ /* delete refs to this node */
            RemSuccLink(p,*a);
            a++;
         }
         b=(Link *)p->succ;
         for (i=1;i<=p->numSucc;i++) {
            RemPredLink(p,*b);
            b++;
         }
         a=(Link *)p->pred;         /* bridge all the links */
         for (i=1;i<=p->numPred;i++) {
            b=(Link *)p->succ;
            for (j=1;j<=p->numSucc;j++) {
               JoinNodes(*a,*b++);
            }
            a++;
         }
         q = p; p = p->chain;
         FreeNode(q);               /* delete the node */
      }
      else {                     /* not glue so add to new chain */
         q = p->chain;
         p->chain = net->chain;
         net->chain = p;
         p = q;
      }
   }
}

/* ReSizeNodes: scan the network and shrink all of its nodes */
static void ReSizeNodes(Network *net)
{
   Link p;
   
   p=net->chain;
   while (p!=NULL) {
      ShrinkNode(p);
      p=p->chain;
   }
}

/* --------------------------- Interface ----------------------------- */

/* EXPORT-> ParseNetwork: parse and build a network */
Network ParseNetwork(char *fname)
{
   Link hd,tl;
   Network theNet;
   
   InitScan(fname);
   PGetSym(); 
   PNetwork(&hd,&tl);
   theNet.entryNode = hd;
   theNet.exitNode = tl;
   theNet.chain = curChain;
   RemoveGlue(&theNet);
   ReSizeNodes(&theNet);
   return theNet;
}


/* EXPORT-> FreeNetwork: free storage allocated for network */
void FreeNetwork(Network *network)
{
   Link p = network->chain;
   Link q;

   while (p!=NULL) {
      q = p->chain;
      FreeNode(p); 
      p = q;
   }
}

/* EXPORT-> PrintNetwork: print network to stdout */
void PrintNetwork(Network *network)
{
   Link p = network->chain;
   
   printf("\n");
   while (p!=NULL) {
      PrintNode(p);
      p=p->chain;
   }
}


/* ------------------------ End of HParse.c ------------------------- */
