/* ----------------------------------------------------------- */
/*                   HTK  -  HMM Toolkit                       */
/* ----------------------------------------------------------- */
/*         Copyright: Steve Young      October 1989            */
/*         Last Update:                September 1990          */
/*                                                             */
/*         Cambridge University Engineering Department         */
/*         Trumpington Street, Cambridge, CB2 1PZ              */
/*                                                             */
/* Permission is hereby granted to copy, use and modify this   */
/* software provided that this banner notice is not removed    */
/*                                                             */
/* ----------------------------------------------------------- */
/*         File: HShell.h:   Interface to the Shell            */
/* ----------------------------------------------------------- */

#include <stdio.h>         /* Standard Libraries */
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <ctype.h>
#include <math.h>

#include "HShell.h"

#ifdef MPW
   #pragma segment hmods
   #include <CursorCtl.h>  /* MPW Libraries */
   /* Needed for Spin Cursor functions */
#endif

/* ------------------ Argument Processing -------------------- */

/*
   Arguments are processed in sequence by calls to Getxxx().
   The enum type ArgKind is used to indicate the kind of the next
   argument 
  
     SWITCHARG - any str starting with '-' char
     STRINGARG - any str starting with a letter 
     INTARG    - any str starting with a digit not containing a dec pt
     FLOATARG  - any str starting with a digit and not an INTARG
     NOARG     - returned when all arguments processed
     
   Each of the Getxxx() functions will return value of next arg in 
   form specified by xxx, an error is reported if next arg is not of 
   the required type except that an int will be converted to a float 
   if requested.
   
   If a script file is set, then the contents of the script file
   are effectively appended to the command line.  However, the
   actual strings returned for args in the script are volatile and
   should be copied by the host program where necessary.

*/

static int argcount;      /* total args = argc */
static int nextarg=1;     /* next arg to return in GetxxxArg */
static char **arglist;    /* actual arg list = argv */
static FILE *script = NULL;   /* script file if any */
static int scriptcount = 0;   /* num words in script */
static char scriptBuf[132];   /* buffer for current script arg */
static char scriptBufLoaded = FALSE;

/* EXPORT-> InitShell:   Called by main to initialise the module. 
   Main also passes over the command line arguments */
void InitShell(int argc, char *argv[])
{
   argcount = argc;
   arglist = argv;
#ifdef MPW
   InitCursorCtl(NULL);
#endif
}

/* ScriptWord: return next word from script */
static char * ScriptWord(void)
{
   char ch;
   int i;
   
   i=0; ch=' ';
   while (isspace(ch)) ch = fgetc(script);
   if (ch==EOF) return NULL;
   if (ch=='\''){
      ch = fgetc(script);
      while (ch != '\'' && ch != EOF) {
         scriptBuf[i++] = ch; 
         ch = fgetc(script);
      }
      if (ch==EOF)
         HError(1,"Closing quote missing in script file");
   } else
      do {
         scriptBuf[i++] = ch; 
         ch = fgetc(script);
      }while (!isspace(ch) && ch != EOF);
   scriptBuf[i] = '\0';
   scriptBufLoaded = TRUE;
   return scriptBuf;
}
         
/* GetNextArg: from either command line or script file */
static char * GetNextArg(int step)
{
   char *s;
   
   if (argcount>nextarg) 
      s = arglist[nextarg];
   else if (scriptBufLoaded)
      s = scriptBuf;
   else
      s = ScriptWord();
   if (step) {
      ++nextarg;
      scriptBufLoaded = FALSE;
   }
   return s;
}

/* EXPORT-> NumArgs: Number of actual cmd line arguments left unprocessed */
int  NumArgs(void)
{
   return argcount+scriptcount-nextarg;
}

/* EXPORT-> NextArg: Kind of next argument */
ArgKind NextArg(void)
{
   char *s;
   
   if (NumArgs() == 0) return NOARG;
   s = GetNextArg(0);
   if (*s=='-') 
      return SWITCHARG;
   if (isdigit(*s) ){
      if (strchr(s,'.') == NULL)
         return INTARG;
      else
         return FLOATARG;
   }
   return STRINGARG;

}

/* ArgError: called when next arg is different to requested arg 
   or no more args are left */
static void ArgError(char *s)
{
   if (NextArg()==NOARG)
      HError(1,"%s arg requested but no args left",s);
   else
      HError(1,"%s arg requested but next is %s",s,GetNextArg(0));
}

/* EXPORT-> GetStrArg: get string arg */
char * GetStrArg(void)
{
   if (NextArg() != STRINGARG) 
      ArgError("String");
   return GetNextArg(1);
}
   
/* EXPORT-> GetSwtArg: get switch arg */
char * GetSwtArg(void)
{
   if (NextArg() != SWITCHARG) 
      ArgError("Switch");
   return GetNextArg(1)+1;
}

/* EXPORT-> GetIntArg: get int arg */
int    GetIntArg(void)
{
   if (NextArg() != INTARG) 
      ArgError("Int");
   return atoi(GetNextArg(1));
}

/* EXPORT-> GetFltArg: get float arg */
float  GetFltArg(void)
{
   int k = NextArg();
   
   if (k == INTARG || k==FLOATARG) 
      return atof(GetNextArg(1));
   ArgError("Float");
}

/* EXPORT->SetScriptFile: open script file and count words in it */
void SetScriptFile(void)
{
   char *fn;
   
   if (script != NULL)
      HError(1,"Script file is already open");
   if (NextArg() != STRINGARG)
      HError(1,"Script file name expected");
   fn = GetStrArg();
   if ((script = fopen(fn,"r")) == NULL)
      HError(2,"Cannot open script file %s",fn);
   while (ScriptWord() != NULL) ++scriptcount;
   rewind(script);
   scriptBufLoaded = FALSE;
}

/* -------------------- Error Reporting -------------------- */

/*
  Prints toolname and then message to stderr.  Form of message and any
  subsequent arguments is similar to 'printf' with just %d %f %e %s being
  recognised.  If status<>0 then tool is aborted and status returned to OS,
  return status is also printed in error message in this case.
*/

/* EXPORT-> HError: print error message on stderr and abort if status<>0 */
void HError(int status, char *message, ...)
{
   va_list ap;             /* Pointer to unnamed args */
   char *p;                /* Message String */
   double dval;            /* Message args */
   int ival;
   char *sval;
   
   fflush(stdout);        /* Flush any pending output */
   va_start(ap,message);
   fprintf(stderr,"%s: ERROR ",arglist[0]);
   for (p=message;*p;p++){
      if (*p != '%'){
         fputc(*p,stderr);
         continue;
      }
      switch(*++p){
      case 'd':
         ival = va_arg(ap,int);
         fprintf(stderr,"%d",ival);
         break;
      case 'f':
         dval = va_arg(ap,double);
         fprintf(stderr,"%f",dval);
         break;
      case 'e':
         dval = va_arg(ap,double);
         fprintf(stderr,"%e",dval);
         break;      
      case 's':
         sval = va_arg(ap,char *);
         fprintf(stderr,"%s",sval);
         break;
      }
   }
   va_end(ap);
   if (status>0) {
      fprintf(stderr," ABORTED [%d]\n",status);
      exit(status);
   }
   fprintf(stderr,"\n");
}

/* -------------------- File Name Handling ------------------- */

/*
   Given a filename of the general form "path/n.x" in fn, the following
   functions return "n.x", "n", "x" and "path", respectively. In each case,
   the string is returned in s which must be large enough and s is returned
   as the function result.
*/

/* EXPORT-> NameOf: name of fn */
char * NameOf(char *fn, char *s)
{
   char *t;
   
   t = strrchr(fn,PATHCHAR);
   if (t == NULL) 
      t = fn;
   else 
      t++;
   return strcpy(s,t);
}

/* EXPORT-> BaseOf: base name part of fn */
char * BaseOf(char *fn, char *s)
{
   char *t;
   
   NameOf(fn,s);
   t = strrchr(s,'.');
   if (t == NULL) return s;
   if (t>=s) *t = '\0';
   return s;
}

/* EXPORT-> PathOf: path part of fn */
char * PathOf(char *fn, char *s)
{
   char *t;
   
   strcpy(s,fn);
   t = strrchr(s,PATHCHAR);
   if (t == NULL) 
      *s='\0';
   else
      *++t = '\0';
   return s;
}

/* EXPORT-> ExtnOf: extension part of fn */
char * ExtnOf(char *fn, char *s)
{
   char *t,buf[100];
   
   NameOf(fn,buf);
   t = strrchr(buf,'.');
   if (t == NULL){
      *s ='\0'; return s;
   }
   return strcpy(s,t+1);
}

/* EXPORT-> MakeFN: construct a filename from fn */
char * MakeFN(char *fn, char *path, char *ext, char *s)
{
   char newPath[255], base[64], newExt[20];  /* components of new fn */
   int i;
   
   BaseOf(fn,base);
   if (path==NULL) 
      PathOf(fn,newPath);
   else
      strcpy(newPath,path);
   if (ext==NULL)
      ExtnOf(fn,newExt);
   else 
      strcpy(newExt,ext);
   i = strlen(newPath);
   if (i>0 && newPath[i-1] != PATHCHAR) {
      newPath[i] = PATHCHAR;
      newPath[i+1] = '\0';
   }
   strcpy(s,newPath);
   strcat(s,base);
   if (strlen(newExt) > 0) {
      strcat(s,".");
      strcat(s,newExt);
   }
   return s;
}

/* EXPORT-> NewString: return a string of given size */   
char *NewString(int size)
{
   char *s;
   if ((s=(char *)malloc(size+1)) == NULL)
      HError(2,"Unable to allocate string");
   *s='\0'; /* make it empty */
   return s;
}


/* ----------------------------------------------------------- */
/*                      END:  HShell.c                         */
/* ----------------------------------------------------------- */
