/* ----------------------------------------------------------- */
/*                   HTK  -  HMM Toolkit                       */
/* ----------------------------------------------------------- */
/*        Copyright: Steve Young      October 1989             */
/*        Last Update:                September 1990           */
/*                                                             */
/*        Cambridge University Engineering Department          */
/*        Trumpington Street, Cambridge, CB2 1PZ               */
/*                                                             */
/*  Permission is hereby granted to copy, use and modify this  */
/*  software provided that this banner notice is not removed   */
/*                                                             */
/* ----------------------------------------------------------- */
/*      File: HCode.c: Encode a Speech File as LPC Parameters  */
/* ----------------------------------------------------------- */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include "HShell.h"
#include "HMath.h"
#include "HSpIO.h"
#include "HSigP.h"

#define VERSION "1.2"

/*
   Data is stored as a contiguous sequence of parameter vectors.
   Qualifier data is added to end of each parameter vector in the
   order Energy, Pitch.
*/

#define MINSAMPPERIOD 50         /* min sample period in usecs */
#define MAXDELWIN  20            /* max +/-limit of delta window */
#define MAXCEPLIFT 1000          /* max value of cep liftering coef */

static int   fOrder  = 12;       /* lp filter order */
static float fPeriod = 10.0;     /* frame period in msecs */
static float fDur   = 25.6;      /* frame size in msecs */
static float preEmp  = 0.0;      /* pre-emphasis coef */
static int   cepLift = 0;        /* cepstral liftering coef */
static int   delWin  = 2;        /* +/-limit of delta window */
static int   hamSet  = 0;        /* Enable Hamming windowing */
static SampKind lpKind = LPC;    /* Kind of out file to create */
static Boolean withE = FALSE;    /* Flag to append log energy */
static float escale = 0.1;       /* Scale factor for log energy */
static Boolean withP = FALSE;    /* Flag to append f0 estimate */
static int   trace   = 0;        /* trace level */
static blipCount = 0;            /* used by Blip */

/* global data */

static int frameSize;   /* num samples per frame */
static int numComps;    /* total number of components per sample */
static int frameRate;   /* frame period in samples */
static int sampSize;    /* number of bytes per sample */
static int totalFrames; /* in data */


/* ---------------- Process Command Line ------------------------- */

void ReportUsage(void)
{
   fprintf(stderr,"\nUSAGE: HCode[V%s] [options] srcfile lpcfile\n\n",VERSION);
   fprintf(stderr," Option                                   Default\n\n");
   fprintf(stderr," -c      Output LP Cepstral Coef             LPC\n");
   fprintf(stderr," -d      Output LP Cepstra+Delta Coef        LPC\n");
   fprintf(stderr," -e      Append Log Energy                   Off\n");
   fprintf(stderr," -f T    Set frame period to T (msecs)      10.0\n");
   fprintf(stderr," -h      Apply Hamming Window                Off\n");
   fprintf(stderr," -i      Output LP Ref Coef in Short form    LPC\n");
   fprintf(stderr," -k X    Set PreEmphasis Coef to X           0.0\n");
   fprintf(stderr," -l N    Set Cepstral Liftering Coef to N    0\n");
   fprintf(stderr," -p N    Set filter order to N              12\n");
   fprintf(stderr," -q N    Set Delta window to -N..+N          2\n");
   fprintf(stderr," -r      Output LP Reflection Coef           LPC\n");
   fprintf(stderr," -s      Scale factor for Log Energy         0.1\n");
   fprintf(stderr," -v      Append f0 estimate                  Off\n");
   fprintf(stderr," -w T    Set window duration to T (msecs)   25.6\n");
   fprintf(stderr," -T N    Set trace level to N                0\n");
   fprintf(stderr," -F fmt  Set src file format to fmt          HTK\n");
   exit(1);
}
   
int GetChkedInt(int min, int max, char * swtname)
{
   int val;
   
   if (NextArg() != INTARG) 
      HError(1,"Integer Arg Required for %s option",swtname);
   val=GetIntArg();
   if (val<min || val>max) 
      HError(1,"Integer arg out of range in %s option",swtname);
   return val;
}

float GetChkedFlt(float min, float max, char * swtname)
{
   float val;
   
   if (NextArg() != INTARG && NextArg() != FLOATARG) 
      HError(1,"Float Arg Required for %s option",swtname);
   val=GetFltArg();
   if (val<min || val>max) 
      HError(1,"Float arg out of range in %s option",swtname);
   return val;
}

main(int argc, char *argv[])
{
   char * s, *srcfn, *tgtfn;
   char temp[20];
   void Convert(char *src, char *tgt);
   FileFormat ff = 255;    /* ie undefined */
   
   InitShell(argc,argv);
   InitMath(FALSE);
   if (NumArgs()<2)
      ReportUsage();
   while (NextArg() == SWITCHARG) {
      s = GetSwtArg();
      if (strlen(s)!=1) 
         HError(1,"Bad switch %s; must be single letter",s);
      switch(s[0]){
         case 'p':
            fOrder = GetChkedInt(1, MAXFEATURES / 2,s); break;
         case 'f':
            fPeriod = GetChkedFlt(0.1,100.0,s); break;
         case 'w':
            fDur = GetChkedFlt(0.1,MAXFRAMESIZE*MINSAMPPERIOD/1000,s); break;
         case 'k':
            preEmp = GetChkedFlt(0.0,1.0,s); break;
         case 'h':
            hamSet = TRUE; break;
         case 'r':
            lpKind = LPREFC; break;
         case 's':
            escale = GetChkedFlt(0.0,10.0,s); break;
         case 'c':
            lpKind = LPCEPSTRA; break;
         case 'd':
            lpKind = LPDELCEP; break;
         case 'e':
            withE = TRUE; break;
         case 'v':
            withP = TRUE; break;
         case 'i':
            lpKind = IREFC; break;
         case 'l':
            cepLift = GetChkedInt(0,MAXCEPLIFT,s); break;
         case 'q':
            delWin = GetChkedInt(1,MAXDELWIN,s); break;
         case 'F':
            if (NextArg() != STRINGARG)
               HError(1,"File format expected");
            if((ff = Str2Format(GetStrArg())) == ALIEN)
               HError(0,"Warning ALIEN src file format set");
            SetFormat(ff);
            break;
         case 'T': 
            trace = GetChkedInt(0,3,s); break;
         default:
            HError(1,"Unknown switch %s",s);
      }
   }
   if (withE) lpKind |= HASENERGY;
   if (withP) lpKind |= HASPITCH;
   if (NextArg()!=STRINGARG)
      HError(1,"Source file name expected");
   srcfn = GetStrArg();
   if (NextArg()!=STRINGARG) 
      HError(1,"Target file name expected");
   tgtfn = GetStrArg();
   if (NumArgs() != 0)
      HError(0,"Extra arguments left unprocessed!");
   if (trace == 3) {
      printf("HCode: %s --> %s\n",srcfn,tgtfn);
      printf(" Filter Order  : %d\n",fOrder);
      printf(" Frame Size    : %.2f\n",fPeriod);
      printf(" Frame Duration: %.2f\n",fDur);
      printf(" PreEmphasis   : %.4f\n",preEmp);
      printf(" Cep Liftering : %d\n",cepLift);
      printf(" Delta Window  : %d\n",delWin);
      printf(" Hamming Window: %s\n",hamSet?"on":"off");
      printf(" Sample Kind   : %s\n",SampKind2Str(lpKind,temp));
      printf(" Src FileFormat: %s\n",
               ff > ALIEN ? "undef" : Format2Str(ff,temp));
      printf(" Trace Level   : %d\n",trace); 
      fflush(stdout);
   }
   Convert(srcfn,tgtfn);
   exit(0);
}
/* ------------------------- Tracing ------------------------- */

/* PrintVector: print vector v labelled as frame n */
void PrintVector(int n, Vector v)
{
   int i,j,imax;
   
   printf("Frame %d:\n",n);
   imax=VectorSize(v); j=0;
   for (i=1;i<=imax;i++) {
      printf("%4d. %8.4f",i,v[i]);
      if (++j == 6){
         printf("\n"); j=0;
      }
   }
   if (j>0) printf("\n");
   fflush(stdout);
}

/* Blip: print a dot on output device */
void Blip()
{
   printf("."); 
   if (++blipCount==64) {
      blipCount=0; printf("\n");
   }
   fflush(stdout);
}
         
/* -------------------- Convert Float Parms to Shorts -------- */

/* Cvt2Short: convert all floats in data to shorts (for IREFC Format) */
void Cvt2Short(DataFile *data)
{
   int i,count;
   short *sp;
   float *fp;
   
   count = data->dataSize / sizeof(float);
   sp = (short *)data->data; 
   fp = (float *)data->data;
   for (i=0;i<count;i++)
      *sp++ = *fp++ * 16368.0;
   data->sampSize = data->sampSize*sizeof(short) / sizeof(float);
   data->dataSize = count * sizeof(short);
}

/* -------------------- Delta Computations ------------------- */
      
/* AddDeltaCoef: Rescan target and calculate delta coefficients */
void AddDeltaCoef(DataFile *tgt, int nFrames, int n2)
{
   int i,lasti,j,t,n;
   float *fp1,*fp2, *back, *forw;
   float sum, sigmaT2;
   
   n = n2 / 2;                      /* n = num data components */
   sigmaT2 = 0.0;
   for (t=1;t<=delWin;t++)
      sigmaT2 += t*t;
   sigmaT2 *= 2.0;
   fp1 = (float *)tgt->data;              /* fp1 will step thru ceps */
   fp2 = fp1 + n;                         /* fp2 will step thru dels */
   lasti = nFrames - delWin;
   for (i=0;i<delWin;i++) {               /* leading edge - use simple */
      for (j=0;j<n;j++){                  /* first order difference */
         *fp2 = *(fp1+n2) - *fp1;
         ++fp1; ++fp2;
      }
      fp1 += n; fp2 += n;
   }
   for (i=delWin;i<lasti;i++){            /* Full regression */
      for (j=0;j<n;j++){
         back = forw = fp1; sum = 0.0;
         for (t=1;t<=delWin;t++) {
            back -= n2; forw += n2;
            sum += t * (*forw - *back);
         }
         *fp2 = sum / sigmaT2;
         ++fp1; ++fp2;
      }
      fp1 += n; fp2 +=n;
   }
   for (i=lasti;i<nFrames;i++) {    /* trailing edge - use simple */
      for (j=0;j<n;j++){            /* first order difference */
         *fp2 = *fp1 - *(fp1-n2);
         ++fp1; ++fp2;
      }
      fp1 += n; fp2 += n;
   }
}

/* ------------------ Fundamental Freq Estimation ------------- */

float F0(void)
{
   HError(1,"Sorry - F0 Estimation not implemented");
   return 0.0;
}

/* ------------------ Log Energy Normalisation ---------------- */

/* NormaliseLogEnergy: normalise energy and then convert to log
         in range 0.0 .. 1.0*/
void NormaliseLogEnergy(DataFile *data)
{
   float *p,max;
   int i;
   Boolean shortRep;    /* will be converted to a short */
   
   shortRep = BaseSampKind(lpKind);
   /* find max log energy */
   SPINBALL
   p = (float*)data->data + fOrder;    /* p -> first energy value */
   max = *p;
   for (i=1;i<totalFrames;i++){
      p += numComps;                   /* step p to next e val */
      if (*p > max) max = *p;
   }
   if (trace>1)
      printf("Log Energy Norm: Max = %.1f\n",max);
      
   /* normalise */
   SPINBALL
   p = (float*)data->data + fOrder;    /* p -> first energy value */
   for (i=0;i<totalFrames;i++){
      *p = (*p - max) * escale + 1.0;
      if (shortRep && *p < -1.0) {
         *p = -1.0;                 /* clip if short rep */
         HError(0,"log energy clipped");
      }
      if (trace==3)
         printf("%d. Normalised Log Energy = %.3f\n",i,*p);
      p += numComps; 
   }
}

/* ------------------ Main conversion routines ---------------- */

/* CvtInit: calculate main conversion parameters */
void CvtInit(DataFile *src)
{
   frameSize = fDur * 10000.0 / src->sampPeriod;
   if (src->nSamples < frameSize)
      HError(2,"Source file too short");
   frameRate = fPeriod * 10000.0 / src->sampPeriod;
   totalFrames = src->nSamples / frameRate;
   numComps = fOrder;
   if (withP) ++numComps;
   if (withE) ++numComps;
   if (BaseSampKind(lpKind) == LPDELCEP) numComps *= 2;
   sampSize = sizeof(float) * numComps;
}

/* GetFrame: Read samples st,..,en from f into s.  If en lies beyond
   end of f then s is padded with zeroes */
void GetFrame(DataFile *f, long st, long en, Vector s)
{
   long i,fend;
   int j=0;
   short sample;
   
   fend = (en>=f->nSamples) ? f->nSamples-1 : en;
   for (i=st;i<=fend;i++){
      GetSample(f,i,&sample);
      s[++j] = sample;
   }
   if (en == fend) return;
   for (i=fend+1;i<=en;i++)
      s[++j] = 0.0;
}

/* Convert: top level conversion routine */
void Convert(char *srcfname, char *tgtfname)
{
   Vector a,k,c,s;
   Vector p;         /* actual output vector */
   DataFile src,tgt;
   int thisFrame,i;
   int lpBytes;
   long sampIdx=0;   /* Index of current source sample */
   float re,te;      /* returned by Wave2LPC but not used */
   
   if (trace >0) printf("Starting Conversion\n");
   SpOpen(srcfname,&src);
   if (src.sampKind != WAVEFORM)
      HError(2,"Source must be a WAVEFORM file");
   CvtInit(&src);
   lpBytes = fOrder*sizeof(float);
   SpCreate(fPeriod*10000,sampSize,lpKind,&tgt);
   a=CreateVector(fOrder); k=CreateVector(fOrder);
   s=CreateVector(frameSize); c=CreateVector(fOrder); 
   p=CreateVector(numComps);
   
   for (thisFrame=0;thisFrame<totalFrames; thisFrame++) { /* main loop */
      GetFrame(&src,sampIdx,sampIdx+frameSize-1,s);
      sampIdx += frameRate;
      if (preEmp>0.0)
         PreEmphasise(s,preEmp);
      if (hamSet)
         Ham(s);
      Wave2LPC(s,a,k,&re,&te);
      switch (BaseSampKind(lpKind)) {
         case LPC:
            memcpy(p+1,a+1,lpBytes);
            if (trace>1) PrintVector(thisFrame,a);
            break;
         case IREFC:
         case LPREFC:
            memcpy(p+1,k+1,lpBytes);
            if (trace>1) PrintVector(thisFrame,k);
            break;
         case LPCEPSTRA:
         case LPDELCEP:
            LPC2Cepstrum(a,c);
            if (cepLift > 0)
               WeightCepstrum(c,1,fOrder,cepLift);
            memcpy(p+1,c+1,lpBytes);
            if (trace>1 && lpKind != LPDELCEP )
               PrintVector(thisFrame,c);
            break;
      }
      i = fOrder;
      if (withE) {
         p[++i] = (te<MINLARG) ? LZERO : log(te);
         if (trace>1) printf("   Frame Log Energy = %.1f\n",p[i]);
      }
      if (withP) {
         p[++i] = F0();
         if (trace>1) printf("   F0 Estimate = %.3f\n",p[i]);
      }
      PutSample(&tgt,thisFrame,p+1);
      if (trace==1) Blip();
      SPINBALL
   }
   if (withE) {
      NormaliseLogEnergy(&tgt);
   }
   if (BaseSampKind(lpKind) == LPDELCEP) {
      if (trace>0 ) {
         blipCount = 0;
         printf("\nAdding delta coefficients \n");
         fflush(stdout);
      }
      AddDeltaCoef(&tgt,totalFrames,numComps);
   } else if (BaseSampKind(lpKind) == IREFC) {
      if (trace >0 ) {
         printf("\nConverting to short form \n");
         fflush(stdout);
      }
      Cvt2Short(&tgt);
   }
   SpSave(tgtfname,&tgt);
   SpClose(&src); SpClose(&tgt);
   FreeVector(a); FreeVector(k);
   FreeVector(c); FreeVector(s);
   FreeVector(p);
   if (trace>0) 
      printf("\nDone\n");
}

/* ----------------------------------------------------------- */
/*                      END:  HCode.c                          */
/* ----------------------------------------------------------- */
