/* ----------------------------------------------------------- */
/*                   HTK  -  HMM Toolkit                       */
/* ----------------------------------------------------------- */
/*         Copyright: Steve Young      December 1989           */
/*         Last Update:                September 1990          */
/*                                                             */
/*         Cambridge University Engineering Department         */
/*         Trumpington Street, Cambridge, CB2 1PZ              */
/*                                                             */
/* Permission is hereby granted to copy, use and modify this   */
/* software provided that this banner notice is not removed    */
/*                                                             */
/* ----------------------------------------------------------- */
/*         File: HERest.c: Embedded B-W ReEstimation           */
/* ----------------------------------------------------------- */

/*
   This program is used to perform a single reestimation of
   the parameters of a set of HMMs using Baum-Welch.  Training
   data consists of one or more utterances each of which has a 
   transcription in the form of a standard label file (segment
   boundaries are ignored).  For each training utterance, a
   composite model is effectively synthesised by concatenating
   the phoneme models given by the transcription.  Each phone
   model has the usual set of accumulators allocated to it,
   these are updated by performing a standard B-W pass over
   each training utterance using the composite model. 
   
   Note that unlike HRest, only the beta matrix and output probs
   are held in memory complete. The alpha matrix is calculated 
   on the fly.

   Processing is in two stages

      1) Either (a)
         
         Each data file contains training data which is 
         processed and accumulators for state occupation, 
         state transition, means and variances are updated.
         
         Or (b)
         
         Each data file contains a dump of the accumulators
         produced by previous runs of the program.  These
         are read in and added together to form a single set
         of accumulators.

      2) Either (a) 

         New estimates for the hmm parameters are calculated

         Or (b)

         The accumulators are dumped into a file.

   Thus, on a single processor the combination 1(a) 2(a) would
   be used.  If N processors are available then the 
   training data would split into N equal groups and the
   n'th group would be used on processor n with switch -p n
   set.  Processor n would then perform 1(a) 2(b) and dump
   its accumulators to file HERn.acc in the target hmm
   directory.  When all processors had finished, the 
   program would be run again with switch
   -p 0 set and HERn.* input as data files.  This would use
   the combination 1(b) 2(a) to complete the reestimation.

*/


#include <stdio.h>      /* Standard C Libraries */
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include "HShell.h"     /* HMM ToolKit Modules */
#include "HMath.h"
#include "HSpIO.h"
#include "HLabel.h"
#include "HModel.h"

#define VERSION   "1.2"
#define NOPRUNE 1.0E20
#define MINMODELPROB 10.0        /* prune thresh for forward pass */

/* Global Settings */

static char * labDir = NULL;     /* label (transcription) file directory */
static char * labExt = "lab";    /* label file extension */
static char * hmmDir = NULL;     /* directory to look for hmm def files */
static char * hmmExt = NULL;     /* hmm def file extension */
static char * newDir = NULL;     /* directory to store new hmm def files */
static char * newExt = NULL;     /* extension of new reestimated hmm files */
static float minVar  = 0.0;      /* minimum variance (diagonal only) */
static int minEgs    = 3;        /* min examples to train a model */
static LogReal pruneThresh = NOPRUNE;  /* pruning threshold */
static short grandVar = FALSE;   /* enable grand variances */
enum _UPDSet{UPMEANS=1,UPVARS=2,UPTRANS=4,UPMIXES=8};
typedef enum _UPDSet UPDSet;
static UPDSet uFlags = UPMEANS|UPVARS|UPTRANS|UPMIXES;   /* update flags */
static int parMode   = -1;       /* enable one of the // modes */
static int  trace    = 0;        /* Trace level */

static long sampPeriod = 0;      /* data sample period in 100us units */
static int  vSize = 0;           /* data vector size */
static int  maxStates = 0;       /* max states in any model */
static int  M = 0;               /* number of mixtures */
static CovKind CK;               /* covariance kind of all units */
static SampKind SK;              /* sample kind of all units */
static int  isTied = FALSE;      /* true if tied mixtures */

/* Global Data Structures */

static int H;              /* number of distinct HMM's (Labels) */
static HMMDef *defList;    /* array[1..H] of HMM definitions */
static char **hmmFn;       /* array[1..H] of HMM file name */
static LabId *idList;      /* array[1..H] of HMM file name id */
static Vector ot;          /* output vector ... */
static Vector ot1;         /* ... and its successor */
static int Q;              /* number of models in transcription */
static int T;              /* number of frames in utterance */
static long totalT=0;      /* total number of frames in training data */
static short *hi;          /* array[1..Q] of HMM modelnumbers 1..H */
static Vector *alfat;      /* array[1..Q][1..Nq] of prob */
static Vector *alfat1;     /* alpha[t-1] */
static Vector **beta;      /* array[1..Q][1..T][1..Nq] of prob */
static Vector **otprob;    /* array[1..Q][1..T][2..Nq-1] of prob */
static short *qHi;         /* top of pruning beam */
static short *qLo;         /* bottom of pruning beam */
static DataFile data;      /* current training data */
static Transcription tr;   /* current transcription */
static LogReal pr;         /* log prob of current utterance */
static LogReal totalPr;    /* total log prob upto current utterance */

typedef struct {
   int numEgs;          /* number of examples of this model */
   Matrix tran;         /* array[1..Nh][1..Nh] of transition count */
   Vector occ;          /* array[1..Nh] of occupation count */
   Vector **mu;         /* array[2..Nh-1][1..M] of MeanVector; */
   Covariance **va;     /* array[2..Nh-1][1..M] of Covariance; 
                     note: only upper triangle of full cov used */            
   Matrix c;            /* array[1..Nh-1][1..M] of MixWeight */
}Accumulator;
static Accumulator *acc;   /* array[1..H]of Accumulator */
static Accumulator *lacc;  /* array[1..H]of Load Accumulator - only
                              created and used when parMode==0 */
static Vector gVar;        /* grand variance */
static Matrix gCov;        /* grand covariance */
static Vector *tMu;        /* array[1..M] of MeanVector  */
static Covariance *tVa;    /* array[1..M] of Covariance */

/* ------------------ Process Command Line ------------------------- */

void ReportUsage(void)
{
   fprintf(stderr,"\nUSAGE: HERest[V%s] [options] hmmList dataFiles...\n\n",
           VERSION);
   fprintf(stderr," Option                                   Default\n\n");
   fprintf(stderr," -d s    dir to find hmm definitions       current\n");
   fprintf(stderr," -e s    dir to store new hmm definitions  current\n");
   fprintf(stderr," -g      enable grand variances             off\n");
   fprintf(stderr," -m N    set min examples needed per model   3\n");
   fprintf(stderr," -o s    extension for new hmm files        as src\n");
   fprintf(stderr," -p N    set parallel mode to N             off\n");
   fprintf(stderr," -t f    set pruning threshold to f         inf\n");
   fprintf(stderr," -u tmvw Update t)rans m)eans v)ars w)ghts  tmvw\n");
   fprintf(stderr," -v f    Set minimum variance to f          0.0\n");
   fprintf(stderr," -x s    extension for hmm files            none\n");
   fprintf(stderr," -F fmt  Set data file format to fmt        HTK\n");
   fprintf(stderr," -G fmt  Set label file format to fmt       HTK\n");
   fprintf(stderr," -L s    dir to find label file(s)         current\n");
   fprintf(stderr," -S f    set script file to f               none\n");
   fprintf(stderr," -T N    set trace level to N                0\n");
   fprintf(stderr," -X s    extension for label files          lab\n");
   exit(1);
}

int GetChkedInt(int min, int max, char * swtname)
{
   int val;
   
   if (NextArg() != INTARG) 
      HError(1,"Integer Arg Required for %s option",swtname);
   val=GetIntArg();
   if (val<min || val>max) 
      HError(1,"Integer arg out of range in %s option",swtname);
   return val;
}

float GetChkedFlt(float min, float max, char * swtname)
{
   float val;
   
   if (NextArg() != INTARG && NextArg() != FLOATARG) 
      HError(1,"Float Arg Required for %s option",swtname);
   val=GetFltArg();
   if (val<min || val>max) 
      HError(1,"Float arg out of range in %s option",swtname);
   return val;
}

void SetuFlags()
{
   char *s;
   
   s=GetStrArg();
   uFlags=0;        
   while (*s != NULL)
      switch (*s++) {
         case 't': uFlags+=UPTRANS;break;
         case 'm': uFlags+=UPMEANS;break;
         case 'v': uFlags+=UPVARS;break;
         case 'w': uFlags+=UPMIXES;break;
         default: HError(2,"Unknown update flag %c",*s);
                  break;
      }
}

main(int argc, char *argv[])
{
   char *datafn, *s;
   FileFormat ff;
   void Initialise(char *hmmListFn);
   void FBFile(char * datafn);
   void UpdateModels(void);
   void LoadAccs(char *fn);
   void DumpAccs(int n);
   
   InitShell(argc,argv);
   InitMath(TRUE); InitLabel();
   if (NumArgs()==0)
      ReportUsage();
   while (NextArg() == SWITCHARG) {
      s = GetSwtArg();
      if (strlen(s)!=1) 
         HError(1,"Bad switch %s; must be single letter",s);
      switch(s[0]){
         case 'd':
            if (NextArg()!=STRINGARG)
               HError(1,"HMM definition directory expected");
            hmmDir = GetStrArg(); break;  
         case 'e':
            if (NextArg()!=STRINGARG)
               HError(1,"HMM definition directory expected");
            newDir = GetStrArg(); break;  
         case 'g':
            grandVar = TRUE; break;
         case 'm':
            minEgs = GetChkedInt(1,1000,s); break;
         case 'o':
            if (NextArg()!=STRINGARG)
               HError(1,"HMM file extension expected");
            newExt = GetStrArg(); break;
         case 'p':
            parMode = GetChkedInt(0,500,s); break;
         case 't':
            pruneThresh = GetChkedFlt(0.0,1.0E20,s); break;
         case 'u':
            SetuFlags(); break;
         case 'v':
            minVar = GetChkedFlt(0.0,10.0,s); break;
         case 'x':
            if (NextArg()!=STRINGARG)
               HError(1,"HMM file extension expected");
            hmmExt = GetStrArg(); break;
         case 'F':
            if (NextArg() != STRINGARG)
               HError(1,"Data File format expected");
            if((ff = Str2Format(GetStrArg())) == ALIEN)
               HError(0,"Warning ALIEN Data file format set");
            SetFormat(ff);
            break;
         case 'G':
            if (NextArg() != STRINGARG)
               HError(1,"Label File format expected");
            if((ff = Str2Format(GetStrArg())) == ALIEN)
               HError(0,"Warning ALIEN Label file format set");
            SetLFormat(ff);
            break;
         case 'L':
            if (NextArg()!=STRINGARG)
               HError(1,"Label file directory expected");
            labDir = GetStrArg(); break;
         case 'S': SetScriptFile();
            break;
         case 'T':
            trace = GetChkedInt(0,3,s); break;
         case 'X':
            if (NextArg()!=STRINGARG)
               HError(1,"Label file extension expected");
            labExt = GetStrArg(); break;
         default:
            HError(1,"Unknown switch %s",s);
      }
   } 
   if (NextArg() != STRINGARG)
      HError(1,"HMM list file name expected");
   Initialise(GetStrArg());
   do {
      if (NextArg()!=STRINGARG)
         HError(1,"data file name expected");
      datafn = GetStrArg();
      if (parMode==0)
         LoadAccs(datafn);
      else
         FBFile(datafn);
   } while (NumArgs()>0);
   if (parMode>0)
      DumpAccs(parMode);
   else
      UpdateModels();
   exit(0);
}


/* ----------------------- HMM Management ---------------------- */

/* CreateShortVec: create a vector indexed 1..n of shorts */
short *CreateShortVec(int n)
{
   short * p;
   
   if ((p=(short *)malloc(n*sizeof(short)))==NULL)
      HError(2,"Cannot create array of shorts");
   --p;
   return p;
}

/* FreeShortVec: create a vector indexed 1..N of shorts */
void FreeShortVec(short * v)
{
   free(v+1);
}

/* LoadHMM: load specified hmm, adding hmmExt if set */
char *LoadHMM(char *name, HMMDef *def)
{
   char fn[255],*hfn;
   int len;
   
   if (hmmDir==NULL)
      fn[0]='\0';
   else {
      strcpy(fn,hmmDir);
      len = strlen(fn);
      if (fn[len-1]!=PATHCHAR) {
         fn[len]=PATHCHAR;
         fn[++len]='\0';
      }
   }
   strcat(fn,name);
   if (hmmExt!=NULL) {
      strcat(fn,".");
      strcat(fn,hmmExt);
   }
   SPINBALL
   HOpen(fn,def);
   if (vSize==0) {
      vSize = def->vecSize;
      M = def->numMixtures;
      CK = def->ckind; SK = def->pkind;
      isTied = def->tstat != FreeT;
   } else {
      if(vSize!=def->vecSize)
         HError(2,"HMMs have incompatible vecSizes");
      if(M!=def->numMixtures)
         HError(2,"HMMs have incompatible numMixtures");
      if (CK!=def->ckind)
         HError(2,"HMMs have different covariance kinds");
      if (SK!=def->pkind)
         HError(2,"HMMs have different sample kinds");
      if (isTied != (def->tstat != FreeT))
         HError(2,"Mixed Free and Tied Distributions");
   }
   if (maxStates<def->numStates)
      maxStates = def->numStates;
   hfn = NewString(strlen(fn));
   strcpy(hfn,fn);
   return hfn;
}

/* LoadHMMDefs: create defList and load it */
void LoadHMMDefs(void)
{
   int h;
   
   if ((defList = (HMMDef *)malloc(H*sizeof(HMMDef))) == NULL)
      HError(2,"Unable to create hmm def array");
   --defList;
   if ((hmmFn = (char **)malloc(H*sizeof(char *))) == NULL)
      HError(2,"Unable to create hmm file name array");
   --hmmFn;
   for (h=1;h<=H;h++)
      hmmFn[h]=LoadHMM(idList[h]->name,defList+h);
}

/* CreateMu:  create an accumulator for means */
Vector *CreateMu(void)
{
   int m;
   Vector *mu,v;

   if ((mu = (Vector *) malloc(M*sizeof(Vector))) == NULL)
      HError(2,"Cannot create array of mu");
   mu -= 1;
   for (m=1;m<=M;m++){
      v = mu[m] = CreateVector(vSize);
      ZeroVector(v);
   }
   return mu;
}

/* CreateVa: create accumulator for square sums */
Covariance *CreateVa(void)
{
   int m;
   Covariance *va;
   Vector v;
   Matrix u;
   
   if ((va = (Covariance *) malloc(M*sizeof(Covariance))) == NULL)
      HError(2,"Cannot create array of va");
   va -= 1;
   for (m=1;m<=M;m++)
      if (CK == FULLC) {
         u = va[m].inv = CreateMatrix(vSize,vSize);
         ZeroMatrix(u);
      } else {
         v = va[m].var = CreateVector(vSize);
         ZeroVector(v);
      }
   return va;
}

/* CreateTiedAccs: create tMu and tVa for tied mixtures */
void CreateTiedAccs(void)
{
   tMu = CreateMu();
   tVa = CreateVa();
}

/* CreateAccumulator: Create an accumulator for the given hmm */
void CreateAccumulator(Accumulator *a, HMMDef *hmm)
{
   int i,N;
   
   SPINBALL
   a->numEgs = 0;
   N = hmm->numStates;
   a->tran = CreateMatrix(N,N);
   ZeroMatrix(a->tran);
   a->occ = CreateVector(N);
   ZeroVector(a->occ);
   if ((a->mu=(Vector **)malloc((N-2)*sizeof(Vector *))) == NULL)
      HError(2,"Cannot create pointer array for a->mu");
   a->mu -= 2;
   for (i=2;i<N;i++)
      a->mu[i] = (isTied)?tMu:CreateMu();
   if ((a->va=(Covariance **)malloc((N-2)*sizeof(Covariance *))) == NULL)
      HError(2,"Cannot create pointer array for a->va");
   a->va -= 2;
   for (i=2;i<N;i++) 
      a->va[i] = (isTied)?tVa:CreateVa();
   if (M>1){
      a->c = CreateMatrix(N-1,M);
      ZeroMatrix(a->c);
   }
}

/* CreateAccs: returns an array of accumulators */
Accumulator *CreateAccs(void)
{
   int h;
   Accumulator *a;
   
   if ((a=(Accumulator *)malloc(H*sizeof(Accumulator)))==NULL)
      HError(2,"Cannot allocate array of accumulators");
   --a;
   for (h=1;h<=H;h++)
      CreateAccumulator(a+h,defList+h);
   return a;
}

/* -------------------------- Initialisation ----------------------- */

void Initialise(char *hmmListFn)
{
   int i;
   
   idList = LoadIdList(hmmListFn,&H);
   LoadHMMDefs();
   if (isTied) CreateTiedAccs();
   acc = CreateAccs();
   if (parMode==0) lacc = CreateAccs();
   if (trace>0) {
      printf("HERest[V%s]  Updating: ",VERSION);
      if (uFlags&UPTRANS) printf("Transitions "); 
      if (uFlags&UPMEANS) printf("Means "); 
      if (uFlags&UPVARS)  printf("Variances "); 
      if (uFlags&UPMIXES && M>1)  printf("MixWeights "); 
      printf("\n             ");
      if (isTied) printf("Tied-Mixtures ");
      if (parMode>=0) printf("Parallel-Mode[%d] ",parMode);
      if (pruneThresh < NOPRUNE) printf("Pruning-On[%.1f]",pruneThresh);
      printf("\n");
      if (trace>1)
         printf(" %d Models Loaded, VecSize=%d\n",H,vSize);
      if (trace==3)
         for (i=1;i<=H;i++)
            printf("  %d. %s\n",i,hmmFn[i]);
      fflush(stdout);
   }
   ot = CreateVector(vSize);
   ot1 = CreateVector(vSize);
   totalPr = 0.0;
}

/* ------------------- Accumulator Load and Dump -------------------- */

/* IncAccumulator: increment a by la */
void IncAccumulator(Accumulator *a, Accumulator *la, HMMDef *hmm)
{
   int N,i,j,m,k,kk;
   Vector v,lv;
   Matrix u,lu;
   
   SPINBALL
   a->numEgs += la->numEgs;
   N = hmm->numStates;
   for (i=1;i<=N;i++)
      for (j=1;j<=N;j++)
         a->tran[i][j] += la->tran[i][j];
   for (i=1;i<=N;i++)
      a->occ[i] += la->occ[i];
   for (i=2;i<N;i++)
      for (m=1;m<=M;m++){
         v = a->mu[i][m]; lv = la->mu[i][m];
         for (k=1;k<=vSize;k++)
            v[k] += lv[k];
      }
   SPINBALL
   for (i=2;i<N;i++)
      for (m=1;m<=M;m++)
         if (CK == FULLC) {
            u = a->va[i][m].inv;
            lu = la->va[i][m].inv;
            for (k=1;k<=vSize;k++)
               for (kk=k;kk<=vSize;kk++)
                  u[k][kk] += lu[k][kk];
         } else {
            v = a->va[i][m].var;
            lv = la->va[i][m].var;
            for (k=1;k<=vSize;k++)
               v[k] += lv[k];
         }
   if (M>1){
      u = a->c; lu = la->c;
      for (i=2;i<N;i++)
         for (m=1;m<=M;m++)
            u[i][m] += lu[i][m];
   }
}

/* LoadAccumulator: load accumulator a with contents of f  */
void LoadAccumulator(FILE *f, Accumulator *a, HMMDef *hmm)
{
   int n,m;
   int N;

   SPINBALL
   N= hmm->numStates;
   if (fread(&(a->numEgs),sizeof(int),1,f) != 1)
      HError(2,"Cannot read numEgs");
   BReadMatrix(f,a->tran);
   BReadVector(f,a->occ);
   for (n=2;n<N;n++)
      for (m=1;m<=M;m++)
         BReadVector(f,a->mu[n][m]);
   for (n=2;n<N;n++)
      for (m=1;m<=M;m++)
         if (CK == DIAGC)
            BReadVector(f,a->va[n][m].var);
         else
            BReadMatrix(f,a->va[n][m].inv);
   if (M>1)
      BReadMatrix(f,a->c);
}

/* LoadAccs: increment program accs by accs stored in fn */
void LoadAccs(char *fn)
{
   FILE *f;
   int h;
   long totT;
   LogReal totP;

   if ((f=fopen(fn,"rb"))==NULL)
      HError(2,"Unable to open dump file %s",fn);
   for (h=1;h<=H;h++)
      LoadAccumulator(f,lacc+h,defList+h);
   if (fread(&totT,sizeof(long),1,f) != 1)
      HError(2,"Cannot read total T");
   totalT += totT;
   if (fread(&totP,sizeof(LogReal),1,f) != 1)
      HError(2,"Cannot read total P");
   totalPr += totP;
   fclose(f);
   for (h=1;h<=H;h++)
      IncAccumulator(acc+h,lacc+h,defList+h);
}

/* DumpAccumulator: dump given acc to file f */
void DumpAccumulator(FILE *f, Accumulator *a, HMMDef *hmm)
{
   int n,m;
   int N,M;

   SPINBALL
   N= hmm->numStates;
   M= hmm->numMixtures;
   if (fwrite(&(a->numEgs),sizeof(int),1,f) != 1)
      HError(2,"Cannot dump numEgs");
   BWriteMatrix(f,a->tran);
   BWriteVector(f,a->occ);
   for (n=2;n<N;n++)
      for (m=1;m<=M;m++)
         BWriteVector(f,a->mu[n][m]);
   for (n=2;n<N;n++)
      for (m=1;m<=M;m++)
         if (CK == DIAGC)
            BWriteVector(f,a->va[n][m].var);
         else
            BWriteMatrix(f,a->va[n][m].inv);
   if (M>1)
      BWriteMatrix(f,a->c);
}

/* DumpAccs: dump program accs to file HERn.acc */ 
void DumpAccs(int n)
{
   char fn[64],num[20],dumpfn[100];
   FILE *f;
   int h;

   strcpy(fn,"HER");
   sprintf(num,"%d",n);
   strcat(fn,num); strcat(fn,".acc");
   MakeFN(fn,newDir,NULL,dumpfn);
   if ((f=fopen(dumpfn,"wb"))==NULL)
      HError(2,"Unable to create dump file %s",dumpfn);
   for (h=1;h<=H;h++)
      DumpAccumulator(f,acc+h,defList+h);
   if (fwrite(&totalT,sizeof(long),1,f) != 1)
      HError(2,"Cannot dump total T");
   if (fwrite(&totalPr,sizeof(LogReal),1,f) != 1)
      HError(2,"Cannot dump total Pr");
   fclose(f);
}

/* --------------------------- Forward-Backward --------------------- */

static int maxBeamWidth;      /* max width of beam in model units */
static LogReal maxAlphaBeta;  /* max alpha/beta product along beam ridge */
static LogReal minAlphaBeta;  /* min alpha/beta product along beam ridge */

/* InitPruneStats: initialise pruning stats */
void InitPruneStats(void)
{
   maxBeamWidth = 0;
   maxAlphaBeta = LZERO;
   minAlphaBeta = 1.0;
}

/* CheckPruning: record peak alfa.beta product and position */
void CheckPruning(int t)
{
   int i,q,Nq,bestq,besti,margin;
   Vector aq,bq;
   HMMDef *hmm;
   LogReal l,maxL;

   bestq = besti = 0;
   maxL = LZERO;
   for (q=qLo[t];q<=qHi[t];q++){
      hmm = defList+hi[q]; Nq = hmm->numStates;   
      aq = alfat[q]; bq=beta[q][t];
      for (i=2;i<Nq;i++){
         if ((l=aq[i]+bq[i])>maxL){
            bestq = q; besti = i; maxL=l;
         }
      }
   }
   if (maxL > maxAlphaBeta) maxAlphaBeta = maxL;
   if (maxL < minAlphaBeta) minAlphaBeta = maxL;
   margin = qHi[t]-qLo[t]+1;
   if (margin>maxBeamWidth) maxBeamWidth = margin;
   if (trace == 3) {
      if (bestq == 0) 
         printf("%3d. No max found in alpha.beta\n",t);
      else
         printf("%3d. Max Alfa.Beta = %9.4f at q=%d i=%d [%s]\n",
                t,maxL,bestq,besti,hmmFn[hi[bestq]]);
   }
}

/* SummarisePruning: print out pruning stats */
void SummarisePruning(void)
{
   long totalQ=0;
   float e;
   int t;
   
   for (t=1;t<=T;t++)
      totalQ += qHi[t]-qLo[t]+1;
   e = (1.0 - (float) totalQ / ((float) T*Q)) * 100.0;
   printf(" Pruning %.1f%%; MaxBeamWidth %d; PeakShortFall %.2f\n",
            e,maxBeamWidth,maxAlphaBeta-minAlphaBeta);
   fflush(stdout);
}

/* CreateInsts: create array of hmm instances */
void CreateInsts(void)
{
   int q;
   
   hi = CreateShortVec(Q);
   for (q=1;q<=Q;q++)
      if ((hi[q] = FindLabId(tr->labels[q-1].labid,idList)) == 0)
         HError(2,"Unknown label %s",tr->labels[q-1].labid->name);
}

/* ZeroAlpha: zero alpha's of given models */
void ZeroAlpha(int qlo, int qhi)
{
   HMMDef *hmm;
   int Nq,j,q;
   Vector aq;
   
   hmm = defList+hi[qlo];     /* qlo model - dont erase carry over */
   Nq = hmm->numStates; 
   aq = alfat[qlo];
   for (j=2;j<=Nq;j++)
      aq[j] = LZERO;
   for (q=qlo+1;q<=qhi;q++) { /* general case - all states LERO */
      hmm = defList+hi[q]; 
      Nq = hmm->numStates; 
      aq = alfat[q];
      for (j=1;j<=Nq;j++)
         aq[j] = LZERO;
   }
   if (qhi<Q)                 /* qhi model - carry over LZERO */
      alfat[qhi+1][1] = LZERO;
}

/* InitAlpha: allocate and initialise alfa columns for time t=1 */
void InitAlpha(void)
{
   int i,j,q,Nq;
   HMMDef *hmm;
   Vector aq,outprob;
   LogReal a,x;
   
   /* Create Storage Space - two columns */
   SPINBALL
   if ((alfat=(Vector *)malloc(Q*sizeof(Vector)))==NULL)
      HError(2,"Cannot create array of alfat vectors");
   --alfat;
   for (q=1;q<=Q;q++)
      alfat[q] = CreateVector((defList+hi[q])->numStates);
   if ((alfat1=(Vector *)malloc(Q*sizeof(Vector)))==NULL)
      HError(2,"Cannot create array of alfat1 vectors");
   --alfat1;
   for (q=1;q<=Q;q++)
      alfat1[q] = CreateVector((defList+hi[q])->numStates);
      
   /* Calculate alfat values for t=1 */
   SPINBALL 
   hmm = defList+hi[1]; Nq = hmm->numStates; aq = alfat[1];
   aq[1] = LZERO;
   if((outprob = otprob[1][1]) == NULL)
      HError(2,"Bug - outprob NULL at time %d in InitAlpha",1);
   for (j=2;j<Nq;j++) {
      a = hmm->transP[1][j];
      aq[j] = (a>LSMALL)?a+outprob[j]:LZERO;
   }
   x = LZERO;
   for (i=2;i<Nq;i++) {
      a = hmm->transP[i][Nq];
      if (a>LSMALL)
         x = LAdd(x,aq[i]+a);
   }
   alfat[2][1] = aq[Nq] = x;
   ZeroAlpha(2,Q);
   if (trace>1 && pruneThresh < NOPRUNE)
      CheckPruning(1);
}

/* MaxModelProb: Calc max probability of being in model q at
   time t, return LZERO if cannot do so */
LogReal MaxModelProb(int q, int t)
{
   Vector aq,bq;
   LogReal maxP,x;
   int Nq,i;
   HMMDef *hmm;

   maxP = LZERO;
   hmm = defList+hi[q]; Nq = hmm->numStates;   
   bq=beta[q][t];
   if (bq != NULL) {
      aq = alfat[q]; 
      for (i=1;i<Nq;i++)
         if ((x=aq[i]+bq[i]) > maxP) maxP = x;
   }
   return maxP;
}

/* StepAlpha: calculate alfat column for time t and return
   forward beam limits in startq and endq */
void StepAlpha(int t, int *startq, int *endq)
{
   Vector aq,laq,*tmp;
   float * outprob;
   int i,j,q,Nq;
   LogReal a,x,y;
   HMMDef *hmm;
   
   *startq = qLo[t-1];
   while (pr-MaxModelProb(*startq,t-1)>MINMODELPROB)
      ++(*startq);
   if (*startq<qLo[t])
      *startq = qLo[t];                  
   *endq = qHi[t-1];
   while (pr-MaxModelProb(*endq,t-1)>MINMODELPROB)
      --(*endq);
   if (*endq>qHi[t])
      *endq = qHi[t];                  
   if (trace == 3)
      printf("%d: Alpha Beam %d->%d \n",t,*startq,*endq);

   tmp = alfat1; alfat1 = alfat; alfat = tmp;
   SPINBALL 

   alfat[1][1] = LZERO;
   if (*startq>1) ZeroAlpha(1,*startq-1);
   for (q = *startq; q <= *endq; q++) {
      hmm = defList+hi[q]; Nq = hmm->numStates; 
      aq = alfat[q]; laq = alfat1[q];
      if((outprob = otprob[q][t]) == NULL)
         HError(2,"Bug - outprob NULL at time %d in StepAlpha",t);
      for (j=2;j<Nq;j++) {
         x = LZERO;
         for (i=1;i<Nq;i++){
            a = hmm->transP[i][j]; y = laq[i];
            if (a>LSMALL && y>LSMALL)
               x = LAdd(x,y+a);
         }
         aq[j] = x + outprob[j];
      }
      x = LZERO;
      for (i=2;i<Nq;i++){
         a = hmm->transP[i][Nq]; y = aq[i];
         if (a>LSMALL && y>LSMALL)
            x = LAdd(x,y+a);
      }
      aq[Nq] = x;
      if (q<Q)
         alfat[q+1][1] = x;
   }
   if (*endq<Q) ZeroAlpha(*endq+1,Q);

   if (trace>1 && pruneThresh < NOPRUNE)
      CheckPruning(t);
   if (trace>1 && t==T) {
      printf(" Forward  Prob = %12.1f\n",x);
      fflush(stdout);
      if (pruneThresh < NOPRUNE)
         SummarisePruning();
   }
}

/* FreeAlpha: free storage allocated to alfat columns */
void FreeAlpha(void)
{
   int q;
   
   for (q=1;q<=Q;q++){
      SPINBALL 
      FreeVector(alfat[q]);
      SPINBALL 
      FreeVector(alfat1[q]);
   }
   free(alfat+1);
   free(alfat1+1);
}

/* CreateBeta: create Q and T pointer arrays for beta */
void CreateBeta(void)
{
   int q,t;
   Vector *v;
   
   if ((beta=(Vector **)malloc(Q*sizeof(Vector *)))==NULL)
      HError(2,"Cannot create array[Q] of beta");
   --beta;
   for (q=1;q<=Q;q++){
      if ((v=(Vector *)malloc(T*sizeof(Vector)))==NULL)
         HError(2,"Cannot create array[T] of beta");
      --v;
      for (t=1;t<=T;t++) v[t] = NULL;
      beta[q] = v;
   }
}

/* CreateOtprob: create Q and T pointer arrays for Otprob */
void CreateOtprob(void)
{
   int q,t;
   Vector *v;
   
   if ((otprob=(Vector **)malloc(Q*sizeof(Vector *)))==NULL)
      HError(2,"Cannot create array[Q] of otprob");
   --otprob;
   for (q=1;q<=Q;q++){
      if ((v=(Vector *)malloc(T*sizeof(Vector)))==NULL)
         HError(2,"Cannot create array[T] of otprob");
      --v;
      for (t=1;t<=T;t++) v[t] = NULL;
      otprob[q] = v;
   }
}

/* NewBetaVec: create prob vector size 1..N */
Vector NewBetaVec(int N)
{
   Vector v;
   
   if ((v=(Vector)malloc(N*sizeof(float)))==NULL)
      HError(2,"Cannot create array[1..N] of beta");
   --v;
   return v;
}

/* NewOtprobVec: create prob vector size 2..N-1 */
Vector NewOtprobVec(int N)
{
   Vector v;
   
   if ((v=(Vector)malloc((N-2)*sizeof(float)))==NULL)
      HError(2,"Cannot create array[2..N-1] of Otprob");
   v -= 2;
   return v;
}

/* Setotprob: allocate and calculate otprob matrix at time t */
void Setotprob(int t)
{
   int q,j,Nq,endq;
   Vector outprob;
   HMMDef *hmm;

   GetSample(&data,t-1,ot+1);
   if (isTied) PreCompute(ot);
   endq = 1==qLo[t]?1:qLo[t]-1;
   for (q=qHi[t];q>=endq;q--) {
      hmm = defList+hi[q]; Nq = hmm->numStates;
      outprob = otprob[q][t] = NewOtprobVec(Nq);
      for (j=2;j<Nq;j++)
         outprob[j] = OutP(ot,hmm,j);
   }
}

/* SetBeta: allocate and calculate beta and otprob matrices */
LogReal SetBeta(char * datafn)
{
   int i,j,t,q,Nq,q_at_gMax,startq,endq;
   Vector bqt,bqt1,outprob,maxP;
   LogReal x,a,y,gMax,lMax;
   HMMDef *hmm;
   
   /* Create Storage Space - excluding actual data arrays */
   SPINBALL
   qHi = CreateShortVec(T);         /* storage for min and max q vals */
   qLo = CreateShortVec(T);
   CreateBeta(); 
   CreateOtprob();
   maxP = CreateVector(Q);          /* for calculating beam width */
   
   /* Last Column t = T */
   SPINBALL
   hmm = defList+hi[Q]; Nq = hmm->numStates;
   bqt = beta[Q][T] = NewBetaVec(Nq);
   bqt[Nq] = 0.0;
   for (i=2;i<Nq;i++) 
      bqt[i] = hmm->transP[i][Nq];
   bqt[1]=LZERO;
   qHi[T] = qLo[T] = Q;
   Setotprob(T);

   /* Columns T-1 -> 1 */
   for (t=T-1;t>=1;t--) {
      SPINBALL
      gMax = LZERO;                    /* max value of beta at time t */
      startq = t>=qHi[t+1]?qHi[t+1]:t;
      endq = 1==qLo[t+1]?1:qLo[t+1]-1;
      for (q=startq;q>=endq;q--) {
         lMax = LZERO;                 /* max value of beta in model q */
         hmm = defList+hi[q]; 
         Nq = hmm->numStates;
         bqt = beta[q][t] = NewBetaVec(Nq);  /* create vec for beta vals */
         bqt1 = beta[q][t+1];                /* this may be NULL !! */
         outprob = otprob[q][t+1];
         bqt[Nq] = (q==startq)?LZERO:beta[q+1][t][1];
         for (i=Nq-1;i>=1;i--){
            x = hmm->transP[i][Nq] + bqt[Nq];
            if (q>=qLo[t+1]&&q<=qHi[t+1])
               for (j=2;j<Nq;j++) {
                  a = hmm->transP[i][j]; y = bqt1[j];
                  if (a>LSMALL && y>LSMALL)
                     x = LAdd(x,a+outprob[j]+y);
               }
            bqt[i] = x;
            if (x>lMax) lMax = x;
            if (x>gMax) {
               gMax = x; q_at_gMax = q;
            }
         }
         maxP[q] = lMax;
      }
      while (gMax-maxP[startq] > pruneThresh) 
         --startq;                   /* lower startq till thresh reached */
      qHi[t] = startq;
      while (gMax-maxP[endq]>pruneThresh && endq<t)
    ++endq;                   /* raise endq till thresh reached */
      qLo[t] = endq;

      Setotprob(t);

      if (trace == 3 && pruneThresh < NOPRUNE)
         printf("%d: Beta Beam %d->%d; gMax=%f at %d\n",
                t,qLo[t],qHi[t],gMax,q_at_gMax);
   }

  /* Finally, calc total prob pr */
   pr = LZERO;
   outprob = otprob[1][1];
   for (j=2;j<Nq;j++) {
      a=hmm->transP[1][j]; y=bqt[j];
      if (a>LSMALL && y>LSMALL)
         pr = LAdd(pr,a+outprob[j]+y);
   }
   FreeVector(maxP);

   if (LZERO>=pr) {
      printf("Calc prob=%f. pruneThresh(%f) too small. datafn=%s\n", LZERO,pruneThresh,datafn);
      return pr;
   }

   totalPr += pr;
   if (trace>0) {
      if (trace>1) printf(" Backward");
      else printf(" Utterance");
      printf(" Prob = %12.1f\n",pr);
      fflush(stdout);
   }
   return pr;
}

/* FreeBetaOt: free storage allocated to beta & otprob matrices 
               (in reverse order to original allocation) */
void FreeBetaOt(void)
{
   int q,t;
   Vector vb, *bq;
   Vector vq, *op;
   
   for (q=1;q<=Q;q++){
      bq = beta[q];
      op = otprob[q];
      for (t=1;t<=T;t++){
         vb = bq[t]; vq = op[t];
         if (vq != NULL) {
            SPINBALL 
            free(vq+2);
         }
         if (vb != NULL) {
            SPINBALL 
            free(vb+1);
         }
      }
   }
   SPINBALL 
   for (q=Q;q>=1;q--)
      free(otprob[q]+1);
   free(otprob+1);
   SPINBALL 
   for (q=Q;q>=1;q--)
      free(beta[q]+1);
   SPINBALL 
   free(beta+1);
   FreeShortVec(qLo);
   FreeShortVec(qHi);
}

/* ---------------------- Parameter Update -------------------- */


/* UpOccCount: update the occupation counters of given acc */
void UpOccCount(Accumulator *a, HMMDef *hmm, int t, int q,
                Vector aqt, Vector bqt)
{
   int i,N;
   Real x;
   
   N=hmm->numStates;
   if (q==1){
      if (t==1) ++a->occ[1];
   } else {
      x = aqt[1]+bqt[1]-pr;
      if (x>MINEARG) a->occ[1] += exp(x);
   }
   for (i=2;i<N;i++) {
      x = aqt[i]+bqt[i]-pr;
      if (x>MINEARG) a->occ[i] += exp(x);
   }
}

/* UpTranParms: update the transition counters of given acc */
void UpTranParms(Accumulator *a, HMMDef *hmm, int t, int q,
                 Vector aqt, Vector bqt, Vector bqt1)
{
   int i,j,N;
   Vector ti,ai,outprob;
   Real x;
    
   N = hmm->numStates;
   outprob = otprob[q][t+1]; /* will be NULL only if bqt1 is */
   if (t<T)
      for (i=1;i<N;i++) {
         ti = a->tran[i];
         ai = hmm->transP[i];
         for (j=2;j<N;j++) {
            if (i==1 && q==1) {
               if (t==1) {
                  x = aqt[j]+bqt[j]-pr;
                  if (x>MINEARG) ti[j] += exp(x);
               }
            } else {
               if (bqt1 != NULL) {
                  x = aqt[i]+ai[j]+outprob[j]+bqt1[j]-pr;
                  if (x>MINEARG) ti[j] += exp(x);
               }
            }
         }
         if (q<Q) {
            x = aqt[i]+ai[N]+bqt[N]-pr;
            if (x>MINEARG) ti[N] += exp(x);  
         }
      }
   else 
   if (q==Q)
      for (i=2;i<N;i++) {
         ti = a->tran[i];
         x = aqt[i]+bqt[i]-pr;
         if (x>MINEARG) ti[N] += exp(x);
      }
}

/* Up1MixParms: update mu/va of given acc for 1 mix case */
void Up1MixParms(Accumulator *a, HMMDef *hmm, Vector aqt, Vector bqt)
{
   int j,k,kk,N;
   Vector mu_j,va_j_var,mean,va_j_inv_k;
   Matrix va_j_inv;
   LogReal x;
   Real rx,zmean,zmean1,zmean2;
   
   N = hmm->numStates;
   for (j=2;j<N;j++) {
      x = aqt[j]+bqt[j]-pr;
      if (x>MINEARG) {
         rx = exp(x);
         mu_j = a->mu[j][1]; 
         va_j_var = a->va[j][1].var;
         va_j_inv = a->va[j][1].inv;
         mean = hmm->stateinfo[j].outp[1].mean;
         for (k=1;k<=vSize;k++) {
            if (uFlags&UPMEANS) 
               mu_j[k] += ot[k]*rx;
            if (uFlags&UPVARS){
               if (CK == DIAGC) {
                  zmean = ot[k]-mean[k];
                  va_j_var[k] += zmean*zmean*rx;
               } else {
                  zmean1 = ot[k]-mean[k];
                  va_j_inv_k = va_j_inv[k];
                  for (kk=k;kk<=vSize;kk++) {
                     zmean2 = ot[kk]-mean[kk];
                     va_j_inv_k[kk] += zmean1*zmean2*rx;
                  }
               }
            }
         }
      }
   }
}

/* UpMMixParms: update mu/va of given acc for M>1 mix case */
void UpMMixParms(int q, Accumulator *a, HMMDef *hmm, int t, Vector aqt1, Vector bqt)
{
   int i,j,k,kk,m,N;
   Vector mu_jm,va_jm_var,mean,va_jm_inv_k;
   Matrix va_jm_inv;
   LogReal x,c_jm,initx;
   Real rx,w,zmean,zmean1,zmean2;

   N = hmm->numStates;
   for (j=2;j<N;j++) {
      if ((t==1) && (q==1))
         initx = hmm->transP[1][j];
      else if ((t==1) && (q!=1))    /* shouldnt happen */
         initx = LZERO; 
      else {
         initx = LZERO;
         for (i=1;i<N;i++) 
            initx=LAdd(initx,aqt1[i]+hmm->transP[i][j]);
      }
      if (initx>LSMALL)
         for (m=1;m<=M;m++) {
            w = hmm->stateinfo[j].outp[m].weight;
            if (w>MINMIX){
               mu_jm = a->mu[j][m]; 
               va_jm_var = a->va[j][m].var;
               va_jm_inv = a->va[j][m].inv;
               mean = hmm->stateinfo[j].outp[m].mean;
               c_jm=log(w);
               x = initx+c_jm+bqt[j];
               if (x>LSMALL) {
                  x += MOutP(ot,hmm,j,m)-pr;
                  if (x>MINEARG) {
                     rx = exp(x);
                     for (k=1;k<=vSize;k++) {
                        if (uFlags&UPMEANS)
                           mu_jm[k] += ot[k]*rx;
                        if (uFlags&UPVARS){
                           if (CK == DIAGC) {
                              zmean=ot[k]-mean[k];
                              va_jm_var[k] += zmean*zmean*rx;
                           } else {
                              va_jm_inv_k = va_jm_inv[k];
                              zmean1=ot[k]-mean[k];               
                              for (kk=k;kk<=vSize;kk++) {
                                 zmean2 = ot[kk]-mean[kk];
                                 va_jm_inv_k[kk] += zmean1*zmean2*rx;
                              }
                           }
                        }
                     }
                     if (uFlags&UPMIXES)
                        a->c[j][m] += rx;
                  }
               }
            }
         }
   }  
}

/* -------------------- Top Level of F-B Updating ---------------- */

/* GetOutVec: Get output vector ot */
void GetOutVec(int t)
{
   Vector tmp;

   if (t==1)
      GetSample(&data,t-1,ot+1);
   else {
      tmp = ot; ot = ot1; ot1 = tmp;
   }
   if (t<T)
      GetSample(&data,t,ot1+1);
}

/* FBFile: apply forward-backward to given utterance */
void FBFile(char * datafn)
{
   char labfn[255];
   int q,t,startq,endq;
   Vector aqt,aqt1,bqt,bqt1;
   Accumulator *a;
   HMMDef *hmm;
   
   MakeFN(datafn,labDir,labExt,labfn);
   if (trace>0) {
      printf(" Processing file %s using transcription %s\n",datafn,labfn);
      fflush(stdout);
   }
   tr = LOpen(labfn);
   Q = tr->nLabels;
   SpOpen(datafn,&data);
   if (vSize*sizeof(float) != data.sampSize)
      HError(2,"Incompatible sample size %d shoud be %d in file %s",
      vSize, data.sampSize/sizeof(float),datafn);
   if (SK != data.sampKind)
      HError(2,"Data file %s has wrong sample kind",datafn);
   T = data.nSamples; totalT += T;
   if (sampPeriod==0)
      sampPeriod = data.sampPeriod;
   else
      if (sampPeriod != data.sampPeriod)
         HError(2,"File %s has inconsistent sample period [%d vs %d]",
            datafn,data.sampPeriod,sampPeriod);
   InitPruneStats();
   CreateInsts();
   if (SetBeta(datafn)<=LZERO) {
      SpClose(&data);       /* cannot find path */
      LClose(tr);
      FreeBetaOt();
      FreeShortVec(hi);
      return;
   }

   InitAlpha();
   for (q=1;q<=Q;q++) 
      ++acc[hi[q]].numEgs;
   for (t=1;t<=T;t++) {
      GetOutVec(t);        /* Get Data */
      if (isTied) 
         PreCompute(ot);   /* Need this for MOutP call in UpMMixParms */
      if (t>1) 
         StepAlpha(t,&startq,&endq);           /* Calculate Alpha */
      else
         startq = endq = 1;
      for (q=startq;q<=endq;q++) {   /* inc accs for each active model */
         a = acc+hi[q];
         hmm = defList+hi[q];
         aqt = alfat[q]; bqt = beta[q][t];
         bqt1 = (t==T) ? NULL:beta[q][t+1];
         aqt1 = (t==1) ? NULL:alfat1[q];
         UpOccCount(a,hmm,t,q,aqt,bqt);
         if (uFlags&(UPMEANS|UPVARS|UPMIXES))
            if (M==1)
               Up1MixParms(a,hmm,aqt,bqt);
            else
               UpMMixParms(q,a,hmm,t,aqt1,bqt);
         SPINBALL
         if (uFlags&UPTRANS)
            UpTranParms(a,hmm,t,q,aqt,bqt,bqt1);
      }
   }
   SpClose(&data);               /* clean up */
   LClose(tr);
   FreeAlpha();
   FreeBetaOt();
   FreeShortVec(hi);
}

/* ------------------------ Grand Variance  --------------------- */

/* BuildGVar: build grand variance vector */
void BuildGVar(void)
{
   Vector sqSum,v;
   Accumulator *a;
   HMMDef *hmm;
   int k,h,i,m,N;
   Real x,occSum,occi,cfloor,c_im;

   gVar   = CreateVector(vSize);
   sqSum  = CreateVector(vSize);
   cfloor = MINMIX;
   ZeroVector(sqSum);
   occSum = 0.0;
   for (h=1;h<=H;h++){
      hmm = defList+h; 
      N = hmm->numStates;
      a = acc+h;
      SPINBALL
      for(i=2;i<N;i++){
         occi=a->occ[i];
         if (occi > 0.0)
            for (m=1;m<=M;m++){
               if (M==1) c_im=occi; else c_im=a->c[i][m];
               if (c_im>cfloor){
                  v = a->va[i][m].var;
                  occSum += c_im;
                  for (k=1;k<=vSize;k++)
                     sqSum[k] += v[k];
               }
            }
      }
   }
   for (k=1;k<=vSize;k++){
      x = sqSum[k]/occSum;
      gVar[k] = (x>minVar)?x:minVar;
   }
   FreeVector(sqSum);
}

/* BuildGCov: build grand covariance matrix */
void BuildGCov(void)
{
   Matrix sqSum,mm;
   Accumulator *a;
   HMMDef *hmm;
   int k,kk,h,i,N,m;
   Real x,occSum,occi,cfloor,c_im;

   gCov   = CreateMatrix(vSize,vSize);
   sqSum  = CreateMatrix(vSize,vSize);
   cfloor = MINMIX;
   ZeroMatrix(sqSum);
   occSum = 0;
   for (h=1;h<=H;h++){
      hmm = defList+h; 
      N = hmm->numStates;
      a = acc+h;
      SPINBALL
      for(i=2;i<N;i++){
         occi=a->occ[i];
         if (occi > 0.0)
            for (m=1;m<=M;m++){
               if (M==1) 
                  c_im = occi; 
               else if ((x = a->c[i][m]) > MINEARG)
                  c_im = exp(x);
               else
                  c_im = 0;
               if (c_im>cfloor){
                  mm = a->va[i][m].inv;
                  occSum += c_im;
                  for (k=1;k<=vSize;k++)
                     for (kk=k;kk<=vSize;kk++)
                        sqSum[k][kk] += mm[k][kk];
               }
            }
      }
   }
   for (k=1;k<=vSize;k++)
      for (kk=k;kk<=vSize;kk++) {
         x = sqSum[k][kk]/occSum;
         if (k == kk && x<minVar) x = minVar;
         gCov[k][kk] = gCov[kk][k] = x;
      }
   FreeMatrix(sqSum);
}

/* --------------------------- Model Update --------------------- */

static Vector tWeight;  /* array[1..M] of totalweight; */

/* UpdateModel: use accumulator values to calc new estimate for hmm */
void UpdateModel(int h, Accumulator *a, HMMDef *hmm)
{

   int i,j,k,kk,m,N;
   Real x,occi,c_im,cfloor;
   Vector mean;
   Covariance covariance;
   Vector mu_im,va_im_var,va_im_inv_k;
   Matrix va_im_inv;
   
   cfloor = MINMIX;
   N = hmm->numStates;
   SPINBALL
   if (uFlags&UPTRANS)
      for (i=1;i<N;i++) {
         occi = a->occ[i];
         if (occi > 0.0) 
            for (j=2;j<=N;j++) {
               x = a->tran[i][j]/occi;
               hmm->transP[i][j] = (x>MINLARG)?log(x):LZERO;
            }
         else
            HError(0,"Model %d: state %d never occupied",h,i);
      }
   if (uFlags&(UPMEANS|UPVARS|UPMIXES)) {
      for (i=2;i<N;i++){
         occi=a->occ[i];
         if (occi > 0.0) 
            for (m=1;m<=M;m++) {
               if (M==1) c_im = occi; else c_im = a->c[i][m];
               x = c_im/occi;
               if (uFlags&UPMIXES) {
                  if (x > 1.0) {
                     if (x > 1.001)   /* this is serious! */
                        HError(0,"Mix too big (x=%f) in Model %s: s=%d m=%d",
                               x,idList[h]->name,i,m);
                     x = 1.0;
                  } 
                  hmm->stateinfo[i].outp[m].weight = x;
               }
               if (x>=cfloor) {
                  if (isTied) 
                     tWeight[m] += c_im;
                  else {
                     mean=hmm->stateinfo[i].outp[m].mean;
                     covariance=hmm->stateinfo[i].outp[m].covariance;
                     mu_im=a->mu[i][m]; 
                     va_im_inv=a->va[i][m].inv;
                     va_im_var=a->va[i][m].var;
                     for (k=1;k<=vSize;k++) {
                        if (uFlags&UPMEANS)
                           mean[k] = mu_im[k]/c_im;
                        if (uFlags&UPVARS) {
                           if (CK == DIAGC) {
                              if (grandVar)
                                 covariance.var[k] = gVar[k];
                              else {
                                 x = va_im_var[k]/c_im;                     
                                 covariance.var[k]=(x>minVar)?x:minVar;
                              }
                           } else {
                              va_im_inv_k = va_im_inv[k];
                              for (kk=k;kk<=vSize;kk++)
                                 if (grandVar)
                                    covariance.inv[k][kk] = 
                                       covariance.inv[kk][k] = gCov[kk][k];
                                 else {
                                    x = va_im_inv_k[kk]/c_im;
                                    if (k == kk && x<minVar) x = minVar;
                                    covariance.inv[k][kk] 
                                       = covariance.inv[kk][k] = x;
                                 }
                           }
                        }
                     }
                     if (CK == FULLC) 
                        CovInvert(covariance.inv,covariance.inv);
                  }
               }
            }
         else
            HError(0,"Model %d: state %d never occupied",h,i);
      }
      if (!isTied) FixGConsts(hmm);
   }
}

/* UpdateTiedMixtures: update tied means and covariances */
void UpdateTiedMixtures()
{
   int k,kk,m,h;
   Real x,c_m;
   Vector mean;
   Covariance covariance;
   Vector mu_m,va_m_var,va_m_inv_k;
   Matrix va_m_inv;
   HMMDef *hmm;
   
   hmm = defList+1;  /* 2nd state of 1st model is as good as any! */
   SPINBALL
   if (uFlags&(UPMEANS|UPVARS)) {
      for (m=1;m<=M;m++) {
         mean=hmm->stateinfo[2].outp[m].mean;
         covariance=hmm->stateinfo[2].outp[m].covariance;
         mu_m=tMu[m]; 
         va_m_inv=tVa[m].inv;
         va_m_var=tVa[m].var;
         c_m = tWeight[m];
         for (k=1;k<=vSize;k++) {
            if (uFlags&UPMEANS)
               mean[k] = mu_m[k]/c_m;
            if (uFlags&UPVARS) {
               if (CK == DIAGC) {
                  x = va_m_var[k]/c_m;                     
                  covariance.var[k]=(x>minVar)?x:minVar;
               } else {
                  va_m_inv_k = va_m_inv[k];
                  for (kk=k;kk<=vSize;kk++)
                     x = va_m_inv_k[kk]/c_m;
                     if (k == kk && x<minVar) x = minVar;
                     covariance.inv[k][kk] 
                        = covariance.inv[kk][k] = x;
               }
            }
         }
         if (CK == FULLC) 
            CovInvert(covariance.inv,covariance.inv);
      }
   }
   SPINBALL
   for (h=1;h<=H;h++)
      FixGConsts(defList+h);
}

/* UpdateModels: update all models and save them in newDir if set,
   new files have newExt if set */
void UpdateModels(void)
{
   int h,n;
   char outfn[255];
   
   if (grandVar)
      if (CK==DIAGC) BuildGVar(); else BuildGCov();
   if (isTied) {
      tWeight = CreateVector(M);
      ZeroVector(tWeight);
   }
   for (h=1;h<=H;h++){
      n = acc[h].numEgs;
      if (n<minEgs && trace==0)
         HError(0,"Insufficient training examples for model %d",h);
      if (n>=minEgs)
         UpdateModel(h,acc+h,defList+h);
      if (trace>0) {
         if (n<minEgs)
            printf("Model %d copied: only %d examples\n",h,n);
         else
            printf("Model %d updated with %d examples\n",h,n);
         fflush(stdout);
      }
   }
   if (isTied)
      UpdateTiedMixtures();
   for (h=1;h<=H;h++){
      MakeFN(hmmFn[h],newDir,newExt,outfn);
      HSave(outfn,defList+h);
   }
   if (trace>0)
      printf("Reestimation complete - average log prob per frame = %e\n",
         totalPr/totalT);
}


/* ----------------------------------------------------------- */
/*                      END:  HERest.c                         */
/* ----------------------------------------------------------- */
