/* ----------------------------------------------------------- */
/*                   HTK  -  HMM Toolkit                       */
/* ----------------------------------------------------------- */
/*         Copyright: David Rainton    September 1991          */
/*         Last Update:                                        */
/*                                                             */
/*         ATR Interpreting Telephony Research Labs            */
/*         Japan                                               */
/*                                                             */
/* Permission is hereby granted to copy, use and modify this   */
/* software provided that this banner notice is not removed    */
/*                                                             */
/*  D. Rainton - Modification History                          */
/*                                                             */
/*                                                             */
/* ----------------------------------------------------------- */
/*         File: HGPD.c: GPD Training Program                  */
/* ----------------------------------------------------------- */


#include <stdio.h>      /* Standard C Libraries */
#include <stdlib.h>
#include <math.h>

#include "HShell.h"     /* HMM ToolKit Modules */
#include "HMath.h"
#include "HSpIO.h"
#include "HLabel.h"
#include "HModel.h"
#include "HParse.h"
#include "HUtilities.h"

#define VERSION   "1.1"
#define SAME  -1

/* Types of loss function */ 
#define POWER 0
#define SIG  1
#define PIECEWISELINEAR 2

#define SIG_ARG "s"
#define POWER_ARG "p"
#define PIECEWISELINEAR_ARG "l"

#define TOWARDS +1
#define AWAY -1

#define LEVELSEP '/'

#define MODEL_NOT_FOUND -1
#define NOT_DEFINED 0.0

/* Global Settings */

static char * labDir = NULL;     /* label file directory */
static char * labExt = "rec";    /* label file extension */

char * hmmDir = NULL;     /* directory to look for hmm def files */
char * hmmExt = NULL;     /* hmm def file extension */

static char * outHmmDir = NULL;    /* directory to place new hmm def files */
static char * pathDir = NULL;      /* directory to look for Viterbi path files */
static char * pathExt = "ip";      /* path file extension */
static double Tot_epochs = 50.0;   /* total number of epochs of training*/
static double  Epoch = 1.0;        /* current epoch */
static double wepsilon = 0.01;     /* weight learning constant */
static double mepsilon = 0.01;     /* mean learning constant */
static double vepsilon = 0.01;     /* covariance learning constant */
static int LossFunctionType = POWER; /* the default loss function */
enum _UPDSet{UPMEANS=1,UPVARS=2,UPTRANS=4,UPMIXES=8};
typedef enum _UPDSet UPDSet;
static UPDSet uFlags = UPMEANS|UPVARS|UPMIXES;     /* update flags */

static double alpha = 1.0;         /* the loss function power */
static HMMParameters hmmP;  
static int H;                      /* number of distinct HMM's (Labels) */
static int N=0;                    /* the number of data files */

int  trace    = 0;                 /* Trace level NOT STATIC */
static float wscale = 1.0;         /* the weight scaling factor */
static float mscale = 1.0;         /* the mean scaling factor */

/* Global data structures */

static HMMDef *HMMdefList;          /* array[1..H] of HMM definitions */
static HMMDef *HMMupdateList;       /* array[1..H] of HMM GPD updates */
static LabId *HMMidList;            /* array[1..H] of HMM file name id */
static DataFile data;               /* structure holding current data file */
static Vector ot;                   /* current observation vector */
static Vector um;                   /* temporary vector to store  mean gradient*/
static  Covariance ucov;            /* temperory union to store covariance gradient*/


/* ------------------ Process Command Line ------------------------- */

void ReportUsage(void)
{
   fprintf(stderr,"\nUSAGE: HGPD[V%s] [options] hmmlist  DataFiles...\n\n",
           VERSION);
   fprintf(stderr," Option                                        Default\n\n");
   fprintf(stderr," -s f    weight scaling factor                     1.0\n");
   fprintf(stderr," -b f    mean scaling factor                       1.0\n");
   fprintf(stderr," -w f    weight learning constant                -1.00\n");
   fprintf(stderr," -m f    mean learning constant                  -1.00\n");
   fprintf(stderr," -a f    loss function power (alpha)             -1.00\n");
   fprintf(stderr," -E f    total number of epochs of training         50\n");
   fprintf(stderr," -e f    current epoch\n");
   fprintf(stderr," -d s    dir to find hmm files                 current\n");
   fprintf(stderr," -D s    extension for hmm files                  none\n");
   fprintf(stderr," -F fmt  Set data file format to fmt               HTK\n");
   fprintf(stderr," -G fmt  Set the label file format to fmt          HTK\n");
   fprintf(stderr," -o s    dir to store new hmm files            current\n"); 
   fprintf(stderr," -p s    dir to find path files                current\n");
   fprintf(stderr," -P s    extension for the  path files              ip\n");
   fprintf(stderr," -l s    dir to find label files                  none\n"); 
   fprintf(stderr," -L s    extension for label files                 rec\n");
   fprintf(stderr," -S f    set script file to f                     none\n");
   fprintf(stderr," -T N    set trace level to N                        0\n");
   fprintf(stderr," -f s    loss function (options l, s, p)             p\n");
   fprintf(stderr," -u tmvw Update t)rans m)eans v)ars w)ghts        tmvw\n");

   exit(1);
}

char * LossFunct2Str(int type)
{
 switch (type)
 {
   case SIG:              return(SIG_ARG); 
                          break;
   case POWER:            return(POWER_ARG); 
                          break;
   case PIECEWISELINEAR:  return(PIECEWISELINEAR_ARG);
                          break;

   default:               HError (1,"Invalid loss function type");
                          break; 
 }
}


int Str2LossFunct(char *s)
{
 if (!strcmp(s, SIG_ARG)) return (SIG);
 else
   if (!strcmp(s, POWER_ARG)) return (POWER);
   else
     if (!strcmp(s, PIECEWISELINEAR_ARG)) return (PIECEWISELINEAR);
     else {
       fprintf(stderr, "Invalid loss function type. The choices are:\n");
       fprintf(stderr, "  %s - power loss function\n", POWER_ARG);
       fprintf(stderr, "  %s - sigm loss function\n", SIG_ARG);
       fprintf(stderr, "  %s - piecewise linear loss function\n", PIECEWISELINEAR_ARG);
       exit(1);
     }
}

void PrintArgs(void)
{
   printf(" -E f    total number of epochs %f\n", Tot_epochs);
   printf(" -e f    current epoch %f\n", Epoch);
   printf(" -s f    weight scaling factor %f\n", wscale);
   printf(" -b f    mean scaling factor %f\n", mscale);
   printf(" -w f    weight learning constant      %e\n",wepsilon);
   printf(" -m f    mean learning constant      %e\n",mepsilon);
   printf(" -a f    loss function power (alpha)      %lf\n",alpha);
   printf(" -f s    loss function (options l, s, p) %s\n", LossFunct2Str(LossFunctionType));
   if (uFlags&UPMEANS)  printf("Updating mixture means\n");   
   if (uFlags&UPMIXES)  printf("Updating mixture weights\n");
   if (uFlags&UPVARS)   printf("Updating mixture covariances\n");
   printf("\n");
   fflush(stdout);
}

void SetuFlags()
{
   char *s;
   
   s=GetStrArg();
   uFlags=0;        
   while (*s != NULL)
      switch (*s++) {
         case 't': uFlags+=UPTRANS;break;
         case 'm': uFlags+=UPMEANS;break;
         case 'v': uFlags+=UPVARS;break;
         case 'w': uFlags+=UPMIXES;break;
         default: HError(2,"Unknown update flag %c",*s);
                  break;
      }
}

main(int argc, char *argv[])
{
   char *datafn, *s;
   FileFormat ff;
   char *listfn;
   void Initialise(char * listfn);
   void GPDFile(char * datafn);
   void GPDUpdate();
   void ScaleUpdates();
   
   InitShell(argc,argv);
   InitMath(FALSE);        /* Doesnt use LAdd */
   InitLabel();
   if (NumArgs()==0)
      ReportUsage();
   while (NextArg() == SWITCHARG) {
      s = GetSwtArg();
      if (strlen(s)!=1) 
         HError(1,"Bad switch %s; must be single letter",s);
      switch(s[0]){
         case 'u':
            SetuFlags();
            break;
         case 's':
            wscale = GetChkedFlt(-10E+99,10E+99,s); 
            break;
         case 'b':
            mscale = GetChkedFlt(-10E+99,10E+99,s); 
            break;
         case 'w':
            wepsilon = GetChkedFlt(-10E+99,10E+99,s); 
            break;
         case 'm':
            mepsilon = GetChkedFlt(-10E+99,10E+99,s); 
            break;
         case 'v':
            vepsilon = GetChkedFlt(-10E+99,10E+99,s); 
            break;
         case 'a':
            alpha = GetChkedFlt(-10E+99,10E+99,s); 
            break;
         case 'E':
            Tot_epochs = GetChkedFlt(-10E+99,10E+99,s); 
	    break;
         case 'e':
            Epoch = GetChkedFlt(-10E+99,10E+99,s); 
	    break;
         case 'd':
            if (NextArg()!=STRINGARG)
               HError(1,"HMM definition directory expected");
            hmmDir = GetStrArg(); break;     
         case 'D':
            if (NextArg()!=STRINGARG)
               HError(1,"HMM file extension expected");
            hmmExt = GetStrArg(); break;
         case 'F':
            if (NextArg() != STRINGARG)
               HError(1,"Data File format expected");
            if((ff = Str2Format(GetStrArg())) == ALIEN)
               HError(0,"Warning ALIEN Data file format set");
            SetFormat(ff);
            break;
         case 'G':
            if (NextArg() != STRINGARG)
               HError(1,"Label File format expected");
            if((ff = Str2Format(GetStrArg())) == ALIEN)
               HError(0,"Warning ALIEN Label file format set");
            SetLFormat(ff);
            break;
         case 'o':
            if (NextArg()!=STRINGARG)
               HError(1,"Output hmm file directory expected");
            outHmmDir = GetStrArg(); break;
         case 'f':
            if (NextArg()!=STRINGARG)
               HError(1,"Loss function type expected");
            LossFunctionType = Str2LossFunct(GetStrArg()); break;
         case 'p':
            if (NextArg()!=STRINGARG)
               HError(1,"Path file directory expected");
            pathDir = GetStrArg(); break;
         case 'P':
            if (NextArg()!=STRINGARG)
               HError(1,"Path file extension expected");
            pathExt = GetStrArg(); break;
         case 'l':
            if (NextArg()!=STRINGARG)
               HError(1,"Label file directory expected");
            labDir = GetStrArg(); break;
         case 'L':
            if (NextArg()!=STRINGARG)
               HError(1,"Label file extension expected");
            labExt = GetStrArg(); break;
         case 'S': SetScriptFile();
            break;
         case 'T':
            trace = GetChkedInt(0,3,s); break;
         default:
            HError(1,"Unknown switch %s",s);
      }
   } 
   if (Epoch == 1.0) PrintArgs();
   if (NextArg()!=STRINGARG)
      HError(1,"hmm list file name expected");
   listfn = GetStrArg();
   Initialise(listfn);

/* set the values for the learning constants to get monotonically decreasing
values */

   wepsilon = wepsilon * (Tot_epochs - (Epoch-1.0))/Tot_epochs;
   mepsilon = mepsilon * (Tot_epochs - (Epoch-1.0))/Tot_epochs;
   vepsilon = vepsilon * (Tot_epochs - (Epoch-1.0))/Tot_epochs;
   printf("Epsilon  values: wts: %f, means %f\n",wepsilon,mepsilon);

   do {
      if (NextArg()!=STRINGARG)
         HError(1,"training data file name expected");
      datafn = GetStrArg();
      N+=1;
      if (trace>0) printf("Loaded data file %s\n", datafn); 
     GPDFile(datafn);
   } while (NumArgs()>0);

   if(N>0) {
     ScaleUpdates();
     GPDUpdate();
     SaveHMMDefs(HMMdefList, HMMidList, outHmmDir, hmmExt, H);
   }
   else HError(0,"No data files update the HMM's with");
}

 
Boolean CorrectlyRecognised(Transcription tr)
{
int i;

if (tr->nLabels != tr->next->nLabels) return (0);
else 
  for (i=0;i<tr->nLabels;i++)
    if (strcmp(tr->labels[i].labid->name,tr->next->labels[i].labid->name)) return(0);

return (1);
}

/*--------------------------------------------------------------*/
/* These routines are for accessing Viterbi path information from 
   the label files. The routines were written because the path
   information is rather awkwadly squeezed into the label struture,
   so direct access would be rather confusing
*/

int GetTime(int t, Label *l) { return(l[t].start); }
int GetState(int t, Label *l) { return(l[t].end); } 
double Getlogbjt(int t, Label *l) { return(l[t].score); }
char *GetLabelModelName(int t, Label *l) { return(l[t].labid->name); }



void FlagDifferences(Transcription path)
{
 int i;
 short N; 
 LabList * TruePath, * RecPath;

 N = NumLevels(path);  
 if(trace >2) printf ("Number of path levels = %d\n", N);
 RecPath = path;
 TruePath = path->next;

 for (i=0;i<path->nLabels;i++) {
   if ((strcmp(RecPath->labels[i].labid->name,
	       TruePath->labels[i].labid->name) == 0)   
       && (GetState(i, RecPath->labels) == GetState(i, TruePath->labels))) {
     RecPath->labels[i].score  = SAME;  
     TruePath->labels[i].score = SAME;
   }
 }
 
}


int Name2HmmIndex(char *name, int H)
{
 int index=MODEL_NOT_FOUND;
 int h;

 for(h=1;h<=H;h++) 
   if(!strcmp(HMMidList[h]->name,name)) index = h;
 /* return an error is the HMM is not found */
 if (index == MODEL_NOT_FOUND) 
   HError(2,"HMM Model %s not found", name);
 else return(index);
}

/*---------------------------------------------------------------*/
/* The following routines compute dP/d(mean), and dP/d(cov) */

void DeltaProbDeltaMean(Vector m,Vector o,double wm,
			Covariance cov,Vector um,double pm)
{
 int i;
 double k;

 k = wm * pm;

 if (hmmP.CK == DIAGC) 
   for(i=1;i<=hmmP.vSize;i++)  {
    if(cov.var[i]==0.0) HError(1,"covariance element is zero");
    um[i] =  k * (o[i]-m[i]) / cov.var[i];
  }

 if (hmmP.CK == FULLC) /*not implemented yet*/;

}

void DeltaProbDeltaCov(double wm, double pm, Vector ot,
		       Vector m, Covariance cov, Covariance ucov)
{
 int i;
 double k1,k2;

 k1 = wm*pm/2;

 if (hmmP.CK == DIAGC) 
   for(i=1;i<=hmmP.vSize;i++) {
    k2 = cov.var[i];
    if(k2==0.0)HError(1,"covariance element is zero");
    ucov.var[i] = k1/k2 * ( pow((ot[i]-m[i]),2)/k2 -  1.0 );
   }

 if (hmmP.CK == FULLC) /*not implemented yet*/;

}

/*---------------------------------------------------------------*/
/* These routines compute dg/d(mean) and dg/d(cov) */

void DeltaDiscDeltaMean(Vector um, double po)
{
 VectorDivide(um,po);
}


void DeltaDiscDeltaCov(Covariance cov, double po)
{

 if (hmmP.CK == DIAGC) VectorDivide(cov.var,po);

 if (hmmP.CK == FULLC) /* not implemented yet */;

}
/*-------------------------------------------------------*/
/* These routines compute dd/d(mean) and dd/d(cov) */


void  DeltaLossDeltaMean(Vector um, double lambda)
{
 VectorMultiply(um,lambda);
}

void  DeltaLossDeltaCov(Covariance cov, double lambda)
{

 if (hmmP.CK == DIAGC) VectorMultiply(cov.var,lambda);

 if (hmmP.CK == FULLC) /* not implemented yet */;

}


int UpdateGradient1(int J, Label *l, double sign, double d)
{
 int t;             /* current frame  */
 int s;             /* current state  */
 int i;
 int j;

 int index;
 HMMDef *hmm=NULL;
 HMMDef *uhmm=NULL;
 char * ModelName;

 double logpo;       /* log observation probability */
 double logpm;       /* log mixture observation probability */
 double po;          /* observation probability */
 double pm;          /* mixture observation probability */

 float  w ;          /* mixture weight */
 double logw;        /* the log of the mixture weight */
 Vector m;           /* mixture mean */ 
 Covariance cov;     /* mixture variance */

 float  uw;             /*  update weight */

 for (j=(J-1);j>=0;j--) { /* note tha j is simply an index and runs backwards in time */
   if (l[j].score != SAME) {
     
     /* find the HMM in the database with ModelName */
     t = GetTime(j,l);
     ModelName = GetLabelModelName(j, l);
     index = Name2HmmIndex(ModelName, H);
     hmm = HMMdefList+index;
     uhmm = HMMupdateList+index;
     s = GetState(j,l); 
     
     if (trace==4)
       printf(" time[%d] Model[%d] name[%s] state[%d]\n",t,index,ModelName, s);
     
     GetSample(&data,t-1,ot+1); /* frame times 1->N and data times 0->N-1 */ 
     logpo = OutP(ot, hmm, s);
     po = L2F(logpo);
     
     if(po != 0.0) {
       for (i=1;i<=hmmP.M;i++) {
         if(HMMMixtureExists(hmm,s,i)) {
	   
           logpm = MOutP(ot, hmm, s, i);
           pm = L2F(logpm);
           w   =  hmm->stateinfo[s].outp[i].weight; 
           m   =  hmm->stateinfo[s].outp[i].mean;
           if(hmmP.CK==DIAGC) cov.var= hmm->stateinfo[s].outp[i].covariance.var;
           else /* unimplemented yet */;
	   
           if (uFlags&UPMEANS)             DeltaProbDeltaMean(m,ot,w,cov,um,pm); 
           if (uFlags&UPMIXES && hmmP.M>1)     uw = pm; 
           if (uFlags&UPVARS)              DeltaProbDeltaCov(w,pm,ot,m,cov,ucov);

           if (trace==4) {
             if (uFlags&UPMIXES && hmmP.M>1) printf("dp/dw = %f\n", uw);
             if (uFlags&UPMEANS) ShowVector("dp/dm",um, hmmP.vSize);
           }

           if (uFlags&UPMEANS)                 DeltaDiscDeltaMean(um,po); 
           if (uFlags&UPMIXES && hmmP.M>1)     uw /= po;
           if (uFlags&UPVARS)                  DeltaDiscDeltaCov(ucov,po);

           if (trace==4) {
             if (uFlags&UPMIXES && hmmP.M>1) printf("dg/dw = %f\n", uw);
             if (uFlags&UPMEANS) ShowVector("dg/dm",um, hmmP.vSize);
	   }

           if (uFlags&UPMEANS)                 DeltaLossDeltaMean(um,sign*d); 
           if (uFlags&UPMIXES && hmmP.M>1)     uw = d*sign*uw;
           if (uFlags&UPVARS)                  DeltaLossDeltaCov(ucov,sign*d);


           if (trace==4) {
             printf("d = %f\n", d);
             if (uFlags&UPMIXES && hmmP.M>1) printf("dl/dw = %f\n", uw);
             if (uFlags&UPMEANS) ShowVector("dl/dm",um, hmmP.vSize);
	   }

           if (uFlags&UPMEANS)   AddVectors(uhmm->stateinfo[s].outp[i].mean, um);
           if (uFlags&UPMIXES&&hmmP.M>1) uhmm->stateinfo[s].outp[i].weight += uw;
           if (uFlags&UPVARS)                  
	     if(hmmP.CK == DIAGC) 
	       AddVectors(uhmm->stateinfo[s].outp[i].covariance.var, ucov.var);
	     else   /* not implemented yet */ ;

           if (trace==4) {
             if (uFlags&UPMIXES && hmmP.M>1) 
	       printf("accumulated weight update = %f\n",
		      uhmm->stateinfo[s].outp[i].weight);
             if (uFlags&UPMEANS)
	       ShowVector("accumulated mean update",
			  uhmm->stateinfo[s].outp[i].mean, hmmP.vSize);
	   }
         }
       }
     }
   }
 }

} /* end of UpdateGradient1() */

void  UpdateGradient(Transcription path, double d)
{
  /* Viterbi Path */
  UpdateGradient1(path->nLabels, path->labels, AWAY, d); 

  /* Hand Labelled Path */
  UpdateGradient1(path->next->nLabels, path->next->labels, TOWARDS, d);
}

double TotalLogProb(int I, Label *l)
{
 int i, s1,s2;
 double logp=0.0;
 int index;
 HMMDef *hmm;
 char *ModelName;

 for(i=(I-1);i>0;i--) {  /* note the labels are backwards in time */
 
   s1 = GetState(i, l);
   s2 = GetState(i-1,l);

   ModelName = GetLabelModelName(i, l);
   index = Name2HmmIndex(ModelName, H);
   hmm = HMMdefList+index;
   
   GetSample(&data,GetTime(i,l)-1,ot+1); /*note data 0->N-1 while time is 1->N*/ 
   logp += OutP(ot, hmm, s1);
   if(s2>=s1) logp += hmm->transP[s1][s2]; 
   else       logp += hmm->transP[s1][s1+1];

 }
 GetSample(&data,I-1,ot+1); 
 logp += OutP(ot, hmm, s2);
 logp += hmm->transP[s2][s2+1];

 return (logp);
}

/*------------------------------------------------------*/
/* routines for implementing the various loss functions */

/* Sig: returns (1/(1+exp(-a(x-b))) */
double Sig (double x, double a, double b)
{
 
 return( 1 / (1 + exp(-a*(x-b))) );

}

/* DeltaSig: returns the first differential of Sig(x,a,b) */
double DeltaSig (double x, double a, double b)
{

 return( pow(Sig(x,a,b),2) *a*exp(-a*(x-b)) );

}

/* DeltaPow: returns the first differential of f(x) where
   f(x) = 0     x<= 0
   f(x) = x^a   x > 0
*/
double DeltaPow(double x, double a)
{
 
 if (x<=0.0) return (0.0); 
 return(a*pow(x,(a-1.0)));
 
}

/*DeltaPieceWiseLinear: returns the first differential of the piecewise linear function f(x)
   where   f(x) = 0  x<= a 
           f(x) = x-a/(b-a) b<x>a
           f(x) = 1  x>= b
*/
double DeltaPieceWiseLinear(double x, double a, double b)
{
 
 if (a==b) HError(1,"Infinite slope");
 if (a>b)  HError(1,"Negative slope");

 if (x<=a || x>=b) return (0.0);
 else              return (1.0/(b-a));

}

/*------------------------------------------------------*/

double Misclass(Transcription path)
{
 double d;

 d = TotalLogProb(path->nLabels,path->labels) /*Viterbi*/ - 
     TotalLogProb(path->next->nLabels,path->next->labels); /*Hand Labelled*/
 if (d<0.0) HError(1,"The misclassifcation function (%f)  is negative",d); 
 if (trace == 3) printf("the misclassification measure is %f\n", d);

 return (d);
}

/* DLossDMisclass: returns the first differential of the loss function wrt
                   the misclassification measure
*/ 
double DLossDMisclass(Transcription path)
{
 double d;

 d = Misclass(path);

 switch (LossFunctionType) 
 {
   case POWER: d = DeltaPow(d,alpha);
               break;

   case  SIG: d = DeltaSig(d, alpha, 0.0);
               break;

   default:  HError(1,"Invalid loss function type - should never get here");
             break;
 }
 
 if( (trace>=1) && (d == 0) ) 
   HError(0,"The first differential of the misclassification measure is zero\n");
 if (trace==4) printf("The first differential of d is %f\n", d);

 return (d);
 
}


void GPDFile(char * datafn)
{
   char labfn[255],pathfn[256];
   Transcription tr,path;    
   int i;
   double d;

   MakeFN(datafn,labDir,labExt,labfn); /* find label file */
   MakeFN(datafn,pathDir,pathExt,pathfn); /* find Viterbi path file */

   if (trace>=2) {
    printf ("The label file name is %s\n", labfn);
    printf ("The path file name is %s\n", pathfn);
    fflush (stdout);
   }
 
   tr = LOpen(labfn);
   if (trace==4) PrintTranscription(tr);

   if (!CorrectlyRecognised(tr)) {  
     if (trace>0) printf ("Viterbi and Hand Labels do not agree in %s \n",labfn);

    SpOpen(datafn,&data);
    CheckData(&data,datafn,hmmP.vSize,hmmP.SK);

    path =  LOpen(pathfn); 
    FlagDifferences(path);
    if (trace>3) PrintTranscription(path);

    d = DLossDMisclass(path);

    UpdateGradient(path,d); 

    SpClose(&data);
  }

  LClose(tr); 

}


/*ZeroNegativeMixtures: implements 'if (w_i<MINMIX) then w_i = 0'.
Mixture weights must be >=0 so all w_i<0 must be set to 0. */
void ZeroNegativeMixtures(OutProb *outp)
{
 int m;

 for(m=1;m<=hmmP.M;m++) if(outp[m].weight<0.0) outp[m].weight = 0.0;
} 


/* SumMixtureWeights2One: normalises the mixture weights using
   w_i = w_i / sum_j=1^N w_j where N is the number of mixtures. 
   Note prior to normalisation negative mixture weights are set to 0. */
void  SumMixtureWeights2One(OutProb *outp)
{
 double sum=0.0;
 int m;

 ZeroNegativeMixtures(outp);
 for(m=1;m<=hmmP.M;m++) sum += outp[m].weight;
 if (sum != 0.0) for(m=1;m<=hmmP.M;m++) outp[m].weight /= sum; 
 else HError(1,"Oops - Model state has disappeared!");

}


void GPDUpdate(void)
{
 int h,s,S,m,i;
 double update;
 HMMDef *updateHMM, *oldHMM;

 for (h=1;h<=H;h++) { 
   oldHMM = HMMdefList+h;
   updateHMM = HMMupdateList+h;

   S = oldHMM->numStates; /* I think this can vary across models */


   for (s=2;s<S;s++){ 
     for (m=1;m<=hmmP.M;m++) {
       if(HMMMixtureExists(oldHMM,s,m)) {

         if (uFlags&UPMEANS) 
	   VectorMultiply(updateHMM->stateinfo[s].outp[m].mean, mepsilon);
         if (uFlags&UPMIXES && hmmP.M>1) 
	   updateHMM->stateinfo[s].outp[m].weight *= wepsilon;
         if (uFlags&UPVARS)
           if (hmmP.CK == DIAGC)  
             VectorMultiply(updateHMM->stateinfo[s].outp[m].covariance.var,
			    vepsilon);
       }
     }
   }

   if(trace>=2) {
     printf("Update HMM parameters for model %s\n", HMMidList[h]->name);
     DRHPrint(updateHMM, HPDETAILS|HPMEANS|HPVARS,2,S-1);
   }

   for (s=2;s<S;s++){ 


    for (m=1;m<=hmmP.M;m++) {
      if(HMMMixtureExists(oldHMM,s,m)) {

        if (uFlags&UPMEANS) 
	  AddVectors(oldHMM->stateinfo[s].outp[m].mean, 
		     updateHMM->stateinfo[s].outp[m].mean);
        if (uFlags&UPMIXES && hmmP.M>1)
	  oldHMM->stateinfo[s].outp[m].weight += 
	    updateHMM->stateinfo[s].outp[m].weight;
        if (uFlags&UPVARS)
          if (hmmP.CK == DIAGC) 
            AddVectors(oldHMM->stateinfo[s].outp[m].covariance.var, 
		       updateHMM->stateinfo[s].outp[m].covariance.var);

      }
    }

    if (uFlags&UPMIXES && hmmP.M>1)
      SumMixtureWeights2One(oldHMM->stateinfo[s].outp);
   
   }

 } /* end of h */

} /* end of GPDUpdate() */


void  ScaleUpdates(void)
{

 int h,s,S,m;
 double update;
 HMMDef *updateHMM,*oldHMM;
 double MaxAbsWeight=0.0, MaxMeanMod=0.0;
 double MaxOldAbsWeight=0.0, MaxOldMeanMod=0.0;

 for (h=1;h<=H;h++) { 
   updateHMM = HMMupdateList+h;
   oldHMM = HMMdefList+h;

   S = oldHMM->numStates; /* I think this can vary across models */

   for (s=2;s<S;s++){ 
     for (m=1;m<=hmmP.M;m++) {

       if(HMMMixtureExists(oldHMM,s,m)) {
         if (fabs(updateHMM->stateinfo[s].outp[m].weight) > MaxAbsWeight)
           MaxAbsWeight = fabs(updateHMM->stateinfo[s].outp[m].weight);
         if (fabs(oldHMM->stateinfo[s].outp[m].weight) > MaxOldAbsWeight)
           MaxOldAbsWeight = fabs(oldHMM->stateinfo[s].outp[m].weight);
         if (Vector2Norm(updateHMM->stateinfo[s].outp[m].mean) > MaxMeanMod)
           MaxMeanMod = Vector2Norm(updateHMM->stateinfo[s].outp[m].mean);
         if (Vector2Norm(oldHMM->stateinfo[s].outp[m].mean) > MaxOldMeanMod)
           MaxOldMeanMod = Vector2Norm(oldHMM->stateinfo[s].outp[m].mean);
       }
     }
   }
 }

 if (uFlags&UPMIXES && hmmP.M>1)
   if(MaxAbsWeight == 0.0) HError(1,"No weight updates");
 if (uFlags&UPMIXES && hmmP.M>1) 
   if(MaxOldAbsWeight == 0.0) HError(1,"No old weight");
 if (uFlags&UPMEANS) if(MaxMeanMod == 0.0)   HError(1,"No mean updates");
 if (uFlags&UPMEANS) if(MaxOldMeanMod == 0.0)   HError(1,"No old means");


 for (h=1;h<=H;h++) { 
   updateHMM = HMMupdateList+h;
   oldHMM = HMMdefList+h;

   S = oldHMM->numStates; /* I think this can vary across models */

   for (s=2;s<S;s++){ 
     for (m=1;m<=hmmP.M;m++) {
       if(HMMMixtureExists(oldHMM,s,m)) {
         if (uFlags&UPMEANS)
	   VectorMultiply(updateHMM->stateinfo[s].outp[m].mean, 
			  MaxOldMeanMod/MaxMeanMod);
         if (uFlags&UPMEANS) 
	   VectorMultiply(updateHMM->stateinfo[s].outp[m].mean, mscale);
         if (uFlags&UPMIXES && hmmP.M>1) 
	   updateHMM->stateinfo[s].outp[m].weight *= 
	     (MaxOldAbsWeight/MaxAbsWeight);
         if (uFlags&UPMIXES && hmmP.M>1) 
	   updateHMM->stateinfo[s].outp[m].weight *= wscale;
       }
     }
   }
 }


}

/* ------------------------ Initialisation ----------------------- */

/* Initialise: initialise the global data structures */
void Initialise(char * listfn)
{
   HMMidList = LoadIdList(listfn,&H);
   if (trace >2) printf("%d HMM id's loaded\n", H);
   HMMdefList = LoadHMMDefs(HMMidList, H, hmmDir, hmmExt, &hmmP);
   HMMupdateList = MallocHMMDefList(HMMdefList, H); 
   ot = CreateVector(hmmP.vSize);
   um = CreateVector(hmmP.vSize); 
   if (hmmP.CK == FULLC)
     ucov.inv = CreateMatrix(hmmP.vSize,hmmP.vSize);
   else
     ucov.var = CreateVector(hmmP.vSize);


}






































































