/* ----------------------------------------------------------- */
/*                   HTK  -  HMM Toolkit                       */
/* ----------------------------------------------------------- */
/*         Copyright: Steve Young      October 1989            */
/*         Last Update:                December 1990           */
/*                                                             */
/*         Cambridge University Engineering Department         */
/*         Trumpington Street, Cambridge, CB2 1PZ              */
/*                                                             */
/* Permission is hereby granted to copy, use and modify this   */
/* software provided that this banner notice is not removed    */
/*                                                             */
/*  D. Rainton - Modification History                          */
/*                                                             */
/* July 2 1991 - subspace code added                           */
/* Aug  9 1991 - N-best code added                             */
/* Aug 14 1991 - internal state path output added              */
/* Oct 21 1991 - output scores to ipath file changed           */
/*               from observation probabilites to              */
/*               accumulated probabilities                     */
/*                                                             */
/*  A. Duchon  - Modification History                          */
/*                                                             */
/* Dec 27 1991 - learning rate change to be linear             */
/*             - files randomly presented for learning         */
/* Jan 06 1992 - program changed to allow updates on each token*/
/* Jan 21 1992 - Previous program was actually performing LVQ3 */
/*               when NBest = 1.                               */
/*               When NBest = 2, it might adjust               */
/*               the top 1 _or_ 2 competing classes.           */
/*               With NBest = 3, the top 2 or 3, etc.          */
/*               Now it is changed to do real NBest in terms   */
/*               of GPD, i.e. cor - Nbest competing classes.   */
/* Feb 28 1992 - Other loss functions added: power and         */
/*               piecewise linear.                             */
/*                                                             */
/* ----------------------------------------------------------- */
/*         File: HGPD3.c: Viterbi Recogniser GPD update        */
/* ----------------------------------------------------------- */
/*
   This program will match a network of HMM's against one or more
   speech files and output a transcription in HTK format.  The 
   match is performed using conventional Viterbi decoding.  The
   network is input via HParse, all nodes are assumed to be
   the names of phone labels.  The HMM definition file 
   corresponding to each label must consist of the label name 
   followed by the specified extension (which may be null).
   The set of all hmm's needed must be listed in an idList.
   
   An optional bigram may be specified via the -b option.  The
   bigram should be a text file stored in the format
   
     <lab1> p11 p12 p13  ... p1H
     <lab2> p21 p22 p23  ... p2H
           ...
     <labH> pH1 pH2 pH3  ... pHH
     
   where <labh> is the name of the h'th label in the idList and
   pij is the bigram probabity P(labj|labi).
   
   Recognition may be controlled by the following optional 'scaling'
   parameters:
      
         propPen  - a fixed inter-model transition probability
         bigScale - scale factor for log(bigram) probabilities
         bjtScale - scale factor for log(bj(ot)) probabilities 

   
   The principle data structure is the network corresponding to
   the regular grammar rules used to constrain the recognition.
   This network is built by HParse.  For each phoneme instance,
   a PhoneInstance struct is created and hung-off the 'user' member
   of the corresponding network node.  Each PhoneInstance contains
   a pointer to a Definition struct containing the HMM definition 
   and space for bj(ot) to save recalculating multiple instances. 
   Definition structs are shared by the PhoneInstances so that 
   there is only one copy of each distinct HMM.  
   
   The recogniser design follows the token passing model.
*/


#include <stdio.h>      /* Standard C Libraries */
#include <stdlib.h>
#include <math.h>
#include <string.h>

#include "HShell.h"     /* HMM ToolKit Modules */
#include "HMath.h"
#include "HSpIO.h"
#include "HLabel.h"
#include "HModel.h"
#include "HParse.h"
#include "HUtilities.h"

#define VERSION   "2.5"
#define GCMARGIN   400
#define MAX_N_BEST 20
#define LINE_LEN 256
#define UNDEFINED 0
#define MAXFILES 900   

/* ------------ Tokens, Definitions and PhoneLink Records -------------- */

typedef struct _Definition Definition;
typedef struct _PLR PLR;
typedef struct _SLR SLR;
typedef struct _PhoneInstance PhoneInstance;
typedef struct _FileNameEntry FileNameEntry;
typedef struct _PathCell PathCell;
typedef struct _PathCellList PathCellList;
typedef PathCellList *Path;

enum _UPDSet{UPMEANS=1,UPVARS=2,UPTRANS=4,UPMIXES=8};
typedef enum _UPDSet UPDSet;
static UPDSet uFlags = UPMEANS|UPVARS|UPMIXES;     /* update flags */

typedef struct {
   PhoneInstance * inst;   /* current owner */
   float logprob;          /* score */
   PLR   *link;            /* link */
   SLR   *slrlink;         /* DR - link to SLR */
} Token;

struct _Definition{
   HMMDef hmmdef;       /* actual HMM def */
   float *bjot;         /* array[2..numStates-1] of bj(ot) */
   LabId id;            /* id of this defn */
   Boolean bjLoaded;    /* true if bjot loaded */
};

struct _PLR{
   Boolean linked;      /* flag used in garbage collection */
   Token tok;           /* token at boundary */
   int time;            /* index of model boundary */
   PLR *chain;          /* chain for keeping track of PLR's */
   PLR *alt;            /* chain for keeping track of PLR alternatives */
};

struct _SLR{
   int time;            /* time of observation */
   int state;           /* the model state */
   float bjot;          /* probability of observation */
   SLR *link;           /* points to the previous SLR*/
   SLR *chain;          /* for house keeping */
   char *name;          /* model name */
   float logprob;       /* the accumulated observation probability */
 };                  

struct _PathCell{
   int time;               /* time of observation */
   int state;              /* the model state */
   float bjot;             /* probability of observation */
   float logprob;          /* the accumulated observation probability */
   PathCell  *link;        /* points to the previous SLR*/
   char *name;             /* model name */
   PathCell *chain;        /* for global freeing of pathcells */
 };                  

struct _PathCellList{
   PathCell *pcell;            /* Array[0..nCells-1]Of PathCell */
   PathCellList *next;         /* Next (lower) level if any */
   PathCellList *chain;        /* for global freeing of pathcelllists */
};


struct _PhoneInstance{
   Definition *def;  /* definition for this instance */
   Link netHook;     /* Network node holding this instance */
   int N;            /* copy of hmm[id].numStates */
   Boolean isActive; /* true if pi is in active list */
   Token *state;     /* array[1..N] of token */
   Token *tbuf;      /* list of NBest tokens */
   int lastplrt;     /* time when last PLR gen'ed from this inst */
   PLR *lastplr;     /* pointer to last PLR gen'ed from this inst */
};


/* this structure is for holding the total expected probability of a model producing the correct
   and wrong frames */

struct ModelStatistics { 
 int CorrectNFrames;
 int WrongNFrames;
 double Correctbjot;
 double Wrongbjot;
};


/* Global Settings */

static char *files[MAXFILES];    /* array of file nums read in */
static int filenum[MAXFILES];    /* array of file order    */
static int Nfiles;               /* total number of files  */
static int NumCorrect;           /* counter for correct recognition at global lev*/
static int InTopN;             /* counter for correct path in the the N best paths*/
static int ITRNS;                /* number of small iterations steps */

static char * ilabDir = NULL;    /* hand label file directory */
static char * labExt = "lab";    /* label file extension */
static char * rlabDir = NULL;    /* recognised label files */
static char   UniquerlabDir[LINE_LEN];
static char * hmmDir = NULL;     /* directory to look for hmm def files */
static char * hmmExt = NULL;     /* hmm def file extension */
static char * outHmmDir = NULL;  /* directory to place new hmm def files */

static char *netfn = NULL;       /* network file name */
static char *listfn = NULL;      /* script file name */
static char *bigfn = NULL;       /* bigram file name */

static float propPen = 1.0;      /* inter model propagation probability */
static float bigScale = 1.0;     /* bigram scale factor */
static float bjtScale = 1.0;     /* bjot scale factor */
static float pruning = 0.0;      /* pruning threshold */
static Boolean fixedPen = FALSE; /* enable flag for fixed propPen */
static Boolean isPruning= FALSE; /* enable flag for pruning */
int  trace    = 0;               /* Trace level */

static long sampPeriod;          /* data sample period in 100us units */
static FileFormat dataff;
static FileFormat labelff;
static Token nullToken = {NULL, -1.0E10 ,NULL,NULL};  
static Token startToken = {NULL,0,NULL,NULL};

/* Global data structures */

static LabId enterId;            /* LabId of the ENTER node name */
static LabId exitId;             /* LabId of the EXIT node name */
static Network theNet;           /* the syntax constraint network from the grammer*/
static Definition *hmm;          /* array[1..H] of HMM definitions */
static LabId *idList;            /* array[1..H] of HMM file name id */
static int H = 0;                /* number of distinct HMM's (Labels) */
static int P = 0;                /* number of distinct phonemes */
static int NPI = 0;             /* number of phone instances */
static Token *tBuf;              /* space to copy largest array of tokens */
static Boolean hasBigram=FALSE;  /* bigram loaded flag */
static Matrix bigram;            /* bigram stats if any */
static int nActive = 0;          /* number of active phone instances */
static PhoneInstance **active=NULL;  /* array[0..nActive-1]of active pi's */
static LogReal maxLogP;          /* max LogP at current time frame */
static int maxPLRUsed=0;         /* max PLR's used */
static Boolean gcollect=FALSE;   /* flag used to activate garbage collection */
static int FIntrvl=1;         /* the number of files read before each GPD update */
static int NIter=1;   /* the number of GPD iterations through the entire data set */
static DataFile data;            /* the speech data */
static HMMParameters hmmP;       /* stores global HMM parameters */  

static SLR *slrHD=NULL;    /* this is used for a global freeing of SLR structures */
static PathCell *pathcellHD=NULL; /*this is used for a global freeing of pathcells*/
static PathCellList *pathHD=NULL;   /* this is used for a global freeing of pathcelllists */

static Vector ot;              /* the current observation vector */
static FileNameEntry *fnameHD = NULL;  /* pointer to head of linked list of data file names */
static PhoneInstance **hlpath=NULL;  /* an array of the hand labelled path phone instances */
static int NBest=0;             /* output the N-Best parses */

static Boolean ChangeHmmDir=FALSE; /* flag for storing hmms from each iteration in different directory */

/*------------HGPD stuff------------------*/

static Vector um;                 /* temporary vector to store  mean gradient*/
static  Covariance ucov;      /* temperory union to store covariance gradient*/
static Definition *grad=NULL;  /* array[1..H] of HMM gpd gradient infomation */
static Definition *tmphmm=NULL;      /* array[1..H] of HMM temporary storage */
static struct ModelStatistics  *ModelStats;  /* interesting model statistics */
static double alpha = 0.5;            /* the smoothing factor in the sigmoid 
				       or power of power loss, or q1 for
				       piecewise linear function    */
static double beta = 0.0;        /* the offset factor in the loss function */
static double eta=2.0;         /*the counter discriminant membership constant*/
static double Epsilon=0.1;      /* adaptation step size iteration multiplier */

/* Types of loss function */ 
#define POWER 0
#define SIGM  1
#define PIECEWISELINEAR 2

#define SIGM_ARG "s"
#define POWER_ARG "p"
#define PIECEWISELINEAR_ARG "l"
static int LossFunctionType = SIGM; /* the default loss function */

/*----------end HGPD stuff----------------*/

/* ------------------ Process Command Line ------------------------- */

void ReportUsage(void)
{
  fprintf(stderr,"\nUSAGE: HGPD1[V%s] [options] hmmList netFile speechFiles...\n\n",
           VERSION);
   fprintf(stderr," Option                                            Default\n\n");
   fprintf(stderr," -b f    load bigram from file f                   none\n");
   fprintf(stderr," -d s    dir to find hmm definitions               current\n");
   fprintf(stderr," -D s    dir to store new hmm files                current\n"); 
   fprintf(stderr," -c      save hmms in dir.n                        FALSE\n"); 
   fprintf(stderr," -o f    bj(ot) scale factor                       1.0\n");
   fprintf(stderr," -p f    inter model trans prob = f                1/Nsucc\n");
   fprintf(stderr," -s f    bigram scale factor                       1.0\n");
   fprintf(stderr," -t f    set pruning threshold                     0.0\n");
   fprintf(stderr," -X s    extension for hmm files                   none\n");
   fprintf(stderr," -F fmt  Set data file format to fmt               HTK\n");
   fprintf(stderr," -G fmt  Set the hand label file format to fmt     HTK\n");
   fprintf(stderr," -S f    set script file to f                      none\n");
   fprintf(stderr," -T N    set trace level to N                      0\n");
   fprintf(stderr," -Y s    extension for hand label files            lab\n");
   fprintf(stderr," -l s    dir to get hand label files from          none\n");  
   fprintf(stderr," -L s    dir to store new recognised label files   none\n");  
   fprintf(stderr," -N i    produce the N-Best parses                 1\n");
   fprintf(stderr," -g      turn on garbage collection\n");
   fprintf(stderr," -I i    number of GPD iterations through all data 1\n");
   fprintf(stderr," -i i    the number of files read before update    all\n");
   fprintf(stderr," -u tmvw Update t)rans m)eans v)ars w)ghts         tmvw\n");
   fprintf(stderr," -f lsp  loss function l)inear, s)igmoid, p)ower   s\n");
   fprintf(stderr," -a f    loss function smoothing factor (alpha)    0.5\n"); 
   fprintf(stderr," -h f    loss function offset factor    (beta)     0.0\n"); 
   fprintf(stderr," -e f    counter discriminant membership const.    20.0\n"); 
   fprintf(stderr," -x f    Beginning epsilon value                   0.1\n",Epsilon);
   exit(1);

}

char * LossFunct2Str(int type)
{
 switch (type)
 {
   case SIGM:             return(SIGM_ARG); 
                          break;
   case POWER:            return(POWER_ARG); 
                          break;
   case PIECEWISELINEAR:  return(PIECEWISELINEAR_ARG);
                          break;

   default:               HError (1,"Invalid loss function type");
                          break; 
 }
}


int Str2LossFunct(char *s)
{
 if (!strcmp(s, SIGM_ARG)) return (SIGM);
 else
   if (!strcmp(s, POWER_ARG)) return (POWER);
   else
     if (!strcmp(s, PIECEWISELINEAR_ARG)) return (PIECEWISELINEAR);
     else {
       fprintf(stderr, "Invalid loss function type. The choices are:\n");
       fprintf(stderr, "  %s - power loss function\n", POWER_ARG);
       fprintf(stderr, "  %s - sigm loss function\n", SIGM_ARG);
       fprintf(stderr, "  %s - piecewise linear loss function\n", PIECEWISELINEAR_ARG);
       exit(1);
     }
}

void PrintArgs(void)
{
   printf(" Option           for HGPD3                          Value\n\n");
   printf(" -N i    produce the N-Best competing parses         %d\n",NBest-1);
   printf(" -d s    HMM directory %s\n",hmmDir);
   printf(" -I i    number of GPD iterations through all data   %d\n",NIter);
   switch (LossFunctionType) 
     {
     case POWER: 
       printf (" -f      loss function type:                         p)ower\n");
       break;
     case  SIGM: 
       printf (" -f      loss function type:                         s)igmoid\n");
       break;
     case PIECEWISELINEAR: 
       printf (" -f      loss function type:                         p-w l)inear\n");
       break;
     default:  HError(1,"Invalid loss function type - should never get here");
       break;
     }
   printf(" -a f    loss function smoothing factor (alpha)      %f\n",alpha);
   printf(" -h f    loss function offset factor    (beta)       %f\n",beta);
   printf(" -e f    counter discriminant membership const.      %f\n",eta);
   printf(" -x f    Beginning epsilon value                     %f\n",Epsilon);
   printf("         files read in for training                  %d\n",Nfiles);
   if (FIntrvl!=UNDEFINED)
     printf(" -i i    the number of files read before update      %d\n",FIntrvl);
   else 
     printf(" -i i    the number of files read before update      all\n");
   if (uFlags&UPMEANS)  printf("Updating mixture means\n");   
   if (uFlags&UPMIXES)  printf("Updating mixture weights\n");
   if (uFlags&UPVARS)   printf("Updating mixture covariances\n");
   printf("\n");
   fflush(stdout);
}


void StrCatDotInt(char *str, int i, char *s)
{
  char istr[LINE_LEN];

  sprintf(istr,"%d", i);
  strcpy(s,str);
  strcat(s,".");
  strcat(s, istr);
}

void NewDir(char *name, char *nname, int i)
{
 char command[LINE_LEN];

 StrCatDotInt(name, i, nname);
 strcpy(command, "mkdir ");
 strcat(command, nname);
 system(command);
}

void SetuFlags(void)
{
   char *s;
   
   s=GetStrArg();
   uFlags=0;        
   while (*s != NULL)
      switch (*s++) {
         case 't': uFlags+=UPTRANS;break;
         case 'm': uFlags+=UPMEANS;break;
         case 'v': uFlags+=UPVARS;break;
         case 'w': uFlags+=UPMIXES;break;
         default: HError(2,"Unknown update flag %c",*s);
                  break;
      }
   if (uFlags==0) HError(1,"Updating Nothing!");
}

int main(int argc, char *argv[])
{
   char *s, *datafn;
   int i = 0;
   void Initialise(char * listfn, char * netfn, char * bigfn);
   void FinalTidyUp(void);
   void CreateFileNameEntry(char *name);
   void GPDUpdate(void);
   
   InitShell(argc,argv);
   InitMath(TRUE);
   InitLabel();
   if (NumArgs()==0)
      ReportUsage();
   while (NextArg() == SWITCHARG) {
      s = GetSwtArg();
      if (strlen(s)!=1) 
         HError(1,"Bad switch %s; must be single letter",s);
      switch(s[0]){
         case 'u':
            SetuFlags();
            break;
         case 'x':
            Epsilon = GetChkedFlt(-10E+99,10E+99,s); 
            break;
         case 'f':
            if (NextArg()!=STRINGARG)
               HError(1,"Loss function type expected");
            LossFunctionType = Str2LossFunct(GetStrArg()); break;
         case 'a':
            alpha = GetChkedFlt(-10E+99,10E+99,s); 
            break;
         case 'h':
            beta = GetChkedFlt(-10E+99,10E+99,s); 
            break;
         case 'e':
            eta = GetChkedFlt(-10E+99,10E+99,s); 
            break;
         case 'b':
            if (NextArg()!=STRINGARG)
               HError(1,"Bigram file name expected");
            bigfn = GetStrArg(); 
            break;
	 case 'g':
            gcollect=TRUE; break;
	 case 'c':
            ChangeHmmDir=TRUE; break;
         case 'd':
            if (NextArg()!=STRINGARG)
               HError(1,"HMM definition directory expected");
            hmmDir = GetStrArg(); break;     
         case 'D':
            if (NextArg()!=STRINGARG)
               HError(1,"Output hmm file directory expected");
            outHmmDir = GetStrArg(); break;
         case 'o':
            bjtScale = GetChkedFlt(0.0,100.0,s); 
            break;
         case 'p':
            propPen = log((double)GetChkedFlt(1.0E-20,1.0E20,s)); 
            fixedPen = TRUE; break;
         case 's':
            bigScale = GetChkedFlt(0.0,100.0,s); 
            break;
         case 't':
            pruning = GetChkedFlt(0,1.0E20,s); 
            isPruning = TRUE; break;
         case 'X':
            if (NextArg()!=STRINGARG)
               HError(1,"HMM file extension expected");
            hmmExt = GetStrArg(); break;
         case 'F':
            if (NextArg() != STRINGARG)
               HError(1,"Data File format expected");
            if((dataff = Str2Format(GetStrArg())) == ALIEN)
               HError(0,"Warning ALIEN Data file format set");
            SetFormat(dataff);
            break;
         case 'G':
            if (NextArg() != STRINGARG)
               HError(1,"Label File format expected");
            if((labelff = Str2Format(GetStrArg())) == ALIEN)
               HError(0,"Warning ALIEN Label file format set");
            SetLFormat(labelff);
            break;
         case 'l':
            if (NextArg()!=STRINGARG)
               HError(1,"Label file directory expected");
            ilabDir = GetStrArg(); break;
         case 'L':
            if (NextArg()!=STRINGARG)
               HError(1,"Label file directory expected");
            rlabDir = GetStrArg(); break;
         case 'S': SetScriptFile();
            break;
         case 'T':
            trace = GetChkedInt(0,5,s); break;
	 case 'i':
            FIntrvl = GetChkedInt(1,MAXFILES,s); break;
         case 'I':
            NIter = GetChkedInt(1,20000,s); break;
         case 'N':
            NBest = GetChkedInt(0,100,s); NBest = NBest+1;break;
         case 'Y':
            if (NextArg()!=STRINGARG)
               HError(1,"Label file extension expected");
            labExt = GetStrArg(); break;
         default:
            HError(1,"Unknown switch %s",s);
      }
   } 

   if (NextArg()!=STRINGARG)
      HError(1,"hmm list file name expected");
   listfn = GetStrArg();
   if (NextArg()!=STRINGARG)
      HError(1,"network file name expected");
   netfn = GetStrArg();
   do {
      if (NextArg()!=STRINGARG)
         HError(1,"training data file name expected");
      datafn = GetStrArg();
      files[i] = (char *)malloc(strlen(datafn));
      strcpy(files[i], datafn);
      i++;
   } while (NumArgs()>0);
   Nfiles = i; 
   PrintArgs();
   Initialise(listfn,netfn,bigfn);
   GPDUpdate();
   FinalTidyUp();
}

/* Tidys up at the very end of the program */
void FinalTidyUp(void)
{
  FreeVector(ot);
}

/*------------------Random array generators --AD --------------------*/

void fill_array(arr, size)
int arr[], size;
{
  int i;
  for (i = 0; i< size; i++)
    arr[i] = i;
}

void jumble(arr, size)
int arr[], size;
{
  int i, p1, p2, temp;
  
  for (i = 0; i < size; i++){
    p1 = RandomValue()*size;
    p2 = RandomValue()*size;
    temp = arr[p1];
    arr[p1] = arr[p2];
    arr[p2] = temp;
  }
}

/* -------------------- Bigram Handling Routines -------------- */

/* LoadBigram: load a bigram from given file, note that row 
              and col ordering is changed to map onto idlist */
void LoadBigram(char *fname)
{
   FILE *f;
   char buf[132];
   LabId id;
   int p,i,j,*map;
   char *checkList;
   Vector v,tmp;
   float x;
   
   bigram = CreateMatrix(P,P);
   tmp = CreateVector(P);
   if ((map=(int *)malloc(P*sizeof(int))) == NULL)
      HError(2,"Cant create bigram map array");
   --map;
   checkList = NewString(P);     /* not chars really */
   for (p=1;p<=P;p++)
      checkList[p] = '\0';
   ZeroMatrix(bigram);
   if ((f = fopen(fname,"r")) == NULL)
      HError(2,"Cannot open bigram file %s",fname);
   j = 0;
   while (ReadLabel(f, buf)) {      /* reorder rows during read in */
      if ((id = GetLabId(buf,FALSE))==NULL)
         HError(2,"Unknown label %s in bigram %s",buf,fname);
      p = id->aux; map[++j] = p;
      if (p < 1  ||  p > P)
         HError(2,"Label %s from bigram %s not in id List",buf,fname);
      if (checkList[p] == '\1')
         HError(2,"Label %s in bigram %s doubly defined",buf,fname);
      v = bigram[p];
      ReadVector(f,v);
      for (i=1;i<=P;i++) {
         x = v[i];
         v[i] = (x<MINLARG)? LZERO : log(x) * bigScale;
      }
      checkList[p] = '\1';
   }
   for (p=1;p<=P;p++)      /* check all rows defined */
      if (checkList[p] == '\0')
         HError(2,"No definition for label %s in bigram %s",
            idList[p]->name,fname);
   for (i=1;i<=P;i++) {          /* reorder columns */
      v = bigram[i];
      for (j=1;j<=P;j++)
         tmp[map[j]] = v[j];
      bigram[i] = tmp; tmp = v;
   }
   hasBigram = TRUE;
   free((void *)(map+1));
   FreeVector(tmp);
}

/* ------------------ Phone Link Record Management ----------------- */

#define PLRTRWIDTH 4

static PLR *freeHd=NULL, *freeTl=NULL;
static PLR *usedHd=NULL, *usedTl=NULL;
static numSinceGC=0;                   /* num PLRs allocated since last GC */
static PLR *plrtrbuf[PLRTRWIDTH];      /* trace of best tokens */
static int trbufUsed;

/* InitPLRs: Initialise pool of PLRs */
void InitPLRs(void)
{
  if (usedHd!=NULL) {
    if (freeHd==NULL) {
      freeHd=usedHd; 
      freeTl=usedTl;
    } else {
      freeTl->chain = usedHd;
      freeTl = usedTl;
    }
  }
  usedHd=NULL; usedTl=NULL;
  numSinceGC = 0;
}
      
/* FreePLRs: DR Free the entire pool of PLR's */
void FreePLRs(void)
{
 PLR *p;

  while (usedHd!=NULL) {
    p = usedHd->chain;   
    free((void *)usedHd);    
    usedHd = p;
  }
  usedHd=NULL; usedTl=NULL;
  while (freeHd!=NULL) {
    p = freeHd->chain;
    free((void *)freeHd);
    freeHd = p;
  }
  freeHd=NULL; freeTl=NULL;
  numSinceGC = 0;
}


/* CreateSLR: Returns a pointer to a new SLR */
SLR *CreateSLR(void)
{
   SLR *p=NULL;

   if ((p = (SLR *)malloc((unsigned)sizeof(SLR))) == NULL)  
      HError(2,"Cannot create SLR");   
   p->chain = slrHD;
   p->link = NULL;
   slrHD = p;
   return p;
}   

/* FreeSLRs: Frees a chain of SLR records */
void FreeSLRs(void)
{
  SLR *next=NULL;
  
  while(slrHD!=NULL) { 
    next = slrHD->chain;
    free((void *)slrHD);
    slrHD=next;
  }
  slrHD=NULL;
}

/*---------------- routines for dealing with the path cells--------*/

/* CreatePathCell: Returns a pointer to a new PathCell */
PathCell *CreatePathCell(void)
{
   PathCell *c;

   if ((c = (PathCell *)malloc(sizeof(PathCell))) == NULL)  
      HError(2,"Cannot create PathCell");   
   c->name    = NULL;
   c->link    = NULL;
   c->time    = 0;
   c->state   = 0;
   c->bjot    = 0.0;
   c->logprob = 0.0;
   c->chain = pathcellHD; 
   pathcellHD = c;
   return c;
}   

/* NumPathLevels: return number of levels in Path p */
int NumPathLevels(Path p)
{
   int count = 0;
   
   while (p!=NULL) {p=p->next; ++count;}
   return count;
}

/* FreePathCell: frees a single path cell */
void FreePathCell(PathCell *c)
{
 free((void *)c->name);
 free((void *)c);
}

/* FreePath: Frees a path of PathCell structures */
Path FreePath(Path p)
{
 PathCell *c, *next_c;
 Path next_p;
 
 while(p != NULL) {
   c=p->pcell;
   while (c != NULL) {
     next_c = c->link;
     FreePathCell(c); 
     c = next_c;
   }
   next_p = p->next;
   free((void *)p);
   p = next_p;
 }
 return NULL;
}

/* SLRToPathCell: Copys a single SLR structure into a new UNLINKED path structure and returns a pointer*/
PathCell *SLRToPathCell(SLR *slr)
{
 PathCell *c;

 if(slr == NULL)
  HError(1,"SLR is empty: nothing to copy");
 c = CreatePathCell();
 c->time  = slr->time;
 c->state = slr->state;
 c->bjot  = slr->bjot;
 c->logprob = slr->logprob;
 c->link  = NULL;
 if((c->name = malloc(strlen(slr->name)))==NULL) 
   HError(1,"could not malloc space for path name");
 strcpy(c->name, slr->name);
 return(c);
}

/*CreatePath: Creates a new path */
Path CreatePath(void)
{
 Path p;

 if ((p = (PathCellList *)malloc(sizeof(PathCellList))) == NULL)
   HError(1,"Cannot create path");
 p->next  = NULL;
 p->pcell = NULL;
 p->chain = pathHD;
 pathHD = p;
 return  p;
}

/* MovePath: moves the _level_ sequence of path cells in p1->p2 */
void MovePath(Path *p1, int level, Path *p2)
{
  int i;
  Path temp;

  if((*p2)!=NULL)
    while((*p2)->next!=NULL)  (*p2) = (*p2)->next;
 
  if (level == 1){
    *p2 = *p1;
    (*p1) = (*p1)->next;
  }
  else{
    temp = *p1;
    for (i = 1; i < level-1; i++){
      if (temp->next != NULL)
	temp = temp->next;
      else
	HError(1,"not enough levels for MovePath");
    }
    *p2 = temp->next;
    temp->next = (*p2)->next;
  }
  (*p2)->next = NULL;
}

/* CopyPathCell: destructively copies c1->c2*/
void CopyPathCell(PathCell *c1, PathCell *c2)
{
 if(c1==NULL) HError(1,"No path cell to copy from");
 if(c2==NULL) HError(1,"No path cell to copy to");
 c2->time    = c1->time;
 c2->state   = c1->state;
 c2->bjot    = c1->bjot;
 c2->logprob = c1->logprob;
 c2->link = NULL;
 free((void *)c2->name);
 if((c2->name=malloc(strlen(c1->name)))==NULL)
   HError(1,"Cannot malloc space for name field");
 strcpy(c2->name,c1->name);
}



/* CopyNPathCells: returns a copy of the first n path cells in p */
PathCell *CopyNPathCells(PathCell *p, int n)
{
  PathCell *c, *c1, *c2;
  
  if(p==NULL) HError(1,"No cells to copy");
  if(n==0) return NULL;
  c = CreatePathCell();
  CopyPathCell(p,c); /* p->c*/
  
  for(p=p->link,n--,c1=c;(p!=NULL)&&(n>0);p=p->link,n--) {
    c2 = CreatePathCell();
    CopyPathCell(p,c2);
    c1->link = c2;
    c1=c2;
  }
  c1->link=NULL;
  if(n>0) HError(1,"Not enough cells to copy");
  return(c);
}

PathCell *JoinPathCells(PathCell *p1, PathCell *p2)
{
  PathCell *p3;

  if(p1==NULL) return p2;
  for(p3=p1;p3->link!=NULL;p3=p3->link) /* empty */;
  p3->link = p2;
  return (p1); 
}

/* returns the n'th pathcell from p */
PathCell *GetNthPathCell(PathCell *p, int n)
{
 for(;(n>0)&&(p!=NULL);p=p->link,n--) /* empty */;
 if(n!=0) HError(1,"Too few path cells");
 return(p);
}

/* Corrects the logprob field in p2 */
void CorrectLogprob(PathCell *p1, PathCell *p2, int n)
{
  double diff;
  PathCell *c1, *c2;

  c1 = GetNthPathCell(p1, n+1); 
  c2 = GetNthPathCell(p2, n+1);
  diff = c2->logprob - c1->logprob;
  for(;(p2!=NULL)&&(n>0);p2=p2->link,n--) p2->logprob += diff;
}


/* Fills in all partial paths in p2 with p1 */
void FillInPaths(Path p1, Path p2)
{
 int n;  

 for(;p2!=NULL;p2=p2->next) { 
   n = p1->pcell->time - p2->pcell->time;
   if(n==0) continue;
   if(n<0) HError(1,"path is too short");
   if(n>0) { 
     p2->pcell = JoinPathCells(CopyNPathCells(p1->pcell,n), p2->pcell);   
     CorrectLogprob(p1->pcell, p2->pcell, n);
   }
 }
}

/* MoveLabels: moves the _level_ sequence of labels in t1->t2 */
void MoveLabels(Transcription *t1, int level, Transcription *t2)
{

  int i;
  Transcription temp;

  if ((*t1)==NULL) HError(1,"No labels to move");
  if((*t2)!=NULL)
    while((*t2)->next!=NULL)  (*t2) = (*t2)->next;
  if (level == 1){
    *t2 = *t1;
    (*t1) = (*t1)->next;
  }
  else{
    temp = *t1;
    for (i = 1; i < level-1; i++){
      if (temp->next != NULL)
	temp = temp->next;
      else
	HError(1,"not enough levels for MoveLabels");
    }
    *t2 = temp->next;
    temp->next = (*t2)->next;
  }
  (*t2)->next = NULL;
}


/* PartialSLRListToPath: Copys in reverse a list of SLR structures to a 
   new linked list of type path a pointer to that new list. */
Path PartialSLRListToPath(SLR *slr, int start)
{
  PathCell *c;
  Path p;
  
 if (slr==NULL) HError(1,"no list of SLR records to copy");
 p = CreatePath();
 p->pcell = SLRToPathCell(slr);
 for(c=p->pcell,slr=slr->link;(slr!=NULL)&&(c->time>start); c=c->link,slr=slr->link)
   c->link = SLRToPathCell(slr);
  return p;
}  

/* PrintPathCell: Prints to stdout a single path structure */
void PrintPathCell(PathCell *p)
{
 if (p==NULL) HError(1,"no path cell to print");
 printf("time[%4d] state[%2d] name[%3s]  P(O)[%f] P(seq)[%f]\n", p->time,
	p->state, p->name, p->bjot, p->logprob); 
 fflush(stdout);
}

/* AppendPath: append path p2 to path p1 */
void AppendPath(Path p1, Path p2)
{
   for(;p1->next!=NULL;p1=p1->next) /* empty */;
   p1->next = p2; 
}

/* PrintPath: Prints to stdout a linked path list */
void PrintPath(Path p)
{
 PathCell *c;

 if (p==NULL) return;
 for(;p!=NULL;p=p->next){ 
   for(c=p->pcell; c!=NULL; c=c->link) PrintPathCell(c);
   if(p->next!=NULL) printf("/\n");
 }
 printf("\n\n");fflush(stdout);
}  

/*--------------------end of path routines------------------------------*/

/* CreatePLR: Returns a pointer to a new PLR */
PLR *CreatePLR(void)
{
   PLR *p;
   
   if (freeHd==NULL) {
      if ((p = (PLR *)malloc(sizeof(PLR))) == NULL)
         HError(2,"Cannot create PLR");
   } else {
      p = freeHd;
      if (freeHd==freeTl)
         freeHd = freeTl = NULL;
      else 
         freeHd = freeHd->chain;
   }
   if (usedHd==NULL) 
      usedHd = usedTl = p;
   else {
      usedTl->chain = p;
      usedTl = p;
   }
   p->chain = NULL;
   p->alt = NULL;
   ++numSinceGC;
   return p;
}

/* GarbageCollectPLRs:  Garbage Collect unlinked PLR's */
void GarbageCollectPLRs(void)
{
  int i,j,plrcollected,plrused;
  PLR *plr,*next;
  PhoneInstance *pi;
  Token t;
  PLR *plr1,*plr2;
  
  if (GCMARGIN <= numSinceGC) {
    plrused = plrcollected = 0;
    for (plr=usedHd; plr!=NULL; plr=plr->chain)
      plr->linked = FALSE;
    for (i=0;i<nActive;i++) {
      pi = active[i];
      for (j=1; j<=pi->N; j++) {
	t = pi->state[j];
	if (t.inst != NULL) { 
	  for (plr=t.link;(plr!=NULL);
	       plr=plr->tok.link) {
	    plr->linked=TRUE;
	    plr1 = plr->alt;
	    while(plr1!=NULL) {
	      for (plr2=plr1;(plr2!=NULL);
		   plr2=plr2->tok.link) 
		plr2->linked=TRUE;
	      plr1 = plr1->alt;
	    }
	  }
	}
      }
    }
    
    plr = usedHd; usedHd = usedTl = NULL;
    while (plr!=NULL) {
      next = plr->chain;
      if (plr->linked) {
	++plrused;
	plr->chain = usedHd;
	usedHd = plr;
	if (NULL==usedTl)
	  usedTl = plr;
      } else {
	++plrcollected;
	plr->chain = freeHd;
	freeHd = plr;
	if (NULL==freeTl)
	  freeTl=plr; 
      }
      plr=next;
    }
    
    if (trace>2) {
      printf("GC: NumSinceGC=%d Collected=%d Used=%d\n",
	     numSinceGC,plrcollected,plrused); 
      fflush(stdout);
    }
    numSinceGC = 0;
    if (plrused>maxPLRUsed) maxPLRUsed = plrused;
  }
}

/* ClearTrBuf: clear the PLR trace buffer */
void ClearTrBuf(void)
{
   trbufUsed = 0;
}

/* Add2TrBuf: add the given PLR to the trace buffer */
void Add2TrBuf(PLR *p)
{
   int i;
   PLR *q;
   LogReal x;
   
   x = p->tok.logprob;
   if (trbufUsed == PLRTRWIDTH) {
      q = plrtrbuf[trbufUsed-1];
      if (q->tok.logprob<x)
         --trbufUsed;
      else 
         return;
   }
   for (i=trbufUsed;i>0;i--)
      if (plrtrbuf[i-1]->tok.logprob < x)
         plrtrbuf[i] = plrtrbuf[i-1];
      else
         break;
   plrtrbuf[i] = p;
   ++trbufUsed;
}

/* PrintTrBuf: print out the contents of the trace buffer */
void PrintTrBuf(int t)
{
   int i;
   Token tok;
   
   printf("PLRS: time %d\n",t);
   for (i=0;i<trbufUsed;i++){
      tok = plrtrbuf[i]->tok;
      printf("id %2s, prob %f",tok.inst->def->id->name,tok.logprob);
      if(tok.link!=NULL)
         printf(" back to frame %d",tok.link->time);
      printf("\n");
   }
   fflush(stdout);
}

/* -------------------- Housekeeping for Pruning ----------------- */

/* CreateActiveList: create a list of pointers to active phone instances */
void CreateActiveList(void)
{
   if ((active=(PhoneInstance **)malloc(NPI*sizeof(PhoneInstance *))) == NULL)
      HError(2,"Cannot create active phone instance list");
}

/* DeActivateAll: deactivate all active phone instances */
void DeActivateAll(void)
{
   int i;
   
   for (i=0;i<nActive;i++)
      active[i]->isActive = FALSE; 
   nActive = 0;
}

/* MakeActive: make given pi active */
void MakeActive(PhoneInstance *pi)
{
   int i;

   if (!pi->isActive){
      pi->isActive = TRUE;
      active[nActive++] = pi;
        for (i=1;i<=pi->N;i++)
           pi->state[i] = nullToken;
   }
}

/* CheckActive: scan all active phones and kill any with no paths in beam */
void CheckActive(void)
{
   int i,j,nAOld;
   PhoneInstance *pi;
   LogReal bestP,threshold;
   Token t;
   
   nAOld = nActive;
   threshold = maxLogP - pruning;
   for (i=0;i<nActive;i++) {
      SPINBALL
      pi = active[i];
      bestP = LZERO;          /* find best path in this pi */
      for (j=2; j<=pi->N; j++){
          t = pi->state[j];
           if (t.inst != NULL && t.logprob>bestP)
              bestP = t.logprob;
      }
      if (bestP<threshold){   /* if below thresh kill it */
         active[i] = NULL;
         pi->isActive = FALSE;
      }
   }
   for (i=0,j=0;i<nActive;i++)   /* remove marked pi's */
      if (active[i] != NULL) {
         active[j++] = active[i];
      }
   nActive = j;
   if (trace>2) {
      printf("   Active Beam Width %d -> %d\n",nAOld,nActive);
      fflush(stdout);
   }
}

/* ------------------------ Network Management ------------------- */

/* IdMap: set-up id->aux to inverse map of idList indices.  If
   hmms are context sensitive then the mapping goes from cs models
   to monophones.  Only used by bigram */
void IdMap(void)
{
   int p,h;
   Boolean isCS=FALSE;
   char *s,buf[100];
   LabId id;
   
   for (h=1;h<=H;h++){
      s = idList[h]->name;
      if (strchr(s,'-') != NULL || strchr(s,'+') != NULL){
         isCS = TRUE;
         break;
      }
   }
   
   if (!isCS) {
     for (h=1;h<=H;h++)
       idList[h]->aux = h;
     P = H;
   } else {
     for (h=1;h<=H;h++){
       strcpy(buf,idList[h]->name);
       TriStrip(buf);
       id = GetLabId(buf,TRUE);
       p = id->aux;
       if (p == 0) id->aux = p = ++P;
       idList[h]->aux = p;
     }
   }
 }

void ZeroHmmP(void)
{
 hmmP.vSize      = 0;            /* data vector size */
 hmmP.M          = 0;            /* number of mixtures */
 hmmP.CK         = DIAGC;        /* covariance kind of all units */
 hmmP.SK         = WAVEFORM;     /* sample kind of all units */
 hmmP.isTied     = FALSE;        /* true if tied mixtures */
 hmmP.maxStates  = 0;            /* max num states in any phone model */
} 


/* LoadHMMDefs1: create hmm linked list structure, loads it and returns pointer*/
Definition *LoadHMMDefs1(LabId *idlist)
{
  int h,n;
  Definition *d, *hmm;
  
  ZeroHmmP();
  
  if ((hmm = (Definition *)malloc(H*sizeof(Definition))) == NULL)
    HError(2,"Unable to create hmm def array");
   --hmm;
  for (h=1;h<=H;h++){
    d = hmm+h;
    LoadHMM(idList[h]->name,&(d->hmmdef), hmmDir, hmmExt, &hmmP);
    if(trace > 2) {
      printf("Loaded phoneme HMM %s\n", idList[h]->name);
      fflush(stdout);
    }
    n = d->hmmdef.numStates-2;
    if ((d->bjot = (float *)malloc(n*sizeof(float))) == NULL)
      HError(2,"Cannot create bjot array");
    d->bjot -= 2;    
    d->id = idList[h];
  }
  return (hmm);
}

/* Zeros all trainable parameters in hmm */
void ZeroHMM(HMMDef *hmm)
{
  int s,m;
  OutProb *outp;; 

  for(s=2;s<=hmm->numStates-1;s++) {
    outp =  hmm->stateinfo[s].outp;
    hmm->stateinfo[s].dur1 = 0.0;
    hmm->stateinfo[s].dur2 = 0.0;
    for(m=1;m<=hmm->numMixtures;m++) {
      outp[m].weight = 0.0;
      outp[m].gConst = 0.0;
      ZeroVector(outp[m].mean);
      if(hmm->ckind==DIAGC)
	ZeroVector(outp[m].covariance.var);
      if(hmm->ckind==FULLC)
	ZeroMatrix(outp[m].covariance.inv);
    }  
  }
  ZeroLogMatrix(hmm->transP);  
}

/* zeros all elements in the hmm list */
void ZeroHMMDefs(Definition *hmm)
{
  int h;
   for (h=1;h<=H;h++) ZeroHMM(&(hmm[h].hmmdef));
}

/* AttachDefinition: look for hmm def with given id and attach it to pi */
void AttachDefinition(PhoneInstance *pi, LabId id, Boolean valid /* DR */)
{
   int h;
   for (h=1;h<=H;h++)
      if (!strcmp(idList[h]->name,id->name)){ 
         pi->def = hmm+h;
         return;
      }
   HError(2,"Cannot find def %s to attach",id->name);
}

/* Nulllastplrt: zero lastplrt in all PhoneInsts */
void Nulllastplrt(Link chain)
{
   Link p;
   PhoneInstance *pi;
   
   p = chain;
   while (p!=NULL) {
      if (p->modelName != enterId && p->modelName != exitId ) {
         pi = (PhoneInstance *) p->user;
         pi->lastplrt = 0;
      }
      p = p->chain;
   }
}

/* AttachPhoneInsts: to each node in theNet */
void AttachPhoneInsts(Network theNet)
{
   Link p;
   PhoneInstance *pi;
   
   p = theNet.chain;
   while (p!=NULL) {
      SPINBALL
      if (p->modelName != enterId && p->modelName != exitId ) {
         if ((pi =(PhoneInstance *)malloc(sizeof(PhoneInstance))) == NULL)
            HError(2,"Unable to create phone instance");
         ++NPI;
         AttachDefinition(pi,p->modelName, FALSE); /* DR "FALSE" added */
         pi->N = pi->def->hmmdef.numStates;
         pi->isActive = FALSE;
         if ((pi->state = (Token *)malloc(sizeof(Token)*(pi->N+1))) == NULL) 
           HError(2,"Unable to create state array for model %d",pi->N);
         if ((pi->tbuf = (Token *)malloc(sizeof(Token)*NBest)) == NULL) 
            HError(2,"Unable to create alt token  array for model %d", NBest);
         p->user = (void *)pi; /* DR */ 
         pi->netHook = p;
      }
      p = p->chain;
   }
   Nulllastplrt(theNet.chain);
   if (trace>2){
      printf("  %d Instances Created of %d HMMs\n",NPI,H);
      fflush(stdout);
   }
}

Boolean match(Network theNet, Link p, Transcription tr, int n)
{
  PhoneInstance *pi;
  LinkSet *l;
  int i;

  if(n > tr->nLabels) return FALSE;
  l = p->succ;
  if(n==tr->nLabels) {
    for (i=0; i<p->numSucc;i++) {
      if((*l)[i] == theNet.exitNode) {
        pi = (PhoneInstance *)p->user;
        hlpath[n+1] = pi;
        return TRUE;   
      }
    }
    return FALSE;
  }
  
  for (i=0; i<p->numSucc;i++) {
    pi = (PhoneInstance *)(*l)[i]->user;
    if(pi!=NULL) {
      if(!strcmp(pi->def->id->name, tr->labels[n].labid->name)) {
	p = (*l)[i];
	if(match(theNet, p, tr, n+1)) {        
	  hlpath[n+1]=pi;
	  return TRUE;
	}
	else return FALSE;
      }
    }
  }
  return FALSE;
}

/* AttachHandLabels: tags the hand labelled path through the network */
void AttachHandLabels(Network theNet,  Transcription tr)
{
   Link p;
   
   if((hlpath = (PhoneInstance **)malloc((tr->nLabels+2)*sizeof(PhoneInstance *))) == NULL)
     HError(1,"Could not malloc space for hlpath array"); 
   hlpath[0] = (PhoneInstance *)tr->nLabels;
   p = theNet.entryNode;
   if(!match(theNet, p, tr, 0))        
     HError(1,"Hand labelled path not found in network");

}

void PrintHandLabels(void)
{
int i;
for(i=1;i<=(int)hlpath[0];i++) printf("%s\n", hlpath[i]->def->id->name);
fflush(stdout);
}

void FreeHandLabels(void)
{
 free((void *)hlpath);
 hlpath = NULL;
}

/* ShowPath: print a brief n char resumee of path */
void ShowPath(PLR *plr, int n)
{
   char buf[132];
   int i,count;
   
   buf[0] = '\0'; count = 0;
   while (plr != NULL && count < n) {
      strcat(buf,plr->tok.inst->def->id->name);
      plr = plr->tok.link;
      if (plr != NULL) strcat(buf,"/");
      count = strlen(buf);
   }
   for (i=count;i<n;i++) buf[i]=' ';
   buf[n] = '\0'; 
   printf("%s",buf);
}

/* ShowInst: print contents of given phone instance */
void ShowInst(PhoneInstance *pi)
{
   int i;
   LogReal x;
   PLR *plr;
   PhoneInstance *ppi;
   
   printf("** Inst %6s[%2d]\n  ",pi->def->id->name,(int)pi % 100);
   for (i=1; i<=pi->N; i++)
      printf("%10d",i);
   printf("\n  ");
   for (i=1; i<=pi->N; i++) {
      ppi = pi->state[i].inst;
      if (ppi==NULL)
         printf("%10s","NULL");
      else
         printf("%6s[%2d]",ppi->def->id->name,(int)ppi % 100);
   }
   printf("\n  ");

   for (i=1; i<=pi->N; i++)
      if (pi->state[i].inst == NULL)
         printf("%10s","NULL");
      else {
         x = pi->state[i].logprob;
         if (x>LSMALL)
            printf(" %9.2f",x);
         else
            printf("%10s","---");
      }
   printf("\n  ");
   for (i=1; i<=pi->N; i++) {
      plr = pi->state[i].link;
      if (plr==NULL)
         printf("%10s","NULL");
      else
         ShowPath(plr,10);
   }
   printf("\n");
}


/* ------------------------ Initialisation ----------------------- */

void ZeroModelStats(void)
{
 int i;

 for(i=1;i<=H;i++) {
   ModelStats[i].CorrectNFrames = 0;
   ModelStats[i].WrongNFrames = 0;
   ModelStats[i].Correctbjot = 0.0;
   ModelStats[i].Wrongbjot = 0.0;
 }
}

/* Initialise: initialise the global data structures */
void Initialise(char * listfn, char * netfn, char * bigfn)
{

   idList = LoadIdList(listfn,&H);
   IdMap();
   if (trace>2) {
      printf("Id Map Built: ");
      if (P==H)
         printf(" %d monophones\n",H);
      else
         printf(" %d c-s phones -> %d monophones\n",H,P);
      fflush(stdout);
   }
   hmm = LoadHMMDefs1(idList); /* this defines vSize */
   if (bigfn != NULL)
      LoadBigram(bigfn);
   enterId = GetLabId("ENTER",TRUE);
   exitId = GetLabId("EXIT",TRUE);  
   theNet = ParseNetwork(netfn);    /* build the theNet */
   if(trace>3) PrintNetwork(&theNet); fflush(stdout);
   AttachPhoneInsts(theNet);
   CreateActiveList();
   if ((tBuf=(Token *)malloc((hmmP.maxStates+1)*sizeof(Token))) == NULL)
      HError(2,"Cannot create token buffer");
   if (trace>2) {
      printf("Network built: %d Models, %d Max States in any Model\n",
             H,hmmP.maxStates);
      fflush(stdout);
   }
   ot = CreateVector(hmmP.vSize); /* DR */

   /*------------HGPD stuff--------------*/

   if((uFlags&UPMIXES)&&(hmmP.M<=1)) 
     HError(1, "Cannot update mixture weights for M=%d\n",hmmP.M);

   grad = LoadHMMDefs1(idList);  
   tmphmm = LoadHMMDefs1(idList);  

   um = CreateVector(hmmP.vSize); 
   if (hmmP.CK == FULLC)
     HError(1,"Sorry - full covariance not implemented yet"); 
   else
     ucov.var = CreateVector(hmmP.vSize);

   if((ModelStats = (struct ModelStatistics *)malloc(H*sizeof(struct ModelStatistics))) == NULL)
     HError(1, "Cannot malloc space for HMM model statistics array");

   --ModelStats;

  /*---------end HGPD stuff-------------*/

}
   
/* --------------------------- Recognition ----------------------- */

/* Setbjot: precompute all 'local distances' needed by active pi's */
void Setbjot(Vector ot)
{
  Definition *d;
  int i,j,h,N;
  
  if (hmmP.isTied) PreCompute(ot);
  for (h=1;h<=H;h++)
    hmm[h].bjLoaded = FALSE;
  for (j=0;j<nActive;j++){
    d = active[j]->def;
    if (!d->bjLoaded) {
      N = d->hmmdef.numStates;
      for (i=2;i<N;i++) 
	d->bjot[i] = OutP(ot,&(d->hmmdef),i) * bjtScale;
      d->bjLoaded = TRUE;
    }
  }
}

/* InitModels: Activate models connected to ENTRY node and inject a start token */
void InitModels(Network theNet, Boolean handlabel)
{
   Link p;
   int i;
   LinkSet *l;
   PhoneInstance *pi;
   
   SPINBALL
   /**DeActivateAll();**/ 
   Nulllastplrt(theNet.chain);
   p = theNet.entryNode; l = p->succ;
   if(!handlabel) {
     for (i=0; i<p->numSucc;i++) {
        pi = (PhoneInstance *)(*l)[i]->user;
        if (pi==NULL)
           HError(2,"Bad Syntax Net - non-unit connected to entry node");
        MakeActive(pi);
        startToken.inst = pi;
        pi->state[1] = startToken;
     }
   }
   else {
     for (i=0; i<p->numSucc;i++) {
        pi = (PhoneInstance *)(*l)[i]->user;
        if (pi==NULL)
           HError(2,"Bad Syntax Net - non-unit connected to entry node");
        if (pi == hlpath[1]) {
          MakeActive(pi);
          startToken.inst = pi;
          pi->state[1] = startToken;
        }
     }
   }
   if (trace>2) {
      printf("%d Models Initially Active\n",nActive);
      fflush(stdout);
    }
   InitPLRs();
}

/* StepModel: step the given model by one time frame */
void StepModel(PhoneInstance *pi, int time)
{
  int i,j,besti;
  LogReal best,x,a;
  Token t;
  Definition *d=NULL;
  SLR *slr=NULL;
  int N_1;
  
  SPINBALL
    d = pi->def;
  for (j=2;j<pi->N;j++){     /* Propagate internal tokens */
    best = LZERO; besti = 0; 
    for (i=j-1;i<=j;i++) {   /* DR - assumes left to right no skip */
      if (pi->state[i].inst != NULL) {
	a = d->hmmdef.transP[i][j];
	if (a>LSMALL) {
	  x =  pi->state[i].logprob + a;
	  if (x>best) {
	    best = x; besti = i;
	  }
	}
      }
    }
    
    /* DR - we've now found the best state to get to j from */
    if (besti>0) {
      tBuf[j] = pi->state[besti];     /* propogate the token into tmp storage */
      x = best + d->bjot[j];
      if (x>maxLogP) maxLogP = x;             /* record global max for pruning */
      tBuf[j].logprob = x;              /* update the logprob slot of the token*/
      
      /* each token just trails a huge tail of slrs behind it */
      slr = CreateSLR(); 
      slr->name = pi->def->id->name;
      slr->time = time;     
      slr->state = j;
      slr->logprob = tBuf[j].logprob; 
      slr->bjot = d->bjot[j];
      slr->link = pi->state[besti].slrlink;
      tBuf[j].slrlink = slr;
    } else
      tBuf[j] = nullToken;
  }    
  
  for (i=2;i<pi->N;i++)      /* Copy tokens into state slots */
    pi->state[i] = tBuf[i];  
  
  /* Find best exit token */
  
  N_1 = pi->N-1;
  
  if (pi->state[N_1].inst != NULL) {
    pi->state[pi->N] = pi->state[N_1]; 
    pi->state[pi->N].logprob  += d->hmmdef.transP[N_1][pi->N];
    if (pi->state[pi->N].logprob>maxLogP) maxLogP = pi->state[pi->N].logprob;
                                         /* record global max for pruning */
  }
  else 
    pi->state[pi->N] = nullToken;
    if (trace>4)  
    ShowInst(pi);
}

/* StepModels: step all models by one time frame */
void StepModels(int time)
{
   int i;
   
   maxLogP = LZERO;
   for (i=0;i<nActive;i++)  
        StepModel(active[i], time);
}

/* NullEntryTokens: Set all entry tokens to null ie zero prob */
void  NullEntryTokens(void)
{
   int i,j;
   
   for (i=0;i<nActive;i++) { 
     active[i]->state[1] = nullToken;
     for(j=0;j<NBest;j++) active[i]->tbuf[j] = nullToken; 
  }
}

/* NumPLRs: returns the number of PLRs in the chain */
int NumPLRs(PLR *plr)
{
 int n;

 for(n=0; plr!=NULL; plr=plr->tok.link,n++) /* empty */ ;
 return (n);
}

/* PropagateExitTokens: propagate exit tokens  */
void PropagateExitTokens(Boolean hlpathon)
{
  Link p,q,*a;
  int i,j,l,m;
  Token t;
  PhoneInstance *pi;
  LogReal x,threshold;
  int fromIdx;
  int n;
  
  SPINBALL
    threshold = (isPruning) ? maxLogP-pruning : LZERO;
  for (i=0;i<nActive;i++) {
    pi = active[i]; p = pi->netHook;
    fromIdx = pi->def->id->aux;
    t = pi->state[pi->N];
    if(hlpathon) n = NumPLRs(t.link)+2;
    if (t.logprob>threshold) {
      if (fixedPen)
	t.logprob += propPen;   
      else if (!hasBigram)
	t.logprob -= (float)log((double)p->numSucc);
      a = (Link *)p->succ;
      for (j=1;j<=p->numSucc;j++) {
	q = *a++;
	if ((PhoneInstance *)q->user != NULL) {
	  if(!hlpathon) {  
	    pi = (PhoneInstance *)q->user;
	    if (!pi->isActive) 
	      MakeActive(pi);
	    x = t.logprob;
	    if (hasBigram)
	      x += bigram[fromIdx][pi->def->id->aux];
	    for(l=0;l<NBest;l++) {
	      if (x>LSMALL && x>pi->tbuf[l].logprob) {
		for(m=NBest-1;m>l;m--) pi->tbuf[m] = pi->tbuf[m-1]; 
		pi->tbuf[l] = t;             
		pi->tbuf[l].logprob = x;
		pi->tbuf[l].slrlink = t.slrlink;
		break;
	      }
	    }
	  }
	  else { 
	    pi = (PhoneInstance *)q->user;
	    if(n<=(int)hlpath[0]) {
	      if(pi == hlpath[n]) {
		if (!pi->isActive) 
		  MakeActive(pi);
		x = t.logprob;
		if (hasBigram)
		  x += bigram[fromIdx][pi->def->id->aux];
		if (x>LSMALL && pi->state[1].logprob < x) {
		  pi->state[1] = t;
		  pi->state[1].logprob = x;
		  pi->state[1].slrlink = t.slrlink;
		}
		break; 
	      }
	    }
	  }       
	} 
      } 
    }
  }
  
  if(!hlpathon) 
    for (i=0;i<nActive;i++) {pi = active[i]; pi->state[1] = pi->tbuf[0];}
}


/* RecordDecisions: create PLR's for all distinct entry nodes */
void RecordDecisions(int t)
{
  int i,j;
  PhoneInstance *pi;
  PLR *plr, *plr1, *plr2;
  Token tok;
  
  if (trace>3) ClearTrBuf();
  for (i=0;i<nActive;i++) {
    SPINBALL
      pi = active[i];
    tok = pi->state[1];
    if(tok.inst != NULL) { 
      if (tok.logprob>LSMALL) {
	if (t != tok.inst->lastplrt) {
	  plr = CreatePLR(); /* Creates an empty PLR */
	  plr->tok  = tok;
	  plr->time = t;
	  tok.inst->lastplrt = t;
	  tok.inst->lastplr = plr;
	  if (trace>3) {
	    if (hasBigram)
	      plr->tok.logprob -= 
		bigram[tok.inst->def->id->aux][pi->def->id->aux];
	    Add2TrBuf(plr);
	  }
	  /****/
	  plr1 = plr;
	  for(j=1;j<NBest;j++) {
	    if(pi->tbuf[j].inst != NULL) {
	      plr2 = CreatePLR();
	      plr2->tok = pi->tbuf[j];
	      plr2->time = t;
	      plr1->alt = plr2; 
	      plr1 = plr2;
	    }
	    else{
	      plr->alt = NULL;
	      break;
	    }
	  }
	  /****/
	} else
	  plr = tok.inst->lastplr;
           pi->state[1].link = plr;
	pi->state[1].inst = pi;
      }
    }
  }
  if (trace>3) PrintTrBuf(t);
}

/* ------------------------ Traceback and Output ---------------------- */

/* StartTime: returns the start time of the first label in the transcription tr */
int StartTime(Transcription tr)
{
 return((int)tr->labels[0].start);
}

void WriteOut(PLR *plr, Transcription tr)
{
  Label l;
  
  if(plr->tok.link != NULL)
    l.start = plr->tok.link->time+1;
  else 
    l.start = 1;
  
  l.end = plr->time;
  l.labid = GetLabId(plr->tok.inst->def->id->name,TRUE);
  
  AddLabel(tr,l);
}

Transcription  TraceBack(PLR *plr, int len)
{
  Transcription tr;
  
  if ((len==0)||(plr==NULL)) {
    tr = (Transcription)CreateLabList();
    return (tr);
  }
  else tr = TraceBack(plr->tok.link, len-1); 
  WriteOut(plr, tr);
  return(tr);
}

Transcription WriteSegment(PLR *plr, int len, Path *path)
{
  Transcription tr, tr1;
  int i,n,t;
  PLR *plr1;
  Path path1;

  tr = TraceBack(plr, -1); /* the Viterbi path */
  *path =  PartialSLRListToPath(plr->tok.slrlink,1 /*start time*/);
  while(plr!=NULL) {
    for(n=1;n<NBest;n++) { 
      plr1 = plr;
      for(i=0;i<n;i++) if(plr1->alt != NULL) plr1 = plr1->alt; else break;
      if(i!=n) break;
      tr1 = TraceBack(plr1, len);
      t = StartTime(tr1);      
      AppendLabList(tr1, tr);      
      path1 =  PartialSLRListToPath(plr1->tok.slrlink,t); 
      AppendPath(*path, path1);
    }
    plr = plr->tok.link;
  }
  return tr;
}

/* TranscribeResults: trace back best predecessor of theNet exit and store in transcription */
void TranscribeResults(int t, Network theNet, Path *path, Transcription  *tr)  
{
  Link p,q,*a;
  PhoneInstance *pi, * bestpi;
  LogReal best;
  int i,j,k;
  Token tok, *besttok;
  PLR *plr, *plr1, *plr2;
  int numactive = 0;

  if((besttok = (Token *)malloc(NBest*sizeof(Token)))==NULL) 
    HError(1,"Cannot malloc space for besttok array");
  
  best=LZERO; bestpi=NULL; 
  for(i=0;i<NBest;i++) besttok[i] = nullToken;
  
  p = theNet.exitNode;
  a = (Link *)p->pred;
  for (i=1;i<=p->numPred;i++) {
    q = *a++;
    if ((PhoneInstance *)q->user != NULL) {
      pi = (PhoneInstance *)q->user;
      if(pi->isActive && pi->state[pi->N].inst!=NULL) numactive += 1;
      tok = pi->state[pi->N];
      for(j=0;j<NBest;j++) {
	if (pi->isActive && tok.inst!=NULL && tok.logprob > besttok[j].logprob){ 
	  for(k=NBest-1;k>j;k--) besttok[k] = besttok[k-1];
	  besttok[j] = tok;
	  break;
	}
      }
    }
  }
  
  /* Okay - lets make a final set of PLRS */
  
  plr = CreatePLR();
  plr->tok = besttok[0];
  plr->time = t;
  plr1 = plr;
  
  for(i=1;i<NBest;i++) {
    if(besttok[i].inst != NULL) {
      plr2 = CreatePLR();
      plr2->tok = besttok[i];
      plr2->time = t;
      plr1->alt = plr2;
      plr1 = plr2;
    }
    else {
      plr1->alt = NULL;
      break;
    }
  }
  if(plr->tok.inst==NULL) 
    HError(1,"No tokens survived to end of input");
  *tr = WriteSegment(plr,-1, path);  
  if(besttok!=NULL) free(besttok);
}

/*--------------------*/

#define LINKCHUNKSIZE 5

/* Viterbi: viterbi recognition */
void Viterbi(Boolean hlpathon)
{
  int t;
  long beamWidthSum;
  float x;
  
  beamWidthSum = nActive;
  
  for (t=1;t<=data.nSamples;t++) {
    if(trace >3) {
      printf ("Number of data samples is %d\n", data.nSamples); 
      printf ("*** time is %d ***\n", t);
      fflush(stdout);
    }
    GetSample(&data,t-1,ot+1);
    Setbjot(ot);
    StepModels(t);
    if (isPruning)
      CheckActive();
    beamWidthSum += nActive;
    NullEntryTokens();
    PropagateExitTokens(hlpathon);
    if(gcollect) GarbageCollectPLRs();
    if (t==data.nSamples) break;
    else
      RecordDecisions(t);
  }
   if (trace>2 && isPruning) {
     x = ((float)beamWidthSum * 100.0)/ 
       ((float) NPI * (float)data.nSamples);
     printf("  Total Pruning = %.2f%%\n",100-x);
     fflush(stdout);
   }
  if (trace>2 && maxPLRUsed>0) {
    printf("  Max PLR's Used = %d\n",maxPLRUsed);
    fflush(stdout);
  }
}

Boolean TrEqual(Transcription a, Transcription b)
{
  int i, level = 0, corLevel;
  Transcription temp;
  Boolean StillTrue=1;

  for (temp = a; temp!=NULL; temp=temp->next){
    if (temp->nLabels != b->nLabels){
      if (temp->next != NULL) temp = temp->next;
      else return(FALSE);
    }
    level++;
    StillTrue = 1;
    for (i=0; i< b->nLabels; i++)
      if(strcmp(temp->labels[i].labid->name,b->labels[i].labid->name)){
	StillTrue = 0;
	break;
      }
    if (StillTrue) return(level);
  }
  return (FALSE);
}

/* TidyUpAfterParse: frees up active list and data structures, SLR, PLR. Also sets
   all active PI's to inactive */ 
void TidyUpAfterParse(void)
{
  FreeSLRs();
  FreePLRs();
  maxPLRUsed = 0;
  DeActivateAll();
  nActive = 0;
}


void PrintLnProbs(Path path)
{
  for(;path!=NULL; path=path->next)
  printf("lnP(o) = %f\n", path->pcell->logprob);  
}

#define SAVLST 1
#define SAVLEN 2
#define SAVLSC 4
#define LEVELSEP '/'

void LSave1(char *fname, Transcription t)
{
   LabList *p;
   int i,flags;
   Label *l;
   FILE *f;
   
   if ((f=fopen(fname,"w")) == NULL)
      HError(2,"Unable to create label file %s",fname);
   p = (LabList *) t;
   while (p != NULL){
      flags = 0;     /* see which optional fields need saving */
      for (i = 0, l=p->labels; i < p->nLabels; i++,l++){
         if (l->start != 0) flags |= SAVLST;
         if (l->end   != 0) flags |= SAVLEN;
         /**if (l->score != 0.0) flags |= SAVLSC;***/
      }
      for (i = 0, l=p->labels; i < p->nLabels; i++,l++){
         if (SAVLST&flags) 
            fprintf(f,"%d ",l->start);
         if (SAVLEN&flags) 
            fprintf(f,"%d ",l->end);
         fprintf(f,"%s ",l->labid->name);
         if (SAVLSC&flags)
            fprintf(f,"%f",l->score);
         fprintf(f,"\n");
      }
      if (p->next != NULL)
         fprintf(f,"%c \n",LEVELSEP);
      p = p->next;
   }
   fclose(f);
}

/* MatchFile: top level level control */
void MatchFile(char *datafn, Path *vpath, Path *cpath,
Transcription *vlabel, Transcription *clabel)
{
  int nbest, corLevel=0;
  char ilabfn[LINE_LEN];
  char rlabfn[LINE_LEN];
  Transcription tr, bogus_tr=NULL;
  Path bogus_path=NULL;
  Boolean hlPathInNBest=FALSE;
  

  OpenAndCheckData(&data,datafn,hmmP.vSize,hmmP.SK);
  sampPeriod = data.sampPeriod;
  if (trace>0) {
    printf("HGPD[V%s]: %s\n", VERSION,datafn);
    fflush(stdout);
  }
  MakeFN(datafn,ilabDir,labExt,ilabfn);
  tr = LOpen(ilabfn);
  if(trace>2) {
    printf("label file name is %s\n", ilabfn);
    PrintTranscription1(tr, FALSE);
  }
/*AD  to do Nbest, we perform Viterbi on the Nbest + 1 paths. If one of these
gives the proper labels then we move it to the HL path, otherwise, we do
a forced viterbi on the HL labels, and remove the last (N+1) path and labels
from the Viterbi in order to keep N paths there

Note: the variable NBest actually holds a value 1 greater than the NBest 
that is input to the program   

 */

  InitModels(theNet, FALSE);
  Viterbi(FALSE);
  TranscribeResults(data.nSamples, theNet, vpath, vlabel);

  /* adapt rlabDir to be new */
/*  AD-- No need to save these right now 

  {
    char rlabfn[LINE_LEN];
    
    MakeFN(datafn,UniquerlabDir,labExt,rlabfn);
    LSave1(rlabfn, *vlabel);   
  }
*/
  
  TidyUpAfterParse();
  
  corLevel = TrEqual(*vlabel, tr);
  if (corLevel != 1) NumCorrect--;
  if (corLevel == NBest|| corLevel == 0) InTopN--;

  if(corLevel == 0) {
    nbest = NBest;
    NBest=1;
    AttachHandLabels(theNet, tr);
    if(trace>2) PrintHandLabels();
    InitModels(theNet, TRUE);
    Viterbi(TRUE);
    TranscribeResults(data.nSamples, theNet, cpath, clabel); 
    FreeHandLabels();
    TidyUpAfterParse();
    NBest = nbest;
    if (NBest > 1){ /* this moves the last path to a bogus position, removing
		       it from the NBest paths */
      MovePath(vpath,NBest,&bogus_path);
      MoveLabels(vlabel,NBest,&bogus_tr);
    }
  }
  else {
    hlPathInNBest = TRUE;
    MovePath(vpath, corLevel, cpath); 
    MoveLabels(vlabel,corLevel,clabel);
  }
  LClose(bogus_tr);
  bogus_path = FreePath(bogus_path);
  LClose(tr);
  FillInPaths(*vpath, *vpath); 
  if(trace>2) {
    printf("Viterbi NBest Path\n");   
    PrintPath(*vpath);
  }
  if(trace>0)  {
    printf("Viterbi NBest Transcriptions\n");   
    PrintLnProbs(*vpath);
    PrintTranscription1(*vlabel, FALSE);
  }
  if(trace>2) {
    printf("Hand Labelled Path");   
    if(hlPathInNBest) printf(" (moved from NBest)");
    printf("\n");
    PrintPath(*cpath); 
  }
  if(trace>0)  {
    printf("Hand Labelled Transcription");   
    if(hlPathInNBest) printf(" (moved from NBest)");
    printf("\n");
    PrintLnProbs(*cpath);
    PrintTranscription1(*clabel, FALSE);
  }
}

/*----------------------------GPD Equations-------------------------------*/

/* g(): computes the discriminant measure (ie lnP(x)) for the observation sequence x*/
double g(Path x)
{
 if(x==NULL) HError(1,"No sequence to compute discriminant measure from");
 return(x->pcell->logprob);  
}

/* see p 10. A theory...... */
/* d_k(): returns the misclassication measure */
double d_k(Path k, Path alli)
{
 double g_k, G_k=0.0, K=1.0, d;
 Path i;

 if(k==NULL) HError(1, "No k path to compute discriminant measure from");
 g_k = g(k);
 for(i=alli; i!=NULL; K+=1.0, i=i->next) { 
   if(K==1.0) G_k = eta*g(i);
   else     G_k = LAdd(G_k, eta*g(i));
 }
 if(G_k != 0.0) {
  G_k += log(1.0/(K-1.0));
  G_k *= 1.0/eta;
 }
 d = -g_k + G_k;   /* page 10, second equation from bottom......*/
 return(d);
}

/* see page 15 eqn 2 A theory..... */
double Dd_Dg(Path alli, Path j)
{
 double g_j, T=0.0;
 Path i;

 g_j = g(j);
 for(i=alli; i!=NULL; i=i->next) 
   T += exp(eta*(g(i)-g_j));
 if(T==0.0) HError(1,"You've fucked up somewhere");
 return (1/T);
}

/* l(): return the loss......*/
double l(double d)
{
double loss;

switch (LossFunctionType) 
  {
  case POWER:
    if (d <= -beta) loss = 0.0;
    else loss = pow(d+beta, alpha);
    if (loss > 1.0) loss = 1.0;
    break;
    
  case SIGM:
    loss = Sigmoid(d,alpha,beta);
    break;

  case PIECEWISELINEAR:
    if (d <= -alpha) loss = 0.0;
    if (d > -alpha && d < beta) loss = (d+alpha)/(beta+alpha);
    if (d >= beta) loss = 1.0;
    break;
    
  default:  HError(1,"Invalid loss function type - should never get here");
    break;
  }
return (loss);      
}

/* computes the first differential of the loss w.r.t. 
   the misclassification measure */
double Dl_Dd(double s, double d)
{
double dloss;

switch (LossFunctionType) 
  {
  case POWER: 
    if (d <= -beta) dloss = 0.0;
    else 
      if (alpha == 1.0) dloss = 1.0;
      else dloss =  alpha*pow(d+beta,(alpha-1.0));
    break;
    
  case SIGM: 
    dloss =  s*s*alpha*exp(-alpha*(d-beta));
    break;
    
  case PIECEWISELINEAR:
    if (-alpha==beta) HError(1,"Infinite slope");
    if (-alpha > beta)  HError(1,"Negative slope");
    if (d <= -alpha || d >= beta) dloss = 0.0;
    else   dloss = 1.0/(beta+alpha);
    break;
    
  default:
    HError(1,"Invalid loss function type - should never get here");
    break;
  }
 
return(dloss);
}

void Dg_Dm(HMMDef *model, HMMDef *update, double wt, Vector o, int s, int m)
{
 int i;
 Covariance cov;    /* mixture variance */
 Vector mean;       /* mixture mean */ 
 double w;          /* mixture weight */
 Vector umean;      /* the update mean */
 double k;

 mean      =  model->stateinfo[s].outp[m].mean;
 w         =  model->stateinfo[s].outp[m].weight;
 cov.var   =  model->stateinfo[s].outp[m].covariance.var;
 umean     =  update->stateinfo[s].outp[m].mean;
 k = w*wt;
 for(i=1;i<=model->vecSize;i++)  {
   if(cov.var[i]==0.0) HError(1,"covariance element is zero");
   umean[i] +=  k*(o[i]-mean[i]) / cov.var[i];
  }
}

/* compute dg/dvar */
void Dg_Dv(HMMDef *model, HMMDef *update, double wt, Vector o, int s, int m)
{
 int i;
 double k;
 Covariance cov;    /* mixture variance */
 Vector mean;       /* mixture mean */ 
 double w;          /* mixture weight */
 Covariance ucov;   /* the update covariance */

 w         =  model->stateinfo[s].outp[m].weight;
 cov.var   =  model->stateinfo[s].outp[m].covariance.var;
 mean      =  model->stateinfo[s].outp[m].mean;
 ucov.var  =  update->stateinfo[s].outp[m].covariance.var;
 k = w*wt/2;
 for(i=1;i<=model->vecSize;i++) {
   if(cov.var[i]==0.0) HError(1,"covariance element is zero");
   ucov.var[i] += (k/cov.var[i] * (pow((ot[i]-mean[i]),2)/cov.var[i] -  1.0));
 }
}

/* computes dg/dw */
void Dg_Dw(HMMDef *update, double w, int s, int m)
{
   update->stateinfo[s].outp[m].weight += w; 
}

/* looks up an HMM by name in the database */
int GetModelIndex(char *name, Definition *modeldbase)
{
 int i;

 for(i=1;i<=H;i++) 
   if(!strcmp(name, modeldbase[i].id->name)) 
     return(i); 
 HError(1,"Model %s not found in the database", name);
}


void Dg_Dlambda(double carry, Path j, Boolean correct)
{
  int mixture, state;
  double logpo,logpm,k_pm_po, pm, po;
  HMMDef *model, *update;
  PathCell *c;
  int index;
  
  if (j==NULL) HError(1,"No path to differentiate"); 
  for(c=j->pcell; c!=NULL; c=c->link) {
    GetSample(&data, c->time-1, ot+1);
    logpo = c->bjot;
    index  = GetModelIndex(c->name, hmm); 
    model  = &(hmm[index].hmmdef);
    update = &(grad[index].hmmdef);
    state = c->state;
    if(correct) {
      ModelStats[index].CorrectNFrames += 1;
      ModelStats[index].Correctbjot += logpo;
    }
    else {
      ModelStats[index].WrongNFrames += 1;
      ModelStats[index].Wrongbjot += logpo;
    }
    if(logpo > LSMALL) {
      for(mixture=1;mixture<=model->numMixtures;mixture++) {
	
	if(MixtureExists(model,state,mixture)) {
	  logpm = MOutP(ot,model,state,mixture);
	  k_pm_po = L2F(logpm-logpo) * carry;
	  if(uFlags&UPMEANS) Dg_Dm(model,update,k_pm_po,ot,state,mixture);
	  if(uFlags&UPMIXES) Dg_Dw(update,k_pm_po,state,mixture);
	}
      }
    }
  }
}

double fgrad(Path alli, Path k) /* vpath, cpath */
{
  double f1,f2;
  double loss=0.0;
  double d;
  Path j;
  
  if(alli!=NULL) { /* only update if wrong examples exist */
    d = d_k(k, alli); if(trace>0)   {printf("    d(%e)\n",d);fflush(stdout);}
    loss = l(d);      if(trace>0) {printf("    l(%e)\n",loss);fflush(stdout);}
    f1 = Dl_Dd(loss,d); if(trace>0) {printf("Dl_Dd(%e)\n",f1);fflush(stdout);}
    f2 = -1.0;
    Dg_Dlambda(f1*f2, k, TRUE);
    for(j=alli; j!=NULL; j=j->next) { 
      f2 = Dd_Dg(alli,j); 
      if(trace>0) {printf("Dd_Dg(%e)\n",f2);fflush(stdout);}
      Dg_Dlambda(f1*f2, j, FALSE); 
    }
  }
  else if(trace>0) {printf("Skipping update on this file\n");fflush(stdout);}
  return(loss);
}

double ELoss(char *filename)
{
 Path cpath=NULL, vpath=NULL;
 Transcription  clabel=NULL, vlabel=NULL;
 double loss=0.0;
 int i;

 if(trace>2) {printf("%s\n", filename);fflush(stdout);}
 MatchFile(filename, &vpath, &cpath, &vlabel, &clabel); 
 loss = fgrad(vpath, cpath);           
 vpath = FreePath(vpath); 
 LClose(vlabel);
 LClose(clabel);
 cpath = FreePath(cpath);
 clabel = NULL;
 vlabel = NULL;
 SpClose(&data);
 return(loss);
}

void SaveHMMDefs1(Definition *model, char *hmmDir, char *hmmExt, int n)
{
  int i;
  char UniquehmmDir[LINE_LEN];
  char hmmfn[LINE_LEN];

  if(ChangeHmmDir) NewDir(hmmDir, UniquehmmDir, n);
  else strcpy(UniquehmmDir, hmmDir);
  for(i=1;i<=H;i++) {
    MakeFN(model[i].id->name, UniquehmmDir,hmmExt,hmmfn);
    if (trace>2) printf("Writing out HMM %s\n", hmmfn);
    HSave(hmmfn, &(model[i].hmmdef));
  }
}

void  ScaleUpdates(Definition *hmm, Definition *grad)
{
  int h,s,m;
  double omaxw=0.0, omaxm=0.0, umaxw=0.0, umaxm=0.0;
  HMMDef *ohmm, *uhmm;
  OutProb *ooutp, *uoutp;
  double tmp;
  Boolean chmean = 1, chmix = 1;

  for (h=1;h<=H;h++) { 
    uhmm = &(grad[h].hmmdef);
    ohmm = &(hmm[h].hmmdef);
    for (s=2;s<ohmm->numStates;s++){ 
      for (m=1;m<=ohmm->numMixtures;m++) {
	if(MixtureExists(ohmm,s,m)) {
	  ooutp = ohmm->stateinfo[s].outp+m;
	  uoutp = uhmm->stateinfo[s].outp+m;
	  tmp = fabs(uoutp->weight); if (tmp > umaxw) umaxw = tmp;
	  tmp = fabs(ooutp->weight); if (tmp > omaxw) omaxw = tmp;
	  tmp = Vector2Norm(uoutp->mean); if (tmp > umaxm) umaxm = tmp;
	  tmp = Vector2Norm(ooutp->mean); if (tmp > omaxm) omaxm = tmp;
	}
      }
    }
  }
  
  if (uFlags&UPMIXES) {
    if(umaxw == 0.0){/* HError(0,"Warning no weight updates");*/ chmix = 0;}
    if(omaxw == 0.0) HError(1,"No old weight");
  }
  if (uFlags&UPMEANS) {
    if(umaxm == 0.0){ /*  HError(0,"Warning no mean updates");*/ chmean = 0;}
    if(omaxm == 0.0)   HError(1,"No old means");
  }
  
  for (h=1;h<=H;h++) { 
    uhmm = &(grad[h].hmmdef);
    ohmm = &(hmm[h].hmmdef); 
    
    for (s=2;s<ohmm->numStates;s++){ 
      for (m=1;m<=ohmm->numMixtures;m++) {
	
	if(MixtureExists(ohmm,s,m)) {
	  
	  ooutp = ohmm->stateinfo[s].outp+m;
	  uoutp = uhmm->stateinfo[s].outp+m;
	  
	  if ((uFlags&UPMEANS) && (chmean))
	    VectorMultiply(uoutp->mean, omaxm/umaxm);
	  if ((uFlags&UPMIXES) && (chmix))
	    uoutp->weight *= (omaxw/umaxw);
	}
      }
    }
 }
}

/*ZeroNegativeMixtures: implements 'if (w_i<MINMIX) then w_i = 0'.
Mixture weights must be >=0 so all w_i<0 must be set to 0. */
void ZeroNegativeMixtures(OutProb *outp)
{
 int m;

 for(m=1;m<=hmmP.M;m++) if(outp[m].weight<0.0) outp[m].weight = 0.0;
} 

/* SumMixtureWeights2One: normalises the mixture weights using
   w_i = w_i / sum_j=1^M w_i where M is the number of mixtures. 
   Note prior to normalisation negative mixture weights are set to 0. */

void  SumMixtureWeights2One(OutProb *outp)
{
 double sum=0.0;
 int m;

 ZeroNegativeMixtures(outp);
 for(m=1;m<=hmmP.M;m++) sum += outp[m].weight;
 if (sum != 0.0) for(m=1;m<=hmmP.M;m++) outp[m].weight /= sum; 
 else HError(1,"Oops - Model state has disappeared!");
}

/* Multiply: Multiply the gradient by the learning constant epsilon */
void Multiply(Definition *dX, double epslon)
{

 OutProb *outp;
 HMMDef *dx;
 int s,m,h;
  
 for (h=1;h<=H;h++) { 
   dx = &(dX[h].hmmdef);  
   for (s=2;s<hmmP.maxStates;s++){ 
     for (m=1;m<=hmmP.M;m++) {
         outp = dx->stateinfo[s].outp+m;
         if (uFlags&UPMEANS) VectorMultiply(outp->mean, epslon);
         if (uFlags&UPMIXES)  outp->weight *= epslon;
     }
   }
 }
}

void PrintAllHMMs(char *title, Definition *def, int flag)
{
 int i;

 printf("%s\n", title);
 for(i=1;i<=H;i++) 
   DRHPrint(&(def[i].hmmdef),flag, 2,def[i].hmmdef.numStates-1);
}

/* replace X by X+dX */
void Add(Definition *X, Definition *dX)
{

  int h,s,m;
  HMMDef *x, *dx;
  OutProb *xoutp, *dxoutp;
  
  if(trace>2) PrintAllHMMs("Adapted Scaled Updates", grad,(UPMEANS|UPMIXES));
  
  for (h=1;h<=H;h++) { 
    x  = &(X[h].hmmdef);
    dx = &(dX[h].hmmdef); 
    for (s=2;s<x->numStates;s++){ 
      for (m=1;m<=x->numMixtures;m++) {
	if(MixtureExists(x,s,m)) {
	  xoutp  =   x->stateinfo[s].outp+m;
	  dxoutp =  dx->stateinfo[s].outp+m;
	  if (uFlags&UPMEANS) AddVectors(xoutp->mean, dxoutp->mean);
	  if (uFlags&UPMIXES) xoutp->weight += dxoutp->weight;
	}
      }
      if (uFlags&UPMIXES) SumMixtureWeights2One(x->stateinfo[s].outp);
    }
  }
}

/* replace X by X + e * dX */
void Adapt(Definition *X, Definition *dX, double epsilon)
{
  Multiply(dX, -1.0*epsilon);
  Add(X, dX);
} 

void CopyHmm(Definition from, Definition to)
{
  int s,m;
  for(s=2; s<from.hmmdef.numStates; s++) {
    for(m=1; m<=from.hmmdef.numMixtures; m++) {
      to.hmmdef.stateinfo[s].outp[m].weight = 
	from.hmmdef.stateinfo[s].outp[m].weight; 
      CopyVector(from.hmmdef.stateinfo[s].outp[m].mean, 
		 to.hmmdef.stateinfo[s].outp[m].mean);
      CopyVector(from.hmmdef.stateinfo[s].outp[m].covariance.var, 
		 to.hmmdef.stateinfo[s].outp[m].covariance.var);
    }
  }
}

void CopyHmms(Definition *from, Definition *to)
{
 int i;
 for(i=1;i<=H;i++) CopyHmm(from[i], to[i]);
}

void PrintStats(int n, double loss)
{
  
  double TECorrect, TEWrong, ECorrect, EWrong;
  int k;
  int CorrectNFrames, WrongNFrames;
  
  printf("Loss at %d'th iteration = %f, %%correct %f, %%in top N %f\n",
	 n+1, loss,(float)NumCorrect/(float)Nfiles, 
	 (float)InTopN/(float)Nfiles);
  fflush(stdout);
  
  TECorrect = 0.0; TEWrong = 0.0;
  for(k=1;k<=H;k++) {
    CorrectNFrames = ModelStats[k].CorrectNFrames;
    WrongNFrames = ModelStats[k].WrongNFrames;
    if(CorrectNFrames == 0) ECorrect = 0.0;
    else  ECorrect = ModelStats[k].Correctbjot/CorrectNFrames;
    if(WrongNFrames == 0) EWrong = 0.0;
    else  EWrong = ModelStats[k].Wrongbjot/WrongNFrames;
    TECorrect += ECorrect;
    TEWrong += EWrong;
    if (trace > 0)
      printf("model [%s] EClogP(o) [%f] EWlogP(o) [%f]\n", 
	     hmm[k].id->name, ECorrect, EWrong);
  }
  if (trace > 0){
    printf("Total EClogP(o) [%f] EWlogP(o) [%f]\n",  TECorrect, TEWrong);
    printf("\n");fflush(stdout);
  }
}

void GPDUpdate(void)
{
  double eloss, epsilon = 0.0, temploss=0.0;
  int startfile=0;
  int i,j, ITRNS, itrns=0;
  Boolean NeedChange = 0;

  if(FIntrvl==UNDEFINED) FIntrvl=Nfiles;
  fill_array(filenum, Nfiles);
  ITRNS = NIter * Nfiles;

  ZeroHMMDefs(grad);      
  ZeroModelStats();
  
  for(i=0;i<NIter;i++) {
    NeedChange = 0;
    jumble(filenum, Nfiles);
    eloss = 0.0;
    NumCorrect = Nfiles;
    InTopN = Nfiles;
    if(ChangeHmmDir) NewDir(rlabDir, UniquerlabDir, i);
    else strcpy(UniquerlabDir, rlabDir);
    for(j=0;j<Nfiles;j++) {
      epsilon = Epsilon * (1.0 - (float)itrns/(float)ITRNS);
      temploss = ELoss(files[filenum[j]]);
      if (trace > 0) printf("%d  loss = %f\n",j,temploss);
      eloss += temploss;
      if (((j+1) % FIntrvl) == 0){
	if (trace > 0) printf("****updating HMMs\n");
	if(trace>2) PrintAllHMMs("Unscaled Updates", grad,(UPMEANS|UPMIXES));
	ScaleUpdates(hmm, grad);      
	if(trace>1) PrintAllHMMs("Scaled Updates", grad, (UPMEANS|UPMIXES));
	Adapt(hmm, grad, epsilon);            
	ZeroHMMDefs(grad);      
      }
      itrns++;
    }
    SaveHMMDefs1(hmm, outHmmDir, hmmExt, i);
    PrintStats(i, eloss);
    ZeroModelStats();
 }
}

/* ----------------------------------------------------------- */
/*                      END:  HGPD3.c                          */
/* ----------------------------------------------------------- */
