
/* ----------------------------------------------------------- */
/*                   HTK  -  HMM Toolkit                       */
/* ----------------------------------------------------------- */
/*         Copyright: Steve Young      October 1989            */
/*         CoAuthor:  Val Beattie                              */
/*         Last Update:                September 1990          */
/*                                                             */
/*         Cambridge University Engineering Department         */
/*         Trumpington Street, Cambridge, CB2 1PZ              */
/*                                                             */
/* Permission is hereby granted to copy, use and modify this   */
/* software provided that this banner notice is not removed    */
/*                                                             */
/* ----------------------------------------------------------- */
/*         File: HInit.c: HMM initialisation program           */
/* ----------------------------------------------------------- */

/*
   This program is used to initialise (or tune) a hidden Markov model

    Performs the following steps:
      1.  Uniformly segment observations 
      2.  cluster observations into K clusters, where K=num of mixtures
      3.  Compute parameters of current model
      4.  Resegment observations via Viterbi for current model
      5.  Cluster observations into K clusters 
      6.  Compute parameters of new model
      7.  If not max iterations and new model != current model then
          current model=new model; goto 4
      8.  Store model in output file
      Clustering is accomplished via a K-means algorithm.

   HInit can also be used on an existing model, eg one which has just 
   come from standard Baum-Welch training, in this case the first three
   steps are skipped.

   In every case, the structure of the HMM to be initialised must be
   defined beforehand (eg using a prototype HMM def).  This program only 
   computes initial parameters for the output distributions.
   
   If a segment label is specified, then for each data file HInit expects
   to find a corresponding label file.  Using these label files, all 
   examples of each segment are cut from the data files and used as
   training tokens.  If no segment label is specified, each file is
   assumed to be a single token.
   
*/


#include <stdio.h>      /* Standard C Libraries */
#include <stdlib.h>
#include <math.h>

#include "HShell.h"     /* HMM ToolKit Modules */
#include "HMath.h"
#include "HSpIO.h"
#include "HLabel.h"
#include "HDbase.h"  
#include "HModel.h"
#include "HKMeans.h"

#define VERSION   "1.2"

/* Global Settings */
static char * segLab = NULL;  /* segment label if any */
static LabId  segId  = NULL;  /* and its id */
static char * labDir = NULL;  /* label file directory */
static char * labExt = "lab"; /* label file extension */
static int  maxIter  = 20;    /* max iterations in parameter estimation */
static int  maxCIter = 16;    /* max iterations in clustering */
static float epsilon = 0.0001; /* convergence criterion */
static float minVar  = 0.01; /* minimum variance */
static int minSeq    = 3;     /* min sequences to train a model */
static int  newModel = TRUE;  /* flag to enable initial uniform segmentation */
static char *hmmfn;           /* HMM definition file name */
static char *outfn=NULL;      /* output definition file name */
static int  trace    = 0;     /* Trace level */

/* For Convenience */
static int N;           /* numStates */
static int M;           /* numMixtures */
static int V;           /* vecSize */
static CovKind ckind;   /* Covariance kind {DIAGC,FULLC} */

/* Major Data Structures */
static HMMDef theHMM;         /* The HMM to be initialised (or updated) */
static VectorPool *trainData; /* array[2..numStates-1] of VectorPool */

/* Storage for mean and covariance stats */
static Matrix       meansum;     /* array[1..numMix][1..vecSize] */
static Covariance  *squaresum;   /* array[1..numMix] of Covariance */

/* Storage for Viterbi Decoding */
static Vector   thisP,lastP;     /* Columns of log probabilities */
static short   **traceBack;      /* array[0..T-1][2..numStates-1] */

/* ------------------ Process Command Line ------------------------- */

void ReportUsage(void)
{
   fprintf(stderr,"\nUSAGE: HInit[V%s] [options] hmmFile trainFiles...\n\n",
         VERSION);
   fprintf(stderr," Option                                   Default\n\n");
   fprintf(stderr," -c N    Set max cluster iterations to N    16\n");
   fprintf(stderr," -e f    Set convergence factor epsilon    1.0E-4\n");
   fprintf(stderr," -i N    Set max iterations to N            20\n");
   fprintf(stderr," -l s    Set segment label to s             None\n");
   fprintf(stderr," -m N    Set min sequences needed             3\n");
   fprintf(stderr," -o fn   Store new hmm def in fn            srcfn\n");
   fprintf(stderr," -u      Update hmm (suppress uniform seg)  off\n");
   fprintf(stderr," -v f    Set minimum variance to f         1.0E-2\n");
   fprintf(stderr," -F fmt  Set data file format to fmt        HTK\n");
   fprintf(stderr," -G fmt  Set label file format to fmt       HTK\n");
   fprintf(stderr," -L s    dir to find label file(s)        current\n");
   fprintf(stderr," -S f    set script file to f               none\n");
   fprintf(stderr," -T N    set trace level to N                0\n");
   fprintf(stderr," -X s    extension for label files          lab\n");
   exit(1);
}

int GetChkedInt(int min, int max, char * swtname)
{
   int val;
   
   if (NextArg() != INTARG) 
      HError(1,"Integer Arg Required for %s option",swtname);
   val=GetIntArg();
   if (val<min || val>max) 
      HError(1,"Integer arg out of range in %s option",swtname);
   return val;
}

float GetChkedFlt(float min, float max, char * swtname)
{
   float val;
   
   if (NextArg() != INTARG && NextArg() != FLOATARG) 
      HError(1,"Float Arg Required for %s option",swtname);
   val=GetFltArg();
   if (val<min || val>max) 
      HError(1,"Float arg out of range in %s option",swtname);
   return val;
}

main(int argc, char *argv[])
{
   char *datafn, *s;
   FileFormat ff;
   int nSeq;
   void Initialise(void);
   void LoadFile(char *fn);
   void EstimateModel(void);
   
   InitShell(argc,argv);
   InitMath(TRUE);
   if (NumArgs()==0)
      ReportUsage();
   while (NextArg() == SWITCHARG) {
      s = GetSwtArg();
      if (strlen(s)!=1) 
         HError(1,"Bad switch %s; must be single letter",s);
      switch(s[0]){
         case 'c':
            maxCIter = GetChkedInt(1,100,s); break;
         case 'e':
            epsilon = GetChkedFlt(0.0,1.0,s); break;
         case 'i':
            maxIter = GetChkedInt(1,100,s); break;
         case 'l':
            if (NextArg() != STRINGARG)
               HError(1,"Segment label expected");
            segLab = GetStrArg();
            break;
         case 'm':
            minSeq = GetChkedInt(1,1000,s); break;
         case 'o':
            outfn = GetStrArg();
            break;
         case 'u':
            newModel = FALSE; break;
         case 'v':
            minVar = GetChkedFlt(0.0,10.0,s); break;
         case 'T':
            trace = GetChkedInt(0,3,s); break;
         case 'F':
            if (NextArg() != STRINGARG)
               HError(1,"Data File format expected");
            if((ff = Str2Format(GetStrArg())) == ALIEN)
               HError(0,"Warning ALIEN Data file format set");
            SetFormat(ff);
            break;
         case 'G':
            if (NextArg() != STRINGARG)
               HError(1,"Label File format expected");
            if((ff = Str2Format(GetStrArg())) == ALIEN)
               HError(0,"Warning ALIEN Label file format set");
            SetLFormat(ff);
            break;
         case 'L':
            if (NextArg()!=STRINGARG)
               HError(1,"Label file directory expected");
            labDir = GetStrArg(); break;
         case 'S': SetScriptFile();
            break;
         case 'X':
            if (NextArg()!=STRINGARG)
               HError(1,"Label file extension expected");
            labExt = GetStrArg(); break;
         default:
            HError(1,"Unknown switch %s",s);
      }
   } 
   if (NextArg()!=STRINGARG)
      HError(1,"source HMM file name expected");
   hmmfn = GetStrArg();
   if (outfn==NULL) outfn = hmmfn;
   Initialise();
   do {
      if (NextArg()!=STRINGARG)
         HError(1,"training data file name expected");
      datafn = GetStrArg();
      LoadFile(datafn);
   } while (NumArgs()>0);
   FixDbase();
   nSeq = NumSequences();
   if (nSeq < minSeq)
      HError(2,"Too Few Observation Sequences [%d]",nSeq);
   if (trace>0) {
      printf("  %d Observations Loaded\n",nSeq);
      fflush(stdout);
   }
   EstimateModel();
   HSave(outfn,&theHMM);
}

/* ------------------------ Initialisation ----------------------- */

void PrintInitialInfo()
{   
   char dummy[6];
   if (newModel) printf("Initialising  "); else printf("Updating  ");
   printf("HMM %s . . . \n",hmmfn);
   printf("    NumStates:  %d\n",N);
   printf("    NumMix   :  %d\n",M);
   printf("    VecSize  :  %d\n",V);
   printf("    CovKind  :  %s\n",CovKind2Str(ckind,dummy));
   if (segLab==NULL)
   printf("    SegLab   :  NONE\n");
   else
   printf("    SegLab   :  %s\n",segLab);
   printf("    maxIter  :  %d\n",maxIter);
   printf("    maxCIter :  %d\n",maxCIter);
   printf("    epsilon  :  %f\n",epsilon);
   fflush(stdout);
}
   
void Initialise(void)
{
   int k;

   SPINBALL
   HOpen(hmmfn,&theHMM);
   N = theHMM.numStates; M = theHMM.numMixtures; 
   V = theHMM.vecSize; ckind = theHMM.ckind;
   if(segLab != NULL) {
      InitLabel();
      segId = GetLabId(segLab,TRUE);
   }
   if (trace>0) PrintInitialInfo();
   InitDbase(V,theHMM.pkind);
   thisP = CreateVector(N);
   lastP = CreateVector(N);
   if ( (trainData=(VectorPool *)malloc((N-2)*sizeof(VectorPool))) == NULL)
      HError(2,"Unable to create trainData");
   trainData -= 2;   /* to allow index 2..N-1 */
   SPINBALL
   meansum=CreateMatrix(M,V);
   if((squaresum=(Covariance *)malloc(M*sizeof(Covariance))) == NULL)
      HError(2,"Cannot create squaresum");
   --squaresum;      /* to allow index 1..M */
   if (ckind==FULLC) 
      for (k=1;k<=M;k++) 
         squaresum[k].inv=CreateMatrix(V,V);
   else              
      for (k=1;k<=M;k++) 
         squaresum[k].var=CreateVector(V);
}

/* --------------------------- Load Database ---------------------- */

/* CheckData: check data file consistent */
void CheckData(DataFile *data, char *fn)
{
   if (BaseSampKind(data->sampKind) >= ANON)   
      HError(2,"Unknown Parameterisation in Training Data %s",fn);
   if (data->sampKind != theHMM.pkind)
      HError(2,"Parameterisation in %s incompatible with hmm %s",fn,hmmfn);
}

/* LoadFile: load either whole file or reqd segments of fn into HDBase */
void LoadFile(char *fn)
{
   char labfn[80];
   Transcription trans;
   Label *labl;
   int vecStartInd,vecEndInd;  
   int i,ncas;
   DataFile data;
   
   SPINBALL
   if (segId == NULL)  {
      SpOpen(fn,&data);
      CheckData(&data,fn);
      PutSequence(0,(data.nSamples-1)*data.sampPeriod,&data);                  
      if (trace > 1) {
         printf(" Training Observations loaded from %s\n",fn);
         fflush(stdout);
      }        
      SpClose(&data);
   }
   else {
      MakeFN(fn,labDir,labExt,labfn);
      trans = LOpen(labfn);
      ncas=NumCases(segId,trans,0);    /* assume single level trans */
      if (ncas>0) {
         SpOpen(fn,&data);
         CheckData(&data,fn);
         for (i=1;i<=ncas;i++) {
            labl=GetCase(segId,i,trans,0);
            vecStartInd=labl->start/data.sampPeriod;
            vecEndInd=labl->end/data.sampPeriod;
            if (vecEndInd-vecStartInd + 1 > N-2)  
               PutSequence(labl->start,labl->end,&data);                  
            else   
               if (trace > 1) 
                  HError(0,"Ignoring label length  %d",vecEndInd-vecStartInd+1); 
         }
         SpClose(&data);
      }
      LClose(trans);
      if (trace > 1) {
         printf(" %d Training Observations loaded from %s\n",ncas,fn);
         fflush(stdout);
      }        
   } 
}

/* -------------------------- Segmentation -------------------------- */

/* NewPools: attach new empty pools to trainData */
void NewPools(void)
{
   int i;
   
   for (i=2;i<N;i++)
      CreateVectorPool(trainData+i);
}

/* ClearPools: clear pools attached to trainData */
void ClearPools(void)
{
   int i;
   
   for (i=2;i<N;i++)
      FreeVectorPool(trainData+i);
}

/* UniformSegment: uniformly segment all training tokens in the DBase */
void UniformSegment()  
{
   float vperSeg;
   int i,j,s;
   int numSeq,numVecs;
   VecInfo vi;

   if (trace > 1) {
      printf("Performing Uniform Segmentation\n");
      fflush(stdout);
   }
   NewPools();
   numSeq=NumSequences();
   for (i=0;i<numSeq;i++) {
      numVecs=SequenceLength(i);
      vperSeg=((float) numVecs)/((float) (N-2));
      if (vperSeg < 1.0)
         HError(2,"Null Uniform Segment in token of %d vectors",numVecs);
      for (j=0;j<numVecs;j++) {
         s = (int)(j/vperSeg) + 2;
         vi.sequence = i; vi.index = j;
         AddToVectorPool(vi,trainData+s);
      }
      SPINBALL    
   } 
}

/* ShowTraceBack: print the Viterbi traceBack matrix */
void ShowTraceBack(int len, short **tB)
{
   int i,s;
   
   printf("Traceback Matrix\n");
   for (s=2;s<=N-1;s++) {
      printf("     %d     ",s);
      for (i=0;i<len;i++)
         printf("%3d",tB[i][s]);
      printf("\n");
   }
   fflush(stdout);
}

/* ShowP: print current column of log probabilites */
void ShowP(int col, Vector colVec)
{
   int s;
      
   printf("     %d     ",col);
   for (s=2;s<=N-1;s++) {
      if (colVec[s]<=LSMALL)
         printf("  ----  ");
      else
         printf("%8.2f",colVec[s]);
   }
   printf("\n"); fflush(stdout);
}

/* MakeTraceBack: create the traceBack matrix */
void MakeTraceBack(int seqLen)
{
   int t;
   short *p;
   
   if ((traceBack = (short **)calloc(seqLen,sizeof(short *))) == NULL)
      HError(2,"Cannot make traceBack pointer array");
   for (t=0;t<seqLen;t++) {
      if ((p = (short *)calloc(N,sizeof(short))) == NULL)
         HError(2,"Cannot make traceBack column");
      traceBack[t] = p-2;
   }
}

/* FreeTraceBack: free the traceBack matrix */
void FreeTraceBack(int seqLen)
{
   int t;
   
   for (t=0;t<seqLen;t++)
      free(traceBack[t]+2);
   free(traceBack);
}

/* DoTraceBack:  traceBack from s and allocate vectors to trainData pools */
void DoTraceBack(int seqLen, int seqNum, int s)
{
   VecInfo vi;
   int t;
   
   vi.sequence = seqNum;
   for (t=seqLen-1;t>=0;t--)  {
      vi.index = t;
      AddToVectorPool(vi,trainData+s);
      s=traceBack[t][s];
   }
}

/* Viterbi: use the current HMM to segment given sequence */
LogReal Viterbi(int seqNum) 
{
   Vector      v;        
   int         s,ss,bestPrevState;
   int         t,seqLen;
   LogReal     bestProb,p,x,y;

   if (trace > 2)   
      printf("   Viterbi State Alignment   \n");
   seqLen = SequenceLength(seqNum);
   MakeTraceBack(seqLen);
   
   /*From entry state 1: Column 1*/
   v = GetDBVector(seqNum,0);
   for (s=2;s<N;s++) {
      x = theHMM.transP[1][s];
      if (x<LSMALL) 
         lastP[s] = LZERO;
      else
         lastP[s] = x + OutP(v,&theHMM,2);
      traceBack[0][s] = 1;
   }
   if (trace > 2) ShowP(1,lastP);
   SPINBALL   
   
   /* Columns[1] -> Columns[seqLen-1] -- this is the general case */
   for (t=1;t<seqLen;t++) {
      v=GetDBVector(seqNum,t);      
      for (s=2;s<N;s++) {
         bestPrevState=2; 
         x = theHMM.transP[2][s]; y = lastP[2];
         bestProb = (x<LSMALL) ? LZERO : x+y;
         for (ss=3;ss<N;ss++) {
            x = theHMM.transP[ss][s]; y = lastP[ss];
            p = (x<LSMALL) ? LZERO : x+y;
            if (p > bestProb) {
               bestPrevState=ss; bestProb=p;
            }
         }
         thisP[s] = (bestProb<LSMALL) ? LZERO : bestProb+OutP(v,&theHMM,s);
         traceBack[t][s]=bestPrevState;
      }
      CopyVector(thisP,lastP);
      if (trace>2)
         ShowP(t+1,lastP);    
   }
   
   /* column[nVecs]--> exit state(numStates) */
   bestPrevState=2;
   x = theHMM.transP[2][N]; y = lastP[2];
   bestProb=(x<LSMALL) ? LZERO : x+y;
   for (ss=3;ss<N;ss++) {
      x = theHMM.transP[ss][N]; y = lastP[ss];
         p = (x<LSMALL) ? LZERO : x+y;
         if (p > bestProb) {
            bestPrevState=ss; bestProb=p;
         }
   }
   SPINBALL    
   
   /* bestPrevState now gives last internal state along best state sequence */
   
   if (bestProb<LSMALL)
      HError(2,"No Viterbi alignment found!");
   if (trace>2)
      ShowTraceBack(seqLen,traceBack);
   if (trace>2) {
      printf(" bestProb = %12.5f via state %d\n",bestProb,bestPrevState);
   }
   DoTraceBack(seqLen,seqNum,bestPrevState);
   SPINBALL    
   FreeTraceBack(seqLen);
   return bestProb;
   
} 

/* ViterbiSegment: use the current HMM to segment the training data */
LogReal ViterbiSegment(void)
{
   LogReal sumP;
   int i;
   int numSeq;
   
   if (trace > 1) {
      printf("Performing Viterbi Segmentation\n");
      fflush(stdout);
   }
   NewPools();
   sumP = 0.0;
   numSeq = NumSequences();
   for (i=0;i<numSeq;i++)
      sumP=sumP + Viterbi(i);
   sumP = sumP/numSeq;
   return sumP;
}



/* -------------------------- Model Estimation ---------------------- */


/* EstimateParameters: estimate stateinfo given clusters of pool */
void EstimateParameters(StateInfo stateinfo, ClusterSet cset, VectorPool *pool)

{
   int i,k,x,y;
   int *itemp;
   Vector v;
   OutProb *outdistp;
   int count,totalcount;
   Real val1,val2;
   
   if (trace > 1) {
      printf("Performing Parameter Estimation\n");
      fflush(stdout);
   }   
   /* Zero mean and square accumulators */
   ZeroMatrix(meansum);
   for (k=1;k<=M;k++) 
      if (ckind==FULLC)
         ZeroMatrix(squaresum[k].inv);
      else
         ZeroVector(squaresum[k].var);
   SPINBALL    

   /*  Estimate Parameters  */
   outdistp=stateinfo.outp;
   totalcount=pool->poolSize;
   for (k=1;k<=M;k++) { 
      count=cset.cl[k-1].numItems;
      if (count==0) 
         HError(2,"No Samples Collected for Mixture %d",k);
      itemp = cset.cl[k-1].items;
      for (i=0;i<count;i++)  {
         v=GetPoolVector(pool->vi[*itemp]); itemp++;
         for (x=1;x<=V;x++) {
            val1=v[x];
            meansum[k][x] += val1;
            if (ckind==FULLC) {
               squaresum[k].inv[x][x] += val1*val1;
               for (y=x+1;y<=V;y++) 
                  squaresum[k].inv[x][y] += val1*v[y];
            }
            else
               squaresum[k].var[x] += val1*val1;
         }
      }
      for (x=1;x<=V;x++) {
         val1 = meansum[k][x]/(float)count;
         outdistp[k].mean[x]=val1;
         if (trace > 1) 
            printf("    Mean[%2d] =%10.4f    ",x,val1); 
         if (ckind==FULLC) {
            for (y=x;y<=V;y++) {
               val2=squaresum[k].inv[x][y]/(float)count -
                          (val1*meansum[k][y])/(float)count; 
               squaresum[k].inv[x][y]=
                  (val2>minVar || x != y) ? val2 : minVar;     
            } 
            for (y=x+1;y<=V;y++)
               squaresum[k].inv[y][x] = squaresum[k].inv[x][y];
            if (trace > 1) printf("\n");
         }
         else {
            val2=squaresum[k].var[x]/(float)count-val1*val1;
            outdistp[k].covariance.var[x]=(val2>minVar) ? val2 :minVar;
            if (trace > 1) 
               printf("    Var[%2d] =%10.4f\n",x,
                           outdistp[k].covariance.var[x]); 
         }
     }
     if (ckind==FULLC) {
         if (trace>1) {
            printf("Covariance Matrix\n");
            WriteMatrix(stdout,squaresum[k].inv);
         }
        CovInvert(squaresum[k].inv,outdistp[k].covariance.inv);
         if (trace>1) {
            printf("Inverse Covariance Matrix\n");
            WriteMatrix(stdout,outdistp[k].covariance.inv);
         }
      }
     outdistp[k].weight=((float) count)/((float) totalcount); 
     if (trace > 1) 
        printf("  Weight=%6.4f=(samps for mixture %2d/total samples %3d)\n",
                          outdistp[k].weight,count,totalcount);
   }
   SPINBALL    
}

/* ----------------- Top Level of Estimation Procedure --------------- */

void EstimateModel()
{
   LogReal maxP,newP,delta;
   int     converged;
   int     iter,s;     
   ClusterSet cset;
   
   maxP=0; iter=0;
   if (newModel) 
      UniformSegment();
   else 
      maxP=ViterbiSegment();
   do {
      for (s=2;s<N;s++) {
         if (trace > 1) {
            printf(" State %d  \n",s);
            printf("  Creating Clusters - Pool Size = %d\n",
                  trainData[s].poolSize);
            fflush(stdout);
         }
         CreateClustSet(&cset,trainData+s,M,maxCIter,V);
         if (trace==3)
            DumpClustSet(&cset,trainData+s);
         EstimateParameters(theHMM.stateinfo[s],cset,trainData+s);
         FreeClustSet(&cset);
      }
      FixGConsts(&theHMM);
      ClearPools();
      newP=ViterbiSegment();
      iter++;
      delta = newP-maxP;
      converged = (fabs(delta) < epsilon);
      maxP = newP;
      if (trace > 0) {
         printf("Iteration %d: Average LogP =%12.5f",iter,maxP);
         if (iter > 1 || !newModel)
            printf("  Change =%12.5f\n",delta);
         else 
            printf("\n");
         fflush(stdout);
      }
   } while (iter<maxIter && !converged);
   if (trace > 0) {
      if (converged) 
         printf("Estimation converged at iteration %d\n",iter);
      else 
         printf("Estimation aborted at iteration %d\n",iter);
      fflush(stdout);
   }
}



/* ----------------------------------------------------------- */
/*                      END:  HInit.c                          */
/* ----------------------------------------------------------- */
