/* ----------------------------------------------------------- */
/*                   HTK  -  HMM Toolkit                       */
/* ----------------------------------------------------------- */
/*         Copyright: Steve Young      October 1989            */
/*         CoAuthor:  Val Beattie                              */
/*         Last Modified:              Sept 1990               */
/*                                                             */
/*         Cambridge University Engineering Department         */
/*         Trumpington Street, Cambridge, CB2 1PZ              */
/*                                                             */
/* Permission is hereby granted to copy, use and modify this   */
/* software provided that this banner notice is not removed    */
/*                                                             */
/* ----------------------------------------------------------- */
/*         File: HKMeans.c: K-Means Clustering Module          */
/* ----------------------------------------------------------- */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
                     
#include "HShell.h"
#include "HMath.h"
#include "HSpIO.h"
#include "HDbase.h"
#include "HKMeans.h"

/* 
   This module provides KMeans clustering for use by HInit.  The KMeans
   procedure has been hacked to make it linear wrt to pool size in both
   time and space, to do this distances are calculated on the fly and
   the criteria for choosing cluster centres and cluster sizes are
   very simplistic (and sensitive to outliers).  The centre is found
   by first finding two vectors opposite to each other on the periphery 
   of the cluster and then finding the vector nearest to the average of 
   these two peripheral vectors.  The cluster size is just the average
   distance of all vectors to the centre.
*/

#define POOLCHUNKSIZE   100      /* space increment for pools */
#define CLUSCHUNKSIZE   100      /* space increment for clusters */

static ClusterSet *theCS;     /* Global pointer to current cluster set */
static VectorPool *thePool;   /* .. and the associated pool */
static Vector vTmp;           /* Temporary Vector */
static Vector vCmp;        /* One of the arguments to Distance Measure */
static int vSize;          /* size of vectors */



/* ----------------- Pool Data Structure Handling ---------------- */

/* Note that the array of VecInfo records in a pool is indexed
   1..N rather than 0..N-1 to allow a direct correspondence with
   the indices of the distance matrix
*/

/* EXPORT->CreateVectorPool: creates a pool with 1 chunk of space */
void CreateVectorPool(VectorPool *vp)
{
   vp->poolSpace = POOLCHUNKSIZE;
   vp->poolSize = 0;
   if ((vp->vi = (VecInfo *)malloc(POOLCHUNKSIZE*sizeof(VecInfo))) == NULL)
      HError(2,"Cannot create pool");
   --vp->vi; /* to allow for indexing from 1 */
}

/* EXPORT->AddToVectorPool: add vi to pool, expanding it if necessary */
void AddToVectorPool(VecInfo vi, VectorPool *vp)
{
   VecInfo *p;
   
   if (vp->poolSize == vp->poolSpace) {
      vp->poolSpace += POOLCHUNKSIZE;
      if ((p = (VecInfo *)
         realloc(vp->vi+1,vp->poolSpace*sizeof(VecInfo))) == NULL)
         HError(2,"Cannot extend pool");
      vp->vi = p-1;
   }
   vp->vi[++vp->poolSize] = vi;
}

/* EXPORT->FreeVectorPool: free space allocated to vp */
void FreeVectorPool(VectorPool *vp)
{
   vp->poolSize = vp->poolSpace = 0;
   free(vp->vi+1);
}

/* -------------------- Cluster Data Structure Handling ----------- */

/* CreateCluster: creates a cluster with 1 chunk of space */
void CreateCluster(Cluster *cl)
{
   cl->itemSpace = CLUSCHUNKSIZE;
   cl->numItems = 0;
   if ((cl->items = (int *)malloc(CLUSCHUNKSIZE*sizeof(int))) == NULL)
      HError(2,"Cannot create cluster");
}

/* AddToCluster: add item to cluster, expanding it if necessary */
void AddToCluster(int item, Cluster *cl)
{
   int *p;
   
   if (cl->items == NULL)
      CreateCluster(cl);
   else if (cl->numItems == cl->itemSpace) {
      cl->itemSpace += CLUSCHUNKSIZE;
      if ((p = (int *)realloc(cl->items,cl->itemSpace*sizeof(int))) == NULL)
         HError(2,"Cannot extend cluster");
      cl->items = p;
   }
   cl->items[cl->numItems++] = item;
}

/* FreeCluster: free space allocated to cl */
void FreeCluster(Cluster *cl)
{
   cl->itemSpace = cl->numItems = 0;
   if (cl->items != NULL) free(cl->items);
   cl->items = NULL;
}

/* -------------------- Cluster Set Handling ------------------------ */

/* EXPORT->CreateClustSet: Create a cluster set of numCl clusters */
void CreateClustSet(ClusterSet *cset, VectorPool *pool, int numCl, 
                     int maxIter, int vecSize)
{
   int i;
   Cluster *cl;
   void KMeansCluster(int maxIter, int vecSize);
   
   if ((cset->cl = (Cluster *)malloc(numCl*sizeof(Cluster))) == NULL)
      HError(2,"Cannot create space for cluster set");
   cset->numClust = numCl;
   for (i=0;i<numCl;i++){
      cl = cset->cl+i;
      cl->itemSpace = cl->numItems = 0;
      cl->items = NULL;
      cl->vCtr = CreateVector(vecSize);
   }
   theCS = cset;
   thePool = pool;
   KMeansCluster(maxIter,vecSize);
}

/* ClearClustSet: clears existing clusters from theCS */
void ClearClustSet(void)
{
   int i;
   
   for (i=0;i<theCS->numClust;i++)
      FreeCluster(theCS->cl+i);
}

/* EXPORT->FreeClustSet: free all space allocated to cset */
void FreeClustSet(ClusterSet *cset)
{
   int i;
   
   for (i=0;i<cset->numClust;i++) {
      FreeVector(cset->cl[i].vCtr); 
      FreeCluster(cset->cl+i);
   }
   free(cset->cl);
}


/* EXPORT->DumpClustSet: print the given cluster set */
void DumpClustSet(ClusterSet *cset, VectorPool *pool)
{
   int i,j,k;
   Cluster *p;
   
   printf("Dump of Set with %d Clusters from Pool of %d items\n",
         cset->numClust,pool->poolSize);
   for (i=0;i<cset->numClust;i++) {
      p = cset->cl+i;
      printf("%d items, %d space, centre=%d\n",
               p->numItems,p->itemSpace,p->curCtr);
      for (j=0;j<p->numItems;j++) {
         k = p->items[j];
         printf("%d. %d [s=%d, i=%d]\n",
            j,k,pool->vi[k].sequence,pool->vi[k].index);
      }
   }
   fflush(stdout);
}


/* -------------------- Distance Calculation ------------------------ */

/* Distance: Euclidean distance measure between v and vector vCmp */
float Distance(Vector v)
{
   Real d,x;
   int i;

   d = 0.0;
   for (i=1;i<=vSize;i++){
      x = vCmp[i]-v[i]; 
      d += x*x;
   }
   return sqrt(d);
}

/* Furthest: returns furthest point from the one given */
int Furthest(int this,int c)
{
   Real max,d;
   int *itemp,numClItems, maxi, i;
   
   numClItems = theCS->cl[c].numItems;
   itemp = theCS->cl[c].items;
   max = 0.0;
   CopyVector(GetPoolVector(thePool->vi[this]),vCmp);
   for (i=0;i<numClItems;i++) {
      if ((d=Distance(GetPoolVector(thePool->vi[*itemp]))) > max) {
         max = d; maxi = *itemp;
      }
      ++itemp;
   }
   return maxi;
}
/* ----------------------- Basic Clustering Operations ------------------- */

/* AllocateVectors: distribute thePool vectors amongst first numCl clusters.  
         Each vector is placed in cluster with nearest centre. 
         Returns index of any empty cluster, 0 otherwise */ 
int AllocateVectors(int numCl)
{
   int i,besti,item;
   float d,min;
   Cluster *cl;
   Vector v;

   SPINBALL
   for (i=0;i<numCl;i++) {
      cl = &(theCS->cl[i]);
      v = GetPoolVector(thePool->vi[cl->curCtr]);
      CopyVector(v, cl->vCtr);
   }
   for (item=1;item<=thePool->poolSize;item++) {
      min=1.0E25;     /* a big number */
      CopyVector(GetPoolVector(thePool->vi[item]),vCmp);
      for (i=0;i<numCl;i++) {
         d = Distance(theCS->cl[i].vCtr);
         if (d < min) {
            min = d;  besti = i;
         }
      }
      AddToCluster(item,theCS->cl+besti);
   }
   for (i=0;i<numCl;i++)
      if (theCS->cl[i].numItems == 0)
         return i+1;
   return 0;
}

/* FindCentres: of the first numCl clusters and save them in newCtr member,
                also sets aveRad and pt1,pt2 at same time */
void FindCentres(int numCl)
{
   int c,i;
   int *itemp,numClItems;
   int pt1,pt2,ctr;
   Real sum,min,d,x;
   Vector v;

  for (c=0;c<numCl;c++) {
      numClItems=theCS->cl[c].numItems;
      if (numClItems == 0) /* should never happen */
         HError(2,"FindCentres: Cluster %d is empty",c);
      theCS->cl[c].aveRad = 0.0;
      theCS->cl[c].newCtr = 0;
      itemp = theCS->cl[c].items;
      if (numClItems == 1){     /* Cluster only has one element */
         theCS->cl[c].newCtr = *itemp;
         theCS->cl[c].pt1 = *itemp;
         theCS->cl[c].pt2 = *itemp;
         SPINBALL
      } else {

         /* find opposite points on perimeter */
         x = RandomValue()*(numClItems - 1);
         i = (int)x;
         pt1 = Furthest(*(itemp+i),c);
         pt2 = Furthest(pt1,c);
         theCS->cl[c].pt1 = pt1;
         theCS->cl[c].pt2 = pt2;

         if (numCl==1) return;
         
         /* set vCmp to average of pt1 and pt2 */
         CopyVector(GetPoolVector(thePool->vi[pt1]),vTmp);
         v = GetPoolVector(thePool->vi[pt2]);
         for (i=1; i<=vSize; i++)
            vCmp[i] = (vTmp[i] + v[i]) / 2.0;

         /* Centre is point nearest vCmp, find ave Rad at same time */

         sum = 0.0;
         min=1.0E5;
         for (i=0;i<numClItems;i++){
            if ((d=Distance(GetPoolVector(thePool->vi[*itemp]))) < min) {
               min=d; ctr = *itemp;
            }
            sum += d;
            ++itemp;
         }
         theCS->cl[c].newCtr = ctr;
         theCS->cl[c].aveRad = sum/numClItems;
         SPINBALL
      }
   }
}

/* TestConvergence: returns true if all newCtrs are same as curCtrs */
int TestConvergence(int NumCl)
{
   int c;

   for (c=0;c<NumCl;c++) 
      if (theCS->cl[c].newCtr != theCS->cl[c].curCtr)
         return FALSE;
   return TRUE;
}

/* BiggestCluster: returns the index of cluster with largest average radius */
int BiggestCluster(int numCl)
{
   float maxDist,d;
   int c;
   int biggest;
   
   maxDist=theCS->cl[0].aveRad; biggest = 0;
   for (c=1;c<numCl;c++) 
      if ((d=theCS->cl[c].aveRad) > maxDist) {
         maxDist=d;
         biggest=c;
      }
   return biggest;
}

/* Split: split the given cluster into two. The two new centres 
      are defined as the two cluster elements which are furthest apart */
void Split(int c, int numCl)
{
   Cluster *cl,*clnew;
   
   cl = theCS->cl+c;
   cl->curCtr = cl->pt1;
   clnew = theCS->cl + numCl;
   clnew->curCtr = cl->pt2;
   clnew->pt1=clnew->pt2=clnew->newCtr = 0;
}

/* ------------------- the KMeans Cluster Algorithm -------------------- */


void KMeansCluster(int maxIter, int vecSize)
{
   Cluster *p;
   int iter,c,clust,ce,i;
   int item,numItems;
   int converged;

   vSize = vecSize;
   numItems=thePool->poolSize;
   if (numItems < theCS->numClust) 
      HError(2,"Cannot cluster %d items into %d clusters",
         numItems,theCS->numClust);

   if (theCS->numClust<=1) {        /* 1 Cluster is easy! */      
      p = theCS->cl;
      for (item=1;item<=thePool->poolSize;item++)  
         AddToCluster(item,p);
      return;
   }
   
   vTmp = CreateVector(vecSize);       /* General case */
   vCmp = CreateVector(vecSize);
   p = theCS->cl;                      /* put all items in clust 1 */
   for (item=1;item<=thePool->poolSize;item++)  
      AddToCluster(item,p);
   FindCentres(1);                     /* characterise 1st cluster */
   Split(0,1);                         /* then split it */
   /* Now iterate centre calculation and reallocate vectors until
      no change, then split biggest and repeat until required number
      of clusters obtained
   */
   for (c=2;c<=theCS->numClust;c++) {
      for (iter=0,converged=FALSE;(iter<maxIter)&&(converged==FALSE);iter++){
         ClearClustSet();
         if (ce=AllocateVectors(c)) {
            fprintf(stderr,"C[%d] empty making CS[%d] at iter %d\n",ce,c,iter);
            fprintf(stderr,"   PoolSize = %d\n",numItems);
            for (i=0;i<c;i++) {
               p = theCS->cl + i;
               fprintf(stderr,"   C[%d]: ctr=%d newctr=%d pt1=%d pt2=%d\n",
                       i,p->curCtr,p->newCtr,p->pt1,p->pt2);
            }
            HError(2,"Clustering Failed");   
         }
         FindCentres(c);
         if ((converged=TestConvergence(c))==FALSE) 
            for (clust=0;clust<c;clust++)
               theCS->cl[clust].curCtr=theCS->cl[clust].newCtr;
         SPINBALL
      }      
      if (c < theCS->numClust) 
         Split(BiggestCluster(c),c);
   }
   FreeVector(vTmp);
   FreeVector(vCmp);
}

/* ------------------------- End of HKMeans.c --------------------------- */







         
   
   

        
                
   
