/* ----------------------------------------------------------- */
/*                   HTK  -  HMM Toolkit                       */
/* ----------------------------------------------------------- */
/*        Copyright: Steve Young      October 1989             */
/*        Last Update:                September 1990           */
/*                                                             */
/*        Cambridge University Engineering Department          */
/*        Trumpington Street, Cambridge, CB2 1PZ               */
/*                                                             */
/*  Permission is hereby granted to copy, use and modify this  */
/*  software provided that this banner notice is not removed   */
/*                                                             */
/* ----------------------------------------------------------- */
/*    File: HLEd:   edit label file(s)                         */
/* ----------------------------------------------------------- */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include "HShell.h"
#include "HSpIO.h"
#include "HLabel.h"

/*
   This program reads in a list of 'editing' commands from a
   script file and then makes an edited copy of one or more 
   label files.  Each edit command in the script file must be
   on a separate line.  The commands supported are as follows
   where the first letter on each line is the command name and
   the remaining letters denote labels.
   
   R X A B C D ...   - replace all occurrences of labels A or B 
                       or C or D ... by the label X
   C X A Y B         - replace Y in the context of A_B by X.  A
                       and/or B may be a "*" to match any context.
   F A Y B           - as for C but just counts num occurrences
                       of pattern AYB
   M X A B C ...     - merge any sequence of labels A B C... and
                       call the new segment X
   D A B C D ...     - delete any occurrences of labels A or B 
                       or C ...
   X A B C D ...     - define the context A as the set of labels
                       B,C,D,...   
   S                 - labels are sorted into time order.
   L [X]             - convert phonemes to Left-context dependent. If
                       X given then first phone becomes X-a other
                       wise it is left unchanged.
   G [X]             - convert phonemes to riGht-context dependent. If
                       X given then last phone becomes z+X other
                       wise it is left unchanged.
   T [X [Y]]         - convert phonemes to Triphones. If
                       X is given then first phone becomes X-a+b other
                       wise it is left unchanged. If Y is given then
                       last phone becomes y-z+Y otherwise if X is given
                       then it becomes y-z+X otherwise it is left 
                       unchanged.
   
   The source label files can be in any supported label format.
   The output files are always written in HTK format. Only level 0
   of multi-level files is processed.
   
*/
#define VERSION   "1.2"
#define MAXARGS  50

static char * newDir    = NULL;     /* dest label file directory */
static char * newExt    = "lab";    /* dest label file extension */
static int trace = 0;               /* trace level */

enum _EdOp{NOCMD, REPLACE, FIND, DEFCON, CHANGE, MERGE, DELETE, SORT,
           LCTXT, RCTXT, TCTXT};
typedef enum _EdOp EdOp;

typedef struct {
   EdOp op;
   short nArgs;
   LabId args[MAXARGS];
}EditCmd;

typedef struct _ScriptItem{
   EditCmd cmd;
   struct _ScriptItem *next;
}ScriptItem;

static ScriptItem *script;          /* linked list of edit commands */
static Transcription t;             /* the current transcript being edited */
static Transcription tnew=NULL;     /* Used by C, L, R and T commands */
static LabId asterix;               /* id of an asterix */
static FILE *newLabs = NULL;        /* list of newly created labels */

/* ------------------ Process Command Line ------------------------- */

void ReportUsage(void)
{
   fprintf(stderr,"\nUSAGE: HLEd[V%s] [options] edCmdFile labFiles...\n\n",
           VERSION);
   fprintf(stderr," Option                                   Default\n\n");
   fprintf(stderr," -n f    Output list of all new labs to f   off\n");
   fprintf(stderr," -G fmt  Set src label file format to fmt   HTK\n");
   fprintf(stderr," -L s    dir to store output label file(s) current\n");
   fprintf(stderr," -S f    set script file to f               none\n");
   fprintf(stderr," -T N    set trace level to N                0\n");
   fprintf(stderr," -X s    ext for output label files         lab\n");
   exit(1);
}


int GetChkedInt(int min, int max, char * swtname)
{
   int val;
   
   if (NextArg() != INTARG) 
      HError(1,"Integer Arg Required for %s option",swtname);
   val=GetIntArg();
   if (val<min || val>max) 
      HError(1,"Integer arg out of range in %s option",swtname);
   return val;
}

main(int argc, char *argv[])
{
   FileFormat ff;
   char * labFn, *scriptFn, *newFn, *s;
   void EditFile(char *labfn);
   void ReadScript(char *scriptFn);
   void ListFinds(void);

   InitShell(argc,argv);
   InitLabel();
   asterix = GetLabId("*",TRUE);
   if (NumArgs()==0)
      ReportUsage();
   while (NextArg() == SWITCHARG) {
      s = GetSwtArg();
      if (strlen(s)!=1) 
         HError(1,"Bad switch %s; must be single letter",s);
      switch(s[0]){
         case 'n':
            if (NextArg() != STRINGARG)
               HError(1,"New labels file name expected");
            newFn = GetStrArg();
            if ((newLabs = fopen(newFn,"w")) == NULL)
               HError(1,"Cannot create new label file %s",newFn);
            break;
         case 'T':
            trace = GetChkedInt(0,3,s); break;
         case 'G':
            if (NextArg() != STRINGARG)
               HError(1,"Label File format expected");
            if((ff = Str2Format(GetStrArg())) == ALIEN)
               HError(0,"Warning ALIEN Label file format set");
            SetLFormat(ff);
            break;
         case 'L':
            if (NextArg()!=STRINGARG)
               HError(1,"Label file directory expected");
            newDir = GetStrArg(); break;
         case 'S':
            SetScriptFile(); break;
         case 'X':
            if (NextArg()!=STRINGARG)
               HError(1,"Label file extension expected");
            newExt = GetStrArg(); break;
         default:
            HError(1,"Unknown switch %s",s);
      }
   }
   if (NextArg()!=STRINGARG)
      HError(1,"Script file name expected");
   scriptFn = GetStrArg();
   ReadScript(scriptFn);
   do {
      if (NextArg()!=STRINGARG)
         HError(1,"label file name expected");
      labFn = GetStrArg();
      EditFile(labFn);
   } while (NumArgs()>0);
   ListFinds();
   if (newLabs != NULL) fclose(newLabs);
}

/* ------------------- Context List Handling ------------------ */

typedef struct _Context{
   short nDefs;
   LabId cname;
   LabId cdef[MAXARGS];
   struct _Context *next;
} Context;

static Context *xlist = NULL;

/* DefineContext:  add given context to list */
void DefineContext(LabId name, LabId *def, int nDefs)
{
   Context *p;

   if ((p=(Context *)malloc(sizeof(Context))) == NULL)
      HError(2,"Cannot allocate new context");
   p->next = xlist; xlist = p;
   p->cname = name; p->nDefs = nDefs;
   memcpy(p->cdef,def,nDefs*sizeof(LabId));
   p->cdef[nDefs] = NULL;
}

/* PrintContexts: print all contexts */
void PrintContexts(void)
{
   Context *p = xlist;
   void PrintIdList(LabId *i);

   if (p == NULL){
      printf("No Contexts Defined\n");
      return;
   }
   printf("Contexts Defined:\n");
   while (p != NULL) {
      printf("   Context %s == ",p->cname->name);
      PrintIdList(p->cdef);
      printf("\n");
      p = p->next;
   }
}
   
/* GetContext: return list of ids defined by name */
LabId *GetContext(LabId name)
{
   Context *p = xlist;

   while (p != NULL) {
      if (p->cname == name)
         return p->cdef;
      p = p->next;
   }
   HError(2,"Context %s not defined",name->name);
}

/* ------------------- New Label Recording -------------------- */

/* PutLab: if given label new then output it to newLabs */
void PutLab(LabId id)
{
   if (id->aux == 0) {  /* not seen this label before */
      id->aux = 1;
      fprintf(newLabs,"%s\n",id->name);
   }
}

/* ------------------- Find Accumulators ---------------------- */

typedef struct _FindAcc{
   LabId *pattern;      /* pattern to find */
   int acc;             /* num of matches */
   struct _FindAcc *next;
} FindAcc;

static FindAcc *facc = NULL;

/* GetFindAccumulator: return a pointer to acc for given pattern */
int *GetFindAccumulator(LabId *pat)
{
   FindAcc *p = facc;

   while (p != NULL) {
      if (p->pattern == pat)
         return &(p->acc);
      p = p->next;
   }
   HError(2,"Acc %s %s %s not defined",
       pat[0]->name,pat[1]->name,pat[2]->name);
}

/* MakeFindAccumulator: create a new find acc for given pattern */
void MakeFindAccumulator(LabId *pat)
{
   FindAcc *p;

   if ((p = (FindAcc *)malloc(sizeof(FindAcc))) == NULL)
      HError(2,"Cannot create find accumulator");
   p->next = facc; facc = p;
   p->pattern = pat; p->acc = 0;
}

/* ListFinds: list accumulated totals in all find accs */
void ListFinds(void)
{
   FindAcc *p = facc;
   LabId *i;

   if (p != NULL) {
      printf("Find Totals:\n");
      while (p != NULL){
         i = p->pattern;
         printf(" %4d  %s-%s+%s\n",p->acc,
            i[0]->name,i[1]->name,i[2]->name);
         p = p->next;
      }
   }
}
/* ------------------- Load and Print Script ------------------ */

/* PrintIdList: print list of ids */
void PrintIdList(LabId *i)
{
   while (*i != NULL) {
      printf(" %s",(*i)->name);
      ++i;
   }
}

/* PrintScript: prints the currently loaded script - for tracing only */
void PrintScript()
{
   ScriptItem *i;
   int j=0;
   LabId src;
   
   i = script;
   
   printf("HLEd[V%s]: script\n",VERSION);
   while (i != NULL) {
      ++j;
      printf("%2d. ",j);
      src = *(i->cmd.args);
      switch (i->cmd.op) {
         case SORT:
            printf("Sort\n"); 
            break;
         case DELETE:
            printf("Delete  ["); PrintIdList(i->cmd.args); printf("]\n");
            break;
         case REPLACE:
            printf("Replace ["); PrintIdList(i->cmd.args+1); 
            printf("] by %s\n",src->name);
            break;
         case CHANGE:
            printf("Change  ["); PrintIdList(i->cmd.args+1); 
            printf("] by %s\n",src->name);
            break;
         case FIND:
            printf("Find  ["); PrintIdList(i->cmd.args); 
            printf("]\n");
            break;
         case MERGE:
            printf("Merge   ["); PrintIdList(i->cmd.args+1); 
            printf("] to %s\n",src->name);
            break;
         case LCTXT:
            printf("Left Context ["); PrintIdList(i->cmd.args);
            printf("]\n");
            break;
         case RCTXT:
            printf("Right Context ["); PrintIdList(i->cmd.args);
            printf("]\n");
            break;
         case TCTXT:
            printf("Triphonise ["); PrintIdList(i->cmd.args);
            printf("]\n");
            break;
      }
      i = i->next;
   }
}

static char ch;      /* current character from input file */
static FILE *f;      /* input file */

/* ReadCmd: returns next command on input f or NOCMD if none */
EdOp ReadCmd(FILE *f)
{
   char cmdch;
   
   while (isspace(ch)) ch = fgetc(f);
   cmdch = ch; ch = fgetc(f);
   switch(cmdch){
      case 'S': return SORT;
      case 'M': return MERGE;
      case 'R': return REPLACE;
      case 'C': return CHANGE;
      case 'D': return DELETE;
      case 'X': return DEFCON;
      case 'F': return FIND;
      case 'L': return LCTXT;
      case 'G': return RCTXT;
      case 'T': return TCTXT;
      case EOF: return NOCMD;
      default:  HError(2,"Unknown command '%c'",cmdch);
   }
}

/* ReadStr: Read the next string from the current line, return NULL if none */
char *ReadStr(FILE *f)
{
   char buf[255];
   int i;
   char *s;
   
   i=0;
   while (ch==' ' || ch=='\t') ch = fgetc(f);
   if (ch==EOF||ch=='\n') return NULL;
   if (ch=='\''){
      ch = fgetc(f);
      while (ch != '\'' && ch != EOF) {
         buf[i++] = ch; 
         ch = fgetc(f);
      }
      if (ch==EOF)
         HError(1,"Closing quote missing in script file");
      ch = fgetc(f);
   } else
      do {
         buf[i++] = ch; 
         ch = fgetc(f);
      }while (!isspace(ch) && ch != EOF);
   buf[i] = '\0';
   if ((s = (char *) malloc(strlen(buf)+1))==NULL)
      HError(2,"Cannot create string");
   strcpy(s,buf);
   return s;
}

/* ReadIdList: Read a list of args and store in argList, return num args */
int ReadIdList(FILE *f,LabId *argList)
{
   int count = 0;
   char *s;
   
   s = ReadStr(f);
   while (s!=NULL) {
      *argList++ = GetLabId(s,TRUE); ++count;
      s = ReadStr(f);
   }
   *argList = NULL;
   return count;   
}

/* ReadScript: read a file of script commands and store them in script */
void ReadScript(char *scriptFn)
{
   ScriptItem *i, *tail=NULL;
   EdOp op;
   LabId args[MAXARGS];
   LabId id;
   int nArgs;

   if ((f=fopen(scriptFn,"r"))==NULL)
      HError(2,"Cannot Open Script %s",scriptFn);
   ch = ' ';
   script = NULL;
   while (TRUE) {
      if ((op = ReadCmd(f)) == NOCMD) break;
      if (op == DEFCON){
         id = GetLabId(ReadStr(f),TRUE);
         nArgs = ReadIdList(f,args);
         DefineContext(id,args,nArgs);
         continue;
      }
      if ((i = (ScriptItem *)malloc(sizeof(ScriptItem))) == NULL)
         HError(2,"Cannot create script item");
      i->cmd.op = op;
      switch (op) {
         case DELETE:
         case REPLACE:
         case MERGE:
            i->cmd.nArgs = ReadIdList(f,i->cmd.args);
            break;
         case CHANGE:
            i->cmd.nArgs = ReadIdList(f,i->cmd.args);
            if (i->cmd.nArgs != 4)
               HError(2,"C must have 4 arguments");
            break;   
         case FIND:
            i->cmd.nArgs = ReadIdList(f,i->cmd.args);
            if (i->cmd.nArgs != 3)
               HError(2,"F must have 3 args");
            MakeFindAccumulator(i->cmd.args);
            break;   
         case LCTXT:
         case RCTXT:
            i->cmd.nArgs = ReadIdList(f,i->cmd.args);
            if (i->cmd.nArgs > 1)
               HError(2,"L/G command has too many args");
            break;
         case TCTXT:
            i->cmd.nArgs = ReadIdList(f,i->cmd.args);
            if (i->cmd.nArgs > 2)
               HError(2,"T command has too many args");
            break;         
         case SORT:
            break;
      }
      i->next = NULL;
      if (tail==NULL) 
         script = tail = i;
      else {
         tail->next = i;
         tail = i;
      }
   }
   fclose(f);
   if (trace>1){
      PrintScript();
      PrintContexts();
   }
}

/* ----------------------- Edit Operations -------------------- */

/* IsMember: return true if id is in idlist */
int IsMember(LabId id, LabId *idlist)
{
   while (*idlist != NULL) {
      if (id == *idlist) return TRUE;
      ++idlist;
   }
   return FALSE;
}

/* SeqMatch: returns true if labels i,i+1,i+2,... match idList */
int SeqMatch(int numIds, LabId *idList, int i)
{
   int ok = TRUE;
   int j;
   
   for (j=i;j<i+numIds;j++) 
      if (*idList++ != t->labels[j].labid) {
         ok = FALSE;
         break;
      }
   return ok;
}

/* MarkDelete: marks the i'th label in t for deletion */
void MarkDelete(int i)
{
   t->labels[i].start = -1;
}

/* DoDeletes: delete all labels in t marked for deletion */
void DoDeletes(void)
{
   int count,i;
   
   count=0;
   for (i=0;i<t->nLabels;i++)
      if (t->labels[i].start>=0) {
         t->labels[count] = t->labels[i];
         ++count;
      }
   t->nLabels = count;
}

/* DeleteOp: scan the current transcription and delete any items 
   in args,  returns number of deletions performed */
int DeleteOp(LabId *args)
{
   int i,count=0,N;
   
   N = t->nLabels;
   for (i=0;i<N;i++)
      if(IsMember(t->labels[i].labid,args)) {
         ++count;
         MarkDelete(i);
      }
   DoDeletes();
   return count;
}

/* ReplaceOp: scan the current transcription and replace any 
   items in args[1..] with arg[0],  returns the number of 
   replaces performed */
int ReplaceOp(LabId *args)
{
   int i,count=0,N;
   LabId item;
   
   N = t->nLabels;
   item = *args++;
   for (i=0;i<N;i++)
      if(IsMember(t->labels[i].labid,args)) {
         ++count;
         t->labels[i].labid = item;
      }
   if (newLabs != NULL && count>0) PutLab(item);
   return count;
}

/* InContext: Match Context against i'th label in t */
int InContext(int i, LabId *lc, LabId *rc, LabId item)
{
   int N = t->nLabels;
   
   if (t->labels[i].labid != item) return FALSE;
   if (i == 0 && lc != NULL) return FALSE;
   if (lc!=NULL &&
       !IsMember(t->labels[i-1].labid,lc)) return FALSE;
   if (i == N-1 && rc != NULL) return FALSE;
   if (rc!=NULL &&
       !IsMember(t->labels[i+1].labid,rc)) return FALSE;
   return TRUE;
}

/* ChangeOp: scan the current transcription and replace any 
   items which match the context given by args[1..3] with arg[0], 
   returns number of changes performed */
int ChangeOp(LabId *args)
{
   int i,count=0,N;
   LabId *lc,*rc,newId;

   N = t->nLabels; 
   if (tnew==NULL) /* if first change clone transcription */
      tnew = LClone(t);
   newId = *args++;
   lc = (*args==asterix)?NULL:GetContext(*args);
   rc = (*(args+2)==asterix)?NULL:GetContext(*(args+2));
   
   for (i=0;i<N;i++)
      if(InContext(i,lc,rc,*(args+1))) {
         ++count;
         tnew->labels[i].labid = newId;
      }
   if (newLabs != NULL && count>0) PutLab(newId);
   return count;
}

/* FindOp: scan the current transcription and count any 
   items which match the context given by args[0..2] */ 
void FindOp( LabId *args)
{
   int i,N;
   int *acc;
   LabId *lc,*rc;

   N = t->nLabels;
   acc = GetFindAccumulator(args);
   lc = (*args==asterix)?NULL:GetContext(*args);
   rc = (*(args+2)==asterix)?NULL:GetContext(*(args+2));
   
   for (i=0;i<N;i++)
      if(InContext(i,lc,rc,*(args+1)))
         ++(*acc);
}

/* MergeOp: scan the current transcription and merge any consecutive
   sequence of items matching args[1..3] with arg[0], returns number of
   merges performed */
int MergeOp(int nArgs, LabId *args)
{
   int i,j,count=0,N;
   
   N = t->nLabels; nArgs--;
   for (i=0;i<=N-nArgs;i++)
      if(SeqMatch(nArgs,args+1,i)) {
         ++count;
         t->labels[i].labid = *args;
         t->labels[i].end = t->labels[i+nArgs-1].end;
         for (j=i+1;j<i+nArgs;j++) MarkDelete(j);
         i += nArgs-1;
      }
   DoDeletes();
   return count;
}

/* SortOp: sort the current transcription into ascending order of start time */

void SortOp(void)
{
   int sti,i,j,N;
   long st,thist;
   Label tmp;
   
   N = t->nLabels;
   if (trace==3)
      printf("Sorting %d labels",N);
   for (i=0;i<N-1;i++) {
      sti=i; st = t->labels[i].start;
      for(j=i+1;j<N;j++){
         thist = t->labels[j].start;
         if (thist<st) {
            st = thist; sti = j;
         }
      }
      if (sti!=i){
         tmp = t->labels[i]; 
         t->labels[i] = t->labels[sti];
         t->labels[sti] = tmp;
      }     
   }
   if (trace==3){
      printf(" - Done\n",N); fflush(stdout);
   }
}

/* MakeTriId:  concatenate args separated by - and +'s and return its id */
LabId MakeTriId(LabId l, LabId c, LabId r)
{
   char buf[100];
   LabId item;
   
   if ( l!= NULL){
      strcpy(buf,l->name); strcat(buf,"-"); 
      strcat(buf,c->name);
   } else 
      strcpy(buf,c->name);
   if (r != NULL){
      strcat(buf,"+"); strcat(buf,r->name);
   }
   item = GetLabId(buf,TRUE);
   if (newLabs != NULL) PutLab(item);
   return item;
}

/* TriPhonise: convert phone labels in transcription t to either
      left, right or triphone (ie both) contexts */
void TriPhonise(int left, int right, LabId stId, LabId enId)
{
   int i,N;
   LabId leftId, centreId, rightId;
   
   N = t->nLabels; 
   if (N<2)
      HError(2,"Transcription too short to triphonise");
   tnew = LClone(t);
   leftId =left?stId:NULL;    /* First Label */
   centreId = t->labels[0].labid;
   rightId = (right && stId != NULL)?t->labels[1].labid:NULL;
   tnew->labels[0].labid = MakeTriId(leftId,centreId,rightId);
   for (i=1;i<N-1;i++) {         /* Middle Labels */
      leftId =left?centreId:NULL;
      centreId = t->labels[i].labid;
      rightId = right?t->labels[i+1].labid:NULL;
      tnew->labels[i].labid = MakeTriId(leftId,centreId,rightId);
   }
   leftId = (left && enId != NULL)?centreId:NULL;     /* Last Label */
   centreId = t->labels[N-1].labid;
   rightId = right?enId:NULL;
   tnew->labels[N-1].labid = MakeTriId(leftId,centreId,rightId);
   LClose(t); t = tnew; tnew = NULL;
}

/* DoLCtxt: Left context command */
void DoLCtxt(int nArgs,LabId *args)
{
   LabId x;
   
   x = nArgs==1?*args:NULL;
   TriPhonise(TRUE,FALSE,x,x);
}

/* DoRCtxt: Right context command */
void DoRCtxt(int nArgs,LabId *args)
{
   LabId x;
   
   x = nArgs==1?*args:NULL;
   TriPhonise(FALSE,TRUE,x,x);
}

/* DoTCtxt: Left and Right context command */
void DoTCtxt(int nArgs,LabId *args)
{
   LabId l,r;
   
   if (nArgs == 1)
      l = r = *args;
   else if (nArgs == 2){
      l = *args; r = *(args+1);
   } else
      l = r = NULL;
   TriPhonise(TRUE,TRUE,l,r);
}

   
   
/* --------------------- Top Level Control -------------------- */

#ifdef MPW

/* MPW ONLY -- Sets output file creator to HTK; file kind to TEXT */

typedef struct  {
    unsigned long fdType;          /*the type of the file*/
    unsigned long fdCreator;       /*file's creator*/
    unsigned short fdFlags;        /*flags ex. hasbundle,invis,locked, etc. */
    long fdLocation;               /*file's location in folder*/
    short fdFldr;                  /*folder containing file*/
}FInfo;

void SetLabFileInfo(char *fn)
{
   FInfo inf;

   if (getfinfo(fn,0,&inf))
      HError(2,"Cannot Get File Info for %s",fn);
   inf.fdType='T'<<24|'E'<<16|'X'<<8|'T';
   inf.fdCreator='H'<<24|'T'<<16|'K'<<8|' ';
   inf.fdFlags |= 6;
   if(setfinfo(fn,0,&inf))
      HError(2,"Cannot Set File Info for %s",fn);
}
#endif

/* Load label file and apply each script command to it */ 
void EditFile(char *labfn)
{
   ScriptItem *i;
   char outfn[255];
   int m,d,r,c;
   
   t = LOpen(labfn);
   if (trace>0) {
      printf("Editing file: %s\n",labfn); fflush(stdout);
   }
   i = script;
   c=m=d=r=0;
   while (i != NULL) {
      SPINBALL
      if (tnew != NULL && i->cmd.op != CHANGE) { 
         LClose(t); t = tnew; tnew = NULL;
      }
      switch (i->cmd.op) {
         case SORT:
            SortOp(); break;
         case DELETE:
            d += DeleteOp(i->cmd.args); break;
         case FIND:
            FindOp(i->cmd.args); break;
         case CHANGE:
            c += ChangeOp(i->cmd.args); break;
         case REPLACE:
            r += ReplaceOp(i->cmd.args); break;
         case MERGE:
            m += MergeOp(i->cmd.nArgs,i->cmd.args); break;
         case LCTXT:
            DoLCtxt(i->cmd.nArgs,i->cmd.args); break;
         case RCTXT:
            DoRCtxt(i->cmd.nArgs,i->cmd.args); break;
         case TCTXT:
            DoTCtxt(i->cmd.nArgs,i->cmd.args); break;
      }
      i = i->next;
   }
   MakeFN(labfn,newDir,newExt,outfn);
   if (tnew != NULL) { 
      LClose(t); t = tnew; tnew = NULL;
   }
   LSave(outfn,t);
   LClose(t);
#ifdef MPW
    SetLabFileInfo(outfn);
#endif
   if (trace>1) {
      printf("   %d deletes; %d replaces; %d changes; %d merges\n",d,r,c,m); 
      fflush(stdout);
   }
}


/* ------------------------------------------------------------ */
/*                         END:  HLEd.c                         */
/* ------------------------------------------------------------ */


