/* ----------------------------------------------------------- */
/*                   HTK  -  HMM Toolkit                       */
/* ----------------------------------------------------------- */
/*        Copyright: Steve Young      October 1989             */
/*        Cambridge University Engineering Department          */
/*        Trumpington Street, Cambridge, CB2 1PZ               */
/*                                                             */
/*  Permission is hereby granted to copy, use and modify this  */
/*  software provided that this banner notice is not removed   */
/*                                                             */
/* ----------------------------------------------------------- */
/*    File: HLab2Net.c  Convert Label File to Recognition Net  */
/* ----------------------------------------------------------- */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include "HShell.h"
#include "HSpIO.h"
#include "HLabel.h"

/*
   This simple program reads converts label files to the
   equivalent recognition network files to allow HVite to
   be used to automatically segment speech with known
   non-time aligned transcription
*/

#define VERSION   "1.2"

static char * newDir    = NULL;     /* dest recnet file directory */
static char * newExt    = "net";    /* dest recnet file extension */
static int trace = 0;               /* trace level */
static Transcription t;       /* the current transcription being converted */
static int N;                 /* num labels in current transcription */

/* ------------------ Process Command Line ------------------------- */

void ReportUsage(void)
{
   fprintf(stderr,"\nUSAGE: HLab2Net[V%s] [options] labFiles...\n\n",
           VERSION);
   fprintf(stderr," Option                                   Default\n\n");
   fprintf(stderr," -d dir  dir to store output net file(s)  current\n");
   fprintf(stderr," -x s    ext for output net file(s)         net\n");
   fprintf(stderr," -G fmt  Set src label file format to fmt   HTK\n");
   fprintf(stderr," -S f    set script file to f               none\n");
   fprintf(stderr," -T N    set trace level to N                0\n");
   exit(1);
}


int GetChkedInt(int min, int max, char * swtname)
{
   int val;
   
   if (NextArg() != INTARG) 
      HError(1,"Integer Arg Required for %s option",swtname);
   val=GetIntArg();
   if (val<min || val>max) 
      HError(1,"Integer arg out of range in %s option",swtname);
   return val;
}

main(int argc, char *argv[])
{
   FileFormat ff;
   char *labFn, *s;
   void ConvFile(char *labfn);

   InitShell(argc,argv);
   InitMath(FALSE); InitLabel();
   if (NumArgs()==0)
      ReportUsage();
   while (NextArg() == SWITCHARG) {
      s = GetSwtArg();
      if (strlen(s)!=1) 
         HError(1,"Bad switch %s; must be single letter",s);
      switch(s[0]){
         case 'd':
            if (NextArg()!=STRINGARG)
               HError(1,"Output file directory expected");
            newDir = GetStrArg(); break;
         case 'x':
            if (NextArg()!=STRINGARG)
               HError(1,"Rec Net file extension expected");
            newExt = GetStrArg(); break;
         case 'T':
            trace = GetChkedInt(0,3,s); break;
         case 'G':
            if (NextArg() != STRINGARG)
               HError(1,"Label File format expected");
            if((ff = Str2Format(GetStrArg())) == ALIEN)
               HError(0,"Warning ALIEN Label file format set");
            SetLFormat(ff);
            break;
         case 'S':
            SetScriptFile(); break;
         default:
            HError(1,"Unknown switch %s",s);
      }
   }
   do {
      if (NextArg()!=STRINGARG)
         HError(1,"label file name expected");
      labFn = GetStrArg();
      ConvFile(labFn);
   } while (NumArgs()>0);
}

#ifdef MPW

/* MPW ONLY -- Sets output file creator to HTK; file kind to TEXT */

typedef struct  {
    unsigned long fdType;          /*the type of the file*/
    unsigned long fdCreator;       /*file's creator*/
    unsigned short fdFlags;        /*flags ex. hasbundle,invis,locked, etc. */
    long fdLocation;               /*file's location in folder*/
    short fdFldr;                  /*folder containing file*/
}FInfo;

void SetLabFileInfo(char *fn)
{
   FInfo inf;

   if (getfinfo(fn,0,&inf))
      HError(2,"Cannot Get File Info for %s",fn);
   inf.fdType='T'<<24|'E'<<16|'X'<<8|'T';
   inf.fdCreator='H'<<24|'T'<<16|'K'<<8|' ';
   inf.fdFlags |= 6;
   if(setfinfo(fn,0,&inf))
      HError(2,"Cannot Set File Info for %s",fn);
}
#endif

void ConvFile(char *labfn)
{
   FILE *f;
   char outfn[255];
   int j;
   
   t = LOpen(labfn);
   N = t->nLabels;
   if (trace>0) {
      printf("Converting file: %s\n",labfn); fflush(stdout);
   }
   MakeFN(labfn,newDir,newExt,outfn);
   if ((f = fopen(outfn,"w")) == NULL)
      HError(2,"Cannot create rec network file %s",outfn);
   fprintf(f,"(\n");
   for (j=0;j<t->nLabels;j++)
      fprintf(f,"  %s\n", t->labels[j].labid->name);
   fprintf(f,")\n");
   fclose(f);
   LClose(t);
#ifdef MPW
    SetLabFileInfo(outfn);
#endif
}


/* ------------------------------------------------------------ */
/*                         END:  HLab2Net.c                     */
/* ------------------------------------------------------------ */

