/* ----------------------------------------------------------- */
/*                   HTK  -  HMM Toolkit                       */
/* ----------------------------------------------------------- */
/*        Copyright: Steve Young      October 1989             */
/*        Last Update:                September 1990           */
/*                                                             */
/*        Cambridge University Engineering Department          */
/*        Trumpington Street, Cambridge, CB2 1PZ               */
/*                                                             */
/*  Permission is hereby granted to copy, use and modify this  */
/*  software provided that this banner notice is not removed   */
/*                                                             */
/* ----------------------------------------------------------- */
/*      File: HList.c: List the contents of a Speech File      */
/* ----------------------------------------------------------- */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "HShell.h"
#include "HMath.h"
#include "HSpIO.h"

#define VERSION "1.2"

static int   hdOnly  = 0;        /* print header info only */
static int   lWidth  = 80;       /* line width */
static long   st,en;             /* range of samples to list */

/* ---------------- Process Command Line ------------------------- */

void ReportUsage(void)
{
   fprintf(stderr,"\nUSAGE: HList[V%s] [options] file\n\n",VERSION);
   fprintf(stderr," Option                                   Default\n\n");
   fprintf(stderr," -h      Print header info only              Off\n");
   fprintf(stderr," -w N    Set line width to N                  80\n");
   fprintf(stderr," -s N    Start at sample N                    0\n");
   fprintf(stderr," -e N    End at sample N                      0\n"); 
   fprintf(stderr," -F fmt  Set src file format to fmt          HTK\n");
   exit(1);
}
   
int GetChkedInt(int min, int max, char * swtname)
{
   int val;
   
   if (NextArg() != INTARG) 
      HError(1,"Integer Arg Required for %s option",swtname);
   val=GetIntArg();
   if (val<min || val>max) 
      HError(1,"Integer arg out of range in %s option",swtname);
   return val;
}

main(int argc, char *argv[])
{
   char * s, *srcfn;
   void ListFile(char *src);
   
   FileFormat ff = 255;    /* ie undefined */
   
   InitShell(argc,argv);
   InitMath(FALSE);
   if (NumArgs()==0)
      ReportUsage();
   st = 0; en = 1000000;
   while (NextArg() == SWITCHARG) {
      s = GetSwtArg();
      if (strlen(s)!=1) 
         HError(1,"Bad switch %s; must be single letter",s);
      switch(s[0]){
         case 'h':
            hdOnly = 1; break;
         case 'w':
            lWidth = GetChkedInt(40,256,s); break;
         case 's':
            st = GetChkedInt(0,en,s); break;
         case 'e':
            en = GetChkedInt(st,1000000,s); break;
         case 'F':
            if (NextArg() != STRINGARG)
               HError(1,"File format expected");
            if((ff = Str2Format(GetStrArg())) == ALIEN)
               HError(0,"Warning ALIEN src file format set");
            SetFormat(ff);
            break;
         default:
            HError(1,"Unknown switch %s",s);
      }
   }
   if (NextArg()!=STRINGARG)
      HError(1,"List file name expected");
   srcfn = GetStrArg();
   if (NumArgs() != 0)
      HError(0,"Extra arguments left unprocessed!");
   ListFile(srcfn);
}

/* ----------------------- Printing Routines --------------------- */

void PrintVector(int n, Vector v, int st, int en, SampKind kind)
{
   int i,j;
   int itemsPerLine;
   
   itemsPerLine = lWidth / 14;
   if (n>=0) printf("Frame %d:\n",n);
   j=0;
   for (i=st;i<=en;i++) {
      printf("%4d. %8.4f",i,v[i]);
      if (++j == itemsPerLine){
         printf("\n"); j=0;
      }
   }
   if (HasEnergy(kind)){
      printf("   E. %8.4f",v[i++]);
      if (++j == itemsPerLine){
         printf("\n"); j=0;
      }
   }
   if (HasPitch(kind)){
      printf("   P. %8.4f",v[i++]);
      if (++j == itemsPerLine){
         printf("\n"); j=0;
      }
   }
   if (j>0) printf("\n");
}

void PrintVectorData(DataFile *f)
{
   long i;
   Vector v;
   int vLen,nc;
   
   nc = SampNComp(f);
   v=CreateVector(nc);
   vLen = SampDimen(f);
   for (i=st;i<=en;i++) {
      GetSample(f,i,v+1);
      PrintVector(i,v,1,vLen,f->sampKind);
      if (BaseSampKind(f->sampKind) == LPDELCEP)
         PrintVector(-1,v,nc/2+1,nc/2+vLen,f->sampKind);
   }
}

void PrintScalarData(DataFile *f)
{
   long i;
   int j,itemsPerLine;
   short sample;
   
   itemsPerLine = (lWidth-7) / 7;
   j=itemsPerLine;
   for (i=st;i<=en;i++) {
      if (j==itemsPerLine) {
         printf("\n%5d: ",i);
         j=0;
      }
      GetSample(f,i,&sample);
      printf("%7d",sample); j++;
   }
   printf("\n");
}

void PrintData(DataFile *f)
{
   if (st>f->nSamples) st=f->nSamples-1;
   if (en<st) en=st; else
   if (en>=f->nSamples) en=f->nSamples-1;
   if (f->sampKind == WAVEFORM)
      PrintScalarData(f);
   else
      PrintVectorData(f);
}

void PrintHeading(DataFile *f, char * name)
{
   char buf[80];
   
   printf("File: %s [%s]\n", name, Format2Str(f->format,buf));
   printf("  Num Samples: %8d          Sample Period: %.1f usecs\n",
      f->nSamples, f->sampPeriod/10.0);
   printf("  Sample Size: %3d bytes         Sample Kind:   %s\n",
      f->sampSize,SampKind2Str(f->sampKind,buf));
}
   
/* -------------------- ListFile Routine  ------------------- */

void ListFile(char *src)
{
   DataFile f;
   
   SpOpen(src, &f);
   PrintHeading(&f,src);
   if (!hdOnly)
      PrintData(&f);
   SpClose(&f);
}

/* ----------------------------------------------------------- */
/*                      END:  HList.c                         */
/* ----------------------------------------------------------- */

