/* ----------------------------------------------------------- */
/*                   HTK  -  HMM Toolkit                       */
/* ----------------------------------------------------------- */
/*        Copyright: Steve Young      October 1989             */
/*        Cambridge University Engineering Department          */
/*        Trumpington Street, Cambridge, CB2 1PZ               */
/*                                                             */
/*  Permission is hereby granted to copy, use and modify this  */
/*  software provided that this banner notice is not removed   */
/*                                                             */
/* ----------------------------------------------------------- */
/*    File: HMixUp.c  Increase number of mixtures in a HMM     */
/* ----------------------------------------------------------- */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <math.h>

#include "HShell.h"
#include "HMath.h"
#include "HSpIO.h"
#include "HModel.h"

/*
   This simple program reads in a HMM definition and creates a
   new one with one more mixture.  This is done by splitting 
   the largest existing mixture into equal weights and then
   perturbing the means by a specified fraction (pertDepth) of one
   standard deviation.
*/

#define VERSION   "1.2"

static char * newDir    = NULL;     /* dest hmm file directory */
static char * newExt    = NULL;     /* dest hmm file extension */
static float pertDepth  = 0.2;      /* perturbation depth in std devs */
static int trace = 0;               /* trace level */
static HMMDef hmm;                  /* the hmm being converted */

static int N,M,V;

/* ------------------ Process Command Line ------------------------- */

void ReportUsage(void)
{
   fprintf(stderr,"\nUSAGE: HMixUp[V%s] [options] hmmFiles...\n\n",
           VERSION);
   fprintf(stderr," Option                                   Default\n\n");
   fprintf(stderr," -d dir  dir to store output hmm file(s)  current\n");
   fprintf(stderr," -p f    perturbation depth in std devs     0.2\n");
   fprintf(stderr," -x s    ext for output hmm file(s)         none\n");
   fprintf(stderr," -S f    set script file to f               none\n");
   fprintf(stderr," -T N    set trace level to N                0\n");
   exit(1);
}


int GetChkedInt(int min, int max, char * swtname)
{
   int val;
   
   if (NextArg() != INTARG) 
      HError(1,"Integer Arg Required for %s option",swtname);
   val=GetIntArg();
   if (val<min || val>max) 
      HError(1,"Integer arg out of range in %s option",swtname);
   return val;
}

float GetChkedFlt(float min, float max, char * swtname)
{
   float val;
   
   if (NextArg() != INTARG && NextArg() != FLOATARG) 
      HError(1,"Float Arg Required for %s option",swtname);
   val=GetFltArg();
   if (val<min || val>max) 
      HError(1,"Float arg out of range in %s option",swtname);
   return val;
}

main(int argc, char *argv[])
{
   char *hmmFn, *s;
   void MixFile(char *labfn);

   InitShell(argc,argv);
   InitMath(FALSE);
   if (NumArgs()==0)
      ReportUsage();
   while (NextArg() == SWITCHARG) {
      s = GetSwtArg();
      if (strlen(s)!=1) 
         HError(1,"Bad switch %s; must be single letter",s);
      switch(s[0]){
         case 'd':
            if (NextArg()!=STRINGARG)
               HError(1,"Output hmm file directory expected");
            newDir = GetStrArg(); break;
         case 'p':
            pertDepth = GetChkedFlt(0.0,5.0,s); break;
         case 'x':
            if (NextArg()!=STRINGARG)
               HError(1,"Output hmm file extension expected");
            newExt = GetStrArg(); break;
         case 'S':
            SetScriptFile(); break;
         case 'T':
            trace = GetChkedInt(0,3,s); break;
         default:
            HError(1,"Unknown switch %s",s);
      }
   }
   do {
      if (NextArg()!=STRINGARG)
         HError(1,"label file name expected");
      hmmFn = GetStrArg();
      MixFile(hmmFn);
   } while (NumArgs()>0);
}

/* -------------------- Mixture Splitting Routine ----------- */

/* IncMix: returns a new outprob record with an extra mixture */
OutProb *IncMix(OutProb *outp)
{
   OutProb *p,*new,*big;
   int k,m,mm;
   float w,x;

   if ((p=(OutProb*)realloc(outp+1,sizeof(OutProb)*M))==NULL)
      HError(2,"Cannot extend outp array");
   --p;
   w = 0.0; mm = 0;   /* find biggest mixture */
   for (m=1;m<M;m++)
      if (p[m].weight>w){
         w = p[m].weight; mm=m;
      }
   if (mm<1)
      HError(2,"Cant find max mixture");
   big = p + mm;
   new = p + M;
   new->mean = CreateVector(V);     /* create new mix */
   new->covariance.var = CreateVector(V);
   CopyVector(big->mean,new->mean); /* copy big -> new */
   CopyVector(big->covariance.var,new->covariance.var);
   new->weight = w/2.0;
   big->weight = w/2.0;
   for (k=1;k<=V;k++){              /* perturb them */
      x = sqrt(big->covariance.var[k])*pertDepth;
      new->mean[k] += x;
      big->mean[k] -= x;
   }
   return p;
}

/* ----------------------- Top Level Control ----------------- */

/* MixFile: increase number of mixtures in given file */
void MixFile(char *hmmFn)
{
   char outfn[255];
   int i;
   StateInfo *s;
   
   HOpen2(hmmFn, &hmm);
   if (hmm.ckind!=DIAGC)
      HError(2,"This program only supports DIAGC covariances");
   N = hmm.numStates;
   V = hmm.vecSize;
   M = hmm.numMixtures+1;
   if (trace>0) {
      printf("Up Mixing file %s from %d -> %d mixtures\n",
         hmmFn,hmm.numMixtures,M);
      fflush(stdout);
   }
   ++hmm.numMixtures; 
   for (i=2;i<N;i++){
      s = hmm.stateinfo+i;
      s->outp = IncMix(s->outp);
   }
   FixGConsts(&hmm);   
   MakeFN(hmmFn,newDir,newExt,outfn);
   HSave2(outfn,&hmm);
   if (trace>0) {
      printf("Up Mixed file saved to %s\n",outfn); fflush(stdout);
   }
}


/* ------------------------------------------------------------ */
/*                         END:  HMixUp.c                       */
/* ------------------------------------------------------------ */

