/* ----------------------------------------------------------- */
/*                   HTK  -  HMM Toolkit                       */
/* ----------------------------------------------------------- */
/*         Copyright: Steve Young      October 1989            */
/*         CoAuthor:  Val Beattie                              */
/*         Last Update:                September 1990          */
/*                                                             */
/*                                                             */
/*         Cambridge University Engineering Department         */
/*         Trumpington Street, Cambridge, CB2 1PZ              */
/*                                                             */
/* Permission is hereby granted to copy, use and modify this   */
/* software provided that this banner notice is not removed    */
/*                                                             */
/* ----------------------------------------------------------- */
/*         File: HRest.c: HMM initialisation program           */
/* ----------------------------------------------------------- */

/*
   This program is used to estimate the parameters of
   a hidden Markov model using Baum-Welch reestimation. Training
   data is specified exactly as for HInit ie either a list of
   whole word tokens or a list of labelled utterances from
   which all examples of one named segment are to be extracted.
*/


#include <stdio.h>      /* Standard C Libraries */
#include <stdlib.h>
#include <math.h>

#include "HShell.h"     /* HMM ToolKit Modules */
#include "HMath.h"
#include "HSpIO.h"
#include "HLabel.h"
#include "HDbase.h"  
#include "HModel.h"

#define VERSION   "1.2"

/* define DOMEANC       define this to enable mean correction */

/* Global Settings */
static char * segLab = NULL;     /* segment label if any */
static LabId  segId  = NULL;     /* and its index */
static char * labDir = NULL;     /* label file directory */
static char * labExt = "lab";    /* label file extension */
static int  maxIter  = 20;       /* max iterations in parameter estimation */
static float epsilon = 0.0001;   /* convergence criterion */
static int minSeq    = 3;        /* min sequences to train a model */
static float minVar  = 0.0;      /* minimum variance */
static char *hmmfn;              /* HMM definition file name */
static char *outfn=NULL;         /* output definition file name */
enum _UPDSet{UPMEANS=1,UPVARS=2,UPTRANS=4,UPMIXES=8};
typedef enum _UPDSet UPDSet;
static UPDSet uFlags = UPMEANS|UPVARS|UPTRANS|UPMIXES;     /* update flags */
static int  trace    = 0;        /* Trace level */

/* For Convenience */
static int N;              /* numStates */
static int M;              /* numMixtures */
static int V;              /* vecSize */
static int R;              /* num training tokens */
static int maxT,T;         /* max and current sequence lengths */
static CovKind ckind;      /* Covariance kind {DIAGC,FULLC} */

/* Global Data Structures */
static Vector seqProb;     /* array[1..NumSequences] OF prob; */
static HMMDef theHMM;      /* The HMM to be initialised (or updated) */
static Matrix alpha;       /* array[1..N-1][1..maxT] of forward prob */
static Matrix beta;        /* array[1..N-1][1..maxT] of backward prob */
static Matrix outprob;     /* array[1..N-1][1..maxT] of output prob */
static Matrix *mixoutp;    /* array[1..N-1][1..M][1..maxT] of mixture prob */
static Matrix tran;        /* array[1..N-1][1..N-1] of transition count */
static Vector occ1;        /* array[1..N] of occupation count (for t=1) */
static Vector occ;         /* array[1..N] of occupation count (for 1<t<T) */
static Vector occT;        /* array[1..N] of occupation count (for t=T) */
static Vector **mu;        /* array[2..N-1][1..M] of MeanVector; */
static Covariance **va;    /* array[2..N-1][1..M] of Covariance; */
static Matrix c;           /* array[1..N][1..M] of MixWeight */

/* ------------------ Process Command Line ------------------------- */

void ReportUsage(void)
{
   fprintf(stderr,"\nUSAGE: HRest[V%s] [options] hmmFile trainFiles...\n\n",
            VERSION);
   fprintf(stderr," Option                                   Default\n\n");
   fprintf(stderr," -e f    Set convergence factor epsilon    1.0E-4\n");
   fprintf(stderr," -i N    Set max iterations to N            20\n");
   fprintf(stderr," -l s    Set segment label to s             None\n");
   fprintf(stderr," -m N    Set min sequences needed             3\n");
   fprintf(stderr," -o fn   Store new hmm def in fn            srcfn\n");
   fprintf(stderr," -u tmvw Update t)rans m)eans v)ars w)ghts  tmvw\n");
   fprintf(stderr," -v f    Set minimum variance to f          0.0\n");
   fprintf(stderr," -F fmt  Set data file format to fmt        HTK\n");
   fprintf(stderr," -G fmt  Set label file format to fmt       HTK\n");
   fprintf(stderr," -L s    dir to find label file(s)        current\n");
   fprintf(stderr," -S f    set script file to f               none\n");
   fprintf(stderr," -T N    set trace level to N                0\n");
   fprintf(stderr," -X s    extension for label files          lab\n");
   exit(1);
}

int GetChkedInt(int min, int max, char * swtname)
{
   int val;
   
   if (NextArg() != INTARG) 
      HError(1,"Integer Arg Required for %s option",swtname);
   val=GetIntArg();
   if (val<min || val>max) 
      HError(1,"Integer arg out of range in %s option",swtname);
   return val;
}

float GetChkedFlt(float min, float max, char * swtname)
{
   float val;
   
   if (NextArg() != INTARG && NextArg() != FLOATARG) 
      HError(1,"Float Arg Required for %s option",swtname);
   val=GetFltArg();
   if (val<min || val>max) 
      HError(1,"Float arg out of range in %s option",swtname);
   return val;
}

void SetuFlags()
{
   char *s;
   
   s=GetStrArg();
   uFlags=0;        
   while (*s != NULL)
      switch (*s++) {
         case 't': uFlags+=UPTRANS;break;
         case 'm': uFlags+=UPMEANS;break;
         case 'v': uFlags+=UPVARS;break;
         case 'w': uFlags+=UPMIXES;break;
         default: HError(2,"Unknown update flag %c",*s);
                  break;
      }
}

main(int argc, char *argv[])
{
   char *datafn, *s;
   FileFormat ff;
   void Initialise1(void);
   void Initialise2(void);
   void LoadFile(char *fn);
   void ReEstimateModel(void);
   
   InitShell(argc,argv);
   InitMath(TRUE);
   if (NumArgs()==0)
      ReportUsage();
   while (NextArg() == SWITCHARG) {
      s = GetSwtArg();
      if (strlen(s)!=1) 
         HError(1,"Bad switch %s; must be single letter",s);
      switch(s[0]){
         case 'm':
	    minSeq = GetChkedInt(1,1000,s); break;
         case 'e':
            epsilon = GetChkedFlt(0.0,1.0,s); break;
         case 'i':
            maxIter = GetChkedInt(1,100,s); break;
         case 'l':
            if (NextArg() != STRINGARG)
               HError(1,"Segment label expected");
            segLab = GetStrArg();
            break;
         case 'o':
            outfn = GetStrArg();
            break;
         case 'u':
            SetuFlags(); break;
         case 'v':
            minVar = GetChkedFlt(0.0,100.0,s); break;
         case 'T':
            trace = GetChkedInt(0,3,s); break;
         case 'F':
            if (NextArg() != STRINGARG)
               HError(1,"Data File format expected");
            if((ff = Str2Format(GetStrArg())) == ALIEN)
               HError(0,"Warning ALIEN Data file format set");
            SetFormat(ff);
            break;
         case 'G':
            if (NextArg() != STRINGARG)
               HError(1,"Label File format expected");
            if((ff = Str2Format(GetStrArg())) == ALIEN)
               HError(0,"Warning ALIEN Label file format set");
            SetLFormat(ff);
            break;
         case 'L':
            if (NextArg()!=STRINGARG)
               HError(1,"Label file directory expected");
            labDir = GetStrArg(); break;
         case 'S': SetScriptFile();
            break;
         case 'X':
            if (NextArg()!=STRINGARG)
               HError(1,"Label file extension expected");
            labExt = GetStrArg(); break;
         default:
            HError(1,"Unknown switch %s",s);
      }
   }
   if (NextArg()!=STRINGARG)
      HError(1,"source HMM file name expected");
   hmmfn = GetStrArg();
   if (outfn==NULL) outfn = hmmfn;
   Initialise1();
   do {
      if (NextArg()!=STRINGARG)
         HError(1,"training data file name expected");
      datafn = GetStrArg();
      LoadFile(datafn);
   } while (NumArgs()>0);
   FixDbase();
   R = NumSequences();
   if (R < minSeq)
      HError(2,"Too Few Observation Sequences [%d]",R);
   Initialise2();
   if (trace>0) {
      printf("%d sequences loaded, Max length = %d\n", R, maxT);
      fflush(stdout);
   }
   ReEstimateModel();
   if (trace>0) 
      printf("Saving Model to File %s\n",outfn);
   HSave(outfn,&theHMM);
}

/* ------------------------ Initialisation ----------------------- */

/* PrintInitialInfo: print a header of program settings */
void PrintInitialInfo()
{   
   char dummy[6];
   printf("Reestimating HMM %s . . . \n",hmmfn);
   printf("    NumStates:  %d\n",N);
   printf("    NumMix   :  %d\n",M);
   printf("    VecSize  :  %d\n",V);
   printf("    CovKind  :  %s\n",CovKind2Str(ckind,dummy));
   if (segLab==NULL)
   printf("    SegLab   :  NONE\n");
   else
   printf("    SegLab   :  %s\n",segLab);
   printf("    maxIter  :  %d\n",maxIter);
   printf("    epsilon  :  %f\n",epsilon);
   printf("    Updating :  ");
   if (uFlags&UPTRANS) printf("Transitions "); 
   if (uFlags&UPMEANS) printf("Means "); 
   if (uFlags&UPVARS)  printf("Variances "); 
   if (uFlags&UPMIXES && M>1)  printf("MixWeights"); 
   printf("\n\n");
   fflush(stdout);
}
   
/* Initialise1: 1st phase of init prior to loading dbase */
void Initialise1(void)
{
   SPINBALL
   HOpen(hmmfn,&theHMM);
   N = theHMM.numStates; M = theHMM.numMixtures; 
   V = theHMM.vecSize; ckind = theHMM.ckind;
   T = maxT = 0;
   if(segLab != NULL) {
      InitLabel();
      segId = GetLabId(segLab,TRUE);
   }
   if (trace>0) PrintInitialInfo();
   InitDbase(V,theHMM.pkind);
}

/* Initialise2: 2nd phase of init after loading dbase */
void Initialise2(void)
{
   int i,j;

   seqProb = CreateVector(R);
   alpha = CreateMatrix(N-1,maxT);
   beta = CreateMatrix(N-1,maxT);
   outprob = CreateMatrix(N-1,maxT);
   if (M>1) {
      if ((mixoutp=(Matrix *)malloc((N-1)*sizeof(Matrix))) == NULL)
         HError(2,"Cannot create pointer array for mixoutp");
      --mixoutp;
      for (i=1;i<N;i++)
         mixoutp[i] = CreateMatrix(M,maxT);
   }
   tran = CreateMatrix(N,N);
   occ1 = CreateVector(N);
   occ = CreateVector(N);
   occT = CreateVector(N);
   if ((mu=(Vector **)malloc((N-2)*sizeof(Vector *))) == NULL)
      HError(2,"Cannot create pointer array for mu");
   mu -= 2;
   for (i=2;i<N;i++) {
      if ((mu[i] = (Vector *) malloc(M*sizeof(Vector))) == NULL)
         HError(2,"Cannot create col %d of mu", i);
      mu[i] -= 1;
      for (j=1; j<=M; j++)
         mu[i][j] = CreateVector(V);
   }
   if ((va=(Covariance **)malloc((N-2)*sizeof(Covariance *))) == NULL)
      HError(2,"Cannot create pointer array for va");
   va -= 2;
   for (i=2;i<N;i++) {
      if ((va[i] = (Covariance *) malloc(M*sizeof(Covariance))) == NULL)
         HError(2,"Cannot create col %d of va", i);
      va[i] -= 1;
      for (j=1; j<=M; j++)
         if (ckind == FULLC)
            va[i][j].inv = CreateMatrix(V,V);
         else
            va[i][j].var = CreateVector(V);
   }
   c = CreateMatrix(N,M);
}

/* --------------------------- Load Database ---------------------- */

/* CheckData: check data file consistent */
void CheckData(DataFile *data, char *fn)
{
   if (BaseSampKind(data->sampKind) >= ANON)   
      HError(2,"Unknown Parameterisation in Training Data %s",fn);
   if (data->sampKind != theHMM.pkind)
      HError(2,"Parameterisation in %s incompatible with hmm %s",fn,hmmfn);
}

/* LoadFile: load either whole file or reqd segments of fn into HDBase */
void LoadFile(char *fn)
{
   char labfn[80];
   Transcription trans;
   Label *labl;
   int vecStartInd,vecEndInd;  
   int i,ncas;
   DataFile data;
   
   SPINBALL
   if (segId == NULL)  {
      SpOpen(fn,&data);
      CheckData(&data,fn);
      PutSequence(0,(data.nSamples-1)*data.sampPeriod,&data);                  
      if (data.nSamples > maxT) 
         maxT=data.nSamples; 
      if (trace > 1) {
         printf(" Training Observations loaded from %s\n",fn);
         fflush(stdout);
      }        
      SpClose(&data);
   }
   else {
      MakeFN(fn,labDir,labExt,labfn);
      trans=LOpen(labfn);
      ncas=NumCases(segId,trans,0);
      if (ncas>0) {
         SpOpen(fn,&data);
         CheckData(&data,fn);
         for (i=1;i<=ncas;i++) {
            labl=GetCase(segId,i,trans,0);
            vecStartInd=labl->start/data.sampPeriod;
            vecEndInd=labl->end/data.sampPeriod;
            if (vecEndInd-vecStartInd + 1 > N-2)  {
               PutSequence(labl->start,labl->end,&data);                  
               if (vecEndInd-vecStartInd+1 > maxT)
                  maxT=vecEndInd-vecStartInd+1; 
           } else   
               if (trace > 1) 
                  HError(0,"Ignoring label length  %d",vecEndInd-vecStartInd+1);
         }
         SpClose(&data);
      }
      LClose(trans);
      if (trace > 1) {
         printf(" %d Training Observations loaded from %s\n",ncas,fn);
         fflush(stdout);
      }        
   } 
}

/* -------------------- Auxiliary Functions ------------------ */

/* DumpProbMatrix: print the given matrix */
void DumpProbMatrix(char *name, Matrix m)
{
   int i,t;
   LogReal x;

   printf("%s\n",name);
   for (i=N-1;i>1;i--) {
      for(t=1;t<=T;t++) {
         x=m[i][t];
         if (x>LSMALL )
            printf("%8.2f ",x);
         else 
            printf("    ---- ");
      }
      printf("\n");
   }
}

/* TotalProb: calc total prob as product if individual sequence probs */
LogReal TotalProb() 
{
   LogReal sum;
   int i;

   sum = 0.0;
   for (i=1;i<=R;i++) 
      sum += seqProb[i];
   return sum;
}

/* SetOutP: Set the output prob matrices */                        
void SetOutP(int seq)
{
   int i,t,m;
   Vector ot;

   for (t=1;t<=T;t++) {
      ot=GetDBVector(seq,t-1);
      for (i=2;i<N;i++) {
         outprob[i][t]=OutP(ot,&theHMM,i);
         if (M>1)
            for (m=1;m<=M;m++) 
               mixoutp[i][m][t]=MOutP(ot,&theHMM,i,m);
      }
   }
   SPINBALL
}

/* SetAlpha: compute the alpha matrix and seq prob for the given sequence */
Real SetAlpha(int seq)
{
   int i,j,t;
   LogReal x,a;

   SPINBALL
   for (j=2;j<N;j++) {              /* col 1 from entry state */
      a=theHMM.transP[1][j];
      if (a<LSMALL)
         alpha[j][1] = LZERO;
      else
         alpha[j][1] = a+outprob[j][1];
   }
   for (t=2;t<=T;t++) {             /* cols 2 to T */
      SPINBALL
      for (j=2;j<N;j++) {
         x=LZERO ;
         for (i=2;i<N;i++) {
            a=theHMM.transP[i][j];
            if (a>LSMALL)
               x = LAdd(x,alpha[i][t-1]+a);
         }
         alpha[j][t]=x+outprob[j][t];
      }
   }
   x = LZERO ;                      /* finally calc seq prob */
   for (i=2;i<N;i++) {
      a=theHMM.transP[i][N];
      if (a>LSMALL)
         x=LAdd(x,alpha[i][T]+a); 
   }  
   seqProb[seq+1]=x;
   if (trace > 2) {
      DumpProbMatrix("Alpha",alpha); 
      printf("LogP= %10.3f\n\n",x);
   }
}

/* SetBeta: compute beta matrix */
void SetBeta(void)
{
   int i,j,t;
   LogReal x,a;

   for (i=2;i<N;i++)                /* Col T from exit state */
      beta[i][T]=theHMM.transP[i][N];
   for (t=T-1;t>=1;t--) {           /* Col t from col t+1 */
      SPINBALL
      for (i=2;i<N;i++)
         beta[i][t]=LZERO ;
      for (j=2;j<N;j++) {
         x=outprob[j][t+1]+beta[j][t+1];
         for (i=2;i<N;i++) {
            a=theHMM.transP[i][j];
            if (a>LSMALL)
               beta[i][t]=LAdd(beta[i][t],x+a);
         }
      }
   }
   SPINBALL
   if (trace > 2) {
      /* Calc logP, should equal seqProb[seq] */
      x=LZERO ;
      for (j=2;j<N;j++) {
         a=theHMM.transP[1][j];
         if (a>LSMALL)
            x=LAdd(x,beta[j][1]+a+outprob[j][1]); 
      }
      DumpProbMatrix("Beta",beta);
      printf("LogP=%10.3f\n\n",x);
   }
}

/* --------------------- Parameter Estimation ---------------- */

/* InitParameters: zero all the accumulators */
void InitParameters(void)
{
   int i,j,k,kk,m;
   Vector mu_im,va_im_var;
   Matrix va_im_inv;
 
   for (i=2;i<N;i++) {
      occ[i] = occ1[i] = occT[i] = 0.0;
      for (j=2;j<N;j++) 
         tran[i][j] = 0.0;
      for (m=1;m<=M;m++) {
         c[i][m]= 0.0;
         mu_im = mu[i][m];
         for (k=1;k<=V;k++) 
            mu_im[k] = 0.0;
         if (ckind == DIAGC)  {
            va_im_var = va[i][m].var;
            for (k=1;k<=V;k++) 
               va_im_var[k] = 0.0;
         } else {
            va_im_inv = va[i][m].inv;
            for (k=1;k<=V;k++)
               for (kk=1;kk<=V;kk++)
                  va_im_inv[k][kk] = 0.0;
         }
      }
   }
}

/* UpTranParms: implements the transition matrix reestimation */
void UpTranParms(LogReal pr)
{
   int i,j,t;
   Vector tran_i,alpha_i,beta_i,beta_j,outprob_j,a_i;
   LogReal x,a_ij;
   
   for (i=2;i<N;i++) {
      SPINBALL
      alpha_i = alpha[i];
      beta_i = beta[i];
      a_i = theHMM.transP[i];
      /* update state occupation counters */
      x = alpha_i[1]+beta_i[1]-pr;
      if (x>MINEARG) occ1[i] += exp(x);
      x = alpha_i[T]+beta_i[T]-pr;
      if (x>MINEARG) occT[i] += exp(x);
      x=LZERO ;
      for (t=2;t<=T-1;t++)
         x=LAdd(x,alpha_i[t]+beta_i[t]);
      x -= pr;
      if (x>MINEARG) occ[i]+=exp(x);
      /* Update Transition counter */
      if (uFlags&UPTRANS)  {
         tran_i = tran[i];
         for (j=2;j<N;j++) {
            a_ij=a_i[j];
            if (a_ij>LSMALL) {
               x=LZERO; beta_j=beta[j]; outprob_j=outprob[j];
               for (t=1;t<=T-1;t++)
                  x=LAdd(x,alpha_i[t]+a_ij+outprob_j[t+1]+beta_j[t+1]);
               x -= pr;
               if (x>MINEARG)
                  tran_i[j] += exp(x);
            }
         }
      }
   }
}


/* Up1MixParms: implements the output prob reestimation for 1 Mixture case */
void Up1MixParms(int seq, LogReal pr)
{
   int j,k,kk,t;
   Vector ot,alpha_j,beta_j,mu_j,va_j_var,mean,va_j_inv_k;
   Matrix va_j_inv;
   LogReal x;
   Real y,zmean,zmean1,zmean2;
   
   for (j=2;j<N;j++) {
      SPINBALL
      alpha_j = alpha[j];
      beta_j = beta[j];
      mu_j = mu[j][1]; 
      va_j_var = va[j][1].var;
      va_j_inv = va[j][1].inv;
      mean = theHMM.stateinfo[j].outp[1].mean;
      for (t=1;t<=T;t++) {
         ot=GetDBVector(seq,t-1);
         x = alpha_j[t]+beta_j[t] - pr;
         if (x>MINEARG) {
            y = exp(x);
            for (k=1;k<=V;k++) {
               mu_j[k] += ot[k]*y;
               if (ckind==DIAGC) {
                  zmean = ot[k]-mean[k];
                  va_j_var[k] += zmean*zmean*y;
               } else {
                  zmean1 = ot[k]-mean[k];
                  va_j_inv_k=va_j_inv[k];
                  for (kk=k;kk<=V;kk++) {
                     zmean2 = ot[kk]-mean[kk];
                     va_j_inv_k[kk] += zmean1*zmean2*y;
                  }
               }
            }
         }
      }
   }
}

/* UpMMixParms: implements the output prob reestimation for M>1 Mixture case */
void UpMMixParms(int seq, LogReal pr)
{
   int i,j,k,kk,m,t;
   Vector alpha_j,beta_j,mu_jm,va_jm_var,mean,va_jm_inv_k,ot;
   Matrix va_jm_inv;
   LogReal x,c_jm;
   Real y,w,zmean,zmean1,zmean2;

   for (j=2;j<N;j++) {
      SPINBALL
      alpha_j = alpha[j];
      beta_j = beta[j];
      for (m=1;m<=M;m++) {
         mu_jm = mu[j][m]; 
         va_jm_var = va[j][m].var;
         va_jm_inv = va[j][m].inv;
         mean = theHMM.stateinfo[j].outp[m].mean;
         w = theHMM.stateinfo[j].outp[m].weight;
         if (w>MINMIX){
            c_jm=log(w);
            for (t=1;t<=T;t++) {
               ot = GetDBVector(seq,t-1);
               if (t==1) 
                  x=theHMM.transP[1][j];
               else {
                  x=LZERO;
                  for (i=2;i<N;i++) 
                     x=LAdd(x,alpha[i][t-1]+theHMM.transP[i][j]);
               }
               if (x>LSMALL) {
                  x += mixoutp[j][m][t]+c_jm+beta_j[t]-pr;
                  if (x>MINEARG) {
                     y = exp(x);
                     for (k=1;k<=V;k++) {
                        mu_jm[k] += ot[k]*y;
                        if (uFlags&UPVARS)
                           if (ckind == DIAGC) {
                              zmean=ot[k]-mean[k];
                              va_jm_var[k] += zmean*zmean*y;
                           } else {
                              va_jm_inv_k = va_jm_inv[k];
                              zmean1=ot[k]-mean[k];               
                              for (kk=k;kk<=V;kk++) {
                                 zmean2 = ot[kk]-mean[kk];
                                 va_jm_inv_k[kk] += zmean1*zmean2*y;
                              }
                           }
                     }
                     if (uFlags&UPMIXES) 
                        c[j][m] += y;
                  }
               }
            }
         }
      }
   }  
}

/* UpdateParameters: implement the reestimation formulae */
void UpdateParameters(int seq)
{
   LogReal pr;

   pr = seqProb[seq+1];
   UpTranParms(pr);
   if (uFlags&(UPMEANS|UPVARS|UPMIXES))
      if (M==1)
         Up1MixParms(seq,pr);
      else
         UpMMixParms(seq,pr);
}

/* --------------------- Model Update ------------------ */

/* UpdateTheModel: use accumulated parm estimates to update model */
void UpdateTheModel()
{
   int i,j,k,kk,m;
   Real cfloor,occi,c_im,x,y;
   Vector mean;
   Covariance covariance;
   Vector mu_im,va_im_var,va_im_inv_k;
   Matrix va_im_inv;
   
   cfloor = MINMIX;
   for (i=2;i<N;i++) {
      SPINBALL
      occi = occ1[i]+occ[i]+occT[i];
      if (uFlags&UPTRANS) {
         x = occ1[i]/R;
         theHMM.transP[1][i] = (x<MINLARG) ? LZERO : log(x);
         for (j=2;j<N;j++) {
            x = tran[i][j]/occi;
            theHMM.transP[i][j] = (x<MINLARG) ? LZERO : log(x);
         }
         x = occT[i]/occi;
         theHMM.transP[i][N] = (x<MINLARG) ? LZERO : log(x);
      }
      SPINBALL
      if (uFlags&(UPMEANS|UPVARS|UPMIXES)) {
         for (m=1;m<=M;m++) {
            if (M==1) c_im = occi; else c_im=c[i][m];
            if (uFlags&UPMIXES) {
               x = c_im/occi;
               if (x>1.0) {
                  fprintf(stderr,
                       " Warning mix too big: s=%d m=%d c_im=%e occi=%e\n",
                              i,m,c_im,occi);
                  x = 1.0;
               }
               if (x<cfloor) {
                  fprintf(stderr,
                      " Warning mix too small: s=%d m=%d c_im=%e occi=%e\n",
                              i,m,c_im,occi);
                  theHMM.stateinfo[i].outp[m].weight=0.0;
               }else
                  theHMM.stateinfo[i].outp[m].weight=x;
            }
            mean=theHMM.stateinfo[i].outp[m].mean;
            covariance=theHMM.stateinfo[i].outp[m].covariance;
            mu_im=mu[i][m]; 
            va_im_inv=va[i][m].inv;
            va_im_var=va[i][m].var;
            if (uFlags&UPVARS) 
               for (k=1;k<=V;k++) {
                  if (ckind==DIAGC) {
                     x = va_im_var[k]/c_im; 
#ifdef DOMEANC
                     y = mean[k] - mu_im[k]/c_im;
                     x -= y*y;
#endif
                     covariance.var[k]=(x>minVar)?x:minVar;
                  } else {
                     va_im_inv_k = va_im_inv[k];
                     for (kk=k;kk<=V;kk++) {
                        y=va_im_inv_k[kk]/c_im;
#ifdef DOMEANC
                        y -= (mean[k] - mu_im[k]/c_im) * 
                             (mean[kk] - mu_im[kk]/c_im);
#endif
                        covariance.inv[k][kk]=y;
                        covariance.inv[kk][k]=y;
                     }
                  }                 
               }
            if (uFlags&UPMEANS)
               for (k=1;k<=V;k++)
                     mean[k] = mu_im[k]/c_im;
            if (ckind==FULLC) 
               CovInvert(covariance.inv,covariance.inv);
         }
      }
   }
   FixGConsts(&theHMM);
}


/* --------------------- Top Level Control ------------- */

/* ReEstimateModel: top level of algorithm */
void ReEstimateModel(void)
{
   LogReal oldP,newP;
   LogReal delta;
   int converged;
   int iteration;
   int i;

   iteration=1;
   oldP=LZERO;
   do {        /*main re-est loop*/      
      if (trace>1) {
         printf("Model prior to iteration %d\n",iteration);
         HPrint(&theHMM,HPALL,2,N-1);
      }
      InitParameters();
      for (i=0;i<R;i++) {
         T=SequenceLength(i);
         SetOutP(i);
         if (SetAlpha(i) > LSMALL){
	   SetBeta();
	   UpdateParameters(i);
	 } else
	   if (trace>1) printf("Example %d skipped\n",i);
      }
      UpdateTheModel();
      newP=TotalProb()/R;
      delta=newP-oldP;
      converged=(fabs(delta)<epsilon);
      oldP=newP;
      if (trace>0) {
         printf("Average Prob after iteration %d = %10.5f",iteration,oldP);
         if (iteration > 1)
            printf("  change = %10.5f",delta);
         printf("\n");
         fflush(stdout);
      }
   } while ((iteration++ < maxIter) && !converged);
   if (trace>0) {
      if (converged)
         printf("Estimation converged at iteration %d\n",iteration);
      else
         printf("Estimation aborted at iteration %d\n",iteration);
      fflush(stdout);
   }
   if (trace>1) {
      printf("Final Model\n");
      HPrint(&theHMM,HPALL,2,N-1);
   }
}

/* ----------------------------------------------------------- */
/*                      END:  HRest.c                          */
/* ----------------------------------------------------------- */
