/* ----------------------------------------------------------- */
/*                   HTK  -  HMM Toolkit                       */
/* ----------------------------------------------------------- */
/*        Copyright: Steve Young      October 1989             */
/*        Last Update:                September 1990           */
/*                                                             */
/*        Cambridge University Engineering Department          */
/*        Trumpington Street, Cambridge, CB2 1PZ               */
/*                                                             */
/*  Permission is hereby granted to copy, use and modify this  */
/*  software provided that this banner notice is not removed   */
/*                                                             */
/* ----------------------------------------------------------- */
/*    File: HResults.c: gather statistics on results           */
/* ----------------------------------------------------------- */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include "HShell.h"
#include "HMath.h"
#include "HSpIO.h"
#include "HLabel.h"

#define VERSION   "1.2"

/*
   This program reads in a set of HTK Label format files (output
   from a recognition tool) and compares them with the corresponding 
   transcription files.  The basic output is recognition statistics
   for the whole file set.  Optional extras are
      (a)   recognition statistics per file  (-f)
      (b)   time-aligned transcriptions      (-t)
      (c)   phoneme statistics               (-p)
   For comparison purposes, it is also possible to assign two
   labels to the same equivalence class (-e).  Label files
   containing triphone labels of the form A-B-C can be optionally
   stripped down to B via the -s option.  Any label of the form
   ??? is ignored.  Transcription files may be multi-level but
   only the top level is considered.
*/

#define NULCLASS  "???"

static short fullResults = FALSE;   /* enable full Results */
static short outTrans   = FALSE;    /* enable transcription output */
static short outPStats  = FALSE;    /* enable phoneme statistics */
static char * labDir    = NULL;     /* label file directory */
static char * labExt    = "lab";    /* label file extension */
static FileFormat ff    = HTK;      /* ff of transcription files */
static char *recfn;                 /* rec file name (test) */
static char labfn[255];             /* lab file name (reference) */
static Transcription ref,test;      /* the Transcriptions being compared */
static short stripContexts = FALSE; /* strip triphone contexts */

typedef struct _Equiv Equiv;        /* list of equivalent labels */
struct _Equiv{
   LabId classId;
   LabId equivId;
   Equiv *next;
};

static Equiv *eqlist=NULL;    /* List of equivalent label ids */
static LabId *labl;           /* List of all label ids */
static int P;                 /* number of distinct labels in labl */
static LabId nulClass;        /* Id of NULCLASS phone label */

/* ------------------ Process Command Line ------------------------- */

void ReportUsage(void)
{
   fprintf(stderr,"\nUSAGE: HResults[V%s] [options] idList recFiles...\n\n",VERSION);
   fprintf(stderr," Option                                   Default\n\n");
   fprintf(stderr," -e s t  Label t is equivalent to s\n");
   fprintf(stderr," -f      Enable full results                 off\n");
   fprintf(stderr," -p      Output phoneme statistics           off\n");
   fprintf(stderr," -s      Strip triphone contexts             off\n");
   fprintf(stderr," -t      Output time aligned transcriptions  off\n");
   fprintf(stderr," -G fmt  Set label file format to fmt        HTK\n");
   fprintf(stderr," -L s    dir to find label file(s)         current\n");
   fprintf(stderr," -S f    set script file to f                none\n");
   fprintf(stderr," -X s    extension for label files           lab\n");
   exit(1);
}


int GetChkedInt(int min, int max, char * swtname)
{
   int val;
   
   if (NextArg() != INTARG) 
      HError(1,"Integer Arg Required for %s option",swtname);
   val=GetIntArg();
   if (val<min || val>max) 
      HError(1,"Integer arg out of range in %s option",swtname);
   return val;
}

main(int argc, char *argv[])
{
   char *s,*e,*c;
   void InitConMat(void);
   void MatchFiles(void);
   void OutputStats(void);
   void AddEquiv(char * cl, char * eq);
   void CheckEqList(void);
   
   InitShell(argc,argv);
   InitMath(FALSE);
   InitLabel();
   nulClass = GetLabId(NULCLASS,TRUE);
   if (NumArgs()==0)
      ReportUsage();
   while (NextArg() == SWITCHARG) {
      s = GetSwtArg();
      if (strlen(s)!=1) 
         HError(1,"Bad switch %s; must be single letter",s);
      switch(s[0]){
         case 'e':
            if (NextArg() != STRINGARG)
               HError(1,"Eq Class Name Expected");
            c = GetStrArg();
            if (NextArg() != STRINGARG)
               HError(1,"Eq Label Name Expected");
            e = GetStrArg();
            AddEquiv(c,e);
            break;
         case 'f':
            fullResults = TRUE; break;
         case 'p':
            outPStats = TRUE; break;
         case 's':
            stripContexts = TRUE; break;
         case 't':
            outTrans = TRUE; break;
         case 'G':
            if (NextArg() != STRINGARG)
               HError(1,"Label File format expected");
            if((ff = Str2Format(GetStrArg())) == ALIEN)
               HError(0,"Warning ALIEN Label file format set");
            break;
         case 'L':
            if (NextArg()!=STRINGARG)
               HError(1,"Label file directory expected");
            labDir = GetStrArg(); break;
         case 'S': SetScriptFile();
            break;
         case 'X':
            if (NextArg()!=STRINGARG)
               HError(1,"Label file extension expected");
            labExt = GetStrArg(); break;
         default:
            HError(1,"Unknown switch %s",s);
      }
   }
   if (NextArg() != STRINGARG)
      HError(1,"Label Id List File expected");
   labl = LoadIdList(GetStrArg(),&P);
   if (stripContexts) LTriStrip(TRUE);
   CheckEqList();
   InitConMat();
   do {
      if (NextArg()!=STRINGARG)
         HError(1,"recognition output file name expected");
      recfn = GetStrArg();
      MatchFiles();
   } while (NumArgs()>0);
   OutputStats();
   exit(0);
}


/* -------------------- Label Equivalences ----------------------- */

/* AddEquiv: Add the equivalent pair (cl,eq) to eqlist */
void AddEquiv(char * cl, char * eq)
{
   Equiv *p;

   if ((p=(Equiv *)malloc(sizeof(Equiv)))==NULL)
      HError(2,"Unable to create equiv class record");
   p->classId = GetLabId(cl,TRUE);
   p->equivId = GetLabId(eq,TRUE);
   p->next = eqlist; eqlist = p;
}

/* CheckEqList: check that all labels in eqlist are in labl */
void CheckEqList(void)
{
   Equiv *p = eqlist;

   while (p != NULL) {
      if (FindLabId(p->classId,labl) == 0 && p->classId != nulClass)
         HError(2,"Unknown Eq Class Name %s",p->classId->name);
      if (FindLabId(p->equivId,labl) == 0)
         HError(2,"Unknown Eq Label Name %s",p->equivId->name);
      p=p->next;
   }
}

/* ConvertEq: convert all equivalent labels to class name */
void ConvertEq(Transcription t)
{
   LabId cl,eq;
   Equiv *p = eqlist;
   int i;
   
   while (p != NULL){
      cl = p->classId; eq = p->equivId;
      for (i=0;i<t->nLabels;i++)
         if (t->labels[i].labid == eq) 
            t->labels[i].labid = cl;
      p = p->next;
   }
}

/* ------------------------ DP Matching ------------------------ */

/* In the DP matching, the recognition output label file is 
   referred to as the test file, the annotated label file is
   referred to as the reference file. Test labels are indexed
   by i and ref labels are indexed by j.  Viewed as a grid,
   the test labels span the horizantal axis and the ref labels
   span the vertical.  The cell in column i and row j represents
   the match between the i'th test label and the j'th ref label
*/

static long ins   = 0;     /* Total insertions */
static long del   = 0;     /* Total deletions */
static long sub   = 0;     /* Total substitutions */
static long hits  = 0;     /* Total hits */
static long nsyms = 0;     /* Total symbols */
static long nphr  = 0;     /* Total phrases */
static long phrcor= 0;     /* Phrase correct */

static const int subPen = 10;    /* error penalties */
static const int delPen = 7;
static const int insPen = 7;

enum _Direction{DIAG,VERT,HOR,NIL};
typedef enum _Direction Direction;

typedef struct {
   int ins,del,sub,hit;
   int score;
   Direction dir;
} Cell;

static Cell **grid;     /* matrix of cells */

/* DumpGrid: for debugging */
void DumpGrid(void)
{
   int i,j;

   printf("Grid -\n");
   for (j=0; j<=ref->nLabels; j++)
    if (0==j) 
       printf("%10s","");
    else
       printf("%5.4s",ref->labels[j-1].labid->name);
   printf("\n");

   for (i=0; i<=test->nLabels; i++) {
    if (0==i) 
       printf("%5s","");
    else
       printf("%5.4s",test->labels[i-1].labid->name);
    for (j=0; j<=ref->nLabels; j++) {
       printf("%4d",grid[i][j].score);
       if (grid[i][j].dir==DIAG)
          printf("d");
       else if (grid[i][j].dir==HOR)
          printf("h");
       else if (grid[i][j].dir==VERT)
          printf("v");
       else 
          printf("n");
    }
    printf("\n");
   }
   fflush(stdout);
}

/* CreateGrid: Create a grid of cells of the given dimensions */
void CreateGrid(void)
{
   int i, testSize, refSize;
   
   testSize = test->nLabels; refSize = ref->nLabels;
   if ((grid = (Cell **)malloc((testSize+1)*sizeof(Cell *))) == NULL)
      HError(2,"Cannot create grid pointer array");
   for (i=0; i<=testSize;i++)
      if ((grid[i] = (Cell *) calloc(refSize+1,sizeof(Cell))) == NULL)
         HError(2,"Cannot create grid column %d",i);
   grid[0][0].score = grid[0][0].ins = grid[0][0].del = grid[0][0].sub 
                    = grid[0][0].hit = 0;
   grid[0][0].dir = NIL;
   for (i=1;i<=testSize;i++) {
      grid[i][0] = grid[i-1][0];
      grid[i][0].dir = HOR;
      if (test->labels[i-1].labid != nulClass) {
         grid[i][0].score += insPen;
         ++grid[i][0].ins;
      }
   }
   for (i=1;i<=refSize;i++) {
      grid[0][i] = grid[0][i-1];
      grid[0][i].dir = VERT;
      if (ref->labels[i-1].labid != nulClass) {
         grid[0][i].score += delPen;
         ++grid[0][i].del;
      }
   }
}

/* FreeGrid: free storage allocated to grid */
void FreeGrid(int testSize)
{
   int i;
   
   for (i=0;i<=testSize;i++)
      free(grid[i]);
   free(grid);
}

/* DoMatch: fill the grid */
void DoMatch(void)
{
   Cell *gridi,*gridi1;
   int h,d,v,i,j;
   Boolean refnull,testnull;

   for (i=1;i<=test->nLabels;i++){
      gridi = grid[i]; gridi1 = grid[i-1];
      testnull = test->labels[i-1].labid == nulClass;
      for (j=1;j<=ref->nLabels;j++) {
         refnull = ref->labels[j-1].labid == nulClass;
         if (refnull && testnull) {    /* both ref and test are null */
            h = gridi1[j].score; 
            d = gridi1[j-1].score; 
            v = gridi[j-1].score;
            if (d<=v && d<=h) {
               gridi[j] = gridi1[j-1]; 
               gridi[j].dir = DIAG;
            } else if (h<v) {
               gridi[j] = gridi1[j]; 
               gridi[j].dir = HOR;
            } else {
               gridi[j] = gridi[j-1];
               gridi[j].dir = VERT;
            }
         } else if (refnull) {         /* ref is null */
            gridi[j] = gridi[j-1]; 
            gridi[j].dir = VERT;
         } else if (testnull) {        /* test is null */
            gridi[j] = gridi1[j];
            gridi[j].dir = HOR;
       } else {                        /* normal case */
            h = gridi1[j].score +insPen;
            d = gridi1[j-1].score;
            if (ref->labels[j-1].labid != test->labels[i-1].labid)
               d += subPen;
            v = gridi[j-1].score + delPen;
            if (d<=h && d<=v) {        /* DIAG = hit or sub */
               gridi[j] = gridi1[j-1];
               gridi[j].score = d;
               gridi[j].dir = DIAG;
               if (ref->labels[j-1].labid == test->labels[i-1].labid)
                  ++gridi[j].hit;
               else
                  ++gridi[j].sub;
            } else if (h<v) {          /* HOR = ins */
               gridi[j] = gridi1[j];
               gridi[j].score = h;
               gridi[j].dir = HOR;
               ++ gridi[j].ins;
            } else {                   /* VERT = del */
               gridi[j] = gridi[j-1];
               gridi[j].score = v;
               gridi[j].dir = VERT;
               ++gridi[j].del;
            }
         }
      }  /* for j */
   }  /* for i */
   grid[0][0].dir = NIL;
}

/* ------------------- Aligned Transcriptions --------------- */

/* AppendItem: appends item padded to width spaces to s */
void AppendItem(char *s, char *item, int len, int width)
{
   char buf[64];
   int i;

   for (i=0; i<len; i++)
      buf[i] = item[i];
   for (i=len; i<width; i++)
      buf[i] = ' ';
   buf[width] = '\0';
   strcat(s,buf);
}

/* AppendPair: append a to linea and b to lineb */
void AppendPair(char *linea, char *a, char *lineb, char *b)
{
   int lena,lenb,width;
   
   lena = strlen(a); lenb = strlen(b);
   width = ((lena>=lenb) ? lena : lenb)+1;
   AppendItem(linea,a,lena,width);
   AppendItem(lineb,b,lenb,width);
}

/* AppendCell: path upto grid[i][j] to tb and rb (recursive) */
void AppendCell(int i, int j, char *tb, char *rb)
{
   char *rlab,*tlab;
   char empty[1];

   if (i<0 || j<0) 
    HError(2,"Trace back failure in AppendCell");
   empty[0] = '\0';
   switch (grid[i][j].dir){
    case DIAG:
       tlab = test->labels[i-1].labid->name;
       rlab = ref->labels[j-1].labid->name;
       AppendCell(i-1,j-1,tb,rb); break;
    case HOR:
       tlab = test->labels[i-1].labid->name;
       rlab = empty;
       AppendCell(i-1,j,tb,rb); break;
    case VERT:
       rlab = ref->labels[j-1].labid->name;
       tlab = empty;
       AppendCell(i,j-1,tb,rb); break;
    case NIL:
       return;
   }
   AppendPair(rb,rlab,tb,tlab);
}

/* OutTrans: output aligned transcriptions using best path in grid */
void OutTrans(void)
{
   char refBuf[255];
   char testBuf[255];
   
   strcpy(refBuf," LAB: ");
   strcpy(testBuf," REC: ");
   AppendCell(test->nLabels,ref->nLabels,testBuf,refBuf);
   printf("Aligned transcription: %s vs %s\n", labfn, recfn);
   printf("%s\n",refBuf);
   printf("%s\n",testBuf);
   fflush(stdout);
}

/* ----------------- Confusion Matrix (Phoneme Stats) ----------- */

#define MAXCONMATSIZE 100
static short **conMat;  /* confusion matrix, conMat[i][j] is the number of
                           times label i was recognised as label j */

/* InitConMat:  allocate and initialise confusion matrix */
void InitConMat(void)
{
   int i,j;
   short *q;
   
   if (P>MAXCONMATSIZE)
      HError(2,"Confusion matrix would be too large");
   if ((conMat = (short **) malloc(P*sizeof(short *))) == NULL )
      HError(2,"Unable to create confusion matrix pointer array");
   --conMat;  /* index is 1..P */
   for (i=1;i<=P;i++){
      if ((q = (short *)malloc(P*sizeof(short))) == NULL)
         HError(2,"cannot allocate column %d of confusion matrix",i);
      --q;  /* index is 1..P */
      conMat[i] = q;
      for (j=1;j<=P;j++) q[j] = 0;
   }  
}

/* OutConMat: output the confusion matrix */
void OutConMat(void)
{
   int i,j,err,rowerr,maxlen;
   char *s,c;

   maxlen = 0;
   for (i=1;i<=P;i++)
      if (strlen(labl[i]->name) > maxlen)
         maxlen = strlen(labl[i]->name);
   if (maxlen>4) maxlen = 4;
   
   printf("Confusion Matrix\n");
   for (j=0;j<maxlen;j++){
      printf("     ");
      for (i=1;i<=P;i++){
         s = labl[i]->name;
         c = (j<strlen(s))?s[j]:' ';
         printf("  %c",c);
      }
      printf("\n");
   }
   for (i=1;i<=P;i++){
      printf("%4s ",labl[i]->name);
      rowerr = 0;
      for (j=1; j<=P; j++){
         err = conMat[i][j];
         if (i!=j) rowerr += err;
         printf("%3d",err);
      }
      printf("  [%3d errs]\n",rowerr);
   }
}     

/* CollectStats: trace back from grid[i][j] collecting phoneme stats */
void CollectStats(int i,int j)
{
   int ri,ti;
   LabId rlab,tlab;

   do {
      switch(grid[i][j].dir) {
         case NIL:   
            return;
         case DIAG:  
            rlab = ref->labels[j-1].labid;
            tlab = test->labels[i-1].labid;
            if (rlab==nulClass || tlab==nulClass) {
               --i; --j;
               break;
            }
            if ((ri=FindLabId(rlab,labl)) == 0)
               HError(2,",Unknown ref label %s",rlab->name);
            if ((ti=FindLabId(tlab,labl)) == 0)
               HError(2,",Unknown test label %s",tlab->name);
            ++conMat[ri][ti];
            --i; --j;
            break;
         case VERT:
            --j; break;
         case HOR:
            --i;  break;
    }
   } while (!(i==0 || j==0));
}

/* --------------------- Top Level Match Routine ------------ */

/* MatchFiles: match recfn (test) against labfn (ref) */
void MatchFiles(void)
{
   Cell *p;
   int nr,nt,nc,error;
   float accuracy,correct;
   
   SetLFormat(HTK);
   test = LOpen(recfn); 
   ConvertEq(test);
   MakeFN(recfn,labDir,labExt,labfn);
   SetLFormat(ff);
   ref = LOpen(labfn);
   ConvertEq(ref);
   nr = ref->nLabels; nt = test->nLabels;
   CreateGrid();
   DoMatch();
   p = &grid[nt][nr];
   del += p->del; /* update global counters */
   ins += p->ins;
   sub += p->sub;
   hits += p->hit;
   nc = p->hit+p->del+p->sub;
   nsyms += nc; /* DR - Reported Bug Fix (was nyms += nr) */
   ++nphr;
   error = !(p->del==0 && p->ins==0 && p->sub==0);
   if (!error)
      ++phrcor;
   if (fullResults) {
      accuracy = 100.0 * ((float) (p->hit-p->ins) / (float) nc);
      correct =  100.0 * ((float) p->hit / (float) nc);
      printf("%s:  %.2f(%.2f)  [H=%d, D=%d, S=%d, I=%d, N=%d]\n",
             recfn, correct, accuracy, p->hit,p->del,p->sub,p->ins,nc);
   }
   if (outTrans&&error) 
      OutTrans();
   if  (outPStats) 
      CollectStats(nt,nr);
   LClose(ref);
   LClose(test);
   FreeGrid(nt);
}

/* --------------- Output Global Stats ------------------- */

/* OutputStats:  output global statistics */
void OutputStats(void)
{
   float accuracy,correct,phrcorrect;

   accuracy = 100.0 * (float) (hits-ins) / (float) nsyms;
   correct = 100.0 * (float) hits / (float) nsyms;
   phrcorrect = 100.0 * ((float) phrcor / (float) nphr);
   printf("\n------------------------ Overall Results -------------------------\n");
   printf("PHRASE: %%Correct=%.2f [H=%d, S=%d, N=%d]\n",
          phrcorrect,phrcor,nphr-phrcor,nphr);
   printf("PHONE:  %%Corr=%.2f, Acc=%.2f",correct,accuracy);
   printf(" [H=%d, D=%d, S=%d, I=%d, N=%d]\n",hits,del,sub,ins,nsyms);
   printf("------------------------------------------------------------------\n");
   if (outPStats) 
      OutConMat();
}

/* ------------------------------------------------------------ */
/*                      END:  HResults.c                        */
/* ------------------------------------------------------------ */

