/* ----------------------------------------------------------- */
/*                   HTK  -  HMM Toolkit                       */
/* ----------------------------------------------------------- */
/*  HVite  Copyright: Steve Young      October 1989            */
/*  HGPD   Copyright: David Rainton    September 1991          */
/*         Last Update:                December 1991           */
/*                                                             */
/*         Cambridge University Engineering Department         */
/*         Trumpington Street, Cambridge, CB2 1PZ              */
/*                                                             */
/* Permission is hereby granted to copy, use and modify this   */
/* software provided that this banner notice is not removed    */
/*                                                             */
/*  D. Rainton - Modification History                          */
/*                                                             */
/* July 2 1991 - subspace code added                           */
/* Aug  9 1991 - N-best code added                             */
/* Aug 14 1991 - internal state path output added              */
/* Nov. 28 1991 - A.  Duchon  combine HVPath and HGPD,         */
/*                            subsace code deleted             */
/*                                                             */
/* ----------------------------------------------------------- */
/*   File: HVNG.c: Viterbi Recogniser and GPD implementation   */
/* ----------------------------------------------------------- */

/*
   This program will match a network of HMM's against one or more
   speech files and output a transcription in HTK format.  The 
   match is performed using conventional Viterbi decoding.  The
   network is input via HParse, all nodes are assumed to be
   the names of phone labels.  The HMM definition file 
   corresponding to each label must consist of the label name 
   followed by the specified extension (which may be null).
   The set of all hmm's needed must be listed in an idList.
   
   An optional bigram may be specified via the -b option.  The
   bigram should be a text file stored in the format
   
     <lab1> p11 p12 p13  ... p1H
     <lab2> p21 p22 p23  ... p2H
           ...
     <labH> pH1 pH2 pH3  ... pHH
     
   where <labh> is the name of the h'th label in the idList and
   pij is the bigram probabity P(labj|labi).
   
   Recognition may be controlled by the following optional 'scaling'
   parameters:
      
         propPen  - a fixed inter-model transition probability
         bigScale - scale factor for log(bigram) probabilities
         bjtScale - scale factor for log(bj(ot)) probabilities 

   
   The principle data structure is the network corresponding to
   the regular grammar rules used to constrain the recognition.
   This network is built by HParse.  For each phoneme instance,
   a PhoneInstance struct is created and hung-off the 'user' member
   of the corresponding network node.  Each PhoneInstance contains
   a pointer to a Definition struct containing the HMM definition 
   and space for bj(ot) to save recalculating multiple instances. 
   Definition structs are shared by the PhoneInstances so that 
   there is only one copy of each distinct HMM.  
   
   The recogniser design follows the token passing model.
*/


#include <stdio.h>      /* Standard C Libraries */
#include <stdlib.h>
#include <math.h>

#include "HShell.h"     /* HMM ToolKit Modules */
#include "HMath.h"
#include "HSpIO.h"
#include "HLabel.h"
#include "HModel.h"
#include "HParse.h"
#include "HUtilities.h"

#define VERSION   "1.0"
#define GCMARGIN   400
#define MAX_N_BEST 20

/* Types of loss function */ 
#define POWER 0
#define SIG  1
#define PIECEWISELINEAR 2

#define SIG_ARG "s"
#define POWER_ARG "p"
#define PIECEWISELINEAR_ARG "l"

#define DOWN -1
#define UP   +1

#define LEVELSEP '/'

#define MODEL_NOT_FOUND -1
#define NOT_DEFINED 0.0
#define MAXFILES 1000

/* ------------ Tokens, Definitions and PhoneLink Records -------------- */

typedef struct _Definition Definition;
typedef struct _PLR PLR;
typedef struct _ILR ILR;
typedef struct _SLR SLR;
typedef struct _PhoneInstance PhoneInstance;

typedef struct {
   PhoneInstance * inst;   /* current owner */
   float logprob;          /* score */
   PLR   *link;            /* link */
   ILR   *ilink;           /* link*/
   SLR   *slrlink;         /* DR - link to SLR */
} Token;

struct _Definition{
   HMMDef hmmdef;       /* actual HMM def */
   float *bjot;         /* array[2..numStates-1] of bj(ot) */
   LabId id;            /* id of this defn */
   Boolean bjLoaded;    /* true if bjot loaded */
};

struct _PLR{
   Boolean linked;      /* flag used in garbage collection */
   Token tok;           /* token at boundary */
   int time;            /* index of model boundary */
   PLR *chain;          /* chain for keeping track of PLR's */
};

struct _ILR{
   int time;            /* time of state entry */
   int state;           /* state the token is in */
   ILR *chain;          /* chain for keeping track of ILR's*/
   ILR *gchain;        /* global chain for global freeing of ILR structures */
};

struct _SLR{
   int time;            /* time of observation */
   int state;           /* the model state */
   float bjot;          /* probability of observation */
   SLR *link;           /* points to the previous SLR*/
   SLR *chain;          /* for house keeping */
   char *name;          /* model name */
 };                  

struct _PhoneInstance{
   Definition *def;  /* definition for this instance */
   Link netHook;     /* Network node holding this instance */
   int N;            /* copy of defList[id].numStates */
   Boolean isActive; /* true if pi is in active list */
   Token *state;     /* array[1..N]of token */
   int lastplrt;     /* time when last PLR gen'ed from this inst */
   PLR *lastplr;     /* pointer to last PLR gen'ed from this inst */
   Boolean hlpath;   /* DR - indicates whether pi is on hand labeled path through network */
};

/* Global Settings  from HGPD   */

static char * outHmmDir = NULL;   /* directory to place new hmm def files */
static int Tot_epochs = 25;    /* total number of epochs of training*/
static double wepsilon = -1.0;    /* weight learning constant */
static double mepsilon = -1.0;    /* mean learning constant */
static double vepsilon = -1.0;    /* covariance learning constant */
static int LossFunctionType = POWER; /* the default loss function */
enum _UPDSet{UPMEANS=1,UPVARS=2,UPTRANS=4,UPMIXES=8};
typedef enum _UPDSet UPDSet;
static UPDSet uFlags = UPMEANS|UPVARS|UPMIXES;     /* update flags */

static double alpha = 1.0;        /* the loss function power */
static HMMParameters hmmP;  
static int N=0;                   /* the number of data files */
static NumCorrect;                /* to keep track of training course */
static int R = 1;                /* number of epochs between recognition runs*/
int  trace    = 0;                /* Trace level NOT STATIC */
static int UF = 0;               /* Update Frequency, number of tokens */
static float AvgLoss = 0.0;       /* to follow the average loss per update*/
/* Global data structures from HGPD   */

static Definition *HMMdefList;    /* array[1..H] of HMM definitions */
static HMMDef *GPDupdateList; /* array[1..H] of HMM GPD updates */
static LabId *HMMidList;      /* array[1..H] of HMM file name id */
static DataFile data;         /* structure holding current data file */
static Vector ot;             /* current observation vector */
static Vector um;             /* temporary vector to store  mean gradient*/
static Covariance ucov;       /* temperory union to store covariance gradient*/

/* Global Settings  from HVPath    */

static char * labDir = NULL;     /* output label file directory */
static char * labExt = "rec";    /* output label file extension */
static char * ilabDir = NULL;    /* hand label file directory */
static char * ilabExt = "lab";   /* hand label file extension */
static char * hmmDir = NULL;     /* directory to look for hmm def files */
static char * hmmExt = NULL;     /* hmm def file extension */
static char * ipathDir = NULL;   /* directory to store internal path files */
static char * ipathExt = "ip";   /* extension for internal path files */
static float propPen = 1.0;      /* inter model propagation probability */
static float bigScale = 1.0;     /* bigram scale factor */
static float bjtScale = 1.0;     /* bjot scale factor */
static float pruning = 0.0;      /* pruning threshold */
static Boolean fixedPen = FALSE; /* enable flag for fixed propPen */
static Boolean isPruning= FALSE; /* enable flag for pruning */
static int NBest = 1;            /* produce NBest parses */ 

static long sampPeriod;          /* data sample period in 100us units */

/* Global data structures from HVPath    */

#define CHUNKSIZE 100             /*  from HLabel --AD */
static int numChunks;            /* .. in current list  from HLabel --AD*/
static LabId enterId;            /* LabId of the ENTER node name */
static LabId exitId;             /* LabId of the EXIT node name */
static Network theNet;           /* the syntax constraint network */
static int H = 0;                /* number of distinct HMM's (Labels) */
static int P = 0;                /* number of distinct phonemes */
static int NPI = 0;              /* number of phone instances */
static Token *tBuf;              /* space to copy largest array of tokens */
static Boolean hasBigram=FALSE;  /* bigram loaded flag */
static Matrix bigram;            /* bigram stats if any */
static int nActive = 0;          /* number of active phone instances */
static PhoneInstance **active;   /* array[0..nActive-1]of active pi's */
static LogReal maxLogP;          /* max LogP at current time frame */
static int maxPLRUsed=0;         /* max PLR's used */

static ILR *ilrHD=NULL;       /* this is used for a global freeing of ILR structures */
static SLR *slrHD=NULL;       /* this is used for a global freeing of SLR structures */

static LogReal tlgpr[MAX_N_BEST+1];  /* AD -- store the log probabilities  */

/* ------------------ Process Command Line ------------------------- */

void ReportUsage(void)
{
fprintf(stderr,"\nUSAGE: HVNG[V%s] [options] hmmList netFile speechFiles...\n\n", VERSION);
   fprintf(stderr," Option                                         Default\n\n");
   fprintf(stderr," -F fmt  Set data file format to fmt            HTK\n");
   fprintf(stderr," -G fmt  Set the hand label file format to fmt  HTK\n");
   fprintf(stderr," -B f    load bigram from file f                none\n");
   fprintf(stderr," -S f    set script file to f                   none\n");
   fprintf(stderr," -H s    dir to find hmm definitions            current\n");
   fprintf(stderr," -h s    extension for hmm files                none\n");
   fprintf(stderr," -L s    dir to store output label file(s)      current\n");
   fprintf(stderr," -l s    extension for output label files       rec\n");
   fprintf(stderr," -I s    dir to store internal path files       current\n");
   fprintf(stderr," -i s    extension for the internal path files  ip\n");
   fprintf(stderr," -D s    dir to get hand label files from       none\n");  
   fprintf(stderr," -d s    extension for hand label files         lab\n");
   fprintf(stderr," -N s    dir to store new hmm files             current\n"); 
   fprintf(stderr," -o f    bj(ot) scale factor                    1.0\n");
   fprintf(stderr," -p f    inter model trans prob = f             1/Nsucc\n");
   fprintf(stderr," -s f    bigram scale factor                    1.0\n");
   fprintf(stderr," -t f    set pruning threshold                  0.0\n");
   fprintf(stderr," -T N    set trace level to N                   0\n");
   fprintf(stderr," -n n    output the N-Best parses               1\n");
   fprintf(stderr," -w f    weight learning constant               0.01\n");
   fprintf(stderr," -m f    mean learning constant                 0.01\n");
   fprintf(stderr," -v f    variances learning constant            0.01\n");
   fprintf(stderr," -E i    Total number of epochs of training     25\n");
   fprintf(stderr," -R i    When to do recognition                 1\n");
   fprintf(stderr," -a f    loss function power (alpha)            1.0\n");
   fprintf(stderr," -f s    loss function (options l, s, p)        s\n");
   fprintf(stderr," -U d    When to update                         1\n");
   fprintf(stderr," -u tmvw Update t)rans m)eans v)ars w)eights    mw\n");
   exit(1);
}

char * LossFunct2Str(int type)
{
 switch (type)
 {
   case SIG:             return(SIG_ARG); 
                          break;
   case POWER:            return(POWER_ARG); 
                          break;
   case PIECEWISELINEAR:  return(PIECEWISELINEAR_ARG);
                          break;

   default:               HError (1,"Invalid loss function type");
                          break; 
 }
}


int Str2LossFunct(char *s)
{
 if (!strcmp(s, SIG_ARG)) return (SIG);
 else
   if (!strcmp(s, POWER_ARG)) return (POWER);
   else
     if (!strcmp(s, PIECEWISELINEAR_ARG)) return (PIECEWISELINEAR);
     else {
       fprintf(stderr, "Invalid loss function type. The choices are:\n");
       fprintf(stderr, "  %s - power loss function\n", POWER_ARG);
       fprintf(stderr, "  %s - sigm loss function\n", SIG_ARG);
       fprintf(stderr, "  %s - piecewise linear loss function\n", PIECEWISELINEAR_ARG);
       exit(1);
     }
}


void PrintArgs(void)
{
   printf(" -E f    total number of epochs          %d\n", Tot_epochs);
   printf(" -w f    weight learning constant        %e\n",wepsilon);
   printf(" -m f    mean learning constant          %e\n",mepsilon);
   printf(" -a f    loss function power (alpha)     %lf\n",alpha);
   printf(" -f s    loss function (options l, s, p) %s\n", LossFunct2Str(LossFunctionType));
   printf(" -n d    N-best paths found              %d\n",NBest);
   printf(" -U d    Updating every %d tokens\n",UF);
   if (uFlags&UPMEANS)  printf("Updating mixture means\n");   
   if (uFlags&UPMIXES)  printf("Updating mixture weights\n");
   if (uFlags&UPVARS)   printf("Updating mixture covariances\n");
   printf("\n");
   fflush(stdout);
}

void SetuFlags()
{
   char *s;
   
   s=GetStrArg();
   uFlags=0;        
   while (*s != NULL)
      switch (*s++) {
         case 't': uFlags+=UPTRANS;break;
         case 'm': uFlags+=UPMEANS;break;
         case 'v': uFlags+=UPVARS;break;
         case 'w': uFlags+=UPMIXES;break;
         default: HError(2,"Unknown update flag %c",*s);
                  break;
      }
}

main(int argc, char *argv[])
{
   char *datafn, *s, *files[MAXFILES];
   FileFormat ff;
   int filenum[MAXFILES];
   int epoch, i, j, itrns=0, ITRNS;
   float w,m,v;
   char *netfn,*listfn,*bigfn = NULL;
   void Initialise(char * listfn, char * netfn, char * bigfn);
   void SaveDefs(Definition *HMMdefList, LabId *HMMidList, char *hmmDir, char *hmmExt, int H);
   void MatchFile(char * datafn);
   void fill_array(int filenum[MAXFILES], int N);
   void jumble(int filenum[MAXFILES], int N);
   void GPDUpdate();
   void ScaleUpdates();

   InitShell(argc,argv);
   InitMath(FALSE);        /* Doesnt use LAdd */
   InitLabel();
   if (NumArgs()==0)
      ReportUsage();
   while (NextArg() == SWITCHARG) {
      s = GetSwtArg();
      if (strlen(s)!=1) 
         HError(1,"Bad switch %s; must be single letter",s);
      switch(s[0]){
         case 'F':
            if (NextArg() != STRINGARG)
               HError(1,"Data File format expected");
            if((ff = Str2Format(GetStrArg())) == ALIEN)
               HError(0,"Warning ALIEN Data file format set");
            SetFormat(ff);
            break;
         case 'G':
            if (NextArg() != STRINGARG)
               HError(1,"Label File format expected");
            if((ff = Str2Format(GetStrArg())) == ALIEN)
               HError(0,"Warning ALIEN Label file format set");
            SetLFormat(ff);
            break;
         case 'B':
            if (NextArg()!=STRINGARG)
               HError(1,"Bigram file name expected");
            bigfn = GetStrArg(); 
            break;
         case 'S': SetScriptFile();
            break;
         case 'H':
            if (NextArg()!=STRINGARG)
               HError(1,"HMM definition directory expected");
            hmmDir = GetStrArg(); break;     
         case 'h':
            if (NextArg()!=STRINGARG)
               HError(1,"HMM file extension expected");
            hmmExt = GetStrArg(); break;
         case 'L':
            if (NextArg()!=STRINGARG)
               HError(1,"Output Label file directory expected");
            labDir = GetStrArg(); break;
         case 'l':
            if (NextArg()!=STRINGARG)
               HError(1,"Label file extension expected");
            labExt = GetStrArg(); break;
         case 'I':
            if (NextArg()!=STRINGARG)
               HError(1,"Internal path file directory expected");
            ipathDir = GetStrArg(); break;
         case 'i':
            if (NextArg()!=STRINGARG)
               HError(1,"Internal path file extension expected");
            ipathExt = GetStrArg(); break;
         case 'D':
            if (NextArg()!=STRINGARG)
               HError(1,"Input Label file directory expected");
            ilabDir = GetStrArg(); break;
         case 'd':
            if (NextArg()!=STRINGARG)
               HError(1,"Hand label file extension expected");
            ilabExt = GetStrArg(); break;
         case 'N':
            if (NextArg()!=STRINGARG)
               HError(1,"Output hmm file directory expected");
            outHmmDir = GetStrArg(); break;
         case 'o':
            bjtScale = GetChkedFlt(0.0,100.0,s); 
            break;
         case 'p':
            propPen = log(GetChkedFlt(1.0E-20,1.0E20,s)); 
            fixedPen = TRUE; break;
         case 's':
            bigScale = GetChkedFlt(0.0,100.0,s); 
            break;
         case 't':
            pruning = GetChkedFlt(0,1.0E20,s); 
            isPruning = TRUE; break;
         case 'T':
            trace = GetChkedInt(0,4,s); break;
         case 'n':
            NBest = GetChkedInt(1,MAX_N_BEST,s); break;
         case 'w':
            w = GetChkedFlt(-10E+99,10E+99,s);
	    wepsilon = w;
            break;
         case 'm':
            m = GetChkedFlt(-10E+99,10E+99,s); 
	    mepsilon = m;
            break;
         case 'v':
            v = GetChkedFlt(-10E+99,10E+99,s); 
	    vepsilon = v;
            break;
         case 'E':
            Tot_epochs = GetChkedInt(0,100,s); 
	    break;
         case 'R':
            R = GetChkedInt(0,100,s); 
	    break;
         case 'a':
            alpha = GetChkedFlt(-10E+99,10E+99,s); 
            break;
         case 'f':
            if (NextArg()!=STRINGARG)
               HError(1,"Loss function type expected");
            LossFunctionType = Str2LossFunct(GetStrArg()); break;
	 case 'U':
	    UF = GetChkedFlt(1,1000,s);
	    break;
         case 'u':
            SetuFlags();
            break;
         default:
            HError(1,"Unknown switch %s",s);
      }
   } 
   
   if (NextArg()!=STRINGARG)
      HError(1,"hmm list file name expected");
   listfn = GetStrArg();
   if (NextArg()!=STRINGARG)
      HError(1,"network file name expected");
   netfn = GetStrArg();

   i=0;
   do {
      if (NextArg()!=STRINGARG)
         HError(1,"training data file name expected");
      datafn = GetStrArg();
      files[i] = (char *)malloc(strlen(datafn));
      strcpy(files[i],datafn);
      i++;
   } while (NumArgs()>0);

   N = i;
   if (UF == 0) UF = N;
   PrintArgs();
   printf("%d files loaded for input\n",N);

   Initialise(listfn,netfn,bigfn);
   fill_array(filenum, N);
   ITRNS = Tot_epochs * N;
   for (epoch = 0; epoch < Tot_epochs; epoch++){
     jumble(filenum,N);
     NumCorrect = N;
     printf("Epoch: %d; Epsilon  values: wts: %f, means %f, vars %f\n",epoch+1,wepsilon,
	    mepsilon,vepsilon);
     for (j = 0;j<N; j++){
       MatchFile(files[filenum[j]]);
       itrns++;
       if (((j+1) % UF) == 0){
	 if (trace >= 0)
	   /*printf("now updating HMMs, token %d  Avg. Loss %f, means epsilon: %f\n",  j, AvgLoss/UF, mepsilon);*/
	 ScaleUpdates();
	 GPDUpdate();
	 AvgLoss = 0.0;
       }
       wepsilon = w * (1- (float)(itrns)/(float)ITRNS);
       mepsilon = m * (1- (float)(itrns)/(float)ITRNS);
       vepsilon = v * (1- (float)(itrns)/(float)ITRNS);
     }
     if (((j) % UF) != 0){
       if (trace > 0) printf("now updating HMMs, token %d\n",j);
       ScaleUpdates();
       GPDUpdate();
       AvgLoss = 0.0;
     }
     printf(" Total % Correct = %f\n",(float)NumCorrect/(float)N);
     if (epoch > 0)
       if ((epoch % R) == 0){
	 /*system("recog.train");
	 system("recog.test");*/
       }
     if(N>0)
       SaveDefs(HMMdefList, HMMidList, outHmmDir, hmmExt, H);
     else HError(0,"No data files to update the HMM's with");
   }
 }

/* -------------------- Bigram Handling Routines -------------- */

/* LoadBigram: load a bigram from given file, note that row 
              and col ordering is changed to map onto HMMidlist */
void LoadBigram(char *fname)
{
   FILE *f;
   char buf[132];
   LabId id;
   int p,i,j,*map;
   char *checkList;
   Vector v,tmp;
   float x;
   
   bigram = CreateMatrix(P,P);
   tmp = CreateVector(P);
   if ((map=(int *)malloc(P*sizeof(int))) == NULL)
      HError(2,"Cant create bigram map array");
   --map;
   checkList = NewString(P);     /* not chars really */
   for (p=1;p<=P;p++)
      checkList[p] = '\0';
   ZeroMatrix(bigram);
   if ((f = fopen(fname,"r")) == NULL)
      HError(2,"Cannot open bigram file %s",fname);
   j = 0;
   while (ReadLabel(f, buf)) {      /* reorder rows during read in */
      if ((id = GetLabId(buf,FALSE))==NULL)
         HError(2,"Unknown label %s in bigram %s",buf,fname);
      p = id->aux; map[++j] = p;
      if (p < 1  ||  p > P)
         HError(2,"Label %s from bigram %s not in id List",buf,fname);
      if (checkList[p] == '\1')
         HError(2,"Label %s in bigram %s doubly defined",buf,fname);
      v = bigram[p];
      ReadVector(f,v);
      for (i=1;i<=P;i++) {
         x = v[i];
         v[i] = (x<MINLARG)? LZERO : log(x) * bigScale;
      }
      checkList[p] = '\1';
   }
   for (p=1;p<=P;p++)      /* check all rows defined */
      if (checkList[p] == '\0')
         HError(2,"No definition for label %s in bigram %s",
            HMMidList[p]->name,fname);
   for (i=1;i<=P;i++) {          /* reorder columns */
      v = bigram[i];
      for (j=1;j<=P;j++)
         tmp[map[j]] = v[j];
      bigram[i] = tmp; tmp = v;
   }
   hasBigram = TRUE;
   free(map+1);
   FreeVector(tmp);
}

/*------------------Random array generators --AD --------------------*/

void fill_array(arr, size)
int arr[], size;
{
int i;
for (i = 0; i< size; i++)
  arr[i] = i;
}

void jumble(arr, size)
int arr[], size;
{
int i, p1, p2, temp;

for (i = 0; i < size; i++){
  p1 = RandomValue()*size;
  p2 = RandomValue()*size;
  temp = arr[p1];
  arr[p1] = arr[p2];
  arr[p2] = temp;
}

}


/* ------------------ Phone Link Record Management ----------------- */

#define PLRTRWIDTH 4

static PLR *freeHd=NULL, *freeTl=NULL;
static PLR *usedHd=NULL, *usedTl=NULL;
static numSinceGC=0;                   /* num PLRs allocated since last GC */
static PLR *plrtrbuf[PLRTRWIDTH];      /* trace of best tokens */
static int trbufUsed;

/* InitPLRs: Initialise pool of PLRs */
void InitPLRs()
{
   if (usedHd!=NULL) {
      if (freeHd==NULL) {
         freeHd=usedHd; 
         freeTl=usedTl;
      } else {
         freeTl->chain = usedHd;
         freeTl = usedTl;
      }
   }
   usedHd = usedTl = NULL;
   numSinceGC = 0;

}
      
/* FreePLRs: DR Free the entire pool of PLR's */
void FreePLRs()
{
 PLR *p;

  while (usedHd!=NULL) {
    p = usedHd->chain;   
    free(usedHd);    
    usedHd = p;
  }

  usedHd=usedTl=NULL;

  while (freeHd!=NULL) {
    p = freeHd->chain;
    free(freeHd);
    freeHd = p;
  }

  freeHd=freeTl=NULL;
  numSinceGC = 0;

}


/* DR - CreateILR: Returns a pointer to a new ILR */
ILR *CreateILR(void)
{
   ILR *p;

   if ((p = (ILR *)malloc(sizeof(ILR))) == NULL)  
      HError(2,"Cannot create ILR");   

   p->gchain = ilrHD;
   ilrHD = p;

   return p;
}   

/* DR - FreeILRs: Frees a chain of ILR records */
void FreeILRs()
{
 ILR *next;

 while (ilrHD->gchain != NULL) {
   next = ilrHD->gchain;
   free(ilrHD); 
   ilrHD = next;
 }

 free(ilrHD);
 ilrHD = NULL;

}

/* DR - CreateSLR: Returns a pointer to a new SLR */
SLR *CreateSLR(void)
{
   SLR *p;

   if ((p = (SLR *)malloc(sizeof(SLR))) == NULL)  
      HError(2,"Cannot create SLR");   

   p->chain = slrHD;
   slrHD = p;

   return p;
}   

/* DR - FreeSLRs: Frees a chain of SLR records */
void FreeSLRs()
{
 SLR *next;

 while (slrHD->chain != NULL) {
   next = slrHD->chain;
   free(slrHD); 
   slrHD = next;
 }

 free(slrHD);
 slrHD = NULL;

}

/* CreatePLR: Returns a pointer to a new PLR */
PLR *CreatePLR(void)
{
   PLR *p;
   
   if (freeHd==NULL) {
      if ((p = (PLR *)malloc(sizeof(PLR))) == NULL)
         HError(2,"Cannot create PLR");
   } else {
      p = freeHd;
      if (freeHd==freeTl)
         freeHd = freeTl = NULL;
      else 
         freeHd = freeHd->chain;
   }
   if (usedHd==NULL) 
      usedHd = usedTl = p;
   else {
      usedTl->chain = p;
      usedTl = p;
   }
   p->chain = NULL;
   ++numSinceGC;
   return p;
}

/* GarbageCollectPLRs:  Garbage Collect unlinked PLR's */
void GarbageCollectPLRs(void)
{
   int i,j,plrcollected,plrused;
   PLR *plr,*next;
   PhoneInstance *pi;
   Token t;

   if (GCMARGIN <= numSinceGC) {
      plrused = plrcollected = 0;

      for (plr=usedHd; plr!=NULL; plr=plr->chain)
         plr->linked = FALSE;

      for (i=0;i<nActive;i++) {
         pi = active[i];
         for (j=1; j<=pi->N; j++) {
            t = pi->state[j];
            if (t.inst != NULL) 
               for (plr=t.link; (plr!=NULL)&&(!plr->linked);
                                           plr=plr->tok.link)
                  plr->linked=TRUE;
         }
      }

      plr = usedHd; usedHd = usedTl = NULL;
      while (plr!=NULL) {
         next = plr->chain;
         if (plr->linked) {
            ++plrused;
            plr->chain = usedHd;
            usedHd = plr;
            if (NULL==usedTl)
               usedTl = plr;
         } else {
            ++plrcollected;
            plr->chain = freeHd;
            freeHd = plr;
            if (NULL==freeTl)
               freeTl=plr; 
         }
         plr=next;
      }

      if (trace>1) {
         printf("GC: NumSinceGC=%d Collected=%d Used=%d\n",
                                          numSinceGC,plrcollected,plrused); 
         fflush(stdout);
      }
      numSinceGC = 0;
      if (plrused>maxPLRUsed) maxPLRUsed = plrused;
   }
}

/* ClearTrBuf: clear the PLR trace buffer */
void ClearTrBuf(void)
{
   trbufUsed = 0;
}

/* Add2TrBuf: add the given PLR to the trace buffer */
void Add2TrBuf(PLR *p)
{
   int i;
   PLR *q;
   LogReal x;
   
   x = p->tok.logprob;
   if (trbufUsed == PLRTRWIDTH) {
      q = plrtrbuf[trbufUsed-1];
      if (q->tok.logprob<x)
         --trbufUsed;
      else 
         return;
   }
   for (i=trbufUsed;i>0;i--)
      if (plrtrbuf[i-1]->tok.logprob < x)
         plrtrbuf[i] = plrtrbuf[i-1];
      else
         break;
   plrtrbuf[i] = p;
   ++trbufUsed;
}

/* PrintTrBuf: print out the contents of the trace buffer */
void PrintTrBuf(int t)
{
   int i;
   Token tok;
   
   printf("PLRS: time %d\n",t);
   for (i=0;i<trbufUsed;i++){
      tok = plrtrbuf[i]->tok;
      printf("id %2s, prob %f",tok.inst->def->id->name,tok.logprob);
      if(tok.link!=NULL)
         printf(" back to frame %d",tok.link->time);
      printf("\n");
   }
   fflush(stdout);
}

/* -------------------- Housekeeping for Pruning ----------------- */

/* CreateActiveList: create a list of pointers to active phone instances */
void CreateActiveList(void)
{
   if ((active=(PhoneInstance **)malloc(NPI*sizeof(PhoneInstance *))) == NULL)
      HError(2,"Cannot create active phone instance list");
}

/* FreeActiveList: DR free the active list  */
void FreeActiveList(void)
{
 free((PhoneInstance **)active);
}

/* ReAllocActiveList: allow the active list to grow in size */
void ReAllocActiveList(void)
{
 PhoneInstance **newactive;

 if ((newactive =(PhoneInstance **)realloc(active, NPI*sizeof(PhoneInstance *))) == NULL)
   HError(2,"Cannot realloc active phone instance list");

 active = newactive;
}


/* DeActivateAll: deactivate all active phone instances */
void DeActivateAll(void)
{
   int i;
   
   for (i=0;i<nActive;i++)
      active[i]->isActive = FALSE; 
   nActive = 0;
}

/* MakeActive: make given pi active */
void MakeActive(PhoneInstance *pi)
{
   int i;
   Token nullToken = {NULL,LZERO,NULL,NULL,NULL}; /* DR - NULL,NULL added */ 

   if (!pi->isActive){
      pi->isActive = TRUE;
      active[nActive++] = pi;
      for (i=1;i<=pi->N;i++)
         pi->state[i] = nullToken;
   }
}

/* CheckActive: scan all active phones and kill any with no paths in beam */
void CheckActive(void)
{
   int i,j,nAOld;
   PhoneInstance *pi;
   LogReal bestP,threshold;
   Token t;
   
   nAOld = nActive;
   threshold = maxLogP - pruning;
   for (i=0;i<nActive;i++) {
      SPINBALL
      pi = active[i];
      bestP = LZERO;          /* find best path in this pi */
      for (j=2; j<=pi->N; j++){
         t = pi->state[j];
         if (t.inst != NULL && t.logprob>bestP)
            bestP = t.logprob;
      }
      if (bestP<threshold){   /* if below thresh kill it */
         active[i] = NULL;
         pi->isActive = FALSE;
      }
   }
   for (i=0,j=0;i<nActive;i++)   /* remove marked pi's */
      if (active[i] != NULL) {
         active[j++] = active[i];
      }
   nActive = j;
   if (trace>1) {
      printf("   Active Beam Width %d -> %d\n",nAOld,nActive);
      fflush(stdout);
   }
}

/* ------------------------ Network Management ------------------- */

/* IdMap: set-up id->aux to inverse map of HMMidList indices.  If
   hmms are context sensitive then the mapping goes from cs models
   to monophones.  Only used by bigram */
void IdMap(void)
{
   int p,h;
   Boolean isCS=FALSE;
   char *s,buf[100];
   LabId id;
   
   for (h=1;h<=H;h++){
      s = HMMidList[h]->name;
      if (strchr(s,'-') != NULL || strchr(s,'+') != NULL){
         isCS = TRUE;
         break;
      }
   }
   
   if (!isCS) {
      for (h=1;h<=H;h++)
         HMMidList[h]->aux = h;
      P = H;
   } else {
      for (h=1;h<=H;h++){
         strcpy(buf,HMMidList[h]->name);
         TriStrip(buf);
         id = GetLabId(buf,TRUE);
         p = id->aux;
         if (p == 0) id->aux = p = ++P;
         HMMidList[h]->aux = p;
      }
   }
}

/* LoadHMM: load specified hmm, adding hmmExt if set */
void vLoadHMM(char *name, HMMDef *def)
{
   char fn[255];
   int len;
   
   if (hmmDir==NULL)
      fn[0]='\0';
   else {
      strcpy(fn,hmmDir);
      len = strlen(fn);
      if (fn[len-1]!=PATHCHAR) {
         fn[len]=PATHCHAR;
         fn[++len]='\0';
      }
   }
   strcat(fn,name);
   if (hmmExt!=NULL) {
      strcat(fn,".");
      strcat(fn,hmmExt);
   }
   SPINBALL
   HOpen(fn,def);
   if (hmmP.vSize==0) {
      hmmP.vSize = def->vecSize;
      hmmP.M = def->numMixtures;
      hmmP.CK = def->ckind; hmmP.SK = def->pkind;
      hmmP.isTied = def->tstat != FreeT;
   } else {
      if (hmmP.vSize!=def->vecSize)
         HError(2,"HMMs have incompatible vecSizes");
      if (hmmP.M!=def->numMixtures)
         HError(2,"HMMs have incompatible numMixtures");
      if (hmmP.CK!=def->ckind)
         HError(2,"HMMs have different covariance kinds");
      if (hmmP.SK!=def->pkind)
         HError(2,"HMMs have different sample kinds");
      if (hmmP.isTied != (def->tstat != FreeT))
         HError(2,"Mixed Free and Tied Distributions");
   }
   if (hmmP.maxStates<def->numStates)
      hmmP.maxStates = def->numStates;
}

/* LoadHMMDefs: create defList and load it */
void vLoadHMMDefs(void)
{
   int h,n;
   Definition *d;
   
   if ((HMMdefList = (Definition *)malloc(H*sizeof(Definition))) == NULL)
      HError(2,"Unable to create hmm def array");
   --HMMdefList;
   for (h=1;h<=H;h++){
      d = HMMdefList+h;
      vLoadHMM(HMMidList[h]->name,&(d->hmmdef));
      if(trace > 1) {
       printf("Loaded phoneme HMM %s\n", HMMidList[h]->name);
       fflush(stdout);
      }
      n = d->hmmdef.numStates-2;
      if ((d->bjot = (float *)malloc(n*sizeof(float))) == NULL)
         HError(2,"Cannot create bjot array");
      d->bjot -= 2;     /* index is 2.. */
      d->id = HMMidList[h];
   }
}

HMMDef * vMallocHMMDefList(Definition * HMMdefList, int H)
{
   int h;
   HMMDef *defList, *d;
   
   if ((defList = (HMMDef *)malloc((H+1)*sizeof(HMMDef))) == NULL)
      HError(2,"Unable to create hmm def array");
   --defList;
   for (h=1;h<=H;h++) {
     d = &(HMMdefList[h].hmmdef);
     HCreate(d->numStates, d->numMixtures, d->vecSize, d->pkind, d->ckind, d->dkind, d->tstat,
              defList+h); 
   }

   return (defList);
}

/* SaveDefs: save the list of HMM definitions in HMMdefList to the directory hmmDir */
void SaveDefs(Definition *HMMdefList, LabId *HMMidList, char *hmmDir, char *hmmExt, int H)
{
  int i;
  char hmmfn[256];
  
  for(i=1;i<=H;i++) {
    MakeFN(HMMidList[i]->name, hmmDir,hmmExt,hmmfn);
    if (trace>2) printf("Writing out HMM %s\n", hmmfn);
    HSave(hmmfn, &(HMMdefList[i].hmmdef));
  }
}

/* AttachDefinition: look for hmm def with given id and attach it to pi */
void AttachDefinition(PhoneInstance *pi, LabId id, Boolean valid /* DR */)
{
   int h;
   for (h=1;h<=H;h++)
      if (!strcmp(HMMidList[h]->name,id->name)){ /* DR */
         pi->def = HMMdefList+h;
         pi->hlpath = valid; /* DR */
         return;
      }
   HError(2,"Cannot find def %s to attach",id->name);
}

/* Nulllastplrt: zero lastplrt in all PhoneInsts */
void Nulllastplrt(void)
{
   Link p;
   PhoneInstance *pi;
   
   p = theNet.chain;
   while (p!=NULL) {
      if (p->modelName != enterId && p->modelName != exitId ) {
         pi = (PhoneInstance *) p->user;
         pi->lastplrt = 0;
      }
      p = p->chain;
   }
}

/* AttachPhoneInsts: to each node in theNet */
void AttachPhoneInsts(void)
{
   Link p;
   PhoneInstance *pi;
   
   p = theNet.chain;
   while (p!=NULL) {
      SPINBALL
      if (p->modelName != enterId && p->modelName != exitId ) {
         if ((pi =(PhoneInstance *)malloc(sizeof(PhoneInstance))) == NULL)
            HError(2,"Unable to create phone instance");
         ++NPI;
         AttachDefinition(pi,p->modelName, FALSE); /* DR "FALSE" added */
         pi->N = pi->def->hmmdef.numStates;
         pi->isActive = FALSE;
         if ((pi->state = (Token *)malloc(pi->N*sizeof(Token))) == NULL)
            HError(2,"Unable to create state array for model %d",pi->N);
         pi->state -= 1;
         (PhoneInstance *)(p->user) = pi; 
         pi->netHook = p;
      }
      p = p->chain;
   }
   Nulllastplrt();
   if (trace>0){
      printf("  %d Instances Created of %d HMMs\n",NPI,H);
      fflush(stdout);
   }
}

/* ShowPath: print a brief n char resumee of path */
void ShowPath(PLR *plr, int n)
{
   char buf[132];
   int i,count;
   
   
   buf[0] = '\0'; count = 0;
   while (plr != NULL && count < n) {
      strcat(buf,plr->tok.inst->def->id->name);
      plr = plr->tok.link;
      if (plr != NULL) strcat(buf,"/");
      count = strlen(buf);
   }
   for (i=count;i<n;i++) buf[i]=' ';
   buf[n] = '\0'; 
   printf("%s",buf);
}

/* ShowInst: print contents of given phone instance */
void ShowInst(PhoneInstance *pi)
{
   int i;
   LogReal x;
   PLR *plr;
   PhoneInstance *ppi;
   
   printf("** Inst %6s[%2d]\n  ",pi->def->id->name,(int)pi % 100);
   for (i=1; i<=pi->N; i++)
      printf("%10d",i);
   printf("\n  ");
   for (i=1; i<=pi->N; i++) {
      ppi = pi->state[i].inst;
      if (ppi==NULL)
         printf("%10s","NULL");
      else
         printf("%6s[%2d]",ppi->def->id->name,(int)ppi % 100);
   }
   printf("\n  ");

   for (i=1; i<=pi->N; i++)
      if (pi->state[i].inst == NULL)
         printf("%10s","NULL");
      else {
         x = pi->state[i].logprob;
         if (x>LSMALL)
            printf(" %9.2f",x);
         else
            printf("%10s","---");
      }
   printf("\n  ");

   for (i=1; i<=pi->N; i++) {
      plr = pi->state[i].link;
      if (plr==NULL)
         printf("%10s","NULL");
      else
         ShowPath(plr,10);
   }
   printf("\n");
}

/* ------------------------ Initialisation ----------------------- */

/* Initialise: initialise the global data structures */
void Initialise(char * listfn, char * netfn, char * bigfn)
{
   HMMidList = LoadIdList(listfn,&H);
   IdMap();
   if (trace>1) {
      printf("Id Map Built: ");
      if (P==H)
         printf(" %d monophones\n",H);
      else
         printf(" %d c-s phones -> %d monophones\n",H,P);
      fflush(stdout);
   }
   if (trace > 2)
     printf("loading HMM definitions\n");
   vLoadHMMDefs();
   if (trace > 2)
     printf("creating update list\n");
   GPDupdateList = vMallocHMMDefList(HMMdefList,H);
   if (trace > 2)
     printf("creating vectors and matrices\n");
   ot = CreateVector(hmmP.vSize);
   um = CreateVector(hmmP.vSize);
   if (hmmP.CK == FULLC)
     ucov.inv = CreateMatrix(hmmP.vSize,hmmP.vSize);
   else
     ucov.var = CreateVector(hmmP.vSize);
   if (bigfn != NULL)
      LoadBigram(bigfn);
   enterId = GetLabId("ENTER",TRUE);
   exitId = GetLabId("EXIT",TRUE);  
   theNet = ParseNetwork(netfn);    /* build the theNet */
   AttachPhoneInsts();
   if ((tBuf=(Token *)malloc(hmmP.maxStates*sizeof(Token))) == NULL)
      HError(2,"Cannot create token buffer");
   --tBuf;
   if (trace>1) {
      printf("Network built: %d Models, %d Max States in any Model\n",
             H,hmmP.maxStates);
      fflush(stdout);
   }
}
   
/* --------------------------- Recognition ----------------------- */

/* Setbjot: precompute all 'local distances' needed by active pi's */
void Setbjot(Vector ot)
{
   Definition *d;
   int i,j,h,N;
   
   if (hmmP.isTied) PreCompute(ot);
   for (h=1;h<=H;h++)
      HMMdefList[h].bjLoaded = FALSE;
   for (j=0;j<nActive;j++){
      d = active[j]->def;
      if (!d->bjLoaded) {
         N = d->hmmdef.numStates;
         for (i=2;i<N;i++)
            d->bjot[i] = OutP(ot,&(d->hmmdef),i) * bjtScale;
         d->bjLoaded = TRUE;
      }
   }

}

/* InitModels: Activate models connected to ENTRY node and inject
               a start token */
void InitModels(void)
{
   Link p;
   int i;
   Token startToken = {NULL,0,NULL,NULL,NULL}; /* DR - "NULL,NULL" added */
   LinkSet *l;
   PhoneInstance *pi;
   
   SPINBALL
  /* DeActivateAll();  moved to MatchFile */
   Nulllastplrt();
   p = theNet.entryNode; l = p->succ;
   for (i=0; i<p->numSucc;i++) {
      pi = (PhoneInstance *)(*l)[i]->user;
      if (pi==NULL)
         HError(2,"Bad Syntax Net - non-unit connected to entry node");
      MakeActive(pi);
      startToken.inst = pi;
      pi->state[1] = startToken;
   }
   if (trace>1) {
      printf("%d Models Initially Active\n",nActive);
      fflush(stdout);
   }
   InitPLRs();
}

/* StepModel: step the given model by one time frame */
void StepModel(PhoneInstance *pi, int time /* DR */)
{

   int i,j,besti;
   LogReal best,x,a;
   Token t,nullToken = {NULL,LZERO,NULL,NULL,NULL}; /* DR - added "NULL,NULL" */
   Definition *d;
   short record[100];
   ILR *ilr;
   SLR *slr;

   for (i=0;i<100;i++) record[i] = FALSE;
   
   SPINBALL
   d = pi->def;
   for (j=2;j<pi->N;j++){     /* Propagate internal tokens */
      best = LZERO; besti = 0;
      for (i=1;i<pi->N;i++) {
         if (pi->state[i].inst != NULL) {
            a = d->hmmdef.transP[i][j];
            if (a>LSMALL) {
               x =  pi->state[i].logprob + a;
               if (x>best) {
                  best = x; besti = i;
               }
            }
         }
      }
      /* DR - we've now found the best state to get to j from */
      if (besti>0) {
         tBuf[j] = pi->state[besti];           /* propogate the token into tmp storage */
         x = best + d->bjot[j];
         if (x>maxLogP) maxLogP = x;          /* record global max for pruning */
         tBuf[j].logprob = x;                 /* update the logprob slot of the token*/

         /* each token just trails a huge tail of slrs behind it */
           slr = CreateSLR(); 
           slr->name = pi->def->id->name;
           slr->time = time;     
           slr->state = j;
           slr->bjot = d->bjot[j];
           slr->link = pi->state[besti].slrlink;
           tBuf[j].slrlink = slr;
     
           if (besti != j) {                      /* if the token has propogated from another state*/
             ilr = CreateILR();                   /* create a new internal link record */
             ilr->time = time;                    /* record the time of creation */
             ilr->state = j;                      /* record the state */
             ilr->chain = pi->state[besti].ilink; /* point back to the previous ilr */
             tBuf[j].ilink = ilr;                 /* point the token at this new ilr */
           }
      } else
         tBuf[j] = nullToken;
   }


   for (i=2;i<pi->N;i++)      /* Copy tokens into state slots */
      pi->state[i] = tBuf[i];

   best = LZERO; besti = 0;      /* Find best exit token */
   for (i=2;i<pi->N;i++) {
      if (pi->state[i].inst != NULL) {
         a = d->hmmdef.transP[i][pi->N];
         if (a>LSMALL) {
            x =  pi->state[i].logprob + a;
            if (x>best) {
               best = x; besti = i;
            }
         }
      }
   }

   if (besti>0){
      t = pi->state[besti];
      t.logprob = best;
      if (best>maxLogP) maxLogP = best;   /* record global max for pruning */

        t.slrlink = pi->state[besti].slrlink;
          
        if (besti != pi->N) {                   /* if the token has propogated */
           ilr = CreateILR();                   /* create a new internal link record */
           ilr->time = time;                    /* record the time of creation */
           ilr->state = pi->N;                  /* record the state */
           ilr->chain = pi->state[besti].ilink; /* point back to the previous ilr */
           t.ilink = ilr;                       /* point the token at this new ilr */
        }
      pi->state[pi->N] = t;
   } else 
      pi->state[pi->N] = nullToken;
   if (trace==3)  
      ShowInst(pi);
}

/* StepModels: step all models by one time frame */
void StepModels(int time /* DR*/)
{
   int i;
   
   maxLogP = LZERO;
   for (i=0;i<nActive;i++)
      StepModel(active[i], time /* DR */);
}

/* NullEntryTokens: Set all entry tokens to null ie zero prob */
void  NullEntryTokens(void)
{
   int i;
   Token nullToken = {NULL,LZERO,NULL,NULL,NULL}; /* DR - "NULL,NULL" added */
   
   for (i=0;i<nActive;i++)
      active[i]->state[1] = nullToken;
}

/* PropagateExitTokens: propagate exit tokens  */
void PropagateExitTokens(void)
{
   Link p,q,*a;
   int i,j;
   Token t;
   PhoneInstance *pi;
   LogReal x,threshold;
   int fromIdx;
   ILR *ilr; /* DR */
    
   SPINBALL
   threshold = (isPruning) ? maxLogP-pruning : LZERO;
   for (i=0;i<nActive;i++) {
      pi = active[i]; p = pi->netHook;
      fromIdx = pi->def->id->aux;
      t = pi->state[pi->N];


      if (t.logprob>threshold) {
         if (fixedPen)
            t.logprob += propPen;   
         else if (!hasBigram)
               t.logprob -= log(p->numSucc);
         a = (Link *)p->succ;
         for (j=1;j<=p->numSucc;j++) {
            q = *a++;
            if (q->user != NULL) {
               pi = (PhoneInstance *)q->user;
               if (!pi->isActive) 
                  MakeActive(pi);
               x = t.logprob;
               if (hasBigram)
                  x += bigram[fromIdx][pi->def->id->aux];
               if (x>LSMALL && pi->state[1].logprob < x) {
                  pi->state[1] = t;
                  pi->state[1].logprob = x;
                  /* start DR */
                  /* check that we are in the correct label */
                    ilr = CreateILR();
                    ilr->state = 1; 
                    ilr->time = pi->state[1].ilink->time+1;           
                    ilr->chain =  pi->state[1].ilink;
                    pi->state[1].ilink = ilr;
                    pi->state[1].slrlink = t.slrlink;
                    /* end DR */
               }
            }
         }
      }
   }
}


/* RecordDecisions: create PLR's for all distinct entry nodes */
void RecordDecisions(int t)
{
   int i;
   PhoneInstance *pi;
   PLR *plr;
   Token tok;
   
   if (trace>1) ClearTrBuf();
   for (i=0;i<nActive;i++) {
      SPINBALL
      pi = active[i];
      tok = pi->state[1];
      if (tok.logprob>LSMALL) {
         if (t != tok.inst->lastplrt) {
            plr = CreatePLR(); /* Creates an empty PLR */
            plr->tok = tok;
            plr->time = t;
            tok.inst->lastplrt = t;
            tok.inst->lastplr = plr;
            if (trace>1) {
               if (hasBigram)
                  plr->tok.logprob -= 
                     bigram[tok.inst->def->id->aux][pi->def->id->aux];
               Add2TrBuf(plr);
            }
         } else
            plr = tok.inst->lastplr;
         pi->state[1].link = plr;
         pi->state[1].inst = pi;
      }
   }
   if (trace>2) PrintTrBuf(t);
}

/*-------------------- extracted from HLabel--AD ------------------------*/


/* CreateLabList: create and return a lablist with one chunk */
static LabList *vCreateLabList()
{
   LabList *l;
   
   if ((l=(LabList *)malloc(sizeof(LabList))) == NULL)
      HError(2,"Cannot create LabList header");
   l->nLabels = 0; l->next = NULL;
   if ((l->labels = (Label *)malloc(sizeof(Label)*CHUNKSIZE)) == NULL)
      HError(2,"Cannot create LabList body");
   numChunks = 1; 
   return l;
}

/* AddLabel: append given label to given lab list */
static void vAddLabel(LabList *l, Label x)
{
  int newSize;
  
  if (l->nLabels == numChunks * CHUNKSIZE) {
    ++numChunks;      /* grow label list */
    newSize =  sizeof(Label)*CHUNKSIZE*numChunks;
    if ((l->labels = (Label *)realloc(l->labels,newSize)) == NULL)
      HError(2,"Cannot grow lab list");
  }
  l->labels[l->nLabels++] = x;
}

/* ShrinkLabList: remove unused label slots from current lab list */
static void vShrinkLabList(LabList *l)
{
   int newSize;
   
   if (l->nLabels == numChunks*CHUNKSIZE) return;
   newSize = sizeof(Label) * l->nLabels;
   if ((l->labels = (Label *)realloc(l->labels,newSize)) == NULL)
      HError(2,"Cannot shrink label list");
}

/* AppendLabList: append lab list l to transcription t */
static void vAppendLabList(LabList *l, Transcription t)
{
   LabList *p = t;
   
   while(p->next != NULL)
      p = p->next;
   p->next = l; l->next = NULL;
}

/* ------------------------ Traceback and Output ---------------------- */

/* OutLine: print label and start and end points */
void OutLine(FILE * f, long st, long en, LabId lab)
{
   SPINBALL
   st *= sampPeriod; en *= sampPeriod;
   fprintf(f,"%10d  %10d   %3s\n",st,en,lab->name);
   if (trace>1){
      printf("%10d  %10d   %3s\n",st,en,lab->name);
      fflush(stdout);
   }
}

/* OutIPath: print out the internal path beginning at ilr

void OutIPath(FILE *f, ILR *ilr, int n)
{
 ilr = ilr->chain;
  
 if (n>2) {
   n--;
   OutIPath(f, ilr, n);
 }

 fprintf (f, "  [%2d] %2d\n", ilr->state, ilr->time * sampPeriod);
}   
*/

/* OutSegment: print segment of path upto p to file f */
void OutSegment(FILE *f, /*DR*/ FILE *ipathfp, PLR *p)
{
   int st;
   int i; /* DR */
   ILR *ilr; /* DR */
     
   if (p!=NULL) {
      OutSegment(f, /* DR */ ipathfp, p->tok.link);
      if (p->tok.link != NULL) 
         st = p->tok.link->time + 1;
      else
         st = 1;
      OutLine(f,st,p->time,p->tok.inst->def->id);
   }
}



static NameCell *NewCell(char *name)  /* AD -- pulled from HLabel */
{
   char *s;
   NameCell *p;

   if ((s=(char *)malloc(strlen(name)+1)) == NULL)
      HError(2,"Cannot malloc space for Label Name in NewCell");
   strcpy(s,name);
   if ((p=(NameCell *)malloc(sizeof(NameCell))) == NULL)
      HError(2,"Cannot malloc space for NameCell");
   p->name = s; p->next = NULL; p->aux = 0;
   return p;
}

/* Outslr: DR prints out the slr chain associated with a given token */
void Outslr(FILE *fp, SLR *slr, LabList *tr)
{
  SLR *p;
  Label x;

  x.labid = NewCell(slr->name);
  while (slr != NULL) {
    p = slr;
    fprintf (fp, "%4d   %4d  %6s  %f\n", p->time, p->state, p->name, p->bjot); 
    x.labid->name = p->name;
    x.start = p->time;
    x.end =   p->state;
    x.score = p->bjot;
    vAddLabel(tr, x);
    slr = p->link;
  }
  vShrinkLabList(tr);
}

/* OutputResults: trace back best predecessor of theNet exit */
/* DR - modified to produce the NBest outputs */
void OutputResults(FILE *f, /* DR */ FILE *ipathfp,  char *fn, int t, Transcription tr, LabList *path)
{
   Link p,q,*a;
   PhoneInstance *pi, *bestpi;
   LogReal best[MAX_N_BEST];
   int i,st,n;
   ILR *ilr; /* DR */

   if (NBest > theNet.exitNode->numPred) {
     HError(0,"NBest value (%d) is greater than number of nodes jointed to network exit (5d)",
	    NBest,theNet.exitNode->numPred);
     HError(0,"NBest is thus being reduced to %d",theNet.exitNode->numPred);
   }
   
   for (n=1;n<=NBest;n++) best[n]=0;

   for (n=1; n<=NBest; n++) {
     bestpi = NULL;
     p = theNet.exitNode;
     a = (Link *)p->pred;
     for (i=1;i<=p->numPred;i++) {
       q = *a++;
       if (q->user != NULL) {
	 pi = (PhoneInstance *)q->user;
	 if (pi->isActive && pi->state[pi->N].inst != NULL &&
	     pi->state[pi->N].logprob > best[n]){ 
	   if (n==1) {
	     best[n] = pi->state[pi->N].logprob;
	     bestpi = pi;
	   }
	   else {
	     if (pi->state[pi->N].logprob < best[n-1]) {
	       best[n] = pi->state[pi->N].logprob;
	       bestpi = pi;
	     }
	   }
	 }
       }
       
     }
     tlgpr[n] = best[n];

     if (bestpi==NULL) {
       HError(0,"No tokens survived to end of input for %d'th parse%s",n,fn);
       return;
     }
     if (trace>1){
       printf("  MaxProb per frame for %d parse = %f\n",n, best[n]/(t+1));
       printf ("\n----The Viterbi path for the %d'st best parse is:---\n\n",n);
       fflush(stdout);
     }
     OutSegment(f, /* DR */ ipathfp, bestpi->state[bestpi->N].link); 
     if (bestpi->state[bestpi->N].link != NULL) 
       st = bestpi->state[bestpi->N].link->time + 1;
     else
       st = 1;
     OutLine(f,st,t,bestpi->def->id);
     
     /* start DR */
     Outslr(ipathfp, bestpi->state[bestpi->N].slrlink,path);
     /* end DR */
     
     if(n < NBest) {
       fprintf(f, "/\n"); /* label level separator */
       fprintf(ipathfp,"/\n");
     }
     if (trace>1)printf("num labels = %d\n",path->nLabels);
     vAppendLabList(path,tr);
     if (n != NBest){
       path->next = vCreateLabList();
       path = path->next;
     }
   } /* end for n */
 }

/* OutputHlResults: trace back the Hand Labelled Path */
void OutputHlResults(FILE *f, /* DR */ FILE *ipathfp,  char *fn, int t, LabList *path)
{
   Link p,q,*a;
   PhoneInstance *pi, *bestpi;
   LogReal best;
   int i,st,n;
   ILR *ilr; /* DR */
   
   bestpi = NULL;
   p = theNet.exitNode;
   a = (Link *)p->pred;
   for (i=1;i<=p->numPred;i++) {
     q = *a++;
     if (q->user != NULL) {
       pi = (PhoneInstance *)q->user;
       if(pi->hlpath == TRUE) {        
          best = pi->state[pi->N].logprob;
          bestpi = pi;
       }
     }
   }
   tlgpr[0] = best;

   if (bestpi==NULL) {
      HError(0,"No hand labelled path survived to end of input for parse%s",fn);
      return;
   }
   if (trace>1){
      printf("  MaxProb per frame for hand label parse = %f\n", best/(t+1));
      printf ("\n----The Viterbi path for the input labelled parse is:---\n\n");
      fflush(stdout);
   }
   OutSegment(f, /* DR */ ipathfp, bestpi->state[bestpi->N].link); 
   if (bestpi->state[bestpi->N].link != NULL) 
      st = bestpi->state[bestpi->N].link->time + 1;
    else
       st = 1;
   OutLine(f,st,t,bestpi->def->id);
   
   /* start DR */
   Outslr(ipathfp, bestpi->state[bestpi->N].slrlink, path); /* AD - path added */
   if(n < NBest) {    /*AD -- the hl path is now the first path */
     fprintf(f, "/\n"); /* label level separator */
     fprintf(ipathfp,"/\n");
   }
   /* end DR */
   
 }

/* CheckData: check given data file for consistency with HMM's */
void vCheckData(DataFile *data, char *fn)
{
   if (data->sampSize / sizeof(float) != hmmP.vSize)
      HError(2,"Incompatible Vec Size in file %s",fn);
   if (data->sampKind != hmmP.SK)
      HError(2,"Incompatible Sample Kind in file %s",fn);
}

#define LINKCHUNKSIZE 5

/* DR AddHlPathToNet: adds a path to the network corresponding to the hand labelled path */
void AddHlPathToNet(Transcription tr)
{

 int i;
 int numSuccMax = LINKCHUNKSIZE;
 int numPredMax = LINKCHUNKSIZE;
 Link a,b;
 PhoneInstance *pi;
 

 a = theNet.entryNode;
 for(i=0;i<tr->nLabels;i++) {
  b = CreateNode(tr->labels[i].labid, &(theNet.chain), numSuccMax, numPredMax);
  JoinNodes(a,b); 
  if((pi = (PhoneInstance *)malloc(sizeof(PhoneInstance)) ) == NULL)
    HError(2,"Unable to create phone instance");
  AttachDefinition(pi,b->modelName, TRUE);
  NPI += 1;
  pi->N = pi->def->hmmdef.numStates;
  pi->isActive = FALSE;
  if ((pi->state = (Token *)malloc(pi->N*sizeof(Token))) == NULL)
    HError(2,"Unable to create state array for model %d",pi->N);
  pi->state -= 1;
  pi->lastplrt=0;
 (PhoneInstance *)(b->user) = pi; 
  pi->netHook = b;

  a = b; 
 }

 b = theNet.exitNode;
 JoinNodes(a,b);

}


/* this unjoins two connected nodes (a->b) in the network */
void UnJoinNodes(Link a, Link b)
{

 if(NotJoined(a,b)) HError(1,"Cannot unjoin nodes if they are not connected");

  RemPredLink(a,b);
  RemSuccLink(b,a);
  ShrinkNode(a);
  ShrinkNode(b); 
} 

void FreeAttachedPI(Link a)
{
 PhoneInstance *pi;

 pi = (PhoneInstance *)(a->user);
 free (pi->state+1); 
 free(pi); 
 NPI -= 1;
}

/* DR removes the hand labelled path from the Network */

void DetachHlPathFromNet(Transcription tr)
{
 
 Link a,b;
 int i;

 a = theNet.chain;

 UnJoinNodes(a,theNet.exitNode);
 b = a;
 a = a->chain;

 for (i=1;i<tr->nLabels;i++) {
   UnJoinNodes(a,b);
   FreeAttachedPI(b);
   FreeNode(b);
   b = a;
   a = a->chain;
 }

 UnJoinNodes(theNet.entryNode,b);
 FreeAttachedPI(b);
 FreeNode(b);
 theNet.chain = a;

}


/*------------------------------------------------------------------------*/
/*  From here to Matchfile are the routines for GPD      -AD              */


/*--------------------------------------------------------------*/
/* These routines are for accessing Viterbi path information from 
   the label files. The routines were written because the path
   information is rather awkwardly squeezed into the label struture,
   so direct access would be rather confusing
*/

int GetTime(int t, Label *l) { return(l[t].start); }
int GetState(int t, Label *l) { return(l[t].end); } 
double Getlogbjt(int t, Label *l) { return(l[t].score); }
char *GetLabelModelName(int t, Label *l) { return(l[t].labid->name); }

int Name2HmmIndex(char *name, int H)
{
 int index=MODEL_NOT_FOUND;
 int h;

 for(h=1;h<=H;h++) 
   if(!strcmp(HMMidList[h]->name,name)) index = h;
 /* return an error is the HMM is not found */
 if (index == MODEL_NOT_FOUND) 
   HError(2,"HMM Model %s not found", name);
 else return(index);
}

/*---------------------------------------------------------------*/
/* The following routines compute dP/d(mean), and dP/d(cov) */

void DeltaProbDeltaMean(Vector m, Vector o, double wm, Covariance cov, Vector um, double pm)
{
 int i;
 double k;

 k = wm * pm;

 if (hmmP.CK == DIAGC) 
   for(i=1;i<=hmmP.vSize;i++)  {
    if(cov.var[i]==0.0) HError(1,"covariance element is zero");
    um[i] =  k * (o[i]-m[i]) / cov.var[i];
  }

 if (hmmP.CK == FULLC) /*not implemented yet*/;

}

void DeltaProbDeltaCov(double wm, double pm, Vector ot, Vector m, Covariance cov, Covariance ucov)
{
 int i;
 double k1,k2;

 k1 = wm*pm/2;

 if (hmmP.CK == DIAGC) 
   for(i=1;i<=hmmP.vSize;i++) {
    k2 = cov.var[i];
    if(k2==0.0)HError(1,"covariance element is zero");
    ucov.var[i] = k1/k2 * ( pow((ot[i]-m[i]),2)/k2 -  1.0 );
   }

 if (hmmP.CK == FULLC) /*not implemented yet*/;

}

/*---------------------------------------------------------------*/
/* These routines compute dg/d(mean) and dg/d(cov) */

void DeltaDiscDeltaMean(Vector um, double po)
{
 VectorDivide(um,po);
}


void DeltaDiscDeltaCov(Covariance cov, double po)
{

 if (hmmP.CK == DIAGC) VectorDivide(cov.var,po);

 if (hmmP.CK == FULLC) /* not implemented yet */;

}
/*-------------------------------------------------------*/
/* These routines compute dd/d(mean) and dd/d(cov) */


void  DeltaLossDeltaMean(Vector um, double lambda)
{
 VectorMultiply(um,lambda);
}

void  DeltaLossDeltaCov(Covariance cov, double lambda)
{

 if (hmmP.CK == DIAGC) VectorMultiply(cov.var,lambda);

 if (hmmP.CK == FULLC) /* not implemented yet */;

}


int UpdateGradient1(int J, Label *l, double sign, double d)
{
  int t;             /* current frame  */
  int s;             /* current state  */
  int i;
  int j;
  
  int index;
  HMMDef *hmm=NULL;
  HMMDef *uhmm=NULL;
  char * ModelName;
  
  double logpo;       /* log observation probability */
  double logpm;       /* log mixture observation probability */
  double po;          /* observation probability */
  double pm;          /* mixture observation probability */
  
  float  w ;          /* mixture weight */
  double logw;        /* the log of the mixture weight */
  Vector m;           /* mixture mean */ 
  Covariance cov;     /* mixture variance */
  
  float  uw;             /*  update weight */
  
  for (j=(J-1);j>=0;j--) { /* note tha j is simply an index and runs backwards in time */
    
    /* find the HMM in the database with ModelName */
    t = GetTime(j,l);
    ModelName = GetLabelModelName(j, l);
    index = Name2HmmIndex(ModelName, H);
    hmm = &(HMMdefList[index].hmmdef);
    uhmm = GPDupdateList+index;
    s = GetState(j,l); 
    
    if (trace>2) printf(" time[%d] Model[%d] name[%s] state[%d]\n",  t, index, ModelName, s);
    
    GetSample(&data,t-1,ot+1); /* frame times 1->N and data times 0->N-1 */ 
    logpo = OutP(ot, hmm, s);
    po = L2F(logpo);
    
    if(po != 0.0) {
      for (i=1;i<=hmmP.M;i++) {
	if(HMMMixtureExists(hmm,s,i)) {
	  
	  logpm = MOutP(ot, hmm, s, i);
	  pm = L2F(logpm);
	  w   =  hmm->stateinfo[s].outp[i].weight; 
	  m   =  hmm->stateinfo[s].outp[i].mean;
	  if(hmmP.CK == DIAGC) cov.var =  hmm->stateinfo[s].outp[i].covariance.var;
	  else /* unimplemented yet */;
	  
	  if (uFlags&UPMEANS)                 DeltaProbDeltaMean(m,ot,w,cov,um,pm); 
	  if (uFlags&UPMIXES && hmmP.M>1)     uw = pm; 
	  if (uFlags&UPVARS)                  DeltaProbDeltaCov(w,pm,ot,m,cov,ucov);
	  
	  if (trace==4) {
	    if (uFlags&UPMIXES && hmmP.M>1) printf("dp/dw = %f\n", uw);
	    if (uFlags&UPMEANS) ShowVector("dp/dm",um, hmmP.vSize);
	  }
	  
	  if (uFlags&UPMEANS)                 DeltaDiscDeltaMean(um,po); 
	  if (uFlags&UPMIXES && hmmP.M>1)     uw /= po;
	  if (uFlags&UPVARS)                  DeltaDiscDeltaCov(ucov,po);
	  
	  if (trace==4) {
	    if (uFlags&UPMIXES && hmmP.M>1) printf("dg/dw = %f\n", uw);
	    if (uFlags&UPMEANS) ShowVector("dg/dm",um, hmmP.vSize);
	  }
	  
	  if (uFlags&UPMEANS)                 DeltaLossDeltaMean(um,sign*d); 
	  if (uFlags&UPMIXES && hmmP.M>1)     uw = d*sign*uw;
	  if (uFlags&UPVARS)                  DeltaLossDeltaCov(ucov,sign*d);
	  
	  
	  if (trace==4) {
	    printf("d = %f\n", d);
	    if (uFlags&UPMIXES && hmmP.M>1) printf("dl/dw = %f\n", uw);
	    if (uFlags&UPMEANS) ShowVector("dl/dm",um, hmmP.vSize);
	  }
	  
	  if (uFlags&UPMEANS)      AddVectors(uhmm->stateinfo[s].outp[i].mean, um);
	  if (uFlags&UPMIXES && hmmP.M>1 )    uhmm->stateinfo[s].outp[i].weight += uw;
	  if (uFlags&UPVARS)
	    if(hmmP.CK == DIAGC) 
	      AddVectors(uhmm->stateinfo[s].outp[i].covariance.var, ucov.var);
	    else   /* not implemented yet */ ;
	  
	  if (trace==4) {
	    if (uFlags&UPMIXES && hmmP.M>1) 
	      printf("accumulated weight update = %f\n", 
		     uhmm->stateinfo[s].outp[i].weight);
	    if (uFlags&UPMEANS)
	      ShowVector("accumulated mean update",
			 uhmm->stateinfo[s].outp[i].mean, hmmP.vSize);
	  }
	}
      }
    }
  }
} /* end of UpdateGradient1() */

/*------------------------------------------------------*/
/* routines for implementing the various loss functions */

/* Sig: returns (1/(1+exp(-a(x-b))) */
double Sig (double x, double a, double b)
{
 
 return( 1 / (1 + exp(-a*(x-b))) );

}

/* DeltaSig: returns the first differential of Sig(x,a,b) */
double DeltaSig (double x, double a, double b)
{

 return( pow(Sig(x,a,b),2) *a*exp(-a*(x-b)) );

}

/* DeltaPow: returns the first differential of f(x) where
   f(x) = 0     x<= 0
   f(x) = x^a   x > 0
*/
double DeltaPow(double x, double a)
{
 
 if (x<=0.0) return (0.0); 
 return(a*pow(x,(a-1.0)));
 
}

/*DeltaPieceWiseLinear: returns the first differential of the piecewise linear function f(x)
   where   f(x) = 0  x<= a 
           f(x) = x-a/(b-a) b<x>a
           f(x) = 1  x>= b
*/
double DeltaPieceWiseLinear(double x, double a, double b)
{
 
 if (a==b) HError(1,"Infinite slope");
 if (a>b)  HError(1,"Negative slope");

 if (x<=a || x>=b) return (0.0);
 else              return (1.0/(b-a));

}

/*------------------------------------------------------*/

/* DLossDMisclass: returns the first differential of the loss function wrt
                   the misclassification measure
*/ 
double DLossDMisclass(double d)
{

 switch (LossFunctionType) 
 {
   case POWER: d = DeltaPow(d,alpha);
               break;

   case  SIG: d = DeltaSig(d, alpha, 0.0);
               break;

   default:  HError(1,"Invalid loss function type - should never get here");
             break;
 }
 

 return (d);
 
}

/* MatchFile: top level control */
void MatchFile(char * datafn)
{
   char outfn[255];
   DataFile data;
   FILE *outf;
   char ipathfn[256]; /* DR */
   FILE *ipathfp; /* DR */
   char ilabfn[256]; /* DR */
   FILE *ilabfp; /* DR */
   int t,n;
   Vector ot;
   long beamWidthSum;
   float x;
   Transcription tr;       /* DR - current hand label file transcription */
   Transcription tottr;  /* AD - internal path information for GPD stuff */
   Transcription move;    /*  a temporary transcription to move through tottr */
   LabList *RecPath;
   int i; /* DR */
   double d;               /* AD - the derivative of the loss */

   tottr = vCreateLabList();
   RecPath = vCreateLabList();

   MakeFN(datafn,labDir,labExt,outfn);
   MakeFN(datafn,ipathDir,ipathExt,ipathfn); /* DR */

/* DR - start */
   MakeFN(datafn,ilabDir,ilabExt,ilabfn);
   if (trace>1) printf("label file name is %s\n", ilabfn);
   tr = LOpen(ilabfn);
   if (trace >1) {
     printf("\n--the input label file contents are:--\n");
     for(i=0; i<tr->nLabels; i++)
       printf("%s\n", tr->labels[i].labid->name);
     printf("----------end of labels---------------\n\n");
   }
   fflush(stdout);

   AddHlPathToNet(tr);
  Nulllastplrt(); /*this is overkill we should NULL only the stuff as we add it*/
/*DR - end */
   CreateActiveList();

  if ((outf = fopen(outfn,"w")) == NULL)
      HError(2,"Cannot create output label file %s",outfn);
   if ((ipathfp = fopen(ipathfn,"w")) == NULL)                /* DR */   
     HError(2,"Cannot create output label file %s",ipathfn); /* DR */
   SpOpen(datafn,&data);
   vCheckData(&data,datafn);
   sampPeriod = data.sampPeriod;
   ot = CreateVector(hmmP.vSize);
   InitModels();
   beamWidthSum = nActive;
   if(trace >1) {
     printf ("Number of data samples is %d\n", data.nSamples); /* DR */
     fflush(stdout); /* DR */
   }
   for (t=1;t<=data.nSamples;t++) {
      GetSample(&data,t-1,ot+1);
      Setbjot(ot);
      StepModels(t);
      if (isPruning)
         CheckActive();
      beamWidthSum += nActive;
      NullEntryTokens();
      PropagateExitTokens();
      GarbageCollectPLRs();
      if (t==data.nSamples) {
	OutputHlResults(outf,ipathfp,outfn,t, tottr); /* DR--AD  paths  added*/
	OutputResults(outf,ipathfp,outfn,t, tottr,RecPath); 
      }   
      else
         RecordDecisions(t);
   }
   if (trace>0 && isPruning) {
      x = ((float)beamWidthSum * 100.0)/ 
            ((float) NPI * (float)data.nSamples);
      printf("  Total Pruning = %.2f%%\n",100-x);
      fflush(stdout);
   }
   if (trace>0 && maxPLRUsed>0) {
      printf(" Max PLR's Used = %d\n",maxPLRUsed);
      fflush(stdout);
   }
   move = tottr->next;
   d = tlgpr[1] - tlgpr[0];
   AvgLoss += d;
   if (d > 0.0){
     if (trace>0) printf("processing data file %s\n", datafn); 
     NumCorrect--;
     if (trace>1){
       printf("transcription of total path is:\n");
       PrintTranscription(tottr);
       PrintTranscription(move);
     }
     if (trace >0)printf("d1 = %f   ",d);
     d = DLossDMisclass(d);
     if (trace>0) printf("The first differential of d is %f\n", d);
     UpdateGradient1(move->nLabels, move->labels,UP, d); /* Viterbi Path */
     UpdateGradient1(tottr->nLabels, tottr->labels,DOWN, d); /* Correct Path*/
   }
   else
     for (n = 1; n < NBest; n++){
       d = tlgpr[1] - tlgpr[n+1];
       if (trace >0)printf("d1%d = %f",n,d);
       d = DLossDMisclass(d);
       if (trace>0) printf("The first differential of d is %f\n", d);
       move = move->next;
       UpdateGradient1(tottr->nLabels, tottr->labels,DOWN, d); /* Correct Path */
       UpdateGradient1(move->nLabels,move->labels,UP,d);/*the (n+1)'th Path*/
     }       


   SpClose(&data);
   DetachHlPathFromNet(tr); /* DR */
   LClose(tr); /* DR */
   LClose(tottr);  /* AD */
   LClose(RecPath);  /* AD */

   FreeILRs();
   FreeSLRs();
   fclose(ipathfp);
   
   fclose(outf);
   FreePLRs();
   DeActivateAll();
   FreeActiveList();
   FreeVector(ot);
}

/*ZeroNegativeMixtures: implements 'if (w_i<MINMIX) then w_i = 0'.
Mixture weights must be >=0 so all w_i<0 must be set to 0. */
void ZeroNegativeMixtures(OutProb *outp)
{
 int m;

 for(m=1;m<=hmmP.M;m++) if(outp[m].weight<0.0) outp[m].weight = 0.0;
} 


/* SumMixtureWeights2One: normalises the mixture weights using w_i = w_i / sum_j=1^N w_j
                          where N is the number of mixtures. Note prior to normalisation
                          negative mixture weights are set to 0. */
void  SumMixtureWeights2One(OutProb *outp)
{
 double sum=0.0;
 int m;

 ZeroNegativeMixtures(outp);
 for(m=1;m<=hmmP.M;m++) sum += outp[m].weight;
 if (sum != 0.0) for(m=1;m<=hmmP.M;m++) outp[m].weight /= sum; 
 else HError(1,"Oops - Model state has disappeared!");

}


void GPDUpdate(void)
{
 int h,s,S,m,i;
 double update, zero=0.0000000000;
 HMMDef *updateHMM, *oldHMM;

 for (h=1;h<=H;h++) { 
   oldHMM = &(HMMdefList[h].hmmdef);
   updateHMM = GPDupdateList+h;
   S = oldHMM->numStates; /* I think this can vary across models */


   for (s=2;s<S;s++){ 
     for (m=1;m<=hmmP.M;m++) {
       if(HMMMixtureExists(oldHMM,s,m)) {
         if (uFlags&UPMEANS)
	   VectorMultiply(updateHMM->stateinfo[s].outp[m].mean, -mepsilon);
         if (uFlags&UPMIXES && hmmP.M>1) 
	   updateHMM->stateinfo[s].outp[m].weight *= -wepsilon;
         if (uFlags&UPVARS)
           if (hmmP.CK == DIAGC)  
             VectorMultiply(updateHMM->stateinfo[s].outp[m].covariance.var, -vepsilon);
       }
     }
   }

   if(trace>1) {
     printf("Update HMM parameters for model %s\n", HMMidList[h]->name);
     DRHPrint(updateHMM, HPDETAILS|HPMEANS|HPVARS,2,S-1);
   }

   for (s=2;s<S;s++){ 
    for (m=1;m<=hmmP.M;m++) {
      if(HMMMixtureExists(oldHMM,s,m)) {

        if (uFlags&UPMEANS)
	  AddVectors(oldHMM->stateinfo[s].outp[m].mean,
		     updateHMM->stateinfo[s].outp[m].mean);
        if (uFlags&UPMIXES && hmmP.M>1) 
	  oldHMM->stateinfo[s].outp[m].weight += 
	    updateHMM->stateinfo[s].outp[m].weight;
        if (uFlags&UPVARS)
          if (hmmP.CK == DIAGC) 
            AddVectors(oldHMM->stateinfo[s].outp[m].covariance.var, 
		       updateHMM->stateinfo[s].outp[m].covariance.var);
      }
    }
    if (uFlags&UPMIXES&&hmmP.M>1) SumMixtureWeights2One(oldHMM->stateinfo[s].outp);
  }
   
   /* AD -- Now that we've used the update HMM, we should zero it for the next
      round of changes */
   
   for (s=2;s<S;s++){ 
     for (m=1;m<=hmmP.M;m++) {
       ZeroVector(updateHMM->stateinfo[s].outp[m].mean);
       updateHMM->stateinfo[s].outp[m].weight =zero;
       if (hmmP.CK == DIAGC) 
	 ZeroVector(updateHMM->stateinfo[s].outp[m].covariance.var);
     }
   }
   
 } /* end of h */
 
} /* end of GPDUpdate() */


void  ScaleUpdates(void)
{

 int h,s,S,m;
 double update;
 HMMDef *updateHMM,*oldHMM;
 double MaxAbsWeight=0.0, MaxMeanMod=0.0;
 double MaxOldAbsWeight=0.0, MaxOldMeanMod=0.0;
Boolean noupdate = 0;

 for (h=1;h<=H;h++) { 
   updateHMM = GPDupdateList+h;
   oldHMM = &(HMMdefList[h].hmmdef);

   S = oldHMM->numStates; /* I think this can vary across models */

   for (s=2;s<S;s++){ 
     for (m=1;m<=hmmP.M;m++) {

       if(HMMMixtureExists(oldHMM,s,m)) {
         if (fabs(updateHMM->stateinfo[s].outp[m].weight) > MaxAbsWeight)
           MaxAbsWeight = fabs(updateHMM->stateinfo[s].outp[m].weight);
         if (fabs(oldHMM->stateinfo[s].outp[m].weight) > MaxOldAbsWeight)
           MaxOldAbsWeight = fabs(oldHMM->stateinfo[s].outp[m].weight);
         if (Vector2Norm(updateHMM->stateinfo[s].outp[m].mean) > MaxMeanMod)
           MaxMeanMod = Vector2Norm(updateHMM->stateinfo[s].outp[m].mean);
         if (Vector2Norm(oldHMM->stateinfo[s].outp[m].mean) > MaxOldMeanMod)
           MaxOldMeanMod = Vector2Norm(oldHMM->stateinfo[s].outp[m].mean);
       }
     }
   }
 }

 if (uFlags&UPMIXES && hmmP.M>1)
   if(MaxAbsWeight == 0.0)     noupdate = 1;
 if (uFlags&UPMIXES && hmmP.M>1) if(MaxOldAbsWeight == 0.0) HError(1,"No old weight");
 if (uFlags&UPMEANS)
   if(MaxMeanMod == 0.0)     noupdate = 1;
 if (uFlags&UPMEANS) if(MaxOldMeanMod == 0.0)   HError(1,"No old means");


if (!noupdate)
  for (h=1;h<=H;h++) { 
    updateHMM = GPDupdateList+h;
    oldHMM = &(HMMdefList[h].hmmdef);
    
    S = oldHMM->numStates; /* I think this can vary across models */
    
    for (s=2;s<S;s++){ 
      for (m=1;m<=hmmP.M;m++) {
	if(HMMMixtureExists(oldHMM,s,m)) {
	  if (uFlags&UPMEANS)
	    VectorMultiply(updateHMM->stateinfo[s].outp[m].mean, 
			   MaxOldMeanMod/MaxMeanMod);
	  if (uFlags&UPMIXES && hmmP.M>1) 
	    updateHMM->stateinfo[s].outp[m].weight *= 
	      (MaxOldAbsWeight/MaxAbsWeight);
	}
      }
    }
  }
 
 
}

/* ----------------------------------------------------------- */
/*                      END:  HVGN.c                          */
/* ----------------------------------------------------------- */
