/* ----------------------------------------------------------- */
/*                   HTK  -  HMM Toolkit                       */
/* ----------------------------------------------------------- */
/*         Copyright: Steve Young      October 1989            */
/*         Last Update:                December 1990           */
/*                                                             */
/*         Cambridge University Engineering Department         */
/*         Trumpington Street, Cambridge, CB2 1PZ              */
/*                                                             */
/* Permission is hereby granted to copy, use and modify this   */
/* software provided that this banner notice is not removed    */
/*                                                             */
/*  D. Rainton - Modification History                          */
/*                                                             */
/* July 2 1991 - subspace code added                           */
/* Aug  9 1991 - N-best code added                             */
/* Aug 14 1991 - internal state path output added              */
/*                                                             */
/* ----------------------------------------------------------- */
/*         File: HVPath.c: Viterbi Recogniser                  */
/* ----------------------------------------------------------- */

/*
   This program will match a network of HMM's against one or more
   speech files and output a transcription in HTK format.  The 
   match is performed using conventional Viterbi decoding.  The
   network is input via HParse, all nodes are assumed to be
   the names of phone labels.  The HMM definition file 
   corresponding to each label must consist of the label name 
   followed by the specified extension (which may be null).
   The set of all hmm's needed must be listed in an idList.
   
   An optional bigram may be specified via the -b option.  The
   bigram should be a text file stored in the format
   
     <lab1> p11 p12 p13  ... p1H
     <lab2> p21 p22 p23  ... p2H
           ...
     <labH> pH1 pH2 pH3  ... pHH
     
   where <labh> is the name of the h'th label in the idList and
   pij is the bigram probabity P(labj|labi).
   
   Recognition may be controlled by the following optional 'scaling'
   parameters:
      
         propPen  - a fixed inter-model transition probability
         bigScale - scale factor for log(bigram) probabilities
         bjtScale - scale factor for log(bj(ot)) probabilities 

   
   The principle data structure is the network corresponding to
   the regular grammar rules used to constrain the recognition.
   This network is built by HParse.  For each phoneme instance,
   a PhoneInstance struct is created and hung-off the 'user' member
   of the corresponding network node.  Each PhoneInstance contains
   a pointer to a Definition struct containing the HMM definition 
   and space for bj(ot) to save recalculating multiple instances. 
   Definition structs are shared by the PhoneInstances so that 
   there is only one copy of each distinct HMM.  
   
   The recogniser design follows the token passing model.
*/


#include <stdio.h>      /* Standard C Libraries */
#include <stdlib.h>
#include <math.h>

#include "HShell.h"     /* HMM ToolKit Modules */
#include "HMath.h"
#include "HSpIO.h"
#include "HLabel.h"
#include "HModel.h"
#include "HParse.h"
#include "HSubS.h"

#define VERSION   "1.2"
#define GCMARGIN   400
#define MAX_N_BEST 20


/* ------------ Tokens, Definitions and PhoneLink Records -------------- */

typedef struct _Definition Definition;
typedef struct _PLR PLR;
typedef struct _ILR ILR;
typedef struct _SLR SLR;
typedef struct _PhoneInstance PhoneInstance;

typedef struct {
   PhoneInstance * inst;   /* current owner */
   float logprob;          /* score */
   PLR   *link;            /* link */
   ILR   *ilink;           /* link*/
   SLR   *slrlink;         /* DR - link to SLR */
} Token;

struct _Definition{
   HMMDef hmmdef;       /* actual HMM def */
   float *bjot;         /* array[2..numStates-1] of bj(ot) */
   LabId id;            /* id of this defn */
   Boolean bjLoaded;    /* true if bjot loaded */
   SubSDef subsdef;     /* the actual subspace def */
};

struct _PLR{
   Boolean linked;      /* flag used in garbage collection */
   Token tok;           /* token at boundary */
   int time;            /* index of model boundary */
   PLR *chain;          /* chain for keeping track of PLR's */
};

struct _ILR{
   int time;            /* time of state entry */
   int state;           /* state the token is in */
   ILR *chain;          /* chain for keeping track of ILR's*/
   ILR *gchain;        /* global chain for global freeing of ILR structures */
};

struct _SLR{
   int time;            /* time of observation */
   int state;           /* the model state */
   float bjot;          /* probability of observation */
   SLR *link;           /* points to the previous SLR*/
   SLR *chain;          /* for house keeping */
   char *name;          /* model name */
 };                  

struct _PhoneInstance{
   Definition *def;  /* definition for this instance */
   Link netHook;     /* Network node holding this instance */
   int N;            /* copy of defList[id].numStates */
   Boolean isActive; /* true if pi is in active list */
   Token *state;     /* array[1..N]of token */
   int lastplrt;     /* time when last PLR gen'ed from this inst */
   PLR *lastplr;     /* pointer to last PLR gen'ed from this inst */
   Boolean hlpath;   /* DR - indicates whether pi is on hand labeled path through network */
};

/* Global Settings */

static char * labDir = NULL;     /* output label file directory */
static char * labExt = "rec";    /* output label file extension */
static char * ilabDir = NULL;    /* hand label file directory */
static char * ilabExt = "lab";   /* hand label file extension */
static char * hmmDir = NULL;     /* directory to look for hmm def files */
static char * hmmExt = NULL;     /* hmm def file extension */
static char * SubSExt = NULL;    /* extension of the subspace files */
static char * SubSDir = NULL;    /* directory to look for subspace def files */
static char * ipathDir = NULL;   /* directory to store internal path files */
static char * ipathExt = "ip";   /* extension for internal path files */
static float propPen = 1.0;      /* inter model propagation probability */
static float bigScale = 1.0;     /* bigram scale factor */
static float bjtScale = 1.0;     /* bjot scale factor */
static float pruning = 0.0;      /* pruning threshold */
static Boolean fixedPen = FALSE; /* enable flag for fixed propPen */
static Boolean isPruning= FALSE; /* enable flag for pruning */
static short rotate=FALSE;       /* flag indicating data rotation */
static int  trace    = 0;        /* Trace level */
static int NBest = 1;            /* produce NBest parses */ 

static long sampPeriod;          /* data sample period in 100us units */
static int vSize = 0;            /* data vector size */
static int maxStates = 0;        /* max num states in any phone model */
static int  M = 0;               /* number of mixtures */
static CovKind CK;               /* covariance kind of all units */
static SampKind SK;              /* sample kind of all units */
static Boolean isTied = FALSE;   /* true if tied mixtures */
static Boolean ipath_on = FALSE; /* true if the internal path information is wanted */
static Boolean handlabel = FALSE;/* true if input label files are present */

/* Global data structures */

static LabId enterId;            /* LabId of the ENTER node name */
static LabId exitId;             /* LabId of the EXIT node name */
static Network theNet;           /* the syntax constraint network */
static Definition *defList;      /* array[1..H] of HMM definitions */
static LabId *idList;            /* array[1..H] of HMM file name id */
static int H = 0;                /* number of distinct HMM's (Labels) */
static int P = 0;                /* number of distinct phonemes */
static int NPI = 0;              /* number of phone instances */
static Token *tBuf;              /* space to copy largest array of tokens */
static Boolean hasBigram=FALSE;  /* bigram loaded flag */
static Matrix bigram;            /* bigram stats if any */
static int nActive = 0;          /* number of active phone instances */
static PhoneInstance **active;   /* array[0..nActive-1]of active pi's */
static LogReal maxLogP;          /* max LogP at current time frame */
static int maxPLRUsed=0;         /* max PLR's used */

static ILR *ilrHD=NULL;        /* this is used for a global freeing of ILR structures */
static SLR *slrHD=NULL;        /* this is used for a global freeing of SLR structures */

/* ------------------ Process Command Line ------------------------- */

void ReportUsage(void)
{
   fprintf(stderr,"\nUSAGE: HVite[V%s] [options] hmmList netFile speechFiles...\n\n",
           VERSION);
   fprintf(stderr," Option                                          Default\n\n");
   fprintf(stderr," -b f    load bigram from file f                 none\n");
   fprintf(stderr," -d s    dir to find hmm definitions             current\n");
   fprintf(stderr," -o f    bj(ot) scale factor                     1.0\n");
   fprintf(stderr," -p f    inter model trans prob = f              1/Nsucc\n");
   fprintf(stderr," -s f    bigram scale factor                     1.0\n");
   fprintf(stderr," -t f    set pruning threshold                   0.0\n");
   fprintf(stderr," -x s    extension for hmm files                 none\n");
   fprintf(stderr," -F fmt  Set data file format to fmt             HTK\n");
   fprintf(stderr," -G fmt  Set the hand label file format to fmt   HTK\n");
   fprintf(stderr," -L s    dir to store output label file(s)       current\n");
   fprintf(stderr," -D s    dir to find subspace file(s)            current\n"); 
   fprintf(stderr," -S f    set script file to f                    none\n");
   fprintf(stderr," -T N    set trace level to N                    0\n");
   fprintf(stderr," -X s    extension for output label files        rec\n");
   fprintf(stderr," -Y s    extension for hand label files          lab\n");
   fprintf(stderr," -E s    extension for the subspace files        none\n");
   fprintf(stderr," -N n    output the N-Best parses                1\n");
   fprintf(stderr," -I s    dir to store internal path files        current\n");
   fprintf(stderr," -i s    extension for the internal path files   ip\n");
   fprintf(stderr," -l s    dir to get hand label files from        none\n");  
   exit(1);
}

int GetChkedInt(int min, int max, char * swtname)
{
   int val;
   
   if (NextArg() != INTARG) 
      HError(1,"Integer Arg Required for %s option",swtname);
   val=GetIntArg();
   if (val<min || val>max)  
      HError(1,"Integer arg out of range(%d - %d) in %s option",min,max,swtname); /* DR - print min,max */
   return val;
}

float GetChkedFlt(float min, float max, char * swtname)
{
   float val;
   
   if (NextArg() != INTARG && NextArg() != FLOATARG) 
      HError(1,"Float Arg Required for %s option",swtname);
   val=GetFltArg();
   if (val<min || val>max) 
      HError(1,"Float arg out of range (%f - %f) in %s option",min,max,swtname); /* DR - print min,max */
   return val;
}

main(int argc, char *argv[])
{
   char *datafn, *s;
   FileFormat ff;
   char *netfn,*listfn,*bigfn = NULL;
   void Initialise(char * listfn, char * netfn, char * bigfn);
   void MatchFile(char * datafn);
   
   InitShell(argc,argv);
   InitMath(FALSE);        /* Doesnt use LAdd */
   InitLabel();
   if (NumArgs()==0)
      ReportUsage();
   while (NextArg() == SWITCHARG) {
      s = GetSwtArg();
      if (strlen(s)!=1) 
         HError(1,"Bad switch %s; must be single letter",s);
      switch(s[0]){
         case 'b':
            if (NextArg()!=STRINGARG)
               HError(1,"Bigram file name expected");
            bigfn = GetStrArg(); 
            break;
         case 'd':
            if (NextArg()!=STRINGARG)
               HError(1,"HMM definition directory expected");
            hmmDir = GetStrArg(); break;     
         case 'o':
            bjtScale = GetChkedFlt(0.0,100.0,s); 
            break;
         case 'p':
            propPen = log(GetChkedFlt(1.0E-20,1.0E20,s)); 
            fixedPen = TRUE; break;
         case 's':
            bigScale = GetChkedFlt(0.0,100.0,s); 
            break;
         case 't':
            pruning = GetChkedFlt(0,1.0E20,s); 
            isPruning = TRUE; break;
         case 'x':
            if (NextArg()!=STRINGARG)
               HError(1,"HMM file extension expected");
            hmmExt = GetStrArg(); break;
         case 'F':
            if (NextArg() != STRINGARG)
               HError(1,"Data File format expected");
            if((ff = Str2Format(GetStrArg())) == ALIEN)
               HError(0,"Warning ALIEN Data file format set");
            SetFormat(ff);
            break;
         case 'G':
            if (NextArg() != STRINGARG)
               HError(1,"Label File format expected");
            if((ff = Str2Format(GetStrArg())) == ALIEN)
               HError(0,"Warning ALIEN Label file format set");
            SetLFormat(ff);
            break;
         case 'L':
            if (NextArg()!=STRINGARG)
               HError(1,"Output Label file directory expected");
            labDir = GetStrArg(); break;
         case 'l':
            if (NextArg()!=STRINGARG)
               HError(1,"Input Label file directory expected");
            handlabel = TRUE;
            ilabDir = GetStrArg(); break;
         case 'D':
            if (NextArg()!=STRINGARG)
               HError(1,"Subspace file directory expected");
            SubSDir = GetStrArg();
            rotate = TRUE; break;
         case 'S': SetScriptFile();
            break;
         case 'T':
            trace = GetChkedInt(0,3,s); break;
         case 'N':
            NBest = GetChkedInt(1,MAX_N_BEST,s); break;
         case 'X':
            if (NextArg()!=STRINGARG)
               HError(1,"Label file extension expected");
            labExt = GetStrArg(); break;
         case 'Y':
            if (NextArg()!=STRINGARG)
               HError(1,"Hand label file extension expected");
            ilabExt = GetStrArg(); break;
         case 'I':
            if (NextArg()!=STRINGARG)
               HError(1,"Internal path file directory expected");
            ipath_on = TRUE;
            ipathDir = GetStrArg(); break;
         case 'i':
            if (NextArg()!=STRINGARG)
               HError(1,"Internal path file extension expected");
            ipathExt = GetStrArg(); break;
         case 'E':
            if (NextArg()!=STRINGARG)
               HError(1,"Subspace file extension expected");
            SubSExt = GetStrArg(); break;
         default:
            HError(1,"Unknown switch %s",s);
      }
   } 
   if (NextArg()!=STRINGARG)
      HError(1,"hmm list file name expected");
   listfn = GetStrArg();
   if (NextArg()!=STRINGARG)
      HError(1,"network file name expected");
   netfn = GetStrArg();
   Initialise(listfn,netfn,bigfn);
   do {
      if (NextArg()!=STRINGARG)
         HError(1,"training data file name expected");
      datafn = GetStrArg();
      MatchFile(datafn);
   } while (NumArgs()>0);
}

/* -------------------- Bigram Handling Routines -------------- */

/* LoadBigram: load a bigram from given file, note that row 
              and col ordering is changed to map onto idlist */
void LoadBigram(char *fname)
{
   FILE *f;
   char buf[132];
   LabId id;
   int p,i,j,*map;
   char *checkList;
   Vector v,tmp;
   float x;
   
   bigram = CreateMatrix(P,P);
   tmp = CreateVector(P);
   if ((map=(int *)malloc(P*sizeof(int))) == NULL)
      HError(2,"Cant create bigram map array");
   --map;
   checkList = NewString(P);     /* not chars really */
   for (p=1;p<=P;p++)
      checkList[p] = '\0';
   ZeroMatrix(bigram);
   if ((f = fopen(fname,"r")) == NULL)
      HError(2,"Cannot open bigram file %s",fname);
   j = 0;
   while (ReadLabel(f, buf)) {      /* reorder rows during read in */
      if ((id = GetLabId(buf,FALSE))==NULL)
         HError(2,"Unknown label %s in bigram %s",buf,fname);
      p = id->aux; map[++j] = p;
      if (p < 1  ||  p > P)
         HError(2,"Label %s from bigram %s not in id List",buf,fname);
      if (checkList[p] == '\1')
         HError(2,"Label %s in bigram %s doubly defined",buf,fname);
      v = bigram[p];
      ReadVector(f,v);
      for (i=1;i<=P;i++) {
         x = v[i];
         v[i] = (x<MINLARG)? LZERO : log(x) * bigScale;
      }
      checkList[p] = '\1';
   }
   for (p=1;p<=P;p++)      /* check all rows defined */
      if (checkList[p] == '\0')
         HError(2,"No definition for label %s in bigram %s",
            idList[p]->name,fname);
   for (i=1;i<=P;i++) {          /* reorder columns */
      v = bigram[i];
      for (j=1;j<=P;j++)
         tmp[map[j]] = v[j];
      bigram[i] = tmp; tmp = v;
   }
   hasBigram = TRUE;
   free(map+1);
   FreeVector(tmp);
}

/* ------------------ Phone Link Record Management ----------------- */

#define PLRTRWIDTH 4

static PLR *freeHd=NULL, *freeTl=NULL;
static PLR *usedHd=NULL, *usedTl=NULL;
static numSinceGC=0;                   /* num PLRs allocated since last GC */
static PLR *plrtrbuf[PLRTRWIDTH];      /* trace of best tokens */
static int trbufUsed;

/* InitPLRs: Initialise pool of PLRs */
void InitPLRs()
{
   if (usedHd!=NULL) {
      if (freeHd==NULL) {
         freeHd=usedHd; 
         freeTl=usedTl;
      } else {
         freeTl->chain = usedHd;
         freeTl = usedTl;
      }
   }
   usedHd = usedTl = NULL;
   numSinceGC = 0;

}
      
/* FreePLRs: DR Free the entire pool of PLR's */
void FreePLRs()
{
 PLR *p;

  while (usedHd!=NULL) {
    p = usedHd->chain;   
    free(usedHd);    
    usedHd = p;
  }

  usedHd=usedTl=NULL;

  while (freeHd!=NULL) {
    p = freeHd->chain;
    free(freeHd);
    freeHd = p;
  }

  freeHd=freeTl=NULL;
  numSinceGC = 0;

}


/* DR - CreateILR: Returns a pointer to a new ILR */
ILR *CreateILR(void)
{
   ILR *p;

   if ((p = (ILR *)malloc(sizeof(ILR))) == NULL)  
      HError(2,"Cannot create ILR");   

   p->gchain = ilrHD;
   ilrHD = p;

   return p;
}   

/* DR - FreeILRs: Frees a chain of ILR records */
void FreeILRs()
{
 ILR *next;

 while (ilrHD->gchain != NULL) {
   next = ilrHD->gchain;
   free(ilrHD); 
   ilrHD = next;
 }

 free(ilrHD);
 ilrHD = NULL;

}

/* DR - CreateSLR: Returns a pointer to a new SLR */
SLR *CreateSLR(void)
{
   SLR *p;

   if ((p = (SLR *)malloc(sizeof(SLR))) == NULL)  
      HError(2,"Cannot create SLR");   

   p->chain = slrHD;
   slrHD = p;

   return p;
}   

/* DR - FreeSLRs: Frees a chain of SLR records */
void FreeSLRs()
{
 SLR *next;

 while (slrHD->chain != NULL) {
   next = slrHD->chain;
   free(slrHD); 
   slrHD = next;
 }

 free(slrHD);
 slrHD = NULL;

}

/* CreatePLR: Returns a pointer to a new PLR */
PLR *CreatePLR(void)
{
   PLR *p;
   
   if (freeHd==NULL) {
      if ((p = (PLR *)malloc(sizeof(PLR))) == NULL)
         HError(2,"Cannot create PLR");
   } else {
      p = freeHd;
      if (freeHd==freeTl)
         freeHd = freeTl = NULL;
      else 
         freeHd = freeHd->chain;
   }
   if (usedHd==NULL) 
      usedHd = usedTl = p;
   else {
      usedTl->chain = p;
      usedTl = p;
   }
   p->chain = NULL;
   ++numSinceGC;
   return p;
}

/* GarbageCollectPLRs:  Garbage Collect unlinked PLR's */
void GarbageCollectPLRs(void)
{
   int i,j,plrcollected,plrused;
   PLR *plr,*next;
   PhoneInstance *pi;
   Token t;

   if (GCMARGIN <= numSinceGC) {
      plrused = plrcollected = 0;

      for (plr=usedHd; plr!=NULL; plr=plr->chain)
         plr->linked = FALSE;

      for (i=0;i<nActive;i++) {
         pi = active[i];
         for (j=1; j<=pi->N; j++) {
            t = pi->state[j];
            if (t.inst != NULL) 
               for (plr=t.link; (plr!=NULL)&&(!plr->linked);
                                           plr=plr->tok.link)
                  plr->linked=TRUE;
         }
      }

      plr = usedHd; usedHd = usedTl = NULL;
      while (plr!=NULL) {
         next = plr->chain;
         if (plr->linked) {
            ++plrused;
            plr->chain = usedHd;
            usedHd = plr;
            if (NULL==usedTl)
               usedTl = plr;
         } else {
            ++plrcollected;
            plr->chain = freeHd;
            freeHd = plr;
            if (NULL==freeTl)
               freeTl=plr; 
         }
         plr=next;
      }

      if (trace>1) {
         printf("GC: NumSinceGC=%d Collected=%d Used=%d\n",
                                          numSinceGC,plrcollected,plrused); 
         fflush(stdout);
      }
      numSinceGC = 0;
      if (plrused>maxPLRUsed) maxPLRUsed = plrused;
   }
}

/* ClearTrBuf: clear the PLR trace buffer */
void ClearTrBuf(void)
{
   trbufUsed = 0;
}

/* Add2TrBuf: add the given PLR to the trace buffer */
void Add2TrBuf(PLR *p)
{
   int i;
   PLR *q;
   LogReal x;
   
   x = p->tok.logprob;
   if (trbufUsed == PLRTRWIDTH) {
      q = plrtrbuf[trbufUsed-1];
      if (q->tok.logprob<x)
         --trbufUsed;
      else 
         return;
   }
   for (i=trbufUsed;i>0;i--)
      if (plrtrbuf[i-1]->tok.logprob < x)
         plrtrbuf[i] = plrtrbuf[i-1];
      else
         break;
   plrtrbuf[i] = p;
   ++trbufUsed;
}

/* PrintTrBuf: print out the contents of the trace buffer */
void PrintTrBuf(int t)
{
   int i;
   Token tok;
   
   printf("PLRS: time %d\n",t);
   for (i=0;i<trbufUsed;i++){
      tok = plrtrbuf[i]->tok;
      printf("id %2s, prob %f",tok.inst->def->id->name,tok.logprob);
      if(tok.link!=NULL)
         printf(" back to frame %d",tok.link->time);
      printf("\n");
   }
   fflush(stdout);
}

/* -------------------- Housekeeping for Pruning ----------------- */

/* CreateActiveList: create a list of pointers to active phone instances */
void CreateActiveList(void)
{
   if ((active=(PhoneInstance **)malloc(NPI*sizeof(PhoneInstance *))) == NULL)
      HError(2,"Cannot create active phone instance list");
}

/* FreeActiveList: DR free the active list  */
void FreeActiveList(void)
{
 free((PhoneInstance **)active);
}

/* ReAllocActiveList: allow the active list to grow in size */
void ReAllocActiveList(void)
{
 PhoneInstance **newactive;

 if ((newactive =(PhoneInstance **)realloc(active, NPI*sizeof(PhoneInstance *))) == NULL)
   HError(2,"Cannot realloc active phone instance list");

 active = newactive;
}


/* DeActivateAll: deactivate all active phone instances */
void DeActivateAll(void)
{
   int i;
   
   for (i=0;i<nActive;i++)
      active[i]->isActive = FALSE; 
   nActive = 0;
}

/* MakeActive: make given pi active */
void MakeActive(PhoneInstance *pi)
{
   int i;
   Token nullToken = {NULL,LZERO,NULL,NULL,NULL}; /* DR - NULL,NULL added */ 

   if (!pi->isActive){
      pi->isActive = TRUE;
      active[nActive++] = pi;
      for (i=1;i<=pi->N;i++)
         pi->state[i] = nullToken;
   }
}

/* CheckActive: scan all active phones and kill any with no paths in beam */
void CheckActive(void)
{
   int i,j,nAOld;
   PhoneInstance *pi;
   LogReal bestP,threshold;
   Token t;
   
   nAOld = nActive;
   threshold = maxLogP - pruning;
   for (i=0;i<nActive;i++) {
      SPINBALL
      pi = active[i];
      bestP = LZERO;          /* find best path in this pi */
      for (j=2; j<=pi->N; j++){
         t = pi->state[j];
         if (t.inst != NULL && t.logprob>bestP)
            bestP = t.logprob;
      }
      if (bestP<threshold){   /* if below thresh kill it */
         active[i] = NULL;
         pi->isActive = FALSE;
      }
   }
   for (i=0,j=0;i<nActive;i++)   /* remove marked pi's */
      if (active[i] != NULL) {
         active[j++] = active[i];
      }
   nActive = j;
   if (trace>1) {
      printf("   Active Beam Width %d -> %d\n",nAOld,nActive);
      fflush(stdout);
   }
}

/* ------------------------ Network Management ------------------- */

/* IdMap: set-up id->aux to inverse map of idList indices.  If
   hmms are context sensitive then the mapping goes from cs models
   to monophones.  Only used by bigram */
void IdMap(void)
{
   int p,h;
   Boolean isCS=FALSE;
   char *s,buf[100];
   LabId id;
   
   for (h=1;h<=H;h++){
      s = idList[h]->name;
      if (strchr(s,'-') != NULL || strchr(s,'+') != NULL){
         isCS = TRUE;
         break;
      }
   }
   
   if (!isCS) {
      for (h=1;h<=H;h++)
         idList[h]->aux = h;
      P = H;
   } else {
      for (h=1;h<=H;h++){
         strcpy(buf,idList[h]->name);
         TriStrip(buf);
         id = GetLabId(buf,TRUE);
         p = id->aux;
         if (p == 0) id->aux = p = ++P;
         idList[h]->aux = p;
      }
   }
}

/* LoadHMM: load specified hmm, adding hmmExt if set */
void LoadHMM(char *name, HMMDef *def)
{
   char fn[255];
   int len;
   
   if (hmmDir==NULL)
      fn[0]='\0';
   else {
      strcpy(fn,hmmDir);
      len = strlen(fn);
      if (fn[len-1]!=PATHCHAR) {
         fn[len]=PATHCHAR;
         fn[++len]='\0';
      }
   }
   strcat(fn,name);
   if (hmmExt!=NULL) {
      strcat(fn,".");
      strcat(fn,hmmExt);
   }
   SPINBALL
   HOpen(fn,def);
   if (vSize==0) {
      vSize = def->vecSize;
      M = def->numMixtures;
      CK = def->ckind; SK = def->pkind;
      isTied = def->tstat != FreeT;
   } else {
      if(vSize!=def->vecSize)
         HError(2,"HMMs have incompatible vecSizes");
      if(M!=def->numMixtures)
         HError(2,"HMMs have incompatible numMixtures");
      if (CK!=def->ckind)
         HError(2,"HMMs have different covariance kinds");
      if (SK!=def->pkind)
         HError(2,"HMMs have different sample kinds");
      if (isTied != (def->tstat != FreeT))
         HError(2,"Mixed Free and Tied Distributions");
   }
   if (maxStates<def->numStates)
      maxStates = def->numStates;
}

/* LoadSubS: load specified subspace, adding SubSExt if set */
void LoadSubS(char *name, SubSDef *def)
{
   char fn[255];
   int len;
   
   if (SubSDir==NULL)
      fn[0]='\0';
   else {
      strcpy(fn,SubSDir);
      len = strlen(fn);
      if (fn[len-1]!=PATHCHAR) {
         fn[len]=PATHCHAR;
         fn[++len]='\0';
      }
   }
   strcat(fn,name);
   if (SubSExt!=NULL) {
      strcat(fn,".");
      strcat(fn,SubSExt);
   }
   SPINBALL

   SubSOpen(fn, def); 

}

/* LoadHMMDefs: create defList and load it */
void LoadHMMDefs(void)
{
   int h,n;
   Definition *d;
   
   if ((defList = (Definition *)malloc(H*sizeof(Definition))) == NULL)
      HError(2,"Unable to create hmm def array");
   --defList;
   for (h=1;h<=H;h++){
      d = defList+h;
      LoadHMM(idList[h]->name,&(d->hmmdef));
      if(trace >= 1) {
       printf("Loaded phoneme HMM %s\n", idList[h]->name);
       fflush(stdout);
      }
      if(rotate) {
        LoadSubS(idList[h]->name,&(d->subsdef));
        if(trace >= 2) {
          printf("Loaded phoneme subspace %s\n", idList[h]->name);
          fflush(stdout);
        }
      }
      n = d->hmmdef.numStates-2;
      if ((d->bjot = (float *)malloc(n*sizeof(float))) == NULL)
         HError(2,"Cannot create bjot array");
      d->bjot -= 2;     /* index is 2.. */
      d->id = idList[h];
   }
}



/* AttachDefinition: look for hmm def with given id and attach it to pi */
void AttachDefinition(PhoneInstance *pi, LabId id, Boolean valid /* DR */)
{
   int h;
   for (h=1;h<=H;h++)
      if (!strcmp(idList[h]->name,id->name)){ /* DR */
         pi->def = defList+h;
         pi->hlpath = valid; /* DR */
         return;
      }
   HError(2,"Cannot find def %s to attach",id->name);
}

/* Nulllastplrt: zero lastplrt in all PhoneInsts */
void Nulllastplrt(void)
{
   Link p;
   PhoneInstance *pi;
   
   p = theNet.chain;
   while (p!=NULL) {
      if (p->modelName != enterId && p->modelName != exitId ) {
         pi = (PhoneInstance *) p->user;
         pi->lastplrt = 0;
      }
      p = p->chain;
   }
}

/* AttachPhoneInsts: to each node in theNet */
void AttachPhoneInsts(void)
{
   Link p;
   PhoneInstance *pi;
   
   p = theNet.chain;
   while (p!=NULL) {
      SPINBALL
      if (p->modelName != enterId && p->modelName != exitId ) {
         if ((pi =(PhoneInstance *)malloc(sizeof(PhoneInstance))) == NULL)
            HError(2,"Unable to create phone instance");
         ++NPI;
         AttachDefinition(pi,p->modelName, FALSE); /* DR "FALSE" added */
         pi->N = pi->def->hmmdef.numStates;
         pi->isActive = FALSE;
         if ((pi->state = (Token *)malloc(pi->N*sizeof(Token))) == NULL)
            HError(2,"Unable to create state array for model %d",pi->N);
         pi->state -= 1;
         (PhoneInstance *)(p->user) = pi; 
         pi->netHook = p;
      }
      p = p->chain;
   }
   Nulllastplrt();
   if (trace>0){
      printf("  %d Instances Created of %d HMMs\n",NPI,H);
      fflush(stdout);
   }
}

/* ShowPath: print a brief n char resumee of path */
void ShowPath(PLR *plr, int n)
{
   char buf[132];
   int i,count;
   
   
   buf[0] = '\0'; count = 0;
   while (plr != NULL && count < n) {
      strcat(buf,plr->tok.inst->def->id->name);
      plr = plr->tok.link;
      if (plr != NULL) strcat(buf,"/");
      count = strlen(buf);
   }
   for (i=count;i<n;i++) buf[i]=' ';
   buf[n] = '\0'; 
   printf("%s",buf);
}

/* ShowInst: print contents of given phone instance */
void ShowInst(PhoneInstance *pi)
{
   int i;
   LogReal x;
   PLR *plr;
   PhoneInstance *ppi;
   
   printf("** Inst %6s[%2d]\n  ",pi->def->id->name,(int)pi % 100);
   for (i=1; i<=pi->N; i++)
      printf("%10d",i);
   printf("\n  ");
   for (i=1; i<=pi->N; i++) {
      ppi = pi->state[i].inst;
      if (ppi==NULL)
         printf("%10s","NULL");
      else
         printf("%6s[%2d]",ppi->def->id->name,(int)ppi % 100);
   }
   printf("\n  ");

   for (i=1; i<=pi->N; i++)
      if (pi->state[i].inst == NULL)
         printf("%10s","NULL");
      else {
         x = pi->state[i].logprob;
         if (x>LSMALL)
            printf(" %9.2f",x);
         else
            printf("%10s","---");
      }
   printf("\n  ");

   for (i=1; i<=pi->N; i++) {
      plr = pi->state[i].link;
      if (plr==NULL)
         printf("%10s","NULL");
      else
         ShowPath(plr,10);
   }
   printf("\n");
}

/* ------------------------ Initialisation ----------------------- */

/* Initialise: initialise the global data structures */
void Initialise(char * listfn, char * netfn, char * bigfn)
{
   idList = LoadIdList(listfn,&H);
   IdMap();
   if (trace>1) {
      printf("Id Map Built: ");
      if (P==H)
         printf(" %d monophones\n",H);
      else
         printf(" %d c-s phones -> %d monophones\n",H,P);
      fflush(stdout);
   }
   LoadHMMDefs(); /* DR - this routine also loads subspaces*/
   if (bigfn != NULL)
      LoadBigram(bigfn);
   enterId = GetLabId("ENTER",TRUE);
   exitId = GetLabId("EXIT",TRUE);  
   theNet = ParseNetwork(netfn);    /* build the theNet */
   AttachPhoneInsts();
   /**CreateActiveList(); DR - moved to MatchFile */
   if ((tBuf=(Token *)malloc(maxStates*sizeof(Token))) == NULL)
      HError(2,"Cannot create token buffer");
   --tBuf;
   if (trace>1) {
      printf("Network built: %d Models, %d Max States in any Model\n",
             H,maxStates);
      fflush(stdout);
   }
}
   
/* --------------------------- Recognition ----------------------- */

/* Setbjot: precompute all 'local distances' needed by active pi's */
void Setbjot(Vector ot, Vector rot)
{
   Definition *d;
   int i,j,h,N;
   
   if (isTied) PreCompute(ot);
   for (h=1;h<=H;h++)
      defList[h].bjLoaded = FALSE;
   for (j=0;j<nActive;j++){
      d = active[j]->def;
      if(rotate) RotateVector(ot, rot, d->subsdef);
      if (!d->bjLoaded) {
         N = d->hmmdef.numStates;
         for (i=2;i<N;i++)
            if (rotate) d->bjot[i] = OutP(rot,&(d->hmmdef),i) * bjtScale; /*DR - ot -> rot*/
            else d->bjot[i] = OutP(ot,&(d->hmmdef),i) * bjtScale;
         d->bjLoaded = TRUE;
      }
   }

}

/* InitModels: Activate models connected to ENTRY node and inject
               a start token */
void InitModels(void)
{
   Link p;
   int i;
   Token startToken = {NULL,0,NULL,NULL,NULL}; /* DR - "NULL,NULL" added */
   LinkSet *l;
   PhoneInstance *pi;
   
   SPINBALL
   /* DeActivateAll(); DR - moved to MatchFile */
   Nulllastplrt();
   p = theNet.entryNode; l = p->succ;
   for (i=0; i<p->numSucc;i++) {
      pi = (PhoneInstance *)(*l)[i]->user;
      if (pi==NULL)
         HError(2,"Bad Syntax Net - non-unit connected to entry node");
      MakeActive(pi);
      startToken.inst = pi;
      pi->state[1] = startToken;
   }
   if (trace>1) {
      printf("%d Models Initially Active\n",nActive);
      fflush(stdout);
   }
   InitPLRs();
}

/* StepModel: step the given model by one time frame */
void StepModel(PhoneInstance *pi, int time /* DR */)
{

   int i,j,besti;
   LogReal best,x,a;
   Token t,nullToken = {NULL,LZERO,NULL,NULL,NULL}; /* DR - added "NULL,NULL" */
   Definition *d;
   short record[100];
   ILR *ilr;
   SLR *slr;

   for (i=0;i<100;i++) record[i] = FALSE;
   
   SPINBALL
   d = pi->def;
   for (j=2;j<pi->N;j++){     /* Propagate internal tokens */
      best = LZERO; besti = 0;
      for (i=1;i<pi->N;i++) {
         if (pi->state[i].inst != NULL) {
            a = d->hmmdef.transP[i][j];
            if (a>LSMALL) {
               x =  pi->state[i].logprob + a;
               if (x>best) {
                  best = x; besti = i;
               }
            }
         }
      }
      /* DR - we've now found the best state to get to j from */
      if (besti>0) {
         tBuf[j] = pi->state[besti];           /* propogate the token into tmp storage */
         x = best + d->bjot[j];
         if (x>maxLogP) maxLogP = x;          /* record global max for pruning */
         tBuf[j].logprob = x;                 /* update the logprob slot of the token*/

         /* each token just trails a huge tail of slrs behind it */
         if(ipath_on) {

           slr = CreateSLR(); 
           slr->name = pi->def->id->name;
           slr->time = time;     
           slr->state = j;
           slr->bjot = d->bjot[j];
           slr->link = pi->state[besti].slrlink;
           tBuf[j].slrlink = slr;
     
           if (besti != j) {                      /* if the token has propogated from another state*/
             ilr = CreateILR();                   /* create a new internal link record */
             ilr->time = time;                    /* record the time of creation */
             ilr->state = j;                      /* record the state */
             ilr->chain = pi->state[besti].ilink; /* point back to the previous ilr */
             tBuf[j].ilink = ilr;                 /* point the token at this new ilr */
           }
         }
      } else
         tBuf[j] = nullToken;
   }


   for (i=2;i<pi->N;i++)      /* Copy tokens into state slots */
      pi->state[i] = tBuf[i];

   best = LZERO; besti = 0;      /* Find best exit token */
   for (i=2;i<pi->N;i++) {
      if (pi->state[i].inst != NULL) {
         a = d->hmmdef.transP[i][pi->N];
         if (a>LSMALL) {
            x =  pi->state[i].logprob + a;
            if (x>best) {
               best = x; besti = i;
            }
         }
      }
   }

   if (besti>0){
      t = pi->state[besti];
      t.logprob = best;
      if (best>maxLogP) maxLogP = best;   /* record global max for pruning */

      if (ipath_on) {
        t.slrlink = pi->state[besti].slrlink;
          
        if (besti != pi->N) {                   /* if the token has propogated */
           ilr = CreateILR();                   /* create a new internal link record */
           ilr->time = time;                    /* record the time of creation */
           ilr->state = pi->N;                  /* record the state */
           ilr->chain = pi->state[besti].ilink; /* point back to the previous ilr */
           t.ilink = ilr;                       /* point the token at this new ilr */
        }
      }
      pi->state[pi->N] = t;
   } else 
      pi->state[pi->N] = nullToken;
   if (trace==3)  
      ShowInst(pi);
}

/* StepModels: step all models by one time frame */
void StepModels(int time /* DR*/)
{
   int i;
   
   maxLogP = LZERO;
   for (i=0;i<nActive;i++)
      StepModel(active[i], time /* DR */);
}

/* NullEntryTokens: Set all entry tokens to null ie zero prob */
void  NullEntryTokens(void)
{
   int i;
   Token nullToken = {NULL,LZERO,NULL,NULL,NULL}; /* DR - "NULL,NULL" added */
   
   for (i=0;i<nActive;i++)
      active[i]->state[1] = nullToken;
}

/* PropagateExitTokens: propagate exit tokens  */
void PropagateExitTokens(void)
{
   Link p,q,*a;
   int i,j;
   Token t;
   PhoneInstance *pi;
   LogReal x,threshold;
   int fromIdx;
   ILR *ilr; /* DR */
    
   SPINBALL
   threshold = (isPruning) ? maxLogP-pruning : LZERO;
   for (i=0;i<nActive;i++) {
      pi = active[i]; p = pi->netHook;
      fromIdx = pi->def->id->aux;
      t = pi->state[pi->N];


      if (t.logprob>threshold) {
         if (fixedPen)
            t.logprob += propPen;   
         else if (!hasBigram)
               t.logprob -= log(p->numSucc);
         a = (Link *)p->succ;
         for (j=1;j<=p->numSucc;j++) {
            q = *a++;
            if (q->user != NULL) {
               pi = (PhoneInstance *)q->user;
               if (!pi->isActive) 
                  MakeActive(pi);
               x = t.logprob;
               if (hasBigram)
                  x += bigram[fromIdx][pi->def->id->aux];
               if (x>LSMALL && pi->state[1].logprob < x) {
                  pi->state[1] = t;
                  pi->state[1].logprob = x;
                  /* start DR */
                  /* check that we are in the correct label */
                  if(ipath_on) {
                    ilr = CreateILR();
                    ilr->state = 1; 
                    ilr->time = pi->state[1].ilink->time+1;           
                    ilr->chain =  pi->state[1].ilink;
                    pi->state[1].ilink = ilr;
                    pi->state[1].slrlink = t.slrlink;
                    /* end DR */
                  }
               }
            }
         }
      }
   }
}


/* RecordDecisions: create PLR's for all distinct entry nodes */
void RecordDecisions(int t)
{
   int i;
   PhoneInstance *pi;
   PLR *plr;
   Token tok;
   
   if (trace>1) ClearTrBuf();
   for (i=0;i<nActive;i++) {
      SPINBALL
      pi = active[i];
      tok = pi->state[1];
      if (tok.logprob>LSMALL) {
         if (t != tok.inst->lastplrt) {
            plr = CreatePLR(); /* Creates an empty PLR */
            plr->tok = tok;
            plr->time = t;
            tok.inst->lastplrt = t;
            tok.inst->lastplr = plr;
            if (trace>1) {
               if (hasBigram)
                  plr->tok.logprob -= 
                     bigram[tok.inst->def->id->aux][pi->def->id->aux];
               Add2TrBuf(plr);
            }
         } else
            plr = tok.inst->lastplr;
         pi->state[1].link = plr;
         pi->state[1].inst = pi;
      }
   }
   if (trace>1) PrintTrBuf(t);
}

/* ------------------------ Traceback and Output ---------------------- */

/* OutLine: print label and start and end points */
void OutLine(FILE * f, long st, long en, LabId lab)
{
   SPINBALL
   st *= sampPeriod; en *= sampPeriod;
   fprintf(f,"%10d  %10d   %3s\n",st,en,lab->name);
   if (trace>1){
      printf("%10d  %10d   %3s\n",st,en,lab->name);
      fflush(stdout);
   }
}

/* OutIPath: print out the internal path beginning at ilr */

void OutIPath(FILE *f, ILR *ilr, int n)
{
 ilr = ilr->chain;
  
 if (n>2) {
   n--;
   OutIPath(f, ilr, n);
 }

 fprintf (f, "  [%2d] %2d\n", ilr->state, ilr->time * sampPeriod);
}   

/* OutSegment: print segment of path upto p to file f */
void OutSegment(FILE *f, /*DR*/ FILE *ipathfp, PLR *p)
{
   int st;
   int i; /* DR */
   ILR *ilr; /* DR */
     
   if (p!=NULL) {
      OutSegment(f, /* DR */ ipathfp, p->tok.link);
      if (p->tok.link != NULL) 
         st = p->tok.link->time + 1;
      else
         st = 1;
      OutLine(f,st,p->time,p->tok.inst->def->id);
      if (ipath_on) { /* DR */
        if(trace>0) {
          OutLine(stdout,st,p->time,p->tok.inst->def->id); 
          ilr = p->tok.ilink->chain;       
          OutIPath(stdout, p->tok.ilink, p->tok.inst->N);  
        }
      }
   }
}


/* Outslr: DR prints out the slr chain associated with a given token */
void Outslr(FILE *fp, SLR *slr)
{
 SLR *p;

 while (slr != NULL) {
  p = slr;
  fprintf (fp, "%4d   %4d  %6s \t %f\n", p->time, p->state, p->name, p->bjot); 
  slr = p->link;
 }

}

/* OutputResults: trace back best predecessor of theNet exit */
/* DR - modified to produce the NBest outputs */
void OutputResults(FILE *f, /* DR */ FILE *ipathfp,  char *fn, int t)
{
   Link p,q,*a;
   PhoneInstance *pi, *bestpi;
   LogReal best[MAX_N_BEST];
   int i,st,n;
   ILR *ilr; /* DR */

   if (NBest > theNet.exitNode->numPred) {
     HError(0,"NBest value (%d) is greater than number of nodes jointed to network exit (5d)",NBest,theNet.exitNode->numPred);
     HError(0,"NBest is thus being reduced to %d",theNet.exitNode->numPred);
   }
   
   for (n=1;n<=NBest;n++) best[n]=0;

   for (n=1; n<=NBest; n++) {

      bestpi = NULL;
      p = theNet.exitNode;
      a = (Link *)p->pred;
      for (i=1;i<=p->numPred;i++) {
         q = *a++;
         if (q->user != NULL) {
            pi = (PhoneInstance *)q->user;
            if (pi->isActive && pi->state[pi->N].inst != NULL &&
                pi->state[pi->N].logprob > best[n]){ 
                if (n==1) {
                   best[n] = pi->state[pi->N].logprob;
                   bestpi = pi;
		}
                else {
                 if (pi->state[pi->N].logprob < best[n-1]) {
                   best[n] = pi->state[pi->N].logprob;
                   bestpi = pi;
                 }
		}
            }
         }
      }
     if (bestpi==NULL) {
        HError(0,"No tokens survived to end of input for %d'th parse%s",n,fn);
        return;
     }
     if (trace>0){
        printf("  MaxProb per frame for %d parse = %f\n",n, best[n]/(t+1));
        printf ("\n----The Viterbi path for the %d'st best parse is:---\n\n",n);
        fflush(stdout);
     }
     OutSegment(f, /* DR */ ipathfp, bestpi->state[bestpi->N].link); 
     if (bestpi->state[bestpi->N].link != NULL) 
        st = bestpi->state[bestpi->N].link->time + 1;
     else
        st = 1;
     OutLine(f,st,t,bestpi->def->id);

     /* start DR */
     if(ipath_on) { 
       if(trace>0) {
         OutLine(stdout,st,t,bestpi->def->id);
         OutIPath(stdout,  bestpi->state[bestpi->N].ilink, bestpi->N-1);      
         fprintf (stdout, "  [%2d] %2d\n\n", bestpi->state[bestpi->N].ilink->state, bestpi->state[bestpi->N].ilink->time * sampPeriod);
       }
       Outslr(ipathfp, bestpi->state[bestpi->N].slrlink);
     }
      /* end DR */

     if(n <= NBest) {
       fprintf(f, "/\n"); /* label level separator */
       if (ipath_on) fprintf(ipathfp,"/\n");
     }
     
   } /* end for n */

}

/* OutputHlResults: trace back the Hand Labelled Path */
void OutputHlResults(FILE *f, /* DR */ FILE *ipathfp,  char *fn, int t)
{
   Link p,q,*a;
   PhoneInstance *pi, *bestpi;
   LogReal best;
   int i,st,n;
   ILR *ilr; /* DR */
   
   bestpi = NULL;
   p = theNet.exitNode;
   a = (Link *)p->pred;
   for (i=1;i<=p->numPred;i++) {
     q = *a++;
     if (q->user != NULL) {
       pi = (PhoneInstance *)q->user;
       if(pi->hlpath == TRUE) {        
          best = pi->state[pi->N].logprob;
          bestpi = pi;
       }
     }
   }

   if (bestpi==NULL) {
      HError(0,"No hand labelled path  survived to end of input for  parse%s",fn);
      return;
   }
   if (trace>0){
      printf("  MaxProb per frame for hand label parse = %f\n", best/(t+1));
      printf ("\n----The Viterbi path for the input labelled parse is:---\n\n");
      fflush(stdout);
   }
   OutSegment(f, /* DR */ ipathfp, bestpi->state[bestpi->N].link); 
   if (bestpi->state[bestpi->N].link != NULL) 
      st = bestpi->state[bestpi->N].link->time + 1;
    else
       st = 1;
   OutLine(f,st,t,bestpi->def->id);

     /* start DR */
     if(ipath_on) { 
       if(trace>0) {
         OutLine(stdout,st,t,bestpi->def->id);
         OutIPath(stdout,  bestpi->state[bestpi->N].ilink, bestpi->N-1);      
         fprintf (stdout, "  [%2d] %2d\n\n", bestpi->state[bestpi->N].ilink->state, bestpi->state[bestpi->N].ilink->time * sampPeriod);
       }
       Outslr(ipathfp, bestpi->state[bestpi->N].slrlink);
     }
      /* end DR */

}

#ifdef MPW

/* MPW ONLY -- Sets output file creator to HTK; file kind to TEXT */

typedef struct  {
    unsigned long fdType;          /*the type of the file*/
    unsigned long fdCreator;       /*file's creator*/
    unsigned short fdFlags;        /*flags: hasbund,invis,locked, etc.*/
    long fdLocation;               /*file's location in folder*/
    short fdFldr;                  /*folder containing file*/
}FInfo;

void SetLabFileInfo(char *fn)
{
   FInfo inf;

   if (getfinfo(fn,0,&inf))
      HError(2,"Cannot Get File Info for %s",fn);
   inf.fdType='T'<<24|'E'<<16|'X'<<8|'T';
   inf.fdCreator='H'<<24|'T'<<16|'K'<<8|' ';
   inf.fdFlags |= 6;
   if(setfinfo(fn,0,&inf))
      HError(2,"Cannot Set File Info for %s",fn);
}
#endif

/* CheckData: check given data file for consistency with HMM's */
void CheckData(DataFile *data, char *fn)
{
   if (data->sampSize / sizeof(float) != vSize)
      HError(2,"Incompatible Vec Size in file %s",fn);
   if (data->sampKind != SK)
      HError(2,"Incompatible Sample Kind in file %s",fn);
}

#define LINKCHUNKSIZE 5

/* DR AddHlPathToNet: adds a path to the network corresponding to the hand labelled path */
void AddHlPathToNet(Transcription tr)
{

 int i;
 int numSuccMax = LINKCHUNKSIZE;
 int numPredMax = LINKCHUNKSIZE;
 Link a,b;
 PhoneInstance *pi;
 

 a = theNet.entryNode;
 for(i=0;i<tr->nLabels;i++) {
  b = CreateNode(tr->labels[i].labid, &(theNet.chain), numSuccMax, numPredMax);
  JoinNodes(a,b); 
  if((pi = (PhoneInstance *)malloc(sizeof(PhoneInstance)) ) == NULL)
    HError(2,"Unable to create phone instance");
  AttachDefinition(pi,b->modelName, TRUE);
  NPI += 1;
  pi->N = pi->def->hmmdef.numStates;
  pi->isActive = FALSE;
  if ((pi->state = (Token *)malloc(pi->N*sizeof(Token))) == NULL)
    HError(2,"Unable to create state array for model %d",pi->N);
  pi->state -= 1;
  pi->lastplrt=0;
 (PhoneInstance *)(b->user) = pi; 
  pi->netHook = b;

  a = b; 
 }

 b = theNet.exitNode;
 JoinNodes(a,b);

}


/* this unjoins two connected nodes (a->b) in the network */
void UnJoinNodes(Link a, Link b)
{

 if(NotJoined(a,b)) HError(1,"Cannot unjoin nodes if they are not connected");

  RemPredLink(a,b);
  RemSuccLink(b,a);
  ShrinkNode(a);
  ShrinkNode(b); 
} 

void FreeAttachedPI(Link a)
{
 PhoneInstance *pi;

 pi = (PhoneInstance *)(a->user);
 free (pi->state+1); 
 free(pi); 
 NPI -= 1;
}

/* DR removes the hand labelled path from the Network */

void DetachHlPathFromNet(Transcription tr)
{
 
 Link a,b;
 int i;

 a = theNet.chain;

 UnJoinNodes(a,theNet.exitNode);
 b = a;
 a = a->chain;

 for (i=1;i<tr->nLabels;i++) {
   UnJoinNodes(a,b);
   FreeAttachedPI(b);
   FreeNode(b);
   b = a;
   a = a->chain;
 }

 UnJoinNodes(theNet.entryNode,b);
 FreeAttachedPI(b);
 FreeNode(b);
 theNet.chain = a;

}

/* MatchFile: top level control */
void MatchFile(char * datafn)
{
   char outfn[255];
   DataFile data;
   FILE *outf;
   char ipathfn[256]; /* DR */
   FILE *ipathfp; /* DR */
   char ilabfn[256]; /* DR */
   FILE *ilabfp; /* DR */
   int t;
   Vector ot;
   Vector rot; /* DR - rotated vector */
   long beamWidthSum;
   float x;
   Transcription tr;       /* DR - current hand label file transcription */
   int i; /* DR */
   
   MakeFN(datafn,labDir,labExt,outfn);
   if (ipath_on) MakeFN(datafn,ipathDir,ipathExt,ipathfn); /* DR */

   if (trace>0) {
      printf("HVite[V%s]: %s -> %s\n", VERSION,datafn,outfn);
      fflush(stdout);
   }

  /* DR - start */
  if (handlabel) {
    MakeFN(datafn,ilabDir,ilabExt,ilabfn);
    if (trace>0) printf("label file name is %s\n", ilabfn);
    tr = LOpen(ilabfn);
    if (trace >=1) {
      printf("\n--the input label file contents are:--\n");
      for(i=0; i<tr->nLabels; i++)
      printf("%s\n", tr->labels[i].labid->name);
      printf("----------end of labels---------------\n\n");
    }
    fflush(stdout);
    AddHlPathToNet(tr);
    Nulllastplrt(); /****this is overkill we should NULL only the stuff as we add it ****/
  }
  /* DR - end */

  CreateActiveList();
  

  if ((outf = fopen(outfn,"w")) == NULL)
      HError(2,"Cannot create output label file %s",outfn);
   if(ipath_on) /* DR */
   if ((ipathfp = fopen(ipathfn,"w")) == NULL)                /* DR */   
      HError(2,"Cannot create output label file %s",ipathfn); /* DR */
   SpOpen(datafn,&data);
   CheckData(&data,datafn);
   sampPeriod = data.sampPeriod;
   ot = CreateVector(vSize);
   if(rotate) rot = CreateVector(vSize);
   InitModels();
   beamWidthSum = nActive;
   for (t=1;t<=data.nSamples;t++) {
      if(trace >1) {
         printf ("Number of data samples is %d\n", data.nSamples); /* DR */
         fflush(stdout); /* DR */
      }
      if(trace>1) { /* DR */
         printf ("*** time is %d ***\n", t);
         fflush(stdout);
      }
      if (trace==3)
         printf("Speech frame %d\n",t);
      GetSample(&data,t-1,ot+1);
      Setbjot(ot, rot);
      StepModels(t);
      if (isPruning)
         CheckActive();
      beamWidthSum += nActive;
      NullEntryTokens();
      PropagateExitTokens();
      GarbageCollectPLRs();
      if (t==data.nSamples) {
         OutputResults(outf,ipathfp,outfn,t);
         if (handlabel) OutputHlResults(outf,ipathfp,outfn,t); /* DR */
      }   
      else
         RecordDecisions(t);
   }
   if (trace>0 && isPruning) {
      x = ((float)beamWidthSum * 100.0)/ 
            ((float) NPI * (float)data.nSamples);
      printf("  Total Pruning = %.2f%%\n",100-x);
      fflush(stdout);
   }
   if (trace>0 && maxPLRUsed>0) {
      printf("  Max PLR's Used = %d\n",maxPLRUsed);
      fflush(stdout);
   }
   SpClose(&data);
   if (handlabel){        
     DetachHlPathFromNet(tr); /* DR */
     LClose(tr); /* DR */
   }

   if (ipath_on) { /* DR */
     FreeILRs();
     FreeSLRs();
     fclose(ipathfp);
   }

   fclose(outf);
   FreePLRs();
   DeActivateAll();
   FreeActiveList();
#ifdef MPW
    SetLabFileInfo(outfn);
#endif
   FreeVector(ot);
   if(rotate) FreeVector(rot);
}


/* ----------------------------------------------------------- */
/*                      END:  HVite.c                          */
/* ----------------------------------------------------------- */

